(defun search-and-kill()
  "Searches for a string and kills region between current point and end of string"
  (interactive)
  (let ((p (point)))
    (isearch-forward)
    (kill-region p (point))))

(global-set-key [f29] 'scroll-down)              ;PgUp
(global-set-key [f35] 'scroll-up)                ;PgDn
(global-set-key [f27] 'beginning-of-buffer)      ;Home
(global-set-key [f33] 'end-of-buffer)            ;End


(defun scroll-and-go-one-down ()
  "Scroll up one line"
  (interactive)
  (scroll-up 1)
  (next-line 1))

(defun scroll-and-go-one-up ()
  "Scroll down one line"
  (interactive)
  (scroll-down 1)
  (previous-line 1))

(defun scroll-one-down ()
  "Scroll up one line"
  (interactive)
  (scroll-up 1))

(defun scroll-one-up ()
  "Scroll down one line"
  (interactive)
  (scroll-down 1))

(global-set-key [S-kp-8] 'scroll-and-go-one-up)
(global-set-key [S-kp-2] 'scroll-and-go-one-down)
(global-set-key [M-up] 'scroll-one-up)
(global-set-key [M-down] 'scroll-one-down)


(defun display-proof ()
  (interactive)
  (display-buffer "*proof*"))

(defun switch-to-db ()
  (interactive)
  (switch-to-buffer "*database*" t))

(defun switch-to-omega ()
  (interactive)
  (switch-to-buffer "*omega*" t))

(defun clear-input ()
  (interactive)
  (let ((mark (point)))
    (backward-char)
    ;(bol-ilisp)    
    (beginning-of-line)
    (comint-skip-prompt)
    (kill-region mark (point))))

(defun stop ()
  (interactive)
  (clear-input)
  (insert ":pop")
  (return-ilisp))

(defun cont ()
  (interactive)
  (clear-input)
  (insert ":cont "))

(defun ctrl-g ()
  (interactive)
  (funcall (key-binding [?\C-g])))

(defun format-comment ()
  "Break a multi-line string into several strings each containing one line"
  (interactive)
  (reindent-lisp)
  (let ((pos (point))
	(end-pos (search-forward "\""))
	(pos2 (search-backward "\""))
	(begin-pos (search-backward "\"")))
    (while (search-forward "
" end-pos 1)
      (replace-match "\"
\"" nil t)
      (indent-line-ilisp))
    (goto-char pos)))

(global-set-key [?\M-Q] 'format-comment)
;(global-set-key [?\C-z ?\o] 'switch-to-proof)
(global-set-key [f10] 'clear-input)              ;F10
(global-set-key [f11] 'ctrl-g)                   ;F11 und Stop
(global-set-key [f14] 'undo)                     ;Undo
(global-set-key [C-M-delete] 'backward-kill-sexp)
(global-set-key [?\C-%] 'replace-string)
(global-set-key [?\C->] 'mark-buffer-end)
(global-set-key [?\C-<] 'mark-buffer-begin)

;(define-key minibuffer-local-map [f11] 'abort-recursive-edit)
;(define-key minibuffer-local-ns-map [f11] 'abort-recursive-edit)
;(define-key minibuffer-local-completion-map [f11] 'abort-recursive-edit)
;(define-key minibuffer-local-must-match-map [f11] 'abort-recursive-edit)
;(define-key buffer-select-minibuffer-map [f11] 'abort-recursive-edit)
;(define-key ilisp-mode-map [f11] 'stop)


(defun umlaut-a  ()
  (interactive)
  (insert "&auml;"))

(defun umlaut-aa  ()
  (interactive)
  (insert "&Auml;"))

(defun umlaut-o  ()
  (interactive)
  (insert "&ouml;"))

(defun umlaut-oo  ()
  (interactive)
  (insert "&Ouml;"))

(defun umlaut-u  ()
  (interactive)
  (insert "&uuml;"))

(defun umlaut-uu  ()
  (interactive)
  (insert "&Uuml;"))

(defun umlaut-s  ()
  (interactive)
  (insert "&suml;"))

(global-set-key [?\C-c ?a] 'umlaut-a)                      ;  C-c a
(global-set-key [?\C-c ?A] 'umlaut-aa)                     ;  C-c A
(global-set-key [?\C-c ?o] 'umlaut-o)                      ;  C-c o
(global-set-key [?\C-c ?O] 'umlaut-oo)                     ;  C-c O
(global-set-key [?\C-c ?u] 'umlaut-u)                      ;  C-c u
(global-set-key [?\C-c ?U] 'umlaut-uu)                     ;  C-c U
(global-set-key [?\C-c ?s] 'umlaut-s)                      ;  C-c s

