;;; -*- Mode: Lisp -*-
;;; Written 1992 in Emacs-Lisp
;;;
;;; Edited  Fri Aug 19 15:27:56 1994  by richts
;;;


(defun exdoc-time ()
  (substring (current-time-string) 11 16))

(defun exdoc-day ()
  (if (string= (substring (current-time-string) 8 9) " ")
      (concat "0" (substring (current-time-string) 9 10))
    (substring (current-time-string) 8 10)))

(defun exdoc-month ()
  (upcase (substring (current-time-string) 4 7)))

(defun exdoc-year ()
  (substring (current-time-string) 20 24))

(defun exdoc-date ()
  (concat (exdoc-day) "-" (exdoc-month) "-" (exdoc-year)))

(defun exdoc-author ()
  (capitalize (user-login-name)))


;;; 'declare'-comment

(defun exdoc-declare  ()
  "Inserts the declare template for comments with the actual date and user."
  (interactive)
  (insert "\n  (declare (edited  \""(exdoc-date)"\")\n\t   (authors "(exdoc-author)")\n")
  (insert "\t   (input   )\n\t   (effect  )\n\t   (value   ))")
  (message "Declare inserted"))

(defun exdoc-comment  ()
  "Inserts the template for comments with the actual date and user."
  (interactive)
  (insert "\n  ;; edited:  " (exdoc-date)
	  "\n  ;; authors: " (exdoc-author))
  (insert "\n  ;; input:   "
	  "\n  ;; effect:  "
	  "\n  ;; value:   ")
  (message "Comment inserted"))

(defun exdoc-update ()
  "Updates the 'authors' and 'edited' slot in the declare comment.
   The actual user and date are added in front of the lists.  If the user allready
   occurs in the authors list (but not in the last position) the name and the
   corresponding time string are deleted in the authors and edited list,
   respectively."
  (interactive)
  (beginning-of-defun)
  (end-of-defun)
  (let ((end-of-def (point)))
    (beginning-of-defun)
    (if (search-forward-regexp "declare \\((edited\\)" end-of-def t)
	(let* ((begin-edited (point))
	       (end-edited (progn (goto-char (match-beginning 1))
				  (forward-sexp 1)
				  (point)))
	       (begin-authors (search-forward "(authors" end-of-def t))
	       (end-authors (progn (goto-char (match-beginning 0))
				   (forward-sexp 1)
				   (point)))
	       (pos nil))
	  (if (not begin-authors)
	      (message "Error: Declare contains no 'authors' slot")
	    (goto-char begin-authors)
	    (skip-chars-forward " ")
	    (insert (exdoc-author) " ")
	    (if (search-forward-regexp (concat "\\b" (exdoc-author) "\\b ") end-of-def t)
	      (progn (backward-kill-word 1)
		     (setq pos (extract-count-words begin-authors))))
	      ;; save the number of words before the killed user and kill the appropriate time string.
	    (goto-char begin-edited)
	    (skip-chars-forward " ")
	    (insert (format "%S " (exdoc-date)))
	    (if pos
		(progn (forward-sexp (- pos 2))
		       (kill-sexp 1)))
	    (message "'Authors' and 'Edited' updated")
	    ))
      (message "Error: Declare is missing or incorrect!!"))))

(defun extract-count-words (begin)
  "Count the number of words (i.e. their beginning) after BEGIN and before point.
   This function modifies the point."
  (if (search-backward-regexp " \\<" begin t)
      (1+ (extract-count-words begin))
    0))

;;; Extract Documentation

(defun extract-interface-documentation (filename)
  "Extracts the TeX documentation of the interface functions of a file.
   The result is a file that can be incuded into a manual."
  (interactive "FExtract file: ")
  (let ((buffer (get-file-buffer filename)))
    (if (and buffer 
	     (buffer-modified-p (get-file-buffer filename))
	     (y-or-n-p (format "Save Buffer first? ")))
	(save-excursion (set-buffer buffer)
			(save-buffer))))
  (ilisp-send (format "(ags:doc~extract %S 'DOC+LATEX-INTERFACE nil \"tex\")" filename) 
	      (format "Extracting interface documentation of %s" filename)
	      'Extract
	      t))

(defun extract-interface-manual (filename)
  "Extracts the TeX manual of the interface functions of a file.
   The result is a file that is a stand-alone LaTeX file."
  (interactive "FExtract file: ")
  (let ((buffer (get-file-buffer filename)))
    (if (and buffer 
	     (buffer-modified-p (get-file-buffer filename))
	     (y-or-n-p (format "Save Buffer first? ")))
	(save-excursion (set-buffer buffer)
			(save-buffer))))
  (ilisp-send (format "(ags:doc~extract %S 'MAN+LATEX-INTERFACE nil \"tex\")" filename) 
	      (format "Extracting interface manual of %s" filename)
	      'Extract
	      t))

(defun extract-complete-documentation (filename)
  "Extracts the TeX documentation of the complete functions of a file.
   The result is a file that can be incuded into a manual."
  (interactive "FExtract file: ")
  (let ((buffer (get-file-buffer filename)))
    (if (and buffer 
	     (buffer-modified-p (get-file-buffer filename))
	     (y-or-n-p (format "Save Buffer first? ")))
	(save-excursion (set-buffer buffer)
			(save-buffer))))
  (ilisp-send (format "(ags:doc~extract %S 'DOC+LATEX-COMPLETE nil \"tex\")" filename) 
	      (format "Extracting complete documentation of %s" filename)
	      'Extract
	      t))

(defun extract-complete-manual (filename)
  "Extracts the TeX manual of the complete functions of a file.
   The result is a file that is a stand-alone LaTeX file."
  (interactive "FExtract file: ")
  (let ((buffer (get-file-buffer filename)))
    (if (and buffer 
	     (buffer-modified-p (get-file-buffer filename))
	     (y-or-n-p (format "Save Buffer first? ")))
	(save-excursion (set-buffer buffer)
			(save-buffer))))
  (ilisp-send (format "(ags:doc~extract %S 'MAN+LATEX-COMPLETE nil \"tex\")" filename) 
	      (format "Extracting complete manual of %s" filename)
	      'Extract
	      t))




;;; Keybindings

;(define-key keim-mode-map [f8 ?d] 'exdoc-declare)                       ;  Hyper-d
;(define-key keim-mode-map [f8 ?s] 'exdoc-update)                ;  Hyper-s

(provide 'exdocu)
