;;; Written  Sat Aug 20 16:21:40 1994  by richts@cs.uni-sb.de
;;;
;;; This file defines the keim-mode, the major mode that is used for editing Keim-programs.

(require 'repair)

(define-derived-mode keim-mode lisp-mode "KEIM" 
  "This major mode is a variant of `lisp-mode', created by `define-derived-mode'.
It inherits all of the parent's attributes, but has its own keymap,
abbrev table and syntax table:

  `keim-mode-map' and `keim-mode-syntax-table'

which more-or-less shadow

  `lisp-mode-map' and `lisp-mode-syntax-table'

\{keim-mode-map}

After running `lisp-mode-hook' and doing some initialization 
`keim-mode-hooks' is run."
  ;;; here is the place for some commands that can be run each time a buffer is switched to keim mode.
  ;;; these commands are run after lisp-mode-hook and before keim-mode-hooks (note the extra 's').
  ;;; User specific customization should not be done here but in keim-mode-hooks.
  ;;
  ;; The following lines should be at the top level but that does not work (i guess because they are evaled
  ;; before the 'keim-mode' command is run the first time and therefore 'keim-mode-map' has not been defined
  ;; yet).
  (define-key keim-mode-map [f8 ?d] 'exdoc-declare)
  (define-key keim-mode-map [f8 ?s] 'exdoc-update)
  (define-key keim-mode-map "\C-zl" 'mk-load-file-lisp)
)
  

(autoload 'exdoc-declare "/usr/local/keim/emacs/keim/exdocu" 
  "Inserts the declare template for comments with the actual date and user." t nil)
(autoload 'exdoc-update "/usr/local/keim/emacs/keim/exdocu"
  "Updates the 'authors' and 'edited' slot in the declare comment." t nil)
(autoload 'extract-interface-documentation "/usr/local/keim/emacs/keim/exdocu" 
  "Extracts the TeX documentation of the interface functions of a file." t nil)
(autoload 'extract-interface-manual "/usr/local/keim/emacs/keim/exdocu" 
  "Extracts the TeX manual of the interface functions of a file." t nil)
(autoload 'extract-complete-documentation "/usr/local/keim/emacs/keim/exdocu" 
  "Extracts the TeX documentation of the complete functions of a file." t nil)
(autoload 'extract-complete-manual "/usr/local/keim/emacs/keim/exdocu" 
  "Extracts the TeX manual of the complete functions of a file." t nil)

(autoload 'rp-start-patch "/usr/local/keim/emacs/keim/exdocu" 
  "Visit the repair file for BUFFER." t nil)
(autoload 'rp-finish-patch "/usr/local/keim/emacs/keim/exdocu" 
  "Save and checkin BUFFER." t nil)
(autoload 'rp-add-definition "/usr/local/keim/emacs/keim/exdocu" 
  "Add current definition to existing repair buffer." t nil)
(autoload 'rp-patch-buffer "/usr/local/keim/emacs/keim/exdocu" 
  "Add all changed definitions of buffer BUFFER to its repair file." t nil)





;;; Example: (add-hook 'lisp-mode-hook 'turn-on-auto-fill) (add-hook 'keim-mode-hooks (lambda () (message
;;; "This is keim mode")))


(defun mk-load-file-lisp  (file-name)
  "Load a lisp file into the current inferior LISP and go there; if the file is
registered by defsystem in the inferior LISP, call operate-on-component in the
inferior LISP, else call load-file-lisp (in Emacs Lisp)."
  (interactive "FLoad Lisp file: ")
  ;;(interactive (comint-get-source "Load Lisp file: " lisp-prev-l/c-dir/file lisp-source-modes nil))
  (comint-check-source file-name)	; Check to see if buffer needs saved.
  (if (rp-get-repair-file-name file-name)
      ;; Remark: This test is true if the file is registered (it is t if there is no repair file)
      (ilisp-send (format "(mk::operate-on-component (mk::get-file %S) :load :all)" file-name)
		  "Loading component" 'load t)
    (load-file-lisp file-name)))
  


