;; Written  Fri Aug 19 12:30:52 1994  by Richts
;;
;; define some useful keys 
;;


(defun search-and-kill()
  "Searches for a string and kills region between current point and end of string"
  (interactive)
  (let ((p (point)))
    (isearch-forward)
    (kill-region p (point))))


;;; using the PgDn, PgUp, Home and End key

(global-set-key [f29] 'scroll-down)              ;PgUp
(global-set-key [f35] 'scroll-up)                ;PgDn
(global-set-key [f27] 'beginning-of-buffer)      ;Home
(global-set-key [f33] 'end-of-buffer)            ;End


;;; defining scrolling of screen with/without moving the cursor

(defun scroll-and-go-one-down ()
  "Scroll up one line"
  (interactive)
  (scroll-up 1)
  (next-line 1))

(defun scroll-and-go-one-up ()
  "Scroll down one line"
  (interactive)
  (scroll-down 1)
  (previous-line 1))

(defun scroll-one-down ()
  "Scroll up one line"
  (interactive)
  (scroll-up 1))

(defun scroll-one-up ()
  "Scroll down one line"
  (interactive)
  (scroll-down 1))

;;; defining keys for scrolling Shift/Meta-Up/Down

(global-set-key [S-kp-8] 'scroll-and-go-one-up)
(global-set-key [S-kp-2] 'scroll-and-go-one-down)
(global-set-key [S-up] 'scroll-and-go-one-up)
(global-set-key [S-down] 'scroll-and-go-one-down)
(global-set-key [M-up] 'scroll-one-up)
(global-set-key [M-down] 'scroll-one-down)


(defun clear-input ()
  (interactive)
  (let ((mark (point)))
    (backward-char)
    ;(bol-ilisp)    
    (beginning-of-line)
    (comint-skip-prompt)
    (kill-region mark (point))))

(defun ctrl-g ()
  (interactive)
  (funcall (key-binding [?\C-g])))

(defun format-comment-old ()
  "Break a multi-line string into several strings each containing one line"
  (interactive)
  (reindent-lisp)
  (let ((pos (point))
	(end-pos (search-forward "\""))
	(pos2 (search-backward "\""))
	(begin-pos (search-backward "\"")))
    (while (search-forward-regexp "
\(\ *\)" end-pos 1)
      (replace-match "\"
\1\"" nil t)
      (indent-line-ilisp))
    (goto-char pos)))

(defun format-comment (rebreak)
  "Break a multi-line string into several strings each containing one line"
  (interactive "P")
  (set-marker lisp-fill-marker (point))
  (if rebreak
      (refill-strings))
  (goto-char lisp-fill-marker)
  (fill-lisp-string)
  (goto-char lisp-fill-marker)
  (let ((end-pos (search-backward "\""))
	;(pos2 (search-forward "\""))
	(begin-pos (search-forward "\"" nil nil 2)))
    (while (and (> (point) end-pos)
		(search-backward-regexp "
" end-pos 1))   ; \\(\\s-*\\)
      (replace-match "\"
\"" nil t)
      (delete-horizontal-space)
      (indent-line-ilisp)
      (forward-line -1)))
  (goto-char lisp-fill-marker)
  (set-marker lisp-fill-marker nil))

(defun fill-lisp-string ()
  (let* ((region (lisp-in-string))
	 (start (1+ (car region))))
    (cond (region
	   (goto-char start)
	   (let ((left-margin (current-column)))
	     (insert "\n")
	     (let* ((region2 (lisp-in-string))
		    (begin (point))
		    (end (car (cdr region2)))
		    (fill-prefix nil))
	       (fill-region begin end nil t))
	     (let ((end (point)))
	       (end-of-line 0)
	       (delete-region (point) end))
	     (goto-char start)
	     (next-line 1)
	     (delete-region start (point))
	     )))
    ))

(defun refill-strings ()
  (let* ((region (lisp-in-string))
	 (begin (car region))
	 (end (car (cdr region))))
    (goto-char end)
    (end-of-line 2)
    (while (search-backward-regexp "\"\\s-*
\\s-*\"" begin t)
      (replace-match " ")
      (setq end (car (cdr (lisp-in-string))))
      (goto-char end)
      (end-of-line 2))
    (goto-char begin)
    (beginning-of-line 0)
    (while (search-forward-regexp "\"\\s-*
\\s-*\"" end t)
      (replace-match " ")
      (setq begin (car (lisp-in-string)))
      (goto-char begin)
      (beginning-of-line 0))
    ))

(global-set-key [?\M-Q] 'format-comment)
(global-set-key [?\A-q] 'format-comment)
(global-set-key [f10] 'clear-input)              ;F10
(global-set-key [f11] 'ctrl-g)                   ;F11 und Stop
(global-set-key [f14] 'undo)                     ;Undo
(global-set-key [C-M-delete] 'backward-kill-sexp)
(global-set-key [?\C-%] 'replace-string)
(global-set-key [?\C->] 'mark-buffer-end)
(global-set-key [?\C-<] 'mark-buffer-begin)


;(define-key minibuffer-local-map [f11] 'abort-recursive-edit)
;(define-key minibuffer-local-ns-map [f11] 'abort-recursive-edit)
;(define-key minibuffer-local-completion-map [f11] 'abort-recursive-edit)
;(define-key minibuffer-local-must-match-map [f11] 'abort-recursive-edit)
;(define-key buffer-select-minibuffer-map [f11] 'abort-recursive-edit)
;(define-key ilisp-mode-map [f11] 'stop)


(defun umlaut-a  ()
  (interactive)
  (insert ""))

(defun umlaut-aa  ()
  (interactive)
  (insert ""))

(defun umlaut-o  ()
  (interactive)
  (insert ""))

(defun umlaut-oo  ()
  (interactive)
  (insert ""))

(defun umlaut-u  ()
  (interactive)
  (insert ""))

(defun umlaut-uu  ()
  (interactive)
  (insert ""))

(defun umlaut-s  ()
  (interactive)
  (insert ""))

(global-set-key [?\A-a] 'umlaut-a)                      ;  C-c a
(global-set-key [?\A-A] 'umlaut-aa)                     ;  C-c A
(global-set-key [?\A-o] 'umlaut-o)                      ;  C-c o
(global-set-key [?\A-O] 'umlaut-oo)                     ;  C-c O
(global-set-key [?\A-u] 'umlaut-u)                      ;  C-c u
(global-set-key [?\A-U] 'umlaut-uu)                     ;  C-c U
(global-set-key [?\A-s] 'umlaut-s)                      ;  C-c s


(global-set-key [?\A-'] 'umlaut-a)                      ;  C-c a
(global-set-key [?\A-"] 'umlaut-aa)                     ;  C-c A
(global-set-key [?\A-\;] 'umlaut-o)                      ;  C-c o
(global-set-key [?\A-:] 'umlaut-oo)                     ;  C-c O
(global-set-key [?\A-\[] 'umlaut-u)                      ;  C-c u
(global-set-key [?\A-{] 'umlaut-uu)                     ;  C-c U
(global-set-key [?\A--] 'umlaut-s)                      ;  C-c s

(global-set-key [?\A-\M-'] 'umlaut-aa)                      ;  C-c a
(global-set-key [?\A-\M-\;] 'umlaut-oo)                      ;  C-c o
(global-set-key [?\A-\M-\[] 'umlaut-uu)                      ;  C-c u



(defun decode-quoted-printable-region ()
  "Decode quoted-printable region."
  (interactive)
  (require 'gnus)
  (save-excursion
    (gnus-mime-decode-quoted-printable (region-beginning) (region-end))))
