
(defun TeX-LaTeX-sentinel (process name)
  "Cleanup TeX output buffer after running LaTeX."
  (cond ((TeX-TeX-sentinel-check process name))
	((and (save-excursion
		(re-search-forward "^LaTeX Warning: Citation" nil t))
;	      (let ((current (current-buffer)))
;		(set-buffer TeX-command-buffer)
;		(prog1 (TeX-check-files (TeX-master-file "bbl")
;					(TeX-style-list)
;					(append TeX-file-extensions
;						BibTeX-file-extensions))
;		  (set-buffer current)))
	      )
	 (message (concat "You should run BibTeX to get citations right, "
                          (TeX-current-pages)))
	 (setq TeX-command-next TeX-command-BibTeX))
	((re-search-forward "^LaTeX Warning: Label(s)" nil t)
	 (message (concat "You should run LaTeX again "
			  "to get references right, "
                          (TeX-current-pages)))
	 (setq TeX-command-next TeX-command-default))
	((re-search-forward "^LaTeX Warning: Reference" nil t)
	 (message (concat name ": there were unresolved references, "
                          (TeX-current-pages)))
	 (setq TeX-command-next TeX-command-Show))
	((re-search-forward "^LaTeX Warning: Citation" nil t)
	 (message (concat name ": there were unresolved citations, "
                          (TeX-current-pages)))
	 (setq TeX-command-next TeX-command-Show))
	((re-search-forward
	  "^\\(\\*\\* \\)?J?I?\\(La\\|Sli\\)TeX\\(2e\\)? \\(Version\\|ver\\.\\|<[0-9/]*>\\)" nil t)
	 (message (concat name ": successfully formatted "
			  (TeX-current-pages)))
	 (setq TeX-command-next TeX-command-Show))
	(t
	 (message (concat name ": problems after "
			  (TeX-current-pages)))
	 (setq TeX-command-next TeX-command-default))))



(defun TeX-insert-quote (arg)
  "Insert quote."
  (interactive "P")
  (self-insert-command (prefix-numeric-value arg)))

(defun TeX-insert-dollar (&optional arg)
  "Insert dollar sign.  

Show matching dollar sign if this dollar sign end the TeX math mode.  
Ensure double dollar signs match up correctly by inserting extra
dollar signs when needed.

With optional ARG, insert that many dollar signs."
  (interactive "P")
  (if arg
      (let ((count (prefix-numeric-value arg)))
	(if (listp arg)
	    (self-insert-command 1)	;C-u always inserts just one
	  (self-insert-command count)))
    (let ((pc (preceding-char))
	  (pos (point))
	  (pt (point))
	  (single t))
      (TeX-dollar-verify)
      (if (= pc (string-to-char TeX-esc))
	  (insert TeX-dollar-sign)
	(if (and (= pc TeX-dollar-sign)
		  (/= (char-after (- (point) 2)) (string-to-char TeX-esc)))
	    (progn
	      (setq single nil)
	      (if (and (> pos 2) (= (char-after (- pos 2)) TeX-dollar-sign))
		  (setq pt (1- pos))	; Doesn't echo 3rd $, if $$ already
		(backward-char 1) 
		(insert TeX-dollar-sign)
		(goto-char (1+ pos))))
	  (insert TeX-dollar-sign))
	(setq TeX-dollar-list
	      (TeX-locate-delimiter pt TeX-dollar-sign TeX-dollar-list))
	(if TeX-symbol-marker
	    (save-excursion
	      (goto-char TeX-symbol-marker-pos)
	      (if (and (= (preceding-char) TeX-dollar-sign)
		       (/= (char-after (- (point) 2)) TeX-dollar-sign))
		  (progn
		    (backward-char 1)
		    (if single
			(save-excursion
			  (goto-char pos)
			  ;JR-patch (insert TeX-dollar-sign)	; $$foo$`$'
			  )))
		(if (not single)
		    (progn
		      ;JR-patch (insert TeX-dollar-sign) ; `$'$foo$$
		      (backward-char 1))))
	      (TeX-bouncing-point (if single 1 2))))))))

(defun reload-auctex-patch ()
  (interactive)
  (load-library "patch-auctex-9.2y.el"))

(provide 'patch-auctex)
