;;;
;;; This file shows examples of some of the things you might want to
;;; do to install or customize ILISP.  You may not want to include all
;;; of them in your .emacs.  For example, the default key binding
;;; prefix for ILISP is C-z and this file changes the default prefix to
;;; C-c.  For more information on things that can be changed, see the
;;; file ilisp.el. 
;;;

;;; If ilisp lives in some non-standard directory, you must tell emacs
;;; where to get it. This may or may not be necessary.
;(setq load-path (cons (expand-file-name "~jones/emacs/ilisp/") load-path))

;;; If you always want partial minibuffer completion
(require 'completer)

;;; If want always want TMC completion
(load "completion")
(initialize-completions)

;;; If you want to redefine popper keys
(setq popper-load-hook
      '(lambda ()
	(define-key global-map "\C-c1" 'popper-bury-output)
	(define-key global-map "\C-cv" 'popper-scroll-output)
	(define-key global-map "\C-cg" 'popper-grow-output)
	(define-key global-map "\C-cb" 'popper-switch)))

;;; If you always want popper windows
(require 'popper)

;;; Autoload based on your LISP.  You only really need the one you use.
;;; If called with a prefix, you will be prompted for a buffer and program.
(autoload 'run-ilisp "ilisp" "Select a new inferior LISP." t)
(autoload 'clisp     "ilisp" "Inferior generic Common LISP." t)
(autoload 'allegro   "ilisp" "Inferior Allegro Common LISP (4.1)." t)
(autoload 'allegro-42   "ilisp" "Inferior Allegro Common LISP (4.2)." t)
(autoload 'lucid     "ilisp" "Inferior Lucid Common LISP." t)
(autoload 'cmulisp   "ilisp" "Inferior CMU Common LISP." t)
(autoload 'kcl       "ilisp" "Inferior Kyoto Common LISP." t)
(autoload 'akcl      "ilisp" "Inferior Austin Kyoto Common LISP." t)
(autoload 'ibcl      "ilisp" "Ibuki Common LISP." t)
(autoload 'scheme    "ilisp" "Inferior generic Scheme." t)
(autoload 'oaklisp   "ilisp" "Inferior Oaklisp Scheme." t)
(autoload 'keim   "ilisp" "Keen-o KEIM Stuff, using Lucid." t)
(autoload 'keim-develop   "ilisp" "Keen-o KEIM Stuff, using Lucid." t)
(autoload 'gentzen   "ilisp" "Gentzen" t)
(autoload 'mkrp      "ilisp" "Mkrp" t)
(autoload 'hades     "ilisp" "Hades" t)
(autoload 'nd     "ilisp" "Keim + ND" t)
(autoload 'keim-acl   "ilisp" "Keen-o KEIM Stuff, using Allegro." t)
(autoload 'keim-develop-acl   "ilisp" "Keen-o KEIM Stuff, using Allegro. Most up-to-date version." t)
(autoload 'nd-acl     "ilisp" "Keim + ND, using Allegro" t)
(autoload 'mkrp-acl   "ilisp" "Keen-o MKRP Stuff, using Allegro." t)
(autoload 'mkrp-post   "ilisp" "Keen-o MKRP Stuff with POST, using Allegro." t)
(autoload 'omega   "ilisp" "Keen-o OMEGA Stuff with POST, using Allegro." t)

;; In /home/omega/emacs/keim (add this path to load-path):
(autoload 'keim-mode "keim-mode" "Mode for KEIM-programs." t)

;;; This makes reading a lisp file load in ilisp.
(set-default 'auto-mode-alist
	     (append '(("\\.lisp$" . lisp-mode)) auto-mode-alist))
(setq lisp-mode-hook (list '(lambda () (require 'ilisp))))


;;; add some often-used and undefined-by-lisp-mode forms
(put 'defclass 'lisp-indent-hook 'defun)
(put 'defgeneric 'lisp-indent-hook 'defun)
(put 'defmethod 'lisp-indent-hook 'defun)

;; doesn't work with ilisp causes infinite loops
(setq  ilisp-site-hook (list '(lambda () (require  'ilisp-local))))

;;; To be honest, I'd suggest everyone disable the popper, bug or no bug.
;;; If you like it great. If you want to disembowel the thing, here's how:
(setq lisp-no-popper t)
(setq popper-load-hook 
      '(lambda ()
        (setq popper-pop-buffers nil)
        (setq popper-buffers-to-skip nil)))
