(in-package "KEIM")


;;; leanTAP

(defvar tap*env nil)

(defun tap~prove (formula &optional (var-limit 2))
  (declare (edited  "24-FEB-1994 12:35")
	   (authors RICHTS)
	   (input   "A formula in negation normal form and a number.")
	   (effect  "None.")
	   (value   "If there is a tableaux proof for FORMULA where at most NUMBER"
		    "variables are instantiated, a list representing the tableaux is"
		    "returned; else NIL."))
  (uni~with-permanent-bindings
   (tap=prove formula nil nil var-limit)))

(defun tap=prove (Fml UnExp Lits VarLim)
  (declare (edited  "24-FEB-1994 12:38")
	   (authors RICHTS)
	   (input   "The arguments represent one branch in a tableaux: The actual Fml,"
		    "a list of Fml that are still unexpanded in this branch, a number"
		    "that serves as a limit to the number of variable instantiations, and a"
		    "list of Lits, i.e. Fmle that are completely expanded.")
	   (effect  "This function uses the backtracking module, i.e. the backtrack"
		    "sequence can be changed. Furthermore some tableaux information are"
		    "stored in the bound variables.")
	   (value   "If this branch and all branches in the backtrack sequence can be"
		    "closed according to the tableaux rules, a list representing the"
		    "tableaux is returned; else NIL."))
  (let ((A (first (fo~arguments Fml)))
	(B (second (fo~arguments Fml))))
    
    (cond ((termc~conjunction-p Fml) (tap=prove A (cons B UnExp) Lits VarLim))
	  
	  ((termc~disjunction-p Fml) (bktrk~sequence
				      (tap=prove A UnExp Lits VarLim)
				      (tap=prove B UnExp Lits VarLim)))
	  
	  ((termc~universal-quantification-p Fml)
	   (if (plusp VarLim)
	       (let* ((Var (termc~quantification-bound-variable Fml))
		      (Fml1 (top~rename (termc~quantification-scope Fml)
					:variables (list var) :return-renaming nil)))
		 (tap=prove Fml1 (append UnExp (list Fml))
			    Lits (1- VarLim)))))
	  
	  ((let ((Neg (termc~negate Fml tap*env)))
	     (some #'(lambda (Lit)
		       (top~new-binding-list!)
		       (prog1 (and (uni~unify Neg Lit) (bktrk~end))
			 (top~unbind-binding-list!)))
		   Lits)))
	  
	  (UnExp
	   (tap=prove (first UnExp) (rest UnExp) (cons Fml Lits) VarLim)))))


;;; Indexing

#|
((let ((Neg (tap=negate Fml)))
   (some #'(lambda (Unifier)
	     (top~new-binding-list!)
	     (subst~bind-substitution Unifier)
	     (prog1 (bktrk~end)
	       (top~unbind-binding-list!)))
	 (nth-value 3 (toset~all-unifiable Lits Neg)))))

((and UnExp
      (tap=prove (first UnExp) (rest UnExp) (toset~add Lits Fml) VarLim)))
|#

(defun tap~i-prove (formula &optional (var-limit 2))
  (declare (edited  "24-FEB-1994 12:35")
	   (authors RICHTS)
	   (input   "A formula in negation normal form and a number.")
	   (effect  "None.")
	   (value   "If there is a tableaux proof for FORMULA where at most NUMBER"
		    "variables are instantiated, a list representing the tableaux is"
		    "returned; else NIL."))
  (uni~with-permanent-bindings
   (tap=i-prove formula nil (toset~create :indexing "Literals" nil) var-limit)))

(defun tap=i-prove (Fml UnExp Lits VarLim)
  (declare (edited  "24-FEB-1994 12:38")
	   (authors RICHTS)
	   (input   "The arguments represent one branch in a tableaux: The actual Fml,"
		    "a list of Fml that are still unexpanded in this branch, a number"
		    "that serves as a limit to the number of variable instantiations, and a"
		    "list of Lits, i.e. Fmle that are completely expanded.")
	   (effect  "This function uses the backtracking module, i.e. the backtrack"
		    "sequence can be changed. Furthermore some tableaux information are"
		    "stored in the bound variables.")
	   (value   "If this branch and all branches in the backtrack sequence can be"
		    "closed according to the tableaux rules, a list representing the"
		    "tableaux is returned; else NIL."))
  (let ((A (first (fo~arguments Fml)))
	(B (second (fo~arguments Fml))))
    
    (cond ((termc~disjunction-p Fml) (bktrk~sequence
				      (tap=i-prove A UnExp Lits VarLim)
				      (tap=i-prove B UnExp Lits VarLim)))
	  
	  ((termc~conjunction-p Fml) (tap=i-prove A (cons B UnExp) Lits VarLim))
	  
	  ((termc~universal-quantification-p Fml)
	   (if (plusp VarLim)
	       (let* ((Var (termc~quantification-bound-variable Fml))
		      (Fml1 (top~rename (termc~quantification-scope Fml)
					:variables (list var) :return-renaming nil)))
		 (tap=i-prove Fml1
			    (append UnExp (list Fml))
			    Lits (1- VarLim)))))
	  
	  ((let ((Neg (termc~negate Fml tap*env)))
	     (some #'(lambda (Unifier)
		       (top~new-binding-list!)
		       (subst~bind-substitution Unifier)
		       (prog1 (bktrk~end)
			 (top~unbind-binding-list!)))
		   (nth-value 3 (toset~all-unifiable Lits Neg)))))

	  ((and UnExp
		(tap=i-prove (first UnExp) (rest UnExp) (toset~add Lits Fml) VarLim))))))


;;; Sorts

#|
((let ((Neg (tap=negate Fml)))
   (some #'(lambda (Lit)
	     (some #'(lambda (Unifier)
		       (top~new-binding-list!)
		       (subst~bind-substitution Unifier)
		       (prog1 (bktrk~end)
			 (top~unbind-binding-list!)))
		   (uni~remaining-substitutions
		    (uni~up-create Neg Lit))))
	 Lits)))

((null (sort~gc-subsort (sort~sort (uni=term-left (car eq-pair-list)))
				(uni=term-right (car eq-pair-list))))
	 nil)
|#


(defun tap~s-prove (formula &optional (var-limit 2))
  (declare (edited  "24-FEB-1994 12:35")
	   (authors RICHTS)
	   (input   "A formula in negation normal form and a number.")
	   (effect  "None.")
	   (value   "If there is a tableaux proof for FORMULA where at most NUMBER"
		    "variables are instantiated, a list representing the tableaux is"
		    "returned; else NIL."))
  (uni~with-permanent-bindings
   (tap=s-prove formula nil nil var-limit)))

(defun tap=s-prove (Fml UnExp Lits VarLim)
  (declare (edited  "24-FEB-1994 12:38")
	   (authors RICHTS)
	   (input   "The arguments represent one branch in a tableaux: The actual Fml,"
		    "a list of Fml that are still unexpanded in this branch, a number"
		    "that serves as a limit to the number of variable instantiations, and a"
		    "list of Lits, i.e. Fmle that are completely expanded.")
	   (effect  "This function uses the backtracking module, i.e. the backtrack"
		    "sequence can be changed. Furthermore some tableaux information are"
		    "stored in the bound variables.")
	   (value   "If this branch and all branches in the backtrack sequence can be"
		    "closed according to the tableaux rules, a list representing the"
		    "tableaux is returned; else NIL."))
  (let ((A (first (fo~arguments Fml)))
	(B (second (fo~arguments Fml))))
    
    (cond ((termc~disjunction-p Fml) (bktrk~sequence
				      (tap=s-prove A UnExp Lits VarLim)
				      (tap=s-prove B UnExp Lits VarLim)))
	  
	  ((termc~conjunction-p Fml) (tap=s-prove A (cons B UnExp) Lits VarLim))
	  
	  ((termc~universal-quantification-p Fml)
	   (if (plusp VarLim)
	       (let* ((Var (termc~quantification-bound-variable Fml))
		      (Fml1 (top~rename (termc~quantification-scope Fml)
					:variables (list var) :return-renaming nil)))
		 (tap=s-prove Fml1
			    (append UnExp (list Fml))
			    Lits (1- VarLim)))))
	  
	  ((let ((Neg (top~insert-bindings! (termc~negate Fml tap*env))))
	     (some #'(lambda (Lit)
		       (some #'(lambda (Unifier)   (format T "~%~A" Unifier)
				 (top~new-binding-list!)
				 (subst~bind-substitution Unifier)
				 (prog1 (bktrk~end)
				   (format T "~%<-")
				   (top~unbind-binding-list!)))
			     (uni~remaining-solutions (uni~up-create Neg (top~insert-bindings! Lit)))))
		   Lits)))
	  
	  ((and UnExp
		(tap=s-prove (first UnExp) (rest UnExp) (cons Fml Lits) VarLim))))))


