(in-package "KEIM")


;;;; Normalization
(defun nnf~normalize (formula env &optional negate)
  (declare (edited  "24-FEB-1994 12:47")
	   (authors RICHTS)
	   (input   "A formula, the environment of this formula and a flag.")
	   (effect  "Some information about the skolemization is stored into the bond variables of formula.")
	   (value   "A formula in skolemized negation normal form that is equivalent to FORMULA."))
  (let ((subterms (fo~arguments formula)))
    (cond ((termc~negation-p formula)
	   (if negate
	       (nnf~normalize (first subterms) env nil)
	       (nnf~normalize (first subterms) env t)))
	  ((termc~disjunction-p formula)
	   (if negate
	       (nnf=conjunction-create (nnf~normalize (first subterms) env t)
				       (nnf~normalize (second subterms) env t) env)
	       (nnf=disjunction-create (nnf~normalize (first subterms) env nil)
				       (nnf~normalize (second subterms) env nil) env)))
	  ((termc~conjunction-p formula)
	   (if negate
	       (nnf=disjunction-create (nnf~normalize (first subterms) env t)
				       (nnf~normalize (second subterms) env t) env)
	       (nnf=conjunction-create (nnf~normalize (first subterms) env nil)
				       (nnf~normalize (second subterms) env nil) env)))
	  ((termc~implication-p formula)
	   (if negate
	       (nnf=conjunction-create (nnf~normalize (first subterms) env nil)
				       (nnf~normalize (second subterms) env t) env)
	       (nnf=disjunction-create (nnf~normalize (first subterms) env t)
				       (nnf~normalize (second subterms) env nil) env)))
	  ((termc~equivalence-p formula)
	   (if negate
	       (nnf=disjunction-create
		(nnf=conjunction-create (nnf~normalize (first subterms) env nil)
					(nnf~normalize (second subterms) env t) env)
		(nnf=conjunction-create (nnf~normalize (first subterms) env t)
					(nnf~normalize (second subterms) env nil) env) env)
	       (nnf=disjunction-create
		(nnf=conjunction-create (nnf~normalize (first subterms) env nil)
					(nnf~normalize (second subterms) env nil) env)
		(nnf=conjunction-create (nnf~normalize (first subterms) env t)
					(nnf~normalize (second subterms) env t) env) env)))
	  ((termc~universal-quantification-p formula)
	   (if negate
	       (nnf~normalize (nnf=skolemize (termc~quantification-bound-variable formula)
				       (termc~quantification-scope formula)
				       env)
			env t)
	       (nnf=universal-quantification-create (termc~quantification-bound-variable formula)
						    (nnf~normalize (termc~quantification-scope formula) env nil)
						    env)))
	  ((termc~existential-quantification-p formula)
	   (if negate
	       (nnf=universal-quantification-create (termc~quantification-bound-variable formula)
						    (nnf~normalize (termc~quantification-scope formula) env t)
						    env)
	       (nnf~normalize (nnf=skolemize (termc~quantification-bound-variable formula)
				       (termc~quantification-scope formula)
				       env)
			env nil)))
	  (t (if negate
		 (nnf=negation-create formula env)
	         formula)))))

(defun nnf~nnormalize (formula env &optional negate)
  (declare (edited  "24-FEB-1994 12:47")
	   (authors RICHTS)
	   (input   "A formula, the environment of this formula and a flag.")
	   (effect  "Some information about the skolemization is stored into the bond variables of formula.")
	   (value   "A formula in skolemized negation normal form that is equivalent to FORMULA."))
  (let* ((subterms (fo~arguments formula))
	 (term1 (first subterms))
	 (term2 (second subterms)))
	(cond ((termc~negation-p formula)
	       (if negate
		   (nnf~nnormalize term1 env nil)
		   (nnf~nnormalize term1 env t)))	      
	      ((termc~implication-p formula)
	       (if negate
		   (nnf~nnormalize (nnf=conjunction-create term1 (nnf=negation-create term2 env) env) env nil)
		   (nnf~nnormalize (nnf=disjunction-create (nnf=negation-create term1 env) term2 env) env nil)))
	      ((termc~equivalence-p formula)
	       (if negate
		   (nnf~nnormalize
		    (nnf=disjunction-create
		     (nnf=conjunction-create term1 (nnf=negation-create term2 env) env)
		     (nnf=conjunction-create (nnf=negation-create term1 env) term2 env) env) env nil)
		   (nnf~nnormalize
		    (nnf=disjunction-create
		     (nnf=conjunction-create term1 term2 env)
		     (nnf=conjunction-create (nnf=negation-create term1 env)
					     (nnf=negation-create term2 env) env) env) env nil)))
	      ((or (and negate (termc~universal-quantification-p formula))
		   (and (not negate) (termc~existential-quantification-p formula)))
	       (nnf~nnormalize (nnf=skolemize (termc~quantification-bound-variable formula)
					     (termc~quantification-scope formula)
					     env) env negate))
	      ((or (termc~universal-quantification-p formula)
		   (termc~existential-quantification-p formula))
	       (multiple-value-bind (nnf paths)
		   (nnf~nnormalize (termc~quantification-scope formula) env negate)
		 (values (nnf=universal-quantification-create
			  (termc~quantification-bound-variable formula) nnf env)
			 paths)))
	      ((or (termc~conjunction-p formula) (termc~disjunction-p formula))
	       (multiple-value-bind (nnf1 paths1)
		   (nnf~nnormalize term1 env negate)
		 (multiple-value-bind (nnf2 paths2)
		     (nnf~nnormalize term2 env negate)
		   (if (eq negate (termc~disjunction-p formula))
		       (values (nnf=conjunction-create nnf1 nnf2 env)
			       (* paths1 paths2))
		       (values (if (> paths1 paths2)
				   (nnf=disjunction-create nnf2 nnf1 env)
				   (nnf=disjunction-create nnf1 nnf2 env))
			       (+ paths1 paths2))))))
	      (t (values (if negate
			     (nnf=negation-create formula env)
			     formula)
			 1)))))


;;;; Creating formulae

(defun nnf=negation-create (term env)
  (appl~create (env~lookup-object `not env)
	       (list term)))

(defun nnf=conjunction-create (term1 term2 env)
  (appl~create (env~lookup-object `and env)
	       (list term1 term2)))

(defun nnf=disjunction-create (term1 term2 env)
  (appl~create (env~lookup-object `or env)
	       (list term1 term2)))

(defun nnf=implication-create (term1 term2 env)
  (appl~create (env~lookup-object `implies env)
	       (list term1 term2)))

(defun nnf=universal-quantification-create (var scope env)
  (appl~poly-create (env~lookup-object `forall env)
		    (list (abstr~create var scope))))


;;;; Skolemization

(defun nnf=skolemize (var scope env)
  (declare (edited  "24-FEB-1994 12:50")
	   (authors RICHTS)
	   (input   "A variable, a formula that is the scope of VAR, and the environment of both.")
	   (effect  "None.")
	   (value   "A formula equvalent to SCOPE where all occurences of VAR have been replaced by its skolem term."))
  (top~replace-free-variables scope (list var) (list (nnf=skolem-term var scope env))))


(defun nnf=skolem-term (var scope env)
  (declare (edited  "24-FEB-1994 13:31")
	   (authors RICHTS)
	   (input   "A variable, a formula that is the scope of VAR, and the environment of both.")
	   (effect  "None")
	   (value   "The skolem term of VAR in SCOPE, i.e. f(x_1,...,x_n) where"
		    "x_1,...,x_n are the free variables in SCOPE and f is a new function symbol."))
  (let ((vars (delete var (top~free-variables scope) :test #'term~equal)))
    (appl~create (nnf=skolem-function var vars env)
		 vars)))


(defun nnf=skolem-function (var vars env)
  (declare (edited  "24-FEB-1994 13:34")
	   (authors RICHTS)
	   (input   "A variable, a list of variables, and an environment.")
	   (effect  "None.")
	   (value   "A function symbol of type \\alpha_1,...,\\alpha_n -> \\alpha where"
		    "\\alpha is the type of VAR and \\alpha_1,...,\\alpha_n are the types"
		    "of VARS."))
  (sym~env-enter-constant
		   (read-from-string (format nil "f-~A" (incf keim::fo*function-counter)))
		   (type~function-create (length vars)) env))

