(in-package "KEIM")

(defun rep~prove-formula (limit-formula &optional (env rep*env))
  (let* ((formula (cdr limit-formula))
	 (clauses (cnf~normalize formula env))
	 (empty-clause (time (rep~prove clauses))))
    (when empty-clause
      (let ((proof (res~proof-create clauses nil (proof~linearize-proof-tree empty-clause) empty-clause env)))
	(prob~set-assumptions! proof nil)
	(prob~set-conclusion! proof formula)
	proof))))


(defun rep~prove (clauses)
  (mapc #'(lambda (clause) (node~set-justification! clause (res~initial-create "intial")))
	clauses)
  (rep=prove clauses nil))


(defun rep=prove (clauses clause-set)
  (when clauses
    (let* ((clause1 (first clauses))
	   (new-clauses
	    (mapcan #'(lambda (clause)
			(rep=resolution clause1 clause))
		    clause-set))
	   (new-factors
	    (mapcan #'(lambda (lit)
			(rep=factoring lit clause1))
		    (rest (cl~literals clause1))))
	   (empty-clause (find-if #'cl~empty-p new-clauses)))
      (or empty-clause
	  (rep=prove (append (rest clauses) new-factors new-clauses)
		     (cons clause1 clause-set))))))

(defvar rep*pos (pos~add-front 1))

(defun rep=resolution (clause1 clause2)
  (let* ((lit1 (first (cl~literals clause1)))
	 (lit2 (first (cl~literals clause2)))
	 (unifier (if (not (eq (lit~positive-p lit1)
			       (lit~positive-p lit2)))
		      (uni~unify (lit~atom lit1) (lit~atom lit2)))))
    (when unifier
      (multiple-value-bind (lit-list domain codomain)
	  (subst~apply-and-rename unifier
				  (if (lit~positive-p lit1)
				      (append (rest (cl~literals clause1)) (rest (cl~literals clause2)))
				      (append (rest (cl~literals clause2)) (rest (cl~literals clause1))))
				  :return-renaming t)
	(list (cl~create lit-list
			 :justification
			 (res~resolution-create (list clause1 clause2)  (list rep*pos rep*pos)
						unifier  (subst~create domain codomain)  (gentemp "STEP-"))))))))

(defun rep=factoring (lit clause)
  (let* ((lit1 (first (cl~literals clause)))
	 (unifier (if (eq (lit~positive-p lit1)
			  (lit~positive-p lit))
		      (uni~unify (lit~atom lit1) (lit~atom lit)))))
    (when unifier
      (multiple-value-bind (lit-list domain codomain)
	  (subst~apply-and-rename unifier (rest (cl~literals clause)))
	(list (cl~create lit-list
			 :justification
			 (res~factoring-create clause  (list rep*pos (pos~add-front (position lit (cl~literals clause))))
					       unifier  (subst~create domain codomain)  (gentemp "STEP-"))))))))
