;;;; **************************************************************
;;;; * ********************************************************** *
;;;; * *                                                        * *
;;;; * *  MODULE:            example.lisp                       * *
;;;; * *                                                        * *
;;;; * *  VERSION:           1.0                                * *
;;;; * *  LAST MODIFICATION: April 1993                         * *
;;;; * *                                                        * *
;;;; * *  AUTHOR:            C. Meyer (meyer@mpi-sb.mpg.de)     * *
;;;; * *                     MPI fuer Informatik                * *
;;;; * *                     Im Stadtwald                       * *
;;;; * *                     6600 Saarbruecken                  * *
;;;; * *                     Germany                            * *
;;;; * *                                                        * *
;;;; * *  CAUTION:           The file term.lisp must be used    * *
;;;; * *                     to manipulate terms.               * *
;;;; * *                                                        * *
;;;; * ********************************************************** *
;;;; **************************************************************

;;;**************************************************************
;;;* ********************************************************** *
;;;* *                                                        * *
;;;* *  EXAMPLE FUNCTIONS FOR FPA                             * *
;;;* *                                                        * *
;;;* ********************************************************** *
;;;**************************************************************

(defvar *my-indexed-set* nil "Set of terms for insertion in the index")
(defvar *my-query-set* nil   "Set of terms for retrieval on the index")
(defvar *my-index* nil       "FPA-Index")
(defvar *symbol-count* 1     "Next unused symbol number")

(defmacro example=time (form)
  ;; time1 returns the total run-time of 'form' as the second value of a multiple-value-list.
  `(second (multiple-value-list (time1 ,form)))
  )

(eval-when (load compile eval)
  (unless (find-package :example-vars)
    (make-package :example-vars)))

#+old(defun example~load-indexed-set (index-filename)
  " **************************************************************
    INPUT:   A file containing a set of terms.
    RETURNS: None.
    SUMMARY: Reads the declaration of symbols and gives each
             symbol a unique number according to its type.
             Then the following terms are read and inserted
             in *my-indexed-set*.
    ************************************************************** "
  
  (setq *my-indexed-set* nil)
  (format t "Build indexed-set: ~S~%" index-filename)
  (let ((*package* (find-package :example-vars)))
    (time (with-open-file (termfile index-filename :direction :input)
	    (let ((term nil)
		  (symbol nil)
		  (decl-list (read termfile)))
	      (setq *symbol-count* 1)
	      (dolist (decl decl-list)
		(setq symbol (first (second decl)))
		(if (string= "CONSTANTS" (first decl))
		    (setf (symbol-value symbol) (symbol~create-symbol symbol (incf *symbol-count* 2)))
		    (setf (symbol-value symbol) (symbol~create-symbol symbol 0))))
	      (loop
	       (if (eq (setq term (read termfile nil :eof)) :eof)
		   (return)
		   (push term *my-indexed-set*)
		   ))))))
  )

(defun example~load-indexed-set (index-filename)
  " **************************************************************
    INPUT:   A file containing a set of terms.
    RETURNS: None.
    SUMMARY: Reads the declaration of symbols and gives each
             symbol a unique number according to its type.
             Then the following terms are read and inserted
             in *my-indexed-set*.
    ************************************************************** "
  
  (setq *my-indexed-set* nil)
  (format t "Build indexed-set: ~S~%" index-filename)
  (let ((*package* (find-package :example-vars))
	(env (env~create)))
    (time (with-open-file (termfile index-filename :direction :input)
	    (let ((decl-list (read termfile)))
	      (setq *symbol-count* 1)
	      ;; read declarations into environment
	      (dolist (decl decl-list)
		(post~read-object decl env nil))
	      ;; put a symbol~ structure on each property list
	      (dolist (sym (env~class-keys env 'sym+const))
		(let ((const (env~lookup-object sym env))
		      (str (symbol~create-symbol sym
						 (incf *symbol-count* 2))))
		  (keim~put (sym~shared-symbol const) :fpa-sym str)))
	      (dolist (sym (env~class-keys env 'sym+var))
		(let ((const (env~lookup-object sym env))
		      (str (symbol~create-symbol sym 0)))
		  (keim~put (sym~shared-symbol const) :fpa-sym str)))
	      
	      (loop
	       (let ((term nil))
	       (if (eq (setq term (read termfile nil :eof)) :eof)
		   (return)
		   (push (term~read term env) *my-indexed-set*)
		   )))))))
  )

#+old(defun example~load-query-set (index-filename query-filename)
  " **************************************************************
    INPUT:   Two files containing a set of terms.
    RETURNS: None.
    SUMMARY: If both filenames are equal the query set is the same
             as the indexed set, otherwise the declaration of
             symbols is read and each symbol is given a unique
             number according to its type.
             Then the following terms are read and inserted
             in *my-query-set*.
    ************************************************************** "
  
  (setq *my-query-set* nil)
  (format t "Build query-set: ~S~%" query-filename)
  (let ((*package* (find-package :example-vars)))
    (time (if (string= index-filename query-filename)
	      (setq *my-query-set* *my-indexed-set*)
	      (with-open-file (termfile query-filename :direction :input)
		(let ((term nil)
		      (symbol nil)
		      (decl-list (read termfile)))
		  (dolist (decl decl-list)
		    (setq symbol (first (second decl)))
		    (if (string= "CONSTANTS" (first decl))
			(setf (symbol-value symbol) (symbol~create-symbol symbol (incf *symbol-count* 2)))
			(setf (symbol-value symbol) (symbol~create-symbol symbol 0))))
		  (loop
		   (if (eq (setq term (read termfile nil :eof)) :eof)
		       (return)
		       (push term *my-query-set*)
		       )))))))
  )


(defun example~load-query-set (index-filename query-filename)
  " **************************************************************
    INPUT:   Two files containing a set of terms.
    RETURNS: None.
    SUMMARY: If both filenames are equal the query set is the same
             as the indexed set, otherwise the declaration of
             symbols is read and each symbol is given a unique
             number according to its type.
             Then the following terms are read and inserted
             in *my-query-set*.
    ************************************************************** "
  
  (setq *my-query-set* nil)
  (format t "Build query-set: ~S~%" query-filename)
  (let ((*package* (find-package :example-vars))
	(env (env~create)))
    (time (if (string= index-filename query-filename)
	      (setq *my-query-set* *my-indexed-set*)
	      (with-open-file (termfile query-filename :direction :input)
		(let ((decl-list (read termfile)))
		  (setq *symbol-count* 1)
		  ;; read declarations into environment
		  (dolist (decl decl-list)
		    (post~read-object decl env nil))
		  ;; put a symbol~ structure on each property list
		  (dolist (sym (env~class-keys env 'sym+const))
		    (let ((const (env~lookup-object sym env))
			  (str (symbol~create-symbol sym
						     (incf *symbol-count* 2))))
		      (keim~put (sym~shared-symbol const) :fpa-sym str)))
		  (dolist (sym (env~class-keys env 'sym+var))
		    (let ((const (env~lookup-object sym env))
			  (str (symbol~create-symbol sym 0)))
		      (keim~put (sym~shared-symbol const) :fpa-sym str)))
		  (loop
		   (let ((term nil))
		   (if (eq (setq term (read termfile nil :eof)) :eof)
		       (return)
		       (push (term~read term env) *my-query-set*)
		       ))))))))
  )

(defun example~build-index ()
  " **************************************************************
    INPUT:   None.
    RETURNS: None.
    SUMMARY: A FPA-Index is build from *my-indexed-set*.
    ************************************************************** "
  
  (setq *my-index* nil)
  (format t "Build FPA index: ")
  (time (progn (setq *my-index* (index~make-index :fpa :path))
	       (dolist (term *my-indexed-set*)
		       (index~create-entry *my-index* term))))
  )

(defun example~retrieval ()
  " **************************************************************
    INPUT:   None.
    RETURNS: None.
    SUMMARY: A retrieval on *my-index* with all terms of
             *my-query-set*.
    ************************************************************** "
  
  (format t "Term retrieval: ")
  (time (dolist (term *my-query-set*)
		(index~retrieve *my-index* *unification* term)))
  )

(defun example~retrieval-statistics (&optional (count 4))
  " **************************************************************
    INPUT:   Optional the number of retrievals.
    RETURNS: None.
    SUMMARY: Outputs the shortest time under the retrievals
             on *my-index*.
    ************************************************************** "
  
  (setq *gc-silence* t) ; no output during a gc.
  (gc)
  
  (format t "Term retrieval: ")
  (let ((shortest-time (example=time (dolist (term *my-query-set*)
					     (index~retrieve *my-index* *unification* term))))
	(current-time 0))
    (format t "~,1F seconds" shortest-time) 
    (dotimes (i count)
      (gc)
      (setq current-time (example=time (dolist (term *my-query-set*)
					 (index~retrieve *my-index* *unification* term))))
      (format t ", ~,1F seconds" current-time) 
      (when (< current-time shortest-time)
	(setq shortest-time current-time)))
    (format t "~%Shortest time: ~,1F seconds" shortest-time))
  (setq *gc-silence* nil) ; reset
  )

(defun example~complete-retrieval (index-filename query-filename)
  " **************************************************************
    INPUT:   Two files containing a set of terms.
    RETURNS: None.
    SUMMARY: Loads an index set, loads a query set, builds the
             index from the index set and runs a complete
             retrieval over all terms of the query set.
    ************************************************************** "
  
  (example~load-indexed-set index-filename)

  (example~load-query-set index-filename query-filename)

  (example~build-index)

  (example~retrieval)
  )

(defun example~indexing-example (index-filename query-filename)
  " **************************************************************
    INPUT:   Two files containing a set of terms.
    RETURNS: None.
    SUMMARY: Loads an index set, loads a query set, builds the
             index from the index set and runs a complete
             retrieval over all terms of the query set.
             Then one hundred terms are deleted from the index
             and another retrieval is done.
    ************************************************************** "
  
  (example~complete-retrieval index-filename query-filename)

  (format t "Delete the first 100 terms of *my-indexed-set* from *my-index*:~%")

  (time (dotimes (i 100)
	  (index~delete-entry *my-index* (1+ i) (pop *my-indexed-set*))))

  (format t "Now repeat the last retrieval:~%")

  (example~retrieval)
  )

;;;**************************************************************
;;;* ********************************************************** *
;;;* *                                                        * *
;;;* *  UNIFICATION TEST                                      * *
;;;* *                                                        * *
;;;* ********************************************************** *
;;;**************************************************************

(defun example=keim-unify-term (indexed-set term2)
  (let ((matches 0)
	(set nil))
    (dolist (term1 indexed-set (values matches set))
      (when (uni~unify term1 term2)
	(push term1 set)
	(incf matches))))
  )

(defun example=fpa-keim-unify-term (index term)
  (example=keim-unify-term (index~retrieve index *unification* term) term)
  )
    
(defun example~unification-test ()
  (format t "Unification test: ")
  (time (dolist (term *my-query-set*)
    (unless (= (example=keim-unify-term *my-indexed-set* term)
	       (example=fpa-keim-unify-term *my-index* term))
      (format t "~%failed for term: ~S~%" term))
    (princ "."))
  ))

(defun example~load-and-test (index-filename query-filename)
  (example~load-indexed-set index-filename)
  
  (example~load-query-set index-filename query-filename)

  (example~build-index)

  (example~unification-test)
  )
