;;;; **************************************************************
;;;; * ********************************************************** *
;;;; * *                                                        * *
;;;; * *  MODULE:            index.lisp                         * *
;;;; * *                                                        * *
;;;; * *  VERSION:           1.0                                * *
;;;; * *  LAST MODIFICATION: April 1993                         * *
;;;; * *                                                        * *
;;;; * *  AUTHOR:            C. Meyer (meyer@mpi-sb.mpg.de)     * *
;;;; * *                     MPI fuer Informatik                * *
;;;; * *                     Im Stadtwald                       * *
;;;; * *                     6600 Saarbruecken                  * *
;;;; * *                     Germany                            * *
;;;; * *                                                        * *
;;;; * ********************************************************** *
;;;; **************************************************************

(defclass index+index ()
  ((index :initarg :index :accessor index=index :documentation "Term index")
   (path-type :initarg :path-type :reader index=path-type :documentation "only defined for :path"))
  (:documentation "General index class"))

(defgeneric index~make-index (index-type path-type)
  (:documentation
   "Before an index can be used it has to be created.
    index-type has to be the constant :fpa to create a fpa index.
    path-type has to be one of the two constants :path and :path-sub
    where :path stands for ordinary path indexing and :path-sub
    will also consider the subterms of a term. (:path-sub not yet implemented)
    The returned value will be needed as a parameter in the following functions.")
  )

(defgeneric index~create-entry (index term)
  (:documentation
   "Adds an entry to the index.
    The entry is referenced by the returned number and is supposed to be
    retrieved under the query term stored in term.")
  )

(defgeneric index~delete-entry (index id term)
  (:documentation
   "Removes indexing data structures for the entry referenced by
    id using the query term stored in the last parameter.")
  )

(defgeneric index~retrieve (index mode term)
  (:documentation
   "First we determine the index which is being used during
   retrieval. Mode contains one of the constants *unification*,
   *generalization*, or *instance*. The last parameter contains
   the query term. The function returns a list of terms with
   the desired feature.")
  )


