;;;; **************************************************************
;;;; * ********************************************************** *
;;;; * *                                                        * *
;;;; * *  MODULE:            symbol.lisp                        * *
;;;; * *                                                        * *
;;;; * *  VERSION:           1.0                                * *
;;;; * *  LAST MODIFICATION: April 1993                         * *
;;;; * *                                                        * *
;;;; * *  AUTHOR:            C. Meyer (meyer@mpi-sb.mpg.de)     * *
;;;; * *                     MPI fuer Informatik                * *
;;;; * *                     Im Stadtwald                       * *
;;;; * *                     6600 Saarbruecken                  * *
;;;; * *                     Germany                            * *
;;;; * *                                                        * *
;;;; * ********************************************************** *
;;;; **************************************************************

(defstruct (symbol=symbol (:constructor symbol=make-symbol)
			  (:conc-name   symbol=))
  ""
  (name     nil :read-only t)
  (number   0   :read-only t)
  
  (codomain     nil)
  (level        0)
  (time         0)
  (next-binding nil)
  )

;;; **************************************************************
;;; * ********************************************************** *
;;; * *                                                        * *
;;; * *  MACROS                                                * *
;;; * *                                                        * *
;;; * ********************************************************** *
;;; **************************************************************

(eval-when (load compile eval)
  (defconstant *symbol-special-integer* 0)

  (defmacro symbol~create-symbol (name number)
    `(symbol=make-symbol :name ,name :number ,number)
    )

  (defmacro symbol~special-integer ()
    *symbol-special-integer*
    )
  
  (defmacro symbol~number (symbol)
    `(symbol=number ,symbol)
    )

  (defmacro symbol~variable-p (symbol)
    `(evenp (symbol~number ,symbol))
    )
  
  (defmacro symbol~equal-p (symbol1 symbol2)
    `(= (symbol~number ,symbol1) (symbol~number ,symbol2))
    )

  (defmacro symbol~binding (symbol)
    `(symbol=codomain ,symbol)
    )

  (defmacro symbol~create-binding (symbol term)
    `(setf (symbol~binding ,symbol) ,term)
    )

  (defmacro symbol~bound-p (symbol)
    `(not (null (symbol~binding ,symbol)))
    )

  (defmacro symbol~mark (symbol)
    `(symbol=level ,symbol)
    )

  (defmacro symbol~create-mark (symbol level)
    `(setf (symbol=level ,symbol) ,level)
    )

  (defmacro symbol~next-binding (symbol)
    `(symbol=next-binding ,symbol)
    )

  (defmacro symbol~create-next-binding (symbol binding)
    `(setf (symbol=next-binding ,symbol) ,binding)
    )
  )

