;;;; **************************************************************
;;;; * ********************************************************** *
;;;; * *                                                        * *
;;;; * *  MODULE:            user.lisp                          * *
;;;; * *                                                        * *
;;;; * *  VERSION:           1.0                                * *
;;;; * *  LAST MODIFICATION: April 1993                         * *
;;;; * *                                                        * *
;;;; * *  AUTHOR:            C. Meyer (meyer@mpi-sb.mpg.de)     * *
;;;; * *                     MPI fuer Informatik                * *
;;;; * *                     Im Stadtwald                       * *
;;;; * *                     6600 Saarbruecken                  * *
;;;; * *                     Germany                            * *
;;;; * *                                                        * *
;;;; * *  CAUTION:           The "interface" might have to be   * *
;;;; * *                     changed when using this modul with * *
;;;; * *                     "terms" different than our own.    * *
;;;; * *                                                        * *
;;;; * ********************************************************** *
;;;; **************************************************************

;;; **************************************************************
;;; * Interface                                                  *
;;; * -> Replace the names in the macros with your own           *
;;; *    function's names.                                       *
;;; **************************************************************

;; Terms are built up with symbols. The first macro returns
;; a unique integer for every symbol. This integer will be 
;; used in order to sort symbols.

(defmacro user~symbol-to-integer (symbol)
  `(symbol~number (keim~get (sym~shared-symbol ,symbol) :fpa-sym))
  )

;; In order to create keys we need a symbol which is different
;; from all other symbols.

(defmacro user~symbol-special-integer ()
  `(symbol~special-integer)
  )

;; We need a macro which returns the symbol which stands
;; at the top of 'term'.

(defmacro user~term-top (term)
  `(term~top ,term)
  )

;; This macro returns a boolean value (either NIL (FALSE) or
;; T (TRUE)) which indicates whether 'term' is a variable or
;; not.

(defmacro user~term-is-variable-p (term)
  `(sym~variable-p ,term)
  )

;; This macro returns a boolean value (either NIL (FALSE) or
;; T (TRUE)) which indicates whether 'term' is an application
;; or not.

(defmacro user~term-is-application-p (term)
  `(appl~p ,term)
  )

;; The next macro returns a boolean value which is TRUE if
;; all the arguments of 'term' are variables.

#+old(defmacro user~term-all-args-are-var-p (term)
  `(term~all-args-are-var-p ,term)
  )

(defmacro user~term-all-args-are-var-p (term)
  `(every #'sym~variable-p (appl~arguments ,term))
  )

;; The next macro returns the list of arguments of 'term'
;; if 'term' is an application.

#+old(defmacro user~term-list-of-args (term)
  `(term~list-of-app-args ,term)
  )

(defmacro user~term-list-of-args (term)
  (let ((gensym (gensym)))
    `(let ((,gensym ,term))
       (when (appl~p ,gensym)
     (appl~arguments ,gensym)
     ))))





