
(in-package :keim)

(defvar uni*var-temp nil)

;-1-
;TermUnification:
;Dem Unifikationsalgoritmus wird zwei Terme b.z.w
;Atome und der Algoritmus entscheidet, ob die
;Terme b.z.w Atome unifizierbar sind, falls
;sie unifizierbar sind, dann wird einen
;allgemeinsten Unifikator ausgegeben, sonst
;nil.

(defun uni~unify-sort (term1 term2)

  (declare (edited  "03-AUG-1993 14:36")
	   (authors ABOZKURT)
	   (input  "zwei Terme" )
	   (effect  "none")
	   (value   "falls die Terme unifizierbar sind, dann liefert sie einen
		    allgemeinsten Unifikator, sonst nil"))

  (setq uni*var-temp nil)
  (uni=list-unify (list (list term1 term2)) (subst~create nil nil)))


;-2-
;ListUnification: 
;Two term-lists of equal length are given, the lists are
;pairwise unified, beginning with the heads of the lists. The output
;is the mgu of the terms occurring in the lists, modulo their
;unification arrangement.


(defun uni~unify-sort-list (list1 list2)
  
  (declare (edited  MARCH)
	   (authors HARAKETI)
	   (input   "Two lists of terms")
	   (effect  "None")
	   (value   "A mgu, treating the terms in both of the lists resp. as direct
                     subterms of an unspecified term, if these two
                     unspecified terms are unifiable or else `nil'."))

  (setq uni*var-temp nil)
  (uni=list-unify (uni=decomp-list list1 list2) (subst~create nil nil)))


;-3-
;MultiListUnification:
;The Input is alist of lists os terms, the value of the MultiListUnification
;is a substituation that is a most general unifier for all terms in each
;termlist in the input, or nil if it does not exists.


(defun uni~unify-sort-multi (l)
  
  (declare (edited  MARCH)
	   (authors HARAKETI)
	   (input   "A list of a lists of terms; the"
		    "lists must have at least 2 terms")
	   (effect  "None")
	   (value   "A substituation that is a most general unifier for all terms in each"
		    "termlist in the input, or nil if it does not exists."))

  (setq uni*var-temp nil)
  (uni=list-unify (uni=decomp-multi l) (subst~create nil nil)))
  

;***********************************************************************
(defun uni=list-unify (eq-pair-list subst-list)
  (declare (edited  "03-AUG-1993 14:40")
	   (authors HARAKETI ABOZKURT)
	   (input  "eine Liste von Termpaaren und eine leere Substitution" )
	   (effect  "none")
	   (value   "falls die Terme unifizierbar, dann liefer sie
		    einen allgemeinsten Unifikator, sonst nil"))
  
  (cond ((null eq-pair-list) (list subst-list))

	((uni=var-reverse-p  (car eq-pair-list))
	 (uni=list-unify     (cons (uni=var-reverse (car eq-pair-list))
				   (cdr eq-pair-list)) subst-list))
	((uni=var-weaken-p   (car eq-pair-list))
	 (uni=list-unify     (append (uni=var-weaken (car eq-pair-list))
				     (cdr eq-pair-list))  subst-list))
	((uni=term-weaken-p  (car eq-pair-list))
	 (mapcan #'(lambda (new-eq-pair-list)
		     (uni=list-unify (append new-eq-pair-list (cdr eq-pair-list)) subst-list))
		 (uni=term-weaken (uni=term-left  (car eq-pair-list))
				  (uni=term-right (car eq-pair-list))))) 
	
	((uni=clash-p        (car eq-pair-list)) nil)                            
        ((uni=dec-p          (car eq-pair-list))                                   
	 (uni=list-unify     (append (uni=dec (car eq-pair-list))
				     (cdr eq-pair-list))  subst-list))
	((uni=tautology-p    (car eq-pair-list))                            
	 (uni=list-unify     (cdr eq-pair-list) subst-list))
	((uni=orientation-p  (car eq-pair-list))                          
	 (uni=list-unify     (cons (uni=orientation (car eq-pair-list))
				   (cdr eq-pair-list))   subst-list))
	((uni=occur-check-p  (car eq-pair-list)) nil)                     
	((uni=binding-p      (car eq-pair-list )) ;anwendbar
	 (uni=list-unify     (uni=subst (car eq-pair-list) (cdr eq-pair-list))
			     (uni=compose (car eq-pair-list)  subst-list)))
	
	(t (error "Unification-Sort: No firing rule! ~A" eq-pair-list))))

;**************************************************************************
(defun uni=var-reverse-p (eq-pair)

  (declare (edited   "JAN-1994")
	   (authors  HARAKETI)
	   (input    "a pair of terms as a list, such '(t1 t2)")
	   (effect   "none")
	   (value    "BOOLEAN: true if t1 and t2 reversable, else nil"))
  
  (and (sym~variable-p (term~top   (uni=term-left  eq-pair)))
       (sym~variable-p (term~top   (uni=term-right eq-pair)))
       (or (uni=subsort-strict     (uni=term-left  eq-pair)
				   (uni=term-right eq-pair))
	   (and (uni=sort-equal    (uni=term-left  eq-pair)
				   (uni=term-right eq-pair))
		(uni=var-temp      (uni=term-left  eq-pair) uni*var-temp)       
		(not (uni=var-temp (uni=term-right eq-pair) uni*var-temp))))))

;---------------------------------------------------------------------------
(defun uni=var-reverse (eq-pair)
  (declare (edited   "JAN-1994")
	   (authors  HARAKETI)
	   (input    "a pair of terms as a list, such '(t1 t2")
	   (effect   "none" )
	   (value    "the reversed pair: '(t2 t1)"))

  (list (uni=term-right eq-pair)
	(uni=term-left  eq-pair)))

;**************************************************************************
(defun uni=var-weaken-p (eq-pair)
  (declare (edited   "JAN-1994")
	   (authors  HARAKETI)
	   (input    "a pair of terms as a list, such '(t1 t2)")
	   (effect   "none")
	   (value    "BOOLEAN: true if t1 and t2 weakeneable, else nil"))

  (and (sym~variable-p (term~top (uni=term-left  eq-pair)))
	   (sym~variable-p (term~top (uni=term-right eq-pair)))
	   (not (uni=subsort-p (uni=term-left  eq-pair) (uni=term-right eq-pair)))
	   (not (uni=subsort-p (uni=term-right eq-pair) (uni=term-left  eq-pair)))
	   (uni=subsort-gc     (uni=term-left  eq-pair) (uni=term-right eq-pair))))

   
;-------------------------------------------------------------------------------
(defun uni=var-weaken (eq-pair)
  (declare (edited   "JAN-1994")
	   (authors  HARAKETI)
	   (input    "a pair of terms as a list, such '(t1 t2) ")
	   (effect   "none" )
	   (value    "two pirs of list: '((t1 tz) (t2 tz)) "))

  (let ((var-deklaration (uni=var-create (uni=subsort-gc (uni=term-left  eq-pair)
							 (uni=term-right eq-pair)))))
    (list (list (uni=term-left   eq-pair)    var-deklaration)
	  (list (uni=term-right  eq-pair)    var-deklaration))))

  
;******************************************************************************
(defun uni=term-weaken-p (eq-pair)
  (declare (edited   "JAN-1994")
	   (authors  HARAKETI)
	   (input    "a pair of terms as a list, such '(t1 t2)")
	   (effect   "none")
	   (value    "BOOLEAN: true if t2 weakeneable, else nil"))

  (and (sym~variable-p (term~top (uni=term-left  eq-pair)) )
	   (not (uni=term-occur-check-p  (uni=term-left  eq-pair)
					 (uni=term-right eq-pair)))
	   (not (sort~subsort-p (sort~sort (uni=term-right eq-pair))
				(sort~sort (uni=term-left eq-pair))))))
	   
;************************************************************************
(defun uni=term-weaken  (var term)
  (declare (edited   "JAN-1994")
	   (authors  HARAKETI)
	   (input    "a pair of terms as a list, such '(t1 t2) ")
	   (effect   "none" )
	   (value    "two pirs of list: '((t1 t2') (s1 arg1)...(sn argn))"))

  (let ((gcs (sort~gc-subsort (sort~sort var) (sort~sort term))))
    (if gcs
	(mapcan #'(lambda (decl)
		    (if (sort~subsort-p (type~n-range decl) gcs)
			(let* ((subterms (fo~arguments term))
			       (new-vars (mapcar #'(lambda (sort)
						     (uni=var-create  sort))
						 (type~n-domain decl)))
			       (t1 (appl~create (term~top term) new-vars)))
			  (list (cons (list var t1) (mapcar #'list new-vars subterms))))
		      nil))
		(sort~declarations (term~top term))))))

	     
;*******************************************************************************
(defun uni=subsort-strict (term1 term2)
  (declare (edited   "JAN-1994")
	   (authors  HARAKETI)
	   (input    "two terms")
	   (effect   "none" )
	   (value    "the strict subsort of both terms"))

  (let ((sort1 (sort~sort term1))
	(sort2 (sort~sort term2)))
    (and (not (sort~equal sort1 sort2))
	 (sort~subsort-p sort1 sort2))))

;*******************************************************************************
(defun uni=subsort-p (term1 term2)
  (declare (edited   "JAN-1994")
	   (authors  HARAKETI)
	   (input    "two terms")
	   (effect   "none" )
	   (value    ""))

  (sort~subsort-p (sort~sort term1) (sort~sort term2)))
;*******************************************************************************
(defun uni=subsort-equal (term1 term2)
  (declare (edited   "JAN-1994")
	   (authors  HARAKETI)
	   (input    "two terms")
	   (effect   "none" )
	   (value    "the strict subsort of both terms"))

  (sort~equal (sort~sort term1) (sort~sort term2)))
  

;*******************************************************************************
(defun uni=subsort-gc (term1 term2)
  (declare (edited   "JAN-1994")
	   (authors  HARAKETI)
	   (input    "two terms")
	   (effect   "none" )
	   (value    "g.c subsort of both terms"))

  (sort~gc-subsort (sort~sort term1) (sort~sort term2)))

;*******************************************************************************
(defun uni=var-create (sort)
  (declare (edited   "JAN-1994")
	   (authors  HARAKETI)
	   (input    "a sort")
	   (effect   "the new var is consed into var-tem" )
	   (value    "creates a new variable of sort <sort> "
		     "in the env and returns it"))

  (car (push (sym~variable-create (gentemp "x-") sort)
	     uni*var-temp)))

;******************************************************************************
(defun uni=var-temp (var  var-temp)
  
  (declare (edited   "JAN-1994")
	   (authors  HARAKETI)
	   (input    "a var and a list of new vars")
	   (effect   "none")
	   (value    "BOOLEAN: true if var belongs to list"))
  
  (cond ((null var-temp) nil)
	((term~equal var (car var-temp)) t)
	(t  (uni=var-temp var (cdr var-temp)))))
  

;************************************************************************  

(defun uni=sort-equal (term1 term2)
  (declare (edited   "JAN-1994")
	   (authors  HARAKETI)
	   (input    "")
	   (effect   "none")
	   (value    "BOOLEAN: nil if sort1 = sort2, else ture"))
  
  (sort~equal (sort~sort term1) (sort~sort term2)))


;************************************************************************  
(defun uni=clash-p (eq-pair)
  (declare (edited  "03-AUG-1993 14:42")
	   (authors ABOZKURT)
	   (input   "ein Termpaar als Liste, also  '(t1 t2)")
	   (effect  "none")
	   (value   "BOOLEAN: falls clash, dann t, sonst nil" ))

  (let ((fn-left  (term~top (uni=term-left  eq-pair)))  
	(fn-right (term~top (uni=term-right eq-pair))))
    
    (not (or (sym~variable-p fn-left)
	     (sym~variable-p fn-right)
	     (term~equal fn-left fn-right)))))
	


(defun uni=term-left (eq-pair)
  (declare (edited  "03-AUG-1993 14:45")
	   (authors ABOZKURT)
	   (input   "ein Termpaar als Liste, also '(t1 t2)")
	   (effect  "none")
	   (value   "liefert den linken Term, also t1" ))

  (first eq-pair))

(defun uni=term-right (eq-pair)
  (declare (edited  "03-AUG-1993 14:47")
	   (authors ABOZKURT)
	   (input   " ein Termpaar als Liste, also '(t1 t2)")
	   (effect  "none")
	   (value   "liefert den rechten Term, also t2 "))

  (first (rest eq-pair)))



  
;----------------------------------------------------------------------------------------------------
(defun uni=dec-p (eq-pair)
  (declare (edited  "03-AUG-1993 14:48")
	   (authors HARAKETI)
	   (input   "ein Termpaar als Liste, also '(t1 t2)")
	   (effect  "none")
	   (value   " BOOLEAN :t falls t1 und t2 dekomponierbar, sonst nil"))

  (let ((fn-left     (term~top (uni=term-left eq-pair)))
	(fn-right    (term~top (uni=term-right eq-pair)))
	(fn-left-p   (appl~p (uni=term-left eq-pair)))
	(fn-right-p  (appl~p (uni=term-right eq-pair))))

    (and (term~equal fn-left fn-right)
	     fn-left-p
	     fn-right-p)))
   
;----------------------------------------------------------------------------------------------------

(defun uni=dec (eq-pair)
  (declare (edited  "03-AUG-1993 14:50")
	   (authors ABOZKURT)
	   (input   " ein Termpaar als Liste, also '(t1 t2)")
	   (effect  "none")
	   (value   "eine liste von Termpaaren, also '((t1` t2`) (t1`` t2``)...)"))
  
  (uni=dec-1 (fo~arguments (uni=term-left eq-pair))
	    (fo~arguments (uni=term-right eq-pair))))

(defun uni=dec-1 (args-left args-right)

  (cond ((null args-left) nil)
	(t (cons (list (car args-left) (car args-right))
		 (uni=dec-1 (cdr args-left) (cdr args-right))))))

;-----------------------------------------------------------------------------------------------------------------

(defun uni=tautology-p (eq-pair)
  (declare (edited  "03-AUG-1993 14:55")
	   (authors ABOZKURT)
	   (input   " ein Termpaar als Liste, also '(t1 t2)")
	   (effect  "none" )
	   (value   "BOOLEAN: t falls t1 und t2 tautologisch sind, sonst nil "))

  (term~equal (uni=term-right eq-pair)
	      (uni=term-left eq-pair)))
   
;------------------------------------------------------------------------------------

(defun uni=binding-p (eq-pair)
  (declare (edited  "03-AUG-1993 14:56")
	   (authors ABOZKURT)
	   (input   "eq-pair: ein Termpaar als Liste, also '(t1 t2)")
	   (effect  "none")
	   (value   "BOOLEAN: t falls t1 nicht in t2 vorkommt, sonnst nil"))

  (let ((left-var   (term~top (uni=term-left eq-pair)))
	(right-term (uni=term-right eq-pair)))
    (and (sym~variable-p left-var)
	 (sort~subsort-p (sort~sort right-term) (sort~sort left-var))
	 (not (uni=term-occur-check-p left-var right-term)))))

;------------------------------------------------------------------------------------
(defun uni=subst (eq-pair eq-pair-list)
  (declare (edited  "03-AUG-1993 15:02")
	   (authors ABOZKURT)
	   (input  " eq-pair: ein Termpaar als Liste, also '(t1 t2)
		    eq-pair-list: eine Liste von Termpaaren" )	          
	   (effect  "none")
	   (value  " eine Liste von Termpaaren: in der eq-pair-list
		    vorkommenden variablen t1 werden durch Term t2 ersetzt"))
  
  (let ((var (uni=term-left eq-pair))
	(term (uni=term-right eq-pair))
	(left-term (uni=term-left (car eq-pair-list)))
	(right-term (uni=term-right (car eq-pair-list))))
    (if (null (car eq-pair-list))
	nil
      (cons (list (top~replace-free-variables left-term (list var) (list term))
		  (top~replace-free-variables right-term (list var) (list term)))
	    (uni=subst eq-pair (cdr eq-pair-list))))))

;----------------------------------------------------------------------------------
(defun uni=compose (eq-pair substitution)
  (declare (edited  "03-AUG-1993 15:07")
	   (authors ABOZKURT)
	   (input   " eq-pair: ein Termpaar als Liste, also '(t1 t2)
		     subtitution: Eine Substitution")
	   (effect  "none")
	   (value  " Eine neue Subsitution: in der "substitution" vorkommenden
		    variablen t1 werden durch den Term t2 ersetzt"))
  (let ((var (uni=term-left eq-pair))
	(term (uni=term-right eq-pair)))
    (subst~insert-component var term
			    (subst~create (subst~domain substitution)
					  (uni=subst-replace var term
							     (subst~codomain substitution))))))


(defun uni=subst-replace (var term codomain)
  
  (if (null codomain)
      nil
    (cons (top~replace-free-variables (first codomain) (list var) (list term))
	  (uni=subst-replace var term (rest codomain)))))

;----------------------------------------------------------------------------------
(defun uni=orientation-p (eq-pair)
  (declare (edited  "03-AUG-1993 15:11")
	   (authors ABOZKURT)
	   (input    "ein Termpaar als Liste, also '(t1 t2)")
	   (effect   "none")
	   (value    "BOOLEAN: t falls, t1 und t2 orientierbar ist, sonst nil"))

  (and (not (sym~variable-p (term~top (uni=term-left eq-pair))))
	   (sym~variable-p (term~top (uni=term-right eq-pair)))))

;---------------------------------------------------------------------------------

(defun uni=orientation (eq-pair)
  (declare (edited  "03-AUG-1993 15:12")
	   (authors ABOZKURT)
	   (input  " ein Termpaar als Liste, also '(t1 t2) ")
	   (effect  "none" )
	   (value   "eine Termpaar als Liste, also '(t2 t1)"))
  
  (list (uni=term-right eq-pair)
	(uni=term-left eq-pair)))

;--------------------------------------------------------------------------------
  
(defun uni=occur-check-p (eq-pair)
  (declare (edited  "03-AUG-1993 15:14")
	   (authors ABOZKURT)
	   (input   "ein Termpaar als Liste, also '(t1 t2)")
	   (effect  "none")
	   (value   "BOOLEAN: t falls t1 in t2 vorkommt, sonst nil"))
  
  (let ((var (term~top (uni=term-left eq-pair)))
	(term (uni=term-right eq-pair)))
    
    (if (sym~variable-p var) (uni=term-occur-check-p var term)
      nil)))


(defun uni=term-occur-check-p (var term)
  (declare (edited  "03-AUG-1993 15:15")
	   (authors ABOZKURT)
	   (input  " eine Variable und ein Term")
	   (effect  "none")
	   (value   "BOOLEAN: t fall Variable "var" in Term "term"
		    vorkommt, sonst nil"))
  (cond ((null term) nil)
	((term~equal var term) t)
	((symbolp term) nil)                        
	(t (uni=term-list-occur-check var (fo~arguments term)))))

(defun uni=term-list-occur-check (var arg-list)
  (cond ((null arg-list) nil)
	((uni=term-occur-check-p var (car arg-list)) t)
	(t (uni=term-list-occur-check var (cdr arg-list)))))

;***********************************************************************

(defun uni=decomp-list (l1 l2)
  (declare (edited  MARCH)
	   (authors HARAKETI)
	   (input   "Two lists of terms")
	   (effect  "None")
	   (value   "a list of pair of lists"))
  
  (if (null l1) nil
    (cons (list (car l1) (car l2))
	  (uni=decomp-list (cdr l1) (cdr l2)))))

;----------------------------------------------------------------------
(defun uni=decomp-multi   (l)
  (declare (edited  MARCH)
	   (authors HARAKETI)
	   (input   "A list of lists of terms")
	   (effect  "None")
	   (value   "A list of pair of lists"))
  
  (if (null l) nil
    (append (uni=decomp-multi-l (car l))
	    (uni=decomp-multi   (cdr l)))))


(defun uni=decomp-multi-l (l)
  (declare (edited  MARCH)
	   (authors HARAKETI)
	   (input   "List l DIFFERENT NIL!!!")
	   (effect  "None")
	   (value   "a list of pair of lists"))
  
  (if (null (cdr l)) nil
    (cons (list (car l)  (cadr l))
	  (uni=decomp-multi-l  (cdr l)))))


(eval-when (load compile eval)
(defclass sort+unification-problem (uni+unification-problem)
  ((solutions :initarg :solutions :accessor sort=solutions))
  (:documentation "The class of sorted unification problems.")))

(defun uni~up-create (term1 term2)
  (make-instance 'sort+unification-problem
		 :initial-problem (list (list term1 term2))
		 :solutions (uni~unify-sort term1 term2)
		 :previous-solutions nil))


(defmethod uni~pop-solution ((up sort+unification-problem))
  (let ((solution (pop (sort=solutions up))))
    (push solution (uni=previous-solutions up))
    solution))

(defmethod uni~top-solution ((up sort+unification-problem))
  (first (sort=solutions up)))

(defmethod uni~remaining-solutions ((up sort+unification-problem))
  (prog1 (sort=solutions up)
    (setf (uni=previous-solutions up) (append (sort=solutions up) (uni=previous-solutions up)))
    (setf (sort=solutions up) nil)))
