;;; -*- Mode: LISP; Base: 10; Syntax: Common-lisp; Package: KEIM -*-
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; ;;
;;                                                                          ;;
;;   Copyright (C) 1993 by AG Siekmann, Fachbereich Informatik,             ;;
;;   Universitaet des Saarlandes, Saarbruecken, Germany.                    ;;
;;   All rights reserved.                                                   ;;
;;   For information about this program, write to:                          ;;
;;     KEIM Project                                                         ;;
;;     AG Siekmann/FB Informatik                                            ;;
;;     Universitaet des Saarlandes                                          ;;
;;     Postfach 1150                                                        ;;
;;     D-66041 Saarbruecken                                                 ;;
;;     Germany                                                              ;;
;;   electronic mail: keim@cs.uni-sb.de                                     ;;
;;                                                                          ;;
;;   The author makes no representations about the suitability of this      ;;
;;   software for any purpose.  It is provided "AS IS" without express or   ;;
;;   implied warranty.  In particular, it must be understood that this      ;;
;;   software is an experimental version, and is not suitable for use in    ;;
;;   any safety-critical application, and the author denies a license for   ;;
;;   such use.                                                              ;;
;;                                                                          ;;
;;   You may use, copy, modify and distribute this software for any         ;;
;;   noncommercial and non-safety-critical purpose.  Use of this software   ;;
;;   in a commercial product is not included under this license.  You must  ;;
;;   maintain this copyright statement in all copies of this software that  ;;
;;   you modify or distribute.                                              ;;
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; ;;

(in-package "KEIM")

(mod~defmod ho-uni :uses (appl keim mod subst sym term top )
	    :documentation "Abstract interface to unification and matching, 
                includes algorithms and the unification (matching)-problem datastructure."
	    :exports nil
	    )

(eval-when (load compile eval)
(defclass ho-uni+unification-problem (uni+unification-problem)
  ((children :initarg :sons
	     :reader ho-uni=problem-children
	     :writer ho-uni=write-problem-children!)
   (current-problem :initarg :current-problem
		    :reader ho-unif=current-problem
		    :writer ho-unif=write-current-problem!))
  (:documentation "The class of higher-order unification problems.")))

(defgeneric ho-unif~imitate! (up)
  (declare (edited  "24-JAN-1994 14:48")
	   (authors KOHLHASE)
	   (input  "A higher-order unification problem UP.")
	   (effect "In UP a new (child)-unification problem that is obtained from UP by imitation is added.")
	   (value  "The new unification problem."))
  (:method ((up ho-uni+unification-problem))
	   (let* ((set-of-pairs (ho-unif=current-problem up))
		  (pair (ho-unif~find-flex-rigid set-of-pairs))
		  (head-flex (term~head (first pair)))
		  (head-rigid (term~head (first pair)))
		  (binding (ho-unif~general-imitation-binding head-rigid (term~type head-flex)))
		  (new-set-of-pairs (cons (set-of-pairs (list head-flex binding))))
		  (new-prob (make-instance 'ho-uni+unification-problem
					   :initial-problem new-set-of-pairs
					   :previous-solutions nil
					   :current-problem (ho-uni~simpl new-set-of-pairs)
					   :sons nil)))
	     (ho-unif~write-current-problem! (cons new-prob (ho-unif~current-problem up)) up)
	     new-prob)))
	     
