;;; -*- Mode: LISP; Base: 10; Syntax: Common-lisp; Package: KEIM -*-
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; ;;
;;                                                                          ;;
;;   Copyright (C) 1993 by AG Siekmann, Fachbereich Informatik,             ;;
;;   Universitaet des Saarlandes, Saarbruecken, Germany.                    ;;
;;   All rights reserved.                                                   ;;
;;   For information about this program, write to:                          ;;
;;     KEIM Project                                                         ;;
;;     AG Siekmann/FB Informatik                                            ;;
;;     Universitaet des Saarlandes                                          ;;
;;     Postfach 1150                                                        ;;
;;     D-66041 Saarbruecken                                                 ;;
;;     Germany                                                              ;;
;;   electronic mail: keim@cs.uni-sb.de                                     ;;
;;                                                                          ;;
;;   The author makes no representations about the suitability of this      ;;
;;   software for any purpose.  It is provided "AS IS" without express or   ;;
;;   implied warranty.  In particular, it must be understood that this      ;;
;;   software is an experimental version, and is not suitable for use in    ;;
;;   any safety-critical application, and the author denies a license for   ;;
;;   such use.                                                              ;;
;;                                                                          ;;
;;   You may use, copy, modify and distribute this software for any         ;;
;;   noncommercial and non-safety-critical purpose.  Use of this software   ;;
;;   in a commercial product is not included under this license.  You must  ;;
;;   maintain this copyright statement in all copies of this software that  ;;
;;   you modify or distribute.                                              ;;
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; ;;

(in-package "KEIM")

(mod~defmod simpl :uses (appl keim mod subst sym term top )
	    :documentation "Abstract interface to unification and matching, 
                includes algorithms and the unification (matching)-problem datastructure."
	    :exports nil
	    )

(eval-when (load compile eval)
(defclass simpl+unification-problem (uni+unification-problem)
  ((children :initarg :sons
	     :reader simpl=problem-children
	     :writer simpl=write-problem-children!)
   (current-problem :initarg :current-problem
		    :reader simpl=current-problem
		    :writer simpl=write-current-problem!))
  (:documentation "The class of higher-order unification problems."))

(defclass simpl+bound-variable (sym+var)
  (:documentation "The class of bound variables, useful for thowing away binders")))

(defgeneric simpl~bound-variable-create (thing type &optional refsym)
  (declare 
   (authors nesmith)
   (input   "A term symbol, a shared symbol or a symbol, and a type."
	    "Optionally a reference symbol (currently ignored).")
   (effect  "none")
   (value   "Creates a new variable using the name of the given symbol and the"
	    "type, and possibly using the properties of the optional reference symbol,"
	    "and returns it.")
   (example "'X I --> X, "
	    "'X I C  --> X"))
  (:method ((symbol shsym+sharedsymbol) (type type+type) &optional refsym)
	   (simpl=declare-variable-bound!
	    (if refsym
		(sym~variable-create type thing refsym)
	      (sym~variable-create type thing)))))

(defun simpl=declare-variable-bound! (var)
  (declare (edited  "27-JAN-1994 10:18")
	   (authors KOHLHASE)
	   (input   "A variable.")
	   (effect  "The class of VAR is changed to SIMPL+BOUND-VARIABLE.")
	   (value   "The bound variable VAR."))
  (change-class var 'simpl+bound-variable))

	   
(defgeneric simpl~imitate! (up)
  (declare (edited  "24-JAN-1994 14:48")
	   (authors KOHLHASE)
	   (input  "A higher-order unification problem UP.")
	   (effect "In UP a new (child)-unification problem that is obtained from UP by imitation is added.")
	   (value  "The new unification problem."))
  (:method ((up simpl+unification-problem))
	   (let* ((set-of-pairs (simpl=current-problem up))
		  (pair (simpl~find-flex-rigid set-of-pairs))
		  (head-flex (term~head (first pair)))
		  (head-rigid (term~head (first pair)))
		  (binding (simpl~general-imitation-binding head-rigid (term~type head-flex)))
		  (new-set-of-pairs (cons (set-of-pairs (list head-flex binding))))
		  (new-prob (make-instance 'simpl+unification-problem
					   :initial-problem new-set-of-pairs
					   :previous-solutions nil
					   :current-problem (simpl~simpl new-set-of-pairs)
					   :sons nil)))
	     (simpl~write-current-problem! (cons new-prob (simpl~current-problem up)) up)
	     new-prob)))
	     
