;;; -*- syntax: common-lisp; package: keim; base: 10; mode: lisp -*-
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; ;;
;;                                                                          ;;
;;   Copyright (C) 1993 by AG Siekmann, Fachbereich Informatik,             ;;
;;   Universitaet des Saarlandes, Saarbruecken, Germany.                    ;;
;;   All rights reserved.                                                   ;;
;;   For information about this program, write to:                          ;;
;;     KEIM Project                                                         ;;
;;     AG Siekmann/FB Informatik                                            ;;
;;     Universitaet des Saarlandes                                          ;;
;;     Postfach 1150                                                        ;;
;;     D-66041 Saarbruecken                                                 ;;
;;     Germany                                                              ;;
;;   electronic mail: keim@cs.uni-sb.de                                     ;;
;;                                                                          ;;
;;   The author makes no representations about the suitability of this      ;;
;;   software for any purpose.  It is provided "AS IS" without express or   ;;
;;   implied warranty.  In particular, it must be understood that this      ;;
;;   software is an experimental version, and is not suitable for use in    ;;
;;   any safety-critical application, and the author denies a license for   ;;
;;   such use.                                                              ;;
;;                                                                          ;;
;;   You may use, copy, modify and distribute this software for any         ;;
;;   noncommercial and non-safety-critical purpose.  Use of this software   ;;
;;   in a commercial product is not included under this license.  You must  ;;
;;   maintain this copyright statement in all copies of this software that  ;;
;;   you modify or distribute.                                              ;;
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; ;;

(in-package :keim)

(mod~defmod post2ad :uses (appl fo keim mod prob sym termix )
	    :documentation "Prints an ADLER syntax representation of a KEIM object"
	    :exports
	    (post2ad~transform
	     post2ad~read-post-problem
	     post2ad~write-object-to-file
	     post2ad-style
	     ))

#{\section{Post to Adler Transformation}
This module implements a transformer for printing a KEIM problem in
the Adler syntax (the syntax adopted as standard by the German Scherpunkt
Deduktion members).

This is very rudimentary.
#}

(defun post2ad~transform (source-file target-file)
  (declare (edited  "07-MAR-1994 09:37")
	   (authors ACSEHN)
	   (input "A source file in POST syntax and a target file."  )
	   (effect "The source file is read and transformed into ADLER syntax and written to target file."  )
	   (value "none."  ))
  (let ((problem (post2ad~read-post-problem source-file)))
    (post2ad~write-object-to-file target-file problem)))

(defun post2ad~read-post-problem (file)
  (declare (edited  "11-MAR-1994 11:32")
	   (authors ACSEHN)
	   (input "a directory and a file. "  )
	   (effect "none." )
	   (value  "the POST problem read form file in that directory."  ))
  (with-open-file (stream file :direction :input)
		  (prob~read (read stream))))

(defun post2ad~write-object-to-file (filename object)
  (declare (edited  "11-MAR-1994 11:33")
	   (authors ACSEHN)
	   (input "A file and an object. "  )
	   (effect "none." )
	   (value  "the object is written to the file."  ))
  (with-open-file (stream filename :direction :output :if-exists :supersede)
    (pp~pprint object 'post2ad-style stream)))

(eval-when (load compile eval)
(defmacro post2ad=normalize-symbol-or-string (obj)
  `(substitute  #\_ #\- (string-downcase (string ,obj))
		:test #'char-equal)))


(defun post2ad=print-prob (stream object)
  (let ((*standard-output* stream))
    (pprint-logical-block 
     (nil nil :prefix "begin_problem("
	  :suffix "end_problem.")
     (write-string (post2ad=normalize-symbol-or-string (keim~name object)))
     (write-string ").")
     (terpri)
     (when (prob~assumptions object)
       (pprint-logical-block
	(nil (prob~assumptions object) :prefix "list_of_axioms."
	     :suffix "end_of_list.")
	(loop (terpri)
	      (pprint-exit-if-list-exhausted)
	      (write-string "  ")
	      (pprint-indent :current 0)
	      (write (pprint-pop))))
       (terpri))
     (pprint-logical-block
      (nil (list (prob~conclusion object)) :prefix "list_of_theorems."
	   :suffix "end_of_list.")
      (loop (terpri)
	    (pprint-exit-if-list-exhausted)
	    (write-string "  ")
	    (pprint-indent :current 0)
	    (write (pprint-pop))))
     (terpri))))

(defun post2ad=print-named-term (stream object)
  (pprint-logical-block 
   (stream (list (termix~term object)
		 (post2ad=normalize-symbol-or-string (keim~name object))
		 )
	   :prefix "formula(" :suffix ").")
   (pprint-indent :block 0)
   (write (pprint-pop))
   (write-string ",")
   (pprint-newline :linear)
   (write (pprint-pop))))

(defun post2ad=print-appl (stream object)
  (let ((*standard-output* stream))
    (if (fo~quantification-p object)
	(let ((quantif-var (fo~quantification-bound-variable object))
	      (scope (fo~quantification-scope object))
	      (head (appl~function object)))
	  (pprint-logical-block 
	   (nil nil)
	   (write head)
	   (pprint-indent :block 1)
	   (write-string "([")
	   (write quantif-var)
	   (write-string "],")
	   (pprint-indent :block 1)
	   (pprint-newline :linear)
	   (write scope)
	   (write-string ")")))
	(let ((head (appl~function object))
	      (rest (appl~arguments object)))
	  (pprint-logical-block
	   (nil nil)
	   (write head)
	   (pprint-logical-block
	    (nil rest :prefix "(" :suffix ")")
	    (pprint-indent :block 0)
	    (loop (pprint-exit-if-list-exhausted)
		  (write (pprint-pop))
		  (pprint-exit-if-list-exhausted)
		  (write-string ",")
		  (pprint-newline :linear)
		  ))
	   )))))

(defun post2ad=print-sym (stream object)
  (write-string (post2ad=normalize-symbol-or-string
		 (if (string-equal (keim~name object) "=")
		     "equal"
		     (keim~name object)))
		stream))

(pp~defstyle post2ad-style
	     :help "A style for printing KEIM objects in Adler syntax."
	     :parent pp+top
	     :pprint-methods
	     ((appl+appl
	       post2ad=print-appl
	       5)
	      (sym+sym
	       post2ad=print-sym
	       5)
	      (termix+named-term
	       post2ad=print-named-term
	       5)
	      (prob+problem
	       post2ad=print-prob
	       5)
	      ))
