(in-package #-symbolics :user #+symbolics :fcl-user)
;;; Here we define a bunch of lisp-implementation dependent stuff. 
;;; *lisp-fasl-type* will allow us (by always using the :output-file
;;; argument to compile-file) to allow several lisps to coexist and not
;;; get their compiled files mixed up.  This will have to be augmented
;;; for any lisps that aren't represented.


(defvar *lisp-fasl-type*
  #+(and :symbolics :imach)         "ibin"
  #+(and :Symbolics (not :imach))   "bin"
  #+(and dec common vax (not ultrix)) "FAS"
  #+(and dec common vax ultrix)       "fas"
  #+KCL                               "o"
  #+Xerox                             "dfasl"
  #+(and Lucid MC68000)               "lbin"
  #+(and Lucid VAX VMS)               "vbin"
  #+(and Lucid sparc unix)            "sbin"
  #+excl                              "fasl"
  #+(and CMU (or SUN4 SPARC))         "sparcf"
  #+(and CMU (not (or SUN4 SPARC)))   "cmufasl"
  #+PRIME                             "pbin"
  #+HP                                "b"
  #+TI                                "xfasl"
  )

(unless (boundp '*lisp-fasl-type*) (setq *lisp-fasl-type* "newfasl"))

;;; Lucid Lisp has a stupid compile-file that ignores the umask we have,
;;; producing compiled files that are not writeable by the group.
;;; We prefer to set compiled files to be readable and writeable by
;;; the group.

#+lucid
(progn
	 
  (def-foreign-function (set-umask (:language :c)
				   (:name "_umask")
				   (:return-type :signed-32bit))
      (number :signed-32bit))

  (defadvice (compile-file force-umask-002) (file &rest args)
    (let (old-umask)
      (unwind-protect
	   (progn (setq old-umask (set-umask #o002))
		  (apply-advice-continue file args))
	(set-umask old-umask))))
  )

;;; Allegro 4.1 locks the COMMON-LISP package, so that none of its symbols
;;; can be redefined.  But we want to do that in defining the logical-pathnames
;;; stuff (which is not yet implemented in the Lisp itself).
#+allegro-v4.1
(setq excl:*enable-package-locked-errors* nil)
;;; Allegro by default loads fasl files as only stubs. The .fasl files then 
;;; must always be where they were when they were loaded, otherwise
;;; the lisp will get really confused.  We'll turn off this feature, which
;;; will make the dumped lisp a lot bigger, but at least more predictable.
;;; Following doesn't really work, though, because this variable is bound
;;; while loading. So when you begin the initial load, you should say
;;; :libfasl nil as optional args to load.
#+allegro-v4.1
(setq excl:*libfasl* nil)

;;; These are declarations we use..
(proclaim '(declaration edited authors input effect value remark example))
(export '(edited authors input effect value remark example))

;;; We'll assume that CLOS or PCL has already been loaded into this
;;; lisp. At the moment there isn't a good portable way to ensure this here.
;;; We may also have to fix up PCL.  See how *omega-boot-files* is defined.

;;; Here we define the top directory of our distribution.  We'll define
;;; other subdirectories relative to this definition.

(defvar *omega-top-dir* 
     (concatenate 'string 
       #+symbolics "js-sfbsun:" #-symbolics ""
       "/home/omega/"))

;;; We have a real simple setup at the moment for this package. 
;;; We can't use logical-pathnames, because we haven't loaded that file yet.
;;; If things get more complicated, then we'll have to change this.

(defvar *omega-boot-dir*
    (concatenate 'string *omega-top-dir* "sys/prog/"))

;;; Here are the files that need to be loaded to get us up and running

(defvar *omega-boot-files* 
    '(#+pcl "pcl-fixes" ; if we are using PCL, load a patch file first.
      "logical-pathnames" "modules"))

;;; Before we load the files, we make the package AGS.

(unless (find-package "AGS")
  (make-package "AGS" :use 
		#+symbolics '("FCL-USER" "FUTURE-COMMON-LISP")
		#+(and (not symbolics) (or kcl akcl)) 
		'("PCL" "LISP" "ITERATE" "WALKER")
		#+(and (not (or symbolics kcl akcl)) lucid) 
		'("USER" "LUCID-COMMON-LISP" "LISP")
		#+cmu
		'("COMMON-LISP-USER" "COMMON-LISP" "PCL")
		#+allegro-v4.1
		'("COMMON-LISP-USER" "COMMON-LISP" "EXCL")
		#-(or symbolics akcl kcl lucid cmu allegro-v4.1) '("COMMON-LISP-USER")))

(unless (find-package "LOGICAL-PATHNAME")
	      (make-package "LOGICAL-PATHNAME" :nicknames '("LP")))

;;; Get the following symbols into the AGS package, then export them to
;;; any package that uses AGS.

(import '(*omega-top-dir* *omega-boot-files* 
	  *omega-boot-dir* *lisp-fasl-type* *ags-modules*)
	(find-package "AGS"))

(export '(*omega-top-dir* *omega-boot-files* 
	  *omega-boot-dir* *lisp-fasl-type* *ags-modules*)
	(find-package "AGS"))


;;; Okay, now load the files.
(let ((*default-pathname-defaults* (pathname *omega-boot-dir*)))
  (dolist (file *omega-boot-files*)
    (load (merge-pathnames (pathname file) *default-pathname-defaults*))))

;;; After things have been loaded, we set up a logical pathname translation
;;; for these files.
(setf (lp:physical-host-type "js-sfbsun") :unix)
(setf (lp:physical-host-type nil) :unix)

(setf (lp:logical-pathname-translations "AGS")
  `(("**;*.SRC"  ,(lp:logical-pathname "AGS:**;*.lisp"))
    ("**;*.FASL" ,(lp:logical-pathname 
                   (concatenate 'string "AGS:**;*." *lisp-fasl-type*)))
    ("SYS;**" ,(concatenate 'string *omega-top-dir* "sys/**"))
    ("BOOT;**" ,(concatenate 'string *omega-boot-dir* "**"))
    ("PROG;**" ,(concatenate 'string *omega-top-dir* "sys/prog/**"))
    ("MAN;**" ,(concatenate 'string *omega-top-dir* "sys/manual/**"))
    ("MANUAL;**" ,(concatenate 'string *omega-top-dir* "sys/manual/**"))
    ("MACROS;**" ,(concatenate 'string *omega-top-dir* "doc/macros/**"))
    ("EXDOCU;**" ,(concatenate 'string *omega-top-dir* "sys/prog/exdoc/**"))
    ("EMACS;**" ,(concatenate 'string *omega-top-dir* "sys/emacs/**"))))

(defvar *ags-modules* '("AGS:BOOT;modules" "AGS:BOOT;check-modules"
			#+symbolics"AGS:EMACS;symbolics"
			"AGS:EXDOCU;docl" "AGS:EXDOCU;doca" "AGS:EXDOCU;doc"))

(defvar *compile-if-source-newer* t
  "Used by the load functions defined in the boot files.  If non-nil, 
source files newer than their compiled versions will be compiled.  Overridden
by *always-compile*.")

(defvar *always-compile* nil
  "Used by the load functions defined in the boot files.  If non-nil, 
every source file will be compiled.  Overrides *compile-if-source-newer*.")


(dolist (module (cdr *ags-modules*))
  (funcall #'(lambda (file) (funcall (symbol-function
			       (find-symbol "MOD~LOAD-FILE" "AGS"))
			      file
			      :always-compile
			      *always-compile* 
			      :compile-if-source-newer 
			      (or *always-compile* *compile-if-source-newer*)))
	    module))

;;; Now that we've loaded AGS, we push this feature onto the list so
;;; it can be seen that AGS is present in this lisp.

(pushnew :ags *features*)

