;;;****************************************************************************
;;;* Test                                                                     *
;;;****************************************************************************

(in-package :keim)

;;;+--------------------------------------------------------------------------+
;;;| universe                                                                 |
;;;+--------------------------------------------------------------------------+

(defvar env)      ; universe
(defvar a)          ; constants 
(defvar b)
(defvar c) 
(defvar c1)
(defvar c2)
(defvar d)
(defvar u)          ; variables
(defvar v)
(defvar v1)
(defvar v2)
(defvar x)
(defvar y)
(defvar z)

(defvar ab)            ; applications
(defvar abc)
(defvar abz)
(defvar ax)
(defvar axb)
(defvar axy)
(defvar ayxz)
(defvar az)
(defvar bc)
(defvar by)
(defvar cd)
(defvar c1c2u)
(defvar uv)
(defvar uva)
(defvar uvcd)
(defvar uv1cd)
(defvar vx)
(defvar vzu)
(defvar v2bc)
(defvar xay)
(defvar xuv)
(defvar xuv1)
(defvar xyv2b)
(defvar xy)
(defvar xyab)
(defvar xyb)
(defvar xz)
(defvar yuac)
(defvar yuv2c)
(defvar yy)
(defvar zv1)
(defvar zv2)
(defvar zxv)
(defvar zxv1)

(defvar tterm)

(defun cpaci~univ-create ()
  (declare (special env))

  (setf env (env~create))
  (post~read-object-list '((type-constants i o)
			   (constants (f (i i i))
				      (a i) (b i) (c i) (d i)
				      (c1 i) (c2 i))
			   (variables (u i) (v i) (x i) (y i) (z i)
				      (v1 i) (v2 i))
			   (signatures (aci (+ (i i i))) (aci (* (i i i)))))
			 env)
                               ; universe filled with individuals
  (setf a (term~read 'a env))          ; constants 
  (setf b (term~read 'b env))
  (setf c (term~read 'c env)) 
  (setf c1 (term~read 'c1 env))
  (setf c2 (term~read 'c2 env))
  (setf d (term~read 'd env))
  (setf u (term~read 'u env))          ; variables
  (setf v (term~read 'v env))
  (setf v1 (term~read 'v1 env))
  (setf v2 (term~read 'v2 env))
  (setf x (term~read 'x env))
  (setf y (term~read 'y env))
  (setf z (term~read 'z env))

  (setf ab (term~read '(f a b) env))            ; applications
  (setf abc (term~read '(f a (f b c)) env))
  (setf abz (term~read '(f a (f b z)) env))
  (setf ax (term~read '(f a x) env))
  (setf axb (term~read '(f a (f x b)) env))
  (setf axy (term~read '(f a (f x y)) env))
  (setf ayxz (term~read '(f (f a y) (f x z)) env))
  (setf az (term~read '(f a z) env))
  (setf bc (term~read '(f b c) env))
  (setf by (term~read '(f b y) env))
  (setf cd (term~read '(f c d) env))
  (setf c1c2u (term~read '(f c1 (f c2 u)) env))
  (setf uv (term~read '(f u v) env))
  (setf uva (term~read '(f u (f v a)) env))
  (setf uvcd (term~read '(f (f u v) (f c d)) env))
  (setf uv1cd (term~read '(f (f u v1) (f c d)) env))
  (setf vx (term~read '(f v x) env))
  (setf vzu (term~read '(f v (f z u)) env))
  (setf v2bc (term~read '(f v2 (f b c)) env))
  (setf xay (term~read '(f x (f a y)) env))
  (setf xuv (term~read '(f x (f u v)) env))
  (setf xuv1 (term~read '(f x (f u v1)) env))
  (setf xyv2b (term~read '(f (f x y) (f v2 b)) env))
  (setf xy (term~read '(f x y) env))
  (setf xyab (term~read '(f (f x y) (f a b)) env))
  (setf xyb (term~read '(f x (f y b)) env))
  (setf xz (term~read '(f x z) env))
  (setf yuac (term~read '(f (f y u) (f a c)) env))
  (setf yuv2c (term~read '(f (f y u) (f v2 c)) env))
  (setf yy (term~read '(f y y) env))
  (setf zv1 (term~read '(f z v1) env))
  (setf zv2 (term~read '(f z v2) env))
  (setf zxv (term~read '(f z (f x v)) env))
  (setf zxv1 (term~read '(f z (f x v1)) env))
  (setf tterm (term~read '(+ a b) env)))

;;;+--------------------------------------------------------------------------+
;;;| test - functions                                                         |
;;;+--------------------------------------------------------------------------+

(defun cpaci~label-check ()     ; list of labels of atomic terms
  (let
      ((cpaci*theory (th~term-theory tterm)))
    (declare (special cpaci*theory))
    (mapcar #' cpaci=label (list a b c c1 c2 d u v v1 v2 x y z))))

(defun cpaci~label-reset! (termlist)    ; reset of labels of atomic terms
  (let
      ((cpaci*theory (th~term-theory tterm)))
    (declare (special cpaci*theory))
    (mapc #' (lambda (term) (cpaci=set-label! term nil))
	     termlist)
    (values)))

(defun cpaci~print-clause (formset clauselist)
  (declare (ignore c))
  (let*
      ((stream T)
       (tab (cpaci=formset-tab formset))
       (vl (cpaci=sel-var_list (cpaci=formset-sel formset)))
       (cl (cpaci=sel-const_list (cpaci=formset-sel formset))))
    (labels
	((print-clause (clause)
		       (if (equal (cpaci=lit-neg
				   (first (cpaci=cl-pos clause)))
				  :false) 
			   (format stream " ~A ~A => False~%"
				   (cpaci=cl-count clause)
					; number of not true evaluated
					; negative literals in clause
				   (premise clause))
			   (format stream " ~A ~A => ~A~%"
				   (cpaci=cl-count clause)
				   (premise clause)
				   (conclusion clause))))
	 (premise (clause)
		  (mapcan #'(lambda (c)
			      (mapcan
			       #'(lambda (v)
				   (when
				       (member
					(cpaci=aref tab v c)
					(cpaci=cl-neg clause))
				     (list (cons v c))))
			       vl)) cl))

	 ;; (premise clause) is a list of literals pointing at clause
      
	 (conclusion (clause)
		     (mapcan #'(lambda (c)
				 (mapcan
				  #'(lambda (v)
				      (when
					  (member
					   (cpaci=aref tab v c)
					   (cpaci=cl-pos clause))
					(list (cons v c))))
				  vl)) cl)))

      ;; (conclusion clause) is the list of literals from
      ;; (cpaci=cl-pos clause) 
      
      ;;=====================================================================
      (mapcar #'print-clause clauselist))))
 


(defun cpaci~theo-init (equations const-restr)
  (let*
      ((variables (top~free-variables equations))
       (gup (unicp~general-up-create variables))
       (theory (th~term-theory tterm)))
    (unicp~theo-init theory gup equations nil nil nil const-restr)))

(defun cpaci~test (key)
  (cpaci~univ-create)
  (case key
   ((1) (cpaci~theo-init
	 (list (list xay ayxz az) (list az y b)) nil))
   ((2) (cpaci~theo-init
	 (list (list xay) (list az)) nil))
   ((3) (cpaci~theo-init
	 (list (list xay) (list az)) (list (list a z))))
   ((4) (unicp~new-constr
	 (cpaci~theo-init (list (list xay) (list az)) nil)
	 nil
	 nil
	 nil
	 nil))
   ((5) (unicp~new-constr
	 (cpaci~theo-init (list (list xay) (list az)) nil)
	 nil
	 nil
	 nil
	 (list (list a z) (list x z) (list y z))))
   ((6) (unicp~new-constr
	 (cpaci~theo-init (list (list xay) (list az)) nil)
	 nil
	 nil
	 nil
	 (list (list a z) (list x y))))
   ((7) (unicp~new-constr
	 (cpaci~theo-init (list (list ab xy uva cd) (list xy zv1 zv2 uv)) nil)
	 nil
	 nil
	 nil
	 nil))
   ((8) (unicp~new-constr
	 (unicp~new-constr
	  (cpaci~theo-init (list (list xy) (list vx)) nil)
   	  nil
	  nil
	  (list v)
	  nil)
	 nil
	 nil
	 (list x)
	 nil))
   ((9) (unicp~new-constr
	 (cpaci~theo-init (list (list xy) (list vx)) nil)
	 nil
	 nil
	 (list v x)
	  nil))
   ((10) (unicp~new-constr
	  (cpaci~theo-init (list (list xy) (list vzu)) nil)
       	  nil
	  nil
	  (list v)
	  (list (list v u) (list v y))))
   ((11) (unicp~new-constr
	  (cpaci~theo-init (list (list xy) (list vzu)) nil)
	  nil
	  nil
	  (list z)
	  (list (list z x))))
   ((12) (unicp~new-constr
	  (cpaci~theo-init (list (list x) (list a)) nil)
	  nil
	  nil
	  nil
	  (list (list a x))))
   ((13) (cpaci~theo-init
	  (list (list x) (list abc)) nil))
   ((14) (unicp~new-constr
	  (cpaci~theo-init (list (list u) (list xy)) nil)
       	  nil
	  nil
	  (list u)
	  nil))
   ((15) (unicp~new-constr
	  (cpaci~theo-init (list (list x) (list uv)) nil)
       	  nil
	  nil
	  (list u v)
	  nil))
   ((16) (unicp~new-constr
	  (cpaci~theo-init (list (list u) (list xy)) nil)
       	  nil
	  nil
	  (list u x)
	  nil))
  ((17) (cpframe~decide-unify
	 (list (term~read '(+ x y) env))
	 (list (term~read '(f u v) env))))
  ((18) (cpcb~decide-unify
	 (list (term~read '(+ x y) env))
	 (list (term~read '(f u v) env))))
  ((19) (unicp~new-constr
	  (unicp~new-constr
	   (unicp~new-constr
	    (cpaci~theo-init (list (list x) (list uv)) nil)
	    nil
	    nil
	    (list u)
	    nil)
	   nil
	   nil
	   (list u)
	   nil)
       	  nil
	  nil
	  nil
	  (list (list u x))))
  ((20) (unicp~new-constr
	 (unicp~new-constr
	  (cpaci~theo-init (list (list x) (list uv)) nil)
	  nil
	  (list (list u v))
	  (list u v)
	  nil)
	 nil
	 nil
	 nil
	 (list (list u x))))
  ((21) (unicp~new-constr
	  (cpaci~theo-init (list (list x) (list uv)) nil)
       	  nil
	  nil
	  (list v)
	  nil))
   ))

;-----------------------------------------------------------------------------

#|
(defun cpaci~const-add-clauses! (formset constants)
  (declare (edited  "22-NOV-1995")
	   (authors Felix)
	   (input   "formset [cpaci+formset], list of variables x with"
		    "ind(x) <> ACI")
	   (effect  "adds for each 'variable' x from constant"
		    "a clause P(x x) => False and a clause"
		    "NIL => P(x y) for each y<>x (dicriminated)")
	   (value   "")
	   (declare (special  cpaci*theory)))
  (let
      ((tab (cpaci=formset-tab formset))
       (vl (cpaci=sel-var_list (cpaci=formset-sel formset)))
       (cpaci*false (cpaci=formset-false formset)))
    (mapc #'
     (lambda (var)
       (when
	   (member var vl :test #'term~equal)
	 (let ((clause (list
			1
			(list (cpaci=aref tab var var))
			(list cpaci*false))))
	   (push clause (cpaci=lit-pos cpaci*false))
	   (push clause (cpaci=formset-clauselist formset))
	   (cpaci=lit-neg-entry! var var clause tab))
	 (mapc #'
	  (lambda (const)
	    (when (not (term~equal const var))
	      (let ((clause (list
			     0
			     nil
			     (list (cpaci=aref tab var const)))))
		(push clause (cpaci=formset-clauselist formset))
		(cpaci=lit-pos-entry! var const clause tab))))
	  (cpaci=sel-const_list (cpaci=formset-sel formset)))))
       constants)
     formset))



(defun cpaci~values! (up)
  (declare (special  cpaci*theory cpaci*falselit))
  (let*
      ((formset (cpaci~up-formset up))
       (tab (cpaci=formset-tab formset))
       (vl (cpaci=sel-var_list (cpaci=formset-sel formset)))
       (cl (cpaci=sel-const_list (cpaci=formset-sel formset)))
       (new-const-restr nil)
       (new-ind-aci nil)
       (new-identified nil)
       (row-flag nil)
       (temp-identified nil))
    (mapc #'
     (lambda (var)
       (setf row-flag 0)
       (mapc #'
	(lambda (const)
	  (let
	      ((lit (cpaci=aref tab var const)))
	    (cond
	      ((equal  (cpaci=lit-tv lit) nil)
	       (setf row-flag 2))
	      ((and
		(eql row-flag 0)
		(equal  (cpaci=lit-tv lit) :false))
	       (setf row-flag 1)
	       (setf temp-identified (unicp~id-create var const)))
	      ((and
		(eql row-flag 1)
		(equal  (cpaci=lit-tv lit) :false))
	       (setf row-flag 2)))
	    (when
		(and
		 (member lit cpaci*falselit)
		 (not (term~equal var const)))
	      (push var new-ind-aci)
	      (push (unicp~cr-create var (list const)) new-const-restr))))
	cl)
       (when
	   (and
	    (eql row-flag 1)
	    (not (term~equal
		  (unicp~id-left temp-identified)
		  (unicp~id-right temp-identified)))
	    (push temp-identified new-identified))))
     vl)
    (values
     up
     new-identified
     nil
     (delete-duplicates new-ind-aci)
     nil
     new-const-restr
     nil)))

|#
