;;; -*- Package: HADES; Syntax: Common-lisp; Mode: LISP -*-


(in-package 'hades)

(defmacro set-term (var val)
  `(progn (term-names ,var)
	  (setf ,var ,val)))

(defun tu.test-unifiers (multi &optional number)
  (let ((up (tu.test=unify.multi multi)))
    (if number
	(do ((i number (1- i))
	     (subst (do-send up :next-solution) (do-send up :next-solution))
	     (res nil (cons subst res)))
	    ((or (null subst) (zerop i)) res))
	(copy-list (uni-substitutions up)))))

(defun tu.test=unify.multi (multi)
  (uni-create.termlists.unification.problem multi))


(defvar tu.test*verbose nil)

(defun tu.test-subst.test (subst-list1 lsubst-list2)
  (declare (edited  "12-SEP-1990 15:19")
		    (authors RICHTS)
		    (input   "A list of substitutions and a list of user-input-substitutions (see tu.test=subst-equal).")
		    (effect  "SUBST-LIST1 ist replaced by (subst-names SUBST-LIST1).")
		    (value   "T, iff the two lists contain equal substitutions and the application of each substitution"
			     "   in SUBST-LIST1 on each multi-equation in MULTI makes a list of equal terms.")
		    (special multi tu.test*verbose))  ;global Variable
  (if tu.test*verbose
      (tu.test-subst.test.verbose subst-list1 lsubst-list2)
      (prog1 
	(or (eq lsubst-list2 t)
	    (and (numberp lsubst-list2) (= (length subst-list1) lsubst-list2))
	    (and (listp lsubst-list2) (= (length subst-list1) (length lsubst-list2))
		 (tu.test=subst-list-equal subst-list1 lsubst-list2)
		 (or (every #'(lambda (subst) 
				(every #'(lambda (term-list)
					   (every #'(lambda (term) (eql term (car term-list)))
						  (cdr term-list)))
				       (uds-subst.apply.to.term subst multi)))
			    subst-list1)
		     (format t "~%Achtung: Vergleichstest erfoglreich, aber Apply-Test nicht: Vergleichsobjekt falsch."))))
	(when subst-list1
	  (let ((subst-name1 (subst-names subst-list1 nil)))
	    (setf (car subst-list1) (car subst-name1)
		  (cdr subst-list1) (cdr subst-name1)))))))

(defun tu.test-subst.test.verbose (subst-list1 lsubst-list2)
  (declare (edited  "12-SEP-1990 15:19")
		    (authors RICHTS)
		    (input   "A list of substitutions and a list of user-input-substitutions (see tu.test=subst-equal).")
		    (effect  "SUBST-LIST1 ist replaced by (subst-names SUBST-LIST1)."
			     "Each substitution in SUBST-LIST1 that is not in LSUBST-LIST2 or doesn't make multi a list"
			     "of equal terms and each substitution in LSUBST-LIST2 that is not in SUBST-LIST1 is printed")
		    (value   "T, iff the two lists contain equal substitutions and the application of each substitution"
			     "   in SUBST-LIST1 on each multi-equation in MULTI makes a list of equal terms.")
		    (special multi))  ;global Variable
  (prog1 
    (let* ((flag t)
	   (test1 (or (eq lsubst-list2 t)
		      (and (numberp lsubst-list2) (= (length subst-list1) lsubst-list2))
		      (and (listp lsubst-list2) ; (= (length subst-list1) (length lsubst-list2))
			   (tu.test=subst-list-equal subst-list1 lsubst-list2 t))))
	   (test2 (every #'identity
			 (mapcar #'(lambda (subst)
				     (let ((new-multi (uds-subst.apply.to.term subst multi)))
				       (if (every #'(lambda (term-list)
						      (every #'(lambda (term) (eql term (car term-list)))
							     (cdr term-list)))
						  new-multi)
					   t
					   (progn
					     (when flag
					       (print "Falsche Loesungen (instanzierte Terme nicht gleich:")
					       (setf flag nil))
					     (format t "~%~A~%  :: ~A" (subst-names subst nil) (term-names new-multi))
					     nil))))
				 subst-list1))))
      (if (and test1 (not test2))
	  (format t "~%Achtung: Vergleichstest erfoglreich, aber Apply-Test nicht: Vergleichsobjekt falsch."))
      (and test1 test2))
    (when subst-list1
      (let ((subst-name1 (subst-names subst-list1 nil)))
	(setf (car subst-list1) (car subst-name1)
	      (cdr subst-list1) (cdr subst-name1))))))

(defun tu.test=list-subst (subst)
  (cond ((null subst) nil)
	((consp subst) (cons (tu.test=list-subst (car subst)) (tu.test=list-subst (cdr subst))))
	(t (cons (term-list (subst-domain subst)) (term-list (subst-codomain subst))))))

(defun tu.test=subst-list-equal (subst-list1 lsubst-list2 &optional verbose)
  (if verbose
      (let ((flag1 t) (flag2 t))
	(mapc #'(lambda (subst) (cond ((member subst lsubst-list2 :test #'tu.test=subst-equal))
				      (t (when flag1
					   (format t "~%Zusaetzliche Loesungen:")
					   (setf flag1 nil))
					 (format t "~%~A" (subst-names subst nil)))))
	      subst-list1)
	(mapc #'(lambda (lsubst) (cond ((member lsubst subst-list1 :test #'(lambda (x y) (tu.test=subst-equal y x))))
				       (t (when flag2
					    (format t "~%Fehlende Loesungen")
					    (setf flag2 nil))
					  (format t "~%~A" lsubst))))
	      lsubst-list2)
	(and flag1 flag2 (= (length subst-list1) (length lsubst-list2))))
      (set= subst-list1 lsubst-list2 :test #'tu.test=subst-equal)))

(defun tu.test=subst-equal (subst1 lsubst2)
  (declare (edited  "12-SEP-1990 15:27")
	   (authors RICHTS)
	   (input   "A substitution and a user-input-substitution of the form:"
		    " ((<var-symbol1> <list-term1>) (<var-symbol2> <list-term2>) ...)")
	   (effect  )
	   (value   "t, iff the two substitutions are equal"))
  (let (var-alist var-list)
    (labels ((alist-injektiv (var1 var2)
	       (let ((old1 (cdr (assoc var1 var-alist)))
		     (old2 (rassoc var2 var-alist)))
		 (cond ;((and (numberp var1) (not (member var2 lsubst2 :test #'tu.test=member-subst))) nil) ;keine neuen Vars
		       ((eq old1 var2) t)
		       (old1 nil)
		       (old2 nil)
		       (t (setq var-alist (acons var1 var2 var-alist))
			  t))))
	     (term-renamed (lterm1 lterm2)
	       (cond ((not (listp var-alist)) 'false)
		     ((or (null lterm1) (null lterm2)) (and (null lterm1) (null lterm2)))
		     ((or (consp lterm1) (consp lterm2))
		      (and (consp lterm1) (consp lterm2)
			   (term-renamed (car lterm1) (car lterm2))
			   (term-renamed (cdr lterm1) (cdr lterm2))))
		     ((not (atom lterm2)) nil)
		     ((numberp lterm2)
		      (and (term-constant-p lterm1)
			   (numberp (term-constant.value lterm1))
			   (= (term-constant.value lterm1) lterm2)))
		     ((eq lterm1 (eval lterm2)) t)
		     ((and (term-variable-p lterm1)
			   (or (not (numberp (eval lterm2))) (term-variable.p (eval lterm2)))
			   (alist-injektiv (eval lterm2) lterm1)) t)
		     (t nil)))
	     (subst-pair-exists (var-term-list)
	       (let* ((var (eval (car var-term-list)))
		      (lterm2 (cadr var-term-list))
		      (lterm1 (term-list (subst-term var subst1))))
		 (cond ((null lterm1) 
			(and (tu.test=variable.p lterm2) (alist-injektiv (eval lterm2) var)))
		       ((term-renamed lterm1 lterm2) (push var var-list) t)
		       (t nil))))
	     (apair-exists (apair)
	       (let ((var1 (car apair))
		     (var2 (cdr apair)))
		 (cond ((or (not (numberp var1)) (eq var1 var2)) t)
		       ((eq (subst-term var1 subst1) var2) (setq var-list (cons var1 var-list)) t)
		       (t nil)))))
      (and (every #'subst-pair-exists lsubst2)
	   (every #'apair-exists var-alist)     ;(not (format t "~%domain: ~A ; var-list: ~A" (subst-domain subst1) var-list))
	   (set= (subst-domain subst1) var-list)))))

(defun tu.test=member-subst (var var-term-list)
  (or (eql var (eval (car var-term-list)))
      (and (atom (cadr var-term-list)) (eql var (eval (cadr var-term-list))))))

(defun tu.test=term-renamed (lterm1 lterm2 var-alist)
  (cond ((not (listp var-alist)) 'false)
	((null lterm1) (if (null lterm2) var-alist 'false))
	((consp lterm1) 
	 (if (consp lterm2) 
	     (tu.test=term-renamed (cdr lterm1) (cdr lterm2) (tu.test=term-renamed (car lterm1) (car lterm2) var-alist))
	     'false))
	((not (atom lterm2)) 'false)
	((eq lterm1 (eval lterm2)) var-alist)
	((and (term-variable-p lterm1) (not (numberp (eval lterm2))))
	 (let ((old-value (cdr (assoc (eval lterm2) var-alist)))
	       (old-key (rassoc lterm1 var-alist)))
	   (cond ((eq old-value lterm1) var-alist)
		 (old-key 'false)
		 ((null old-value) (acons (eval lterm2) lterm1 var-alist))
		 (t 'false))))
	(t 'false)))

(defun tu.test=variable.p (symbol)
  (and (atom symbol) (term-variable.p (eval symbol))))

#|
(defun au.test=equivalent.subst.lists (substlist1 substlist2)
  (cond
    (substlist2
     (and (>= (list-length substlist1) (list-length substlist2))
	  (every
	    (function (lambda (subst)
			(let ((term (uds-subst.apply.to.term subst au.test*term1)))
			  (and
			    (cond (au.test*term4
				   (and (uds-term.= term (uds-subst.apply.to.term subst au.test*term2))
					(uds-term.= term (uds-subst.apply.to.term subst au.test*term3))
					(uds-term.= term (uds-subst.apply.to.term subst au.test*term4))))
				  (au.test*term3
				   (and (uds-term.= term (uds-subst.apply.to.term subst au.test*term2))
					(uds-term.= term (uds-subst.apply.to.term subst au.test*term3))))
				  (t
				   (uds-term.= term (uds-subst.apply.to.term subst au.test*term2))))
			    (cond (au.test*term11
				   (uds-term.= (uds-subst.apply.to.term subst au.test*term11)
					       (uds-subst.apply.to.term subst au.test*term12)))
				  (t
				   t))
			    (cond (au.test*term21
				   (uds-term.= (uds-subst.apply.to.term subst au.test*term21)
					       (uds-subst.apply.to.term subst au.test*term22)))
				  (t
				   t))
			    (cond (au.test*term31
				   (uds-term.= (uds-subst.apply.to.term subst au.test*term31)
					       (uds-subst.apply.to.term subst au.test*term32)))
				  (t
				   t))))))
	    substlist1)))
    (t
     (not substlist1))))
|#