(in-package "KEIM")

(defun dioadd~solve-equations (matrix &optional (constrained nil))
  (declare (ignore constrained))
  (let ((result (mapcan #'dioadd=solve-equations (dioadd=split-matrix matrix))))
    (format t "~%Solutions for diophantine equations (~A x ~A): ~A"
	    (length (first matrix)) (length matrix) (length result))
    result))

(defun dioadd=solve-equations (matrix)
  (multiple-value-bind (compressed-matrix pos-vector)
      (dioadd=compress-matrix matrix)
    (let ((n (length (first matrix))))
      (mapcar #'(lambda (solution)
		  (dioadd=expand-solution solution pos-vector n))
	      (diocd-solve.equations compressed-matrix)))))

(defun dioadd=compress-matrix (matrix-rest &optional (pos 0)
					   (compressed-matrix (make-list (length matrix-rest)))
					   pos-vector)
  (cond ((some #'null matrix-rest)
	 (values compressed-matrix pos-vector))
	((some #'(lambda (row) (not (zerop (first row))))
	       matrix-rest)
	 (dioadd=compress-matrix (mapcar #'cdr matrix-rest) (1+ pos)
				 (mapcar #'(lambda (row compressed-row)
					     (cons (first row) compressed-row))
					 matrix-rest compressed-matrix)
				 (cons pos pos-vector)))
	(t
	 (dioadd=compress-matrix (mapcar #'cdr matrix-rest) (1+ pos)
				 compressed-matrix
				 pos-vector))))

(defun dioadd=expand-solution (solution pos-vector n)
  (let ((result (make-sequence 'simple-vector n :initial-element 0)))
    (do ((i 0 (1+ i))
	 (pos-rest pos-vector (rest pos-rest)))
	((null pos-rest))
      (setf (elt result (first pos-rest)) (elt solution i)))
    result))
  


(defun dioadd=matrix-to-bitvector (matrix)
  (mapcar #'(lambda (row)
	      (map 'simple-bit-vector #'(lambda (x) (abs (signum x))) row))
	  matrix))

(defun dioadd=split-matrix (matrix)
  (let* ((vectors (dioadd=matrix-to-bitvector matrix))
	 (test-vector (make-array (length (first matrix)) :element-type 'bit :initial-element 0)))
    (dioadd=split (first vectors)
		  (list (first matrix))
		  (cdr vectors)
		  (copy-list (cdr matrix))
		  nil
		  test-vector)))

(defun dioadd=split (res-vector res-matrix vectors-rest matrix-rest splitted-matrix test-vector)
  (if (null matrix-rest)
      (cons res-matrix splitted-matrix)
      (let* ((found-vector-row
	      (some #'(lambda (vector row)
			(if (not (dioadd=vector-zero-p
				  (bit-and res-vector vector test-vector)))
			    (cons vector row)))
		    vectors-rest matrix-rest))
	     (found-vector (car found-vector-row))
	     (found-row (cdr found-vector-row)))
;	(format t "~%res-m: ~A ~% m-rest: ~A ~% split: ~A~% ~A:  ~A ~A~% "
;		res-matrix matrix-rest splitted-matrix
;		(map 'list #'identity res-vector)
;		(map 'list #'identity found-vector) found-row)
	(if found-vector
	    (dioadd=split (bit-ior res-vector found-vector t)
			  (cons found-row res-matrix)
			  (delete found-vector vectors-rest :count 1)
			  (delete found-row matrix-rest :count 1)
			  splitted-matrix
			  test-vector)
	    (dioadd=split (first vectors-rest)
			  (list (first matrix-rest))
			  (cdr vectors-rest)
			  (cdr matrix-rest)
			  (cons res-matrix splitted-matrix)
			  test-vector)))))

			 

(defun dioadd=vector-zero-p (vector)
  (every #'zerop vector))



;;;;; from cpap.lisp

(defun cpac=new-solutions! (up)
  (declare (special cpac*theory)
           (edited  "21-AUG-1995")
	   (authors Neskakis)
	   (input   "unification-problem [cpac+up]")
	   (effect  "None")
	   (value   "a list, first element: a list of the occuring variables and constants"
		    "        rest: lists representing solution-vectors of the diophantine equations"))
  (let* ((argl-1    (mapcar #'cpac=term-arglist (car (unicp~equations up))))
	 (argl-2    (mapcar #'cpac=term-arglist (cadr (unicp~equations up))))
	 (argl-3    (cpac=args argl-1 argl-2))
	 (args      (progn (cpac=clear-positions! (unicp~variables up))
			   argl-3))
	 (dioeq     (cpac=build-diophant-equations argl-1 argl-2 args))
	 (matrix    (cpac=redundant-rows&vectors (cons args dioeq) (length args)))
	 (newdioeq  (cdr matrix))
	 (newargs   (cpac=set-positions! (car matrix)))
	 (constrows (cpac=constant-rows newargs))
	 (vectors   (dioadd~solve-equations newdioeq NIL))) ;constrows)))
    (setf (cpac=solutions-p up) t
          (cpac=solutions up) vectors
	  (cpac=arguments up) newargs
	  (cpac=permanent-solutions up) NIL)))





(defun dioadd=split-eqs (vars res-eqs eqs-rest splitted-eqs)
  (if (null eqs-rest)
      (cons res-eqs splitted-eqs)
      (let* ((found-eq
	      (some #'(lambda (equ)
			(if (intersection vars (top~free-variables equ) :test #'unicp~vars-identified-p)
			    equ))
		    eqs-rest)))
	(if found-eq
	    (dioadd=split-eqs (union vars (top~free-variables found-eq) :test #'unicp~vars-identified-p)
			      (cons found-eq res-eqs)
			      (delete found-eq eqs-rest :count 1)
			      splitted-eqs)
	    (dioadd=split-eqs (top~free-variables (first eqs-rest))
			      (list (first eqs-rest))
			      (cdr eqs-rest)
			      (cons res-eqs splitted-eqs))))))

(defun dioadd=split-equations (eqs)
  (dioadd=split-eqs (top~free-variables (first eqs))
		    (list (first eqs))
		    (cdr eqs)
		    nil))
