(in-package "KEIM")

(defvar prv*clause-counter 0)

(defun prv~prove (problem)
  (setf prv*clause-counter 0)
  (mapc #'prv=make-step! (proof~steps problem)))

(defun prv=make-step! (step)
  (cond ((res~paramodulation-p step)
	 (prv=make-paramod! step))
	((res~resolution-p step)
	 (prv=make-res! step))
	(t (error "~A is not a known step" step))))

(defun prv=make-paramod! (step)
  (setf term*variable-counter 0)
  (incf prv*clause-counter)
  (multiple-value-bind (mother-unifier-list old-vars new-vars)
      (top~rename (list (res~paramod-mother step)
			(res~just-unifier (res~paramod-mother step))))
    (let* ((mother (first mother-unifier-list))
	   (mother-pos (res~paramod-mother-position step))
	   (mother-unifier (second mother-unifier-list))
	   (father (res~paramod-father step))
	   (father-pos (res~paramod-father-position step))
	   (father-unifier (res~just-unifier father)))
      (multiple-value-bind (left right)
	  (let ((args (fo~arguments (lit~atom (term~at-position father father-pos)))))
	    (cond ((eq (res~paramod-direction step) 'LR)
		   (values (first args) (second args)))
		  ((eq (res~paramod-direction step) 'RL)
		   (values (second args) (first args)))
		  (t (error "Undefined direction ~A in step ~A"
			    (res~paramod-direction step) step))))
	(prv=update-step!
	 step
	 (top~replace-at-position mother mother-pos right)
	 (uni~up-multi-create
	  (append (if mother-unifier (uni~initial-problem mother-unifier))
		  (if father-unifier (uni~initial-problem father-unifier))
		  (list (list (term~at-position mother mother-pos) left))))
	 (subst~create old-vars new-vars))))))

(defun prv=make-res! (step)
  (setf term*variable-counter 0)
  (incf prv*clause-counter)
  (multiple-value-bind (clause1-unifier-list old-vars new-vars)
      (let ((old-clause1 (first (res~resolution-clauses step))))
	(top~rename (list old-clause1 (res~just-unifier old-clause1))))
    (let* ((clause1 (first clause1-unifier-list))
	   (clause2 (second (res~resolution-clauses step)))
	   (clause1-lit (term~at-position clause1 (first  (res~resolution-positions step))))
	   (clause2-lit (term~at-position clause2 (second (res~resolution-positions step))))
	   (clause1-args (fo~arguments (lit~atom clause1-lit)))
	   (clause2-args (fo~arguments (lit~atom clause2-lit)))
	   (clause1-unifier (second clause1-unifier-list))
	   (clause2-unifier (res~just-unifier clause2)))
      (when (eq (lit~positive-p clause1-lit)
		(lit~positive-p clause2-lit))
	(error "Resolution not possible; literals ~A and ~A have same sign."
	       clause1-lit clause2-lit))
      (prv=update-step!
       step
       (cl~create (append (remove clause1-lit (cl~literals clause1))
			  (remove clause2-lit (cl~literals clause2))))
       (uni~up-multi-create
	(append (if clause1-unifier (uni~initial-problem clause1-unifier))
		(if clause2-unifier (uni~initial-problem clause2-unifier))
		(mapcar #'list clause1-args clause2-args)))
       (subst~create old-vars new-vars)))))

(defun prv=update-step! (step new-clause constraints renaming)
  (cl~set-literals! step (cl~literals new-clause))
  (res=justification-write-unifier! constraints (node~justification step))
  (res=justification-write-renaming! renaming (node~justification step)))



(defmethod sym~variable-create ((symbol sym+sym) (type type+type) &optional refsym)
      (declare (ignore refsym))
      (let ((shsym 
	     (make-instance 'shsym+sharedsymbol
			    :name (prv~new-var-name)
			    :type type)))
	(make-instance 'sym+var :symbol shsym :type type)))

(defun prv~new-var-name ()
  ;(format nil "x~A-~A" prv*clause-counter (incf term*variable-counter))
  (format nil "x-~A" (incf fo*variable-counter))
  )
  

(defun prv~extract-up (step &optional (stream t))
  (setf fo*variable-counter 0)
  (let* ((up (top~rename (res~just-unifier step) :return-renaming nil))
	 (vars (top~free-variables up)))
    (format stream "~%(setf prv*env (env~~create))")
    (format stream "~%(post~~read-object-list '((type-constants i)")
    (format stream "~%                         (constants (n (i i)))")
    (format stream "~%                         (variables ")
    (post~print-declaration vars stream)
    (format stream ")~%                         (signatures (ac (+ (i i i))) (ac (* (i i i))))")
    (format stream "~%                         ) prv*env)~%")
    (format stream "~%(prv==unif ~12T(list ~{~18T(term~~read '~A prv*env)~^~%~})"
	    (mapcar #'first (uni~initial-problem up)))
    (format stream "~%~12T(list ~{~18T(term~~read '~A prv*env)~^~%~}))"
	    (mapcar #'second (uni~initial-problem up)))
    ))

(defmacro prv==unif (list1 list2)
  `(cpcb~decide-unify ,list1 ,list2))



;;;;;;;;;;;;;;;;;
