(in-package "KEIM")
(defvar cpa*debug nil)
(defvar *cpa-th* nil)
(defvar cpa*changes nil)

;;; +---------------------------------------------------------------------------
;;;  cpa+up (+ funs operating on cpa+up)
;;; +---------------------------------------------------------------------------

(defclass cpa+up (unicp+elementary-up)
      ((a-words-left  :initform NIL :accessor cpa=words-left)
       (a-words-right :initform NIL :accessor cpa=words-right)
       (stack         :initform NIL :accessor cpa=stack))
      (:documentation "This is the class of A Unifikation."))


(defun cpa=create-up (theory general-up equations)
  (declare (special cpa*theory))
  (let*
      ((up (make-instance 'cpa+up
			  :general-up general-up
			  :equations equations
			  :theory theory))
       (terms-left  (unicp~eq-lhss (unicp~equations up)))
       (terms-right (unicp~eq-rhss (unicp~equations up))))
      ;;; (if cpa*debug (cpa=print-equ terms-left terms-right))
      (setf (cpa=words-left up)  (mapcar #'cpa=term-to-word terms-left))
      (setf (cpa=words-right up) (mapcar #'cpa=term-to-word terms-right))
      up))

(defun cpa=term-to-word (term)
  (declare (edited  "16-JAN-1997")
	   (authors Doelfer)
	   (input   "A keim term")
	   (effect  "transforms term in word over A")
	   (value   "the word representing term"))
  (if (appl~p term)
      (mapcan #'cpa=term-to-word (appl~arguments term))
      (list term)))

(defun cpa=equal-words-p (word1 word2)
  (declare (edited  "28-FEB-1997")
	   (authors Doelfer)
	   (input   "word1, word2 - terms wich have been transformed into"
                    "               words via cpa=term-to-word")
	   (effect  "none")
	   (value   "true if word1 is equal to word2"))
  (cond ((and (null word1) (null word2)) t)
        ((or  (null word1) (null word2)) nil)
        (t (if (term~equal (first word1) (first word2))
	       (cpa=equal-words-p (rest word1) (rest word2))
	     nil))))

;;; +-----------------------------------------------------------------------
;;;  unicp~theo-init, unicp~new-constr
;;; +------------------------------------------------------------------------

(defmethod unicp~theo-init ((theory th+a) general-up equations
			    identified discriminated constants const-restr)
  (declare (edited  "15-JAN-1997")
	   (authors Doelfer)
	   (input   "theory"
                    "general-up"
                    "equations"
                    "%"
                    "%"
                    "%")
	   (effect  "")
	   (value   "The cpa+up object is created (therefore the equations are"
                    "transformed into words over A). Then the algorithm for"
                    "associative unification is called."
                    "It returns NIL if the unification-problem is not solveable."
		    "Otherwise a multiple value is returned."
                    "See the documentation for unicp~new-constr for more!!"))

  (let ((cpa*theory theory))
    (declare (special cpa*theory))
    (setq *cpa-th* cpa*theory)
    (let* ((up (cpa=create-up theory general-up equations)))
      ;;; (if cpa*debug (format t "Unicp Theo-Init ~%=============== ~%"))
      (unicp~new-constr up identified discriminated constants const-restr))))


(defmethod unicp~new-constr ((up cpa+up) identified discriminated constants const-restr)
  (declare (edited  "15-JAN-1997")
	   (authors Doelfer)
	   (input   "unification problem (cpa+up)"
                    "%"
		    "%"
		    "%"
		    "%")
	   (effect  "")
	   (value   "The elementary algorithm for the theory ELEMENTARY-UP"
		    "belongs to is called. It returns NIL if the unification"
		    "problem is not solvable with these new restrictions."
		    "Otherwise a multiple value is returned that contains"
		    "some new constraints wich must be met if the unification"
		    "problem should be solvable:"
		    "1.*UP, indicating that the unification problem is still"
		    "   solvable."
		    "2.*A list of new identification of variables (i.e. sets"
		    "   of variables)."
		    "3.*A list of pairs of variables which must not be"
		    "   unified."
		    "4.*A list of variables which must not be constants."
		    "5. A list of variables which need not to be variables."
		    "6.*A list of const-rest $(x y_1 ... y_n)$ where"
		    "   $\sigma(y_i)$ is a subterm of $\sigma(x)$ for a"
		    "   unifier $\sigma$."
		    "7. (A list of pairs $(x . y)$ of variables where no"
		    "   solution is lost if $y$ is in the constant"
		    "   restrictions of $x$)."))
 (let
      ((cpa*theory (unicp~theory up)))  
   (declare (special cpa*theory))
   (if cpa*debug (format t "Unicp Find new constraints ~%========================== ~%"))
   (multiple-value-bind (new-up new-ident new-discr new-ass new-crest)
       (cpa=find-new-constraints up identified discriminated
				 nil constants const-restr)  
     (if (null cpa*changes)
	 new-up  ;;; nothing has changed -> return the input
	 (progn
	   (setq cpa*changes nil)  ;;; reset flag	   
	   (values new-up new-ident new-discr new-ass nil new-crest nil))))))


;;;+--------------------------------------------------------------------------
;;;  save/restore-choice-point                                                
;;;+--------------------------------------------------------------------------

(defmethod unicp~up-save-choice-point ((up cpa+up))
  (let ((konfiguration (cpa=make-config (cpa=words-left up) (cpa=words-right up)))) 
    (push konfiguration (cpa=stack up))
    up))

(defmethod unicp~up-restore-choice-point ((up cpa+up))
  (let ((konfiguration (pop (cpa=stack up))))  
    (setf (cpa=words-left  up) (first  konfiguration))
    (setf (cpa=words-right up) (second konfiguration))
    up))

(defun cpa=make-config (words-left words-right)
  (list words-left words-right))

;;; +-----------------------------------------------------------------------------------
;;;   unicp-solvable-p (calls the old a.lisp routine)
;;; +-----------------------------------------------------------------------------------

(defmethod unicp~solvable-p ((up cpa+up))
  (let* ((wl   (mapcar #'cpa=identify-word (cpa=words-left up)))
	 (wr   (mapcar #'cpa=identify-word (cpa=words-right up)))
         (vars (top~free-variables (unicp~equations up)))
         (vars-as-consts (cpa=return-consts vars nil))
	 (restr          (cpa=return-rest vars-as-consts nil)))
    (if cpa*debug (cpa=print-solveable wl wr (unicp~equations up) vars-as-consts restr
				       (a=decide-unify wl wr vars-as-consts restr NIL)))
    (a=decide-unify wl wr vars-as-consts restr NIL)))

(defun cpa=identify-word (w)
  (declare (edited  "15-MAY-1997")
	   (authors Doelfer)
	   (input   "A word over a")
	   (effect  "None")
	   (value   "Each character of w (e.g. a variable) is substitued by its repr."))
  (mapcar #'unicp~var-label-repres w))

(defun cpa=return-consts (v akku)
  (declare (edited  "12-MAY-1997")
	   (authors Doelfer)
	   (input   "A Theory th, A list of vars v")
	   (effect  "None")
	   (value   "All elements of v wich should be treated as consts"))
  (if (null v)
      akku
      (if (unicp~constant-p (first v) *cpa-th*)
	  (cpa=return-consts (rest v) (cons (first v) akku))
	  (cpa=return-consts (rest v) akku))))

(defun cpa=return-rest (v akku)
  (declare (edited  "12-MAY-1997")
	   (authors Doelfer)
	   (input   "A list of vars v")
	   (effect  "None")
	   (value   "All restrictions of the variables in v"))
  (declare (special rest-list))
  (if cpa*debug (cpa=print-rest-list v akku rest-list))
  (if (null v)
      akku
      (let ((rest-list (unicp~var-label-less (first v))))
	(if (null rest-list)
	    (cpa=return-rest (rest v) akku)
	    (cpa=return-rest (rest v) (cons (cons (first v) rest-list) akku))))))


;;; +------------------------------------------------------------------------------------
;;;   Der "Main"-Algorithmus (cpa=find-new-constraints)
;;; +------------------------------------------------------------------------------------

(defun cpa=find-new-constraints (up ident discr ass const con-rest)
  (declare (edited  "26-FEB-1997")
	   (authors Doelfer)
	   (input   "rule - the rule wich should be applied to the up,"
                    "up - the unification problem up,"
                    "identified - list of identfication objects"
		    "discriminated - list of pairs of variables wich should be discriminated"
		    "ass - list of vars wich are associative fun-symbols"
		    "constants - list of vars wich should be treated as constant symbols"
		    "const-restr - list of constant-restriction")
	   (effect  "none")
	   (value  "Multiple Value: "
		   "(new) up"
		   "(new) identified"
		   "(new) discriminated"
		   "(new) ass (variables)"
		   "(new) const-restr"))
  (let
      ((cpa*theory (unicp~theory up)))
    (declare (special cpa*theory) (special new-up) (special flag) (special return-wert))
    (if cpa*debug (cpa=print-new-constraints-dbg up ident discr ass const con-rest))
    (cond
      ((cpa=rule-p 1 up ident discr const con-rest)
       (multiple-value-bind (new-up  new-id)
	   (cpa=apply-rule 1 up ident discr const con-rest)
	 (let* ((new-ident (cons new-id ident)))
	   (setq cpa*changes t) 
	   (cpa=find-new-constraints new-up new-ident discr ass
				     const con-rest))))
      
      ((cpa=rule-p 2 up ident discr const con-rest)
       (multiple-value-bind (new-up new-a new-cr new-dis) 
	   (cpa=apply-rule 2 up ident discr const con-rest)
	 (let* ((new-ass   (cons (first new-a) ass))
		(new-crest (cons new-cr con-rest))
		(new-disc  (cons (first new-dis) discr)))
	   (setq cpa*changes t)
	   (cpa=find-new-constraints new-up ident new-disc new-ass const
				     new-crest))))
      
      ((cpa=rule-p 3 up ident discr const con-rest)
       (multiple-value-bind (new-up new-id)
	   (cpa=apply-rule 3 up ident discr const con-rest)
	 (let* ((new-ident (cons new-id ident)))
	   (setq cpa*changes t)
	   (cpa=find-new-constraints new-up new-ident discr ass
				     const con-rest))))
      
      ((cpa=rule-p 4 up ident discr const con-rest)
       (multiple-value-bind (new-up new-a new-cr) 
	   (cpa=apply-rule 4 up ident discr const con-rest)
	 (let* ((new-ass   (cons (first new-a) ass))
		(new-crest (cons new-cr con-rest)))
	   (setq cpa*changes t)
	   (cpa=find-new-constraints new-up ident discr new-ass
				     const new-crest))))
      
      ((cpa=rule-p 5 up ident discr const con-rest)
       (let* ((return-wert (cpa=apply-rule 5 up ident discr const con-rest))
	      (new-crest (cons return-wert con-rest)))
	 (setq cpa*changes t)
	 (cpa=find-new-constraints up ident discr ass
				   const new-crest)))
      
      (t
       (if cpa*debug
	   (format *terminal-io*
		   "cpa=find-new-constraints :: returning cr ~S ~%" con-rest))
       (values up ident discr ass con-rest))
	  ;;; if no rule could be applied, then the (old) constraints are given back
      )))




	
;;; +-----------------------------------------------------------------------
;;;   cpa=rule-p (checking, if a specified rule could be applied to the up)
;;; +-----------------------------------------------------------------------

(defun cpa=rule-p (rule up ident discr const const-rest)

  (declare (edited  "10-MAR-1997")
	   (authors Doelfer)
	   (input   "rule - Integer in {1..5} indicating, wich rule should be tested"
		    "up   - a associative unification problem"
		    "ident, discr, const, const-rest - the well-known constraints")
	   (effect  "if the specified rule is applicable on the up, then the corresponding"
		    "equation is made the first equation of the problem")
	   (value   "nil - the rule rule is not applicable"
		    "up  - the modified (see effect) up on wich the rule is applicable"))

  
  (let* ((words-left  (cpa=words-left up))
	 (words-right (cpa=words-right up))
	 (result      (cpa=find-case rule words-left words-right ident discr const const-rest)))
    (if cpa*debug (format t "   Checking Rule ~A : ~S ~%" rule result))
    (if (not result)
	nil
	(let ((wl (cpa=remove-list words-left  (first  result)))
	      (wr (cpa=remove-list words-right (second result))))
	  (setf (cpa=words-left  up) (cons (first  result) wl))
	  (setf (cpa=words-right up) (cons (second result) wr))
	  up))))

(defun cpa=find-case (rule words-left words-right ident discr const const-rest)

  (declare (edited  "12-MAR-1997")
	   (authors Doelfer)
	   (input   "rule - integer indicating wich rule should be checked"
		    "words-left, words-right - equations"
		    "ident, discr, const, const-rest - constraints")
	   (effect  "None")
	   (value   "(lhs rhs) - equation on wich rule rule could be applied"
		    "nil       - there is no equation on wich the rule could be applied"))  

  (if (null words-left)
      nil
      (let ((result (cpa=check-for-case-p rule (first words-left) (first words-right)
					  ident discr const const-rest)))
	(if result
	    (list (first words-left) (first words-right))
	    (cpa=find-case rule (cdr words-left) (cdr words-right)
			   ident discr const const-rest)))))


(defun cpa=check-for-case-p (rule word-left word-right ident discr const const-rest)

  (declare (edited  "12-MAR-1997")
	   (authors Doelfer)
	   (input   "rule - integer indicating wich rule should be checked"
		    "word-left, word-right - a equation"
		    "ident, discr, const, const-rest - constraints")
	   (effect  "None")
	   (value   "t - if the rule rule is applicable to the equation"
		    "nil - if the rule rule is NOT applicable to the equation"))

  (cond ((= rule 1)
	 (and (= (length word-left) 1) (= (length word-right) 1)))
	((= rule 2)
	 (or (and (= (length word-left) 1)  (> (length word-right) 1))
	     (and (= (length word-right) 1) (> (length word-left)  1))))
	((= rule 3)
	 (let* ((x1 (first word-left))  (x2 (first (last word-left)))
		(y1 (first word-right)) (y2 (first (last word-right))))
	   (or (or (cpa=are-identified-p x1 y1 ident) (and (cpa=is-constant-p x1 const)
					   		   (cpa=is-constant-p y1 const)))
	       (or (cpa=are-identified-p x2 y2 ident) (and (cpa=is-constant-p x2 const)
							   (cpa=is-constant-p y2 const))))))
	((= rule 4)
	 (let* ((x1 (first word-left))  (x2 (first (last word-left)))
		(y1 (first word-right)) (y2 (first (last word-right))))
	   (or (and (cpa=are-discr-p x1 y1 discr) (or (cpa=is-constant-p y1 const)
					              (cpa=are-rest-p y1 x1 const-rest)))
               (and (cpa=are-discr-p x1 y1 discr) (or (cpa=is-constant-p x1 const)
					              (cpa=are-rest-p x1 y1 const-rest)))
	       (and (cpa=are-discr-p x2 y2 discr) (or (cpa=is-constant-p y2 const)
					              (cpa=are-rest-p y2 x2 const-rest)))
               (and (cpa=are-discr-p x2 y2 discr) (or (cpa=is-constant-p x2 const)
					              (cpa=are-rest-p x2 y2 const-rest))))))
	((= rule 5)
	 (let* ((y1 (first word-right)) (y2 (first (last word-right)))
		(x1 (first word-left))  (x2 (first (last word-left))))
	   (or (and (cpa=is-constant-p y1 const) (not (cpa=are-rest-p x1 y1 const-rest)))
               (and (cpa=is-constant-p x1 const) (not (cpa=are-rest-p y1 x1 const-rest)))
	       (and (cpa=is-constant-p y2 const) (not (cpa=are-rest-p x2 y2 const-rest)))
	       (and (cpa=is-constant-p x2 const) (not (cpa=are-rest-p y2 x2 const-rest))))))))




;;; +-------------------------------------------------------
;;;   cpa=apply-rule (applys a specified rule on the up)
;;; +-------------------------------------------------------

(defun cpa=apply-rule (rule up ident discr const const-rest)

  (declare (edited  "01-DEC-1997")
	   (authors Doelfer)
	   (input   "rule - number of a rule (1..5)"
		    "up   - a unification problem"
		    "ident, discr, const, const-rest - the usual constraints on up")
	   (effect  "None")
	   (value   ""))
  
  (let ((words-left  (cpa=words-left  up))
	(words-right (cpa=words-right up)))

    (cond ((= rule 1)
	   (let ((x  (first (first words-left))) (y (first (first words-right)))
		 (wl (rest words-left)) (wr (rest words-right)))
	     (setf (cpa=words-left up)
		   (mapcar #'(lambda (z) (substitute x y z :test #'term~equal)) 
		           wl))
             (setf (cpa=words-right up)
		   (mapcar #'(lambda (z) (substitute x y z :test #'term~equal))
			   wr))
             (values up (unicp~id-create x y))))

	  ((= rule 2)
	   (let* ((fall-1 (= 1 (length (first words-left))))
		  (x  (if fall-1 (first (first words-left)) (first (first words-right))))
		  (y  (if fall-1 (first words-right) (first words-left)))
		  (wl (if fall-1 (rest words-left) (rest words-right)))
		  (wr (if fall-1 (rest words-right) (rest words-left))))
	     (setf (cpa=words-left up)
		   (mapcar #'(lambda (z) (cpa=replace-word-for-char x y z)) wl))
	     (setf (cpa=words-right up)
		   (mapcar #'(lambda (z) (cpa=replace-word-for-char x y z)) wr))
	     (values up                        ;;; modified Unification problem
		     (list x)                  ;;; Constraint: x in A
		     (cpa=make-const-rest y x) ;;; Constraint: w <= x
		     (cpa=make-discr y x))))   ;;; Constraint: w != x

          ((= rule 3)
	   (multiple-value-bind (x x1 v w)
	       (cpa=wich-case-3 (first words-left) (first words-right) ident const)
	     (let ((y  (list x1))
		   (wl (cons w (rest words-left)))
		   (wr (cons v (rest words-right))))
	       (setf (cpa=words-left up)
		     (mapcar #'(lambda (z) (cpa=replace-word-for-char x y z)) wl))
	       (setf (cpa=words-right up)
		     (mapcar #'(lambda (z) (cpa=replace-word-for-char x y z)) wr))
	       (values up (unicp~id-create x x1)))))

	 ((= rule 4)
	  (let* ((var (fo~variable-create)))
	    (multiple-value-bind (x y x1 x2 r)
		(cpa=wich-case-4 (first words-left) (first words-right)
				 var discr const const-rest)
	      (setf (cpa=words-left up)
		    (mapcar #'(lambda (z) (cpa=replace-word-for-char x x1 z))
			    (cons x2 (rest words-left)))
		    (cpa=words-right up)
		    (mapcar #'(lambda (z) (cpa=replace-word-for-char x x1 z))
			    (cons r (rest words-right))))
	      (values up (list x) (cpa=make-const-rest y x))
;;;              (let* ((v  (first r))
;;;                     (wl (cons x2 (rest words-left)))
;;;                     (wr (cons v  (rest words-right))))
;;;                (format t "l: ~A r: ~A~%" (cpa=words-left up) (cpa=words-right up))
;;;                (setf (cpa=words-left up)
;;;                      (mapcar #'(lambda (z) (cpa=replace-word-for-char x x1 z)) wl))
;;;                (setf (cpa=words-right up)
;;;                      (mapcar #'(lambda (z) (cpa=replace-word-for-char x x1 z)) wr))
;;;                (values up (list x) (cpa=make-const-rest y x)))
	      )))
	 
	 ((= rule 5)
	  (multiple-value-bind (y x)
	      (cpa=wich-case-5 (first words-left) (first words-right) const const-rest)
	    (unicp~cr-create x y))))))
	 

;; 
;; Die Regeln 3,4 und 5 unterteilen sich jeweils in 2 Faelle. Diese Fallunterscheidung
;; wird von den Routinen cpa=wich-case-X gemanaged.
;; 


(defun cpa=wich-case-3 (wl wr ident const)
  (let ((xa (first wl))
	(xb (first (last wl)))
	(ya (first wr))
	(yb (first (last wr)))
	(wa (rest wl))
	(wb (butlast wl))
	(va (rest wr))
	(vb (butlast wr)))
    (if (or (cpa=are-identified-p xa ya ident) (and (cpa=is-constant-p xa const)
						    (cpa=is-constant-p ya const)))
       (values xa ya wa va)
       (values xb yb wb vb))))

(defun cpa=wich-case-4 (wl wr var disc const const-rest)
  (let* ((x1 (first wl))
	 (x2 (first (last wl)))
	 (y1 (first wr))
	 (y2 (first (last wr)))
	 (w1 (cdr wl))
	 (w2 (butlast wl))
	 (v1 (cdr wr))
	 (v2 (butlast wr)))
    (cond ((and (cpa=are-discr-p x1 y1 disc) (or (cpa=is-constant-p y1 const)
					         (cpa=are-rest-p y1 x1 const-rest)))
	   (values x1 y1 (list y1 var) (append (list var) w1) v1))
	  ((and (cpa=are-discr-p x1 y1 disc) (or (cpa=is-constant-p x1 const)
					         (cpa=are-rest-p x1 y1 const-rest)))
      	   (values y1 x1 (list x1 var) (append (list var) v1) w1))
	  ((and (cpa=are-discr-p x2 y2 disc) (or (cpa=is-constant-p y2 const)
					         (cpa=are-rest-p y2 x2 const-rest)))
           (values x2 y2 (list var y2) (append w2 (list var)) v2))
	  ((and (cpa=are-discr-p x2 y2 disc) (or (cpa=is-constant-p x2 const)
					         (cpa=are-rest-p x2 y2 const-rest)))
	   (values y2 x2 (list var x2) (append v2 (list var)) w2)))))

(defun cpa=wich-case-5 (wl wr const crest)
 (let* ((x1 (first wl))
	(x2 (first (last wl)))
	(y1 (first wr))
	(y2 (first (last wr))))
  (cond ((and (cpa=is-constant-p y1 const) (not (cpa=are-rest-p x1 y1 crest))) (values y1 x1))
        ((and (cpa=is-constant-p x1 const) (not (cpa=are-rest-p y1 x1 crest))) (values x1 y1))
	((and (cpa=is-constant-p y2 const) (not (cpa=are-rest-p x2 y2 crest))) (values y2 x2))
	(t (values x2 y2)))))



;;; +-----------------------------------------------------------------------------------
;;;   Hilfsfunktionen
;;; +-----------------------------------------------------------------------------------

(defun cpa=remove-list (liste list-to-be-removed)
  (declare (edited  "20-FEB-1997" "18-SEP-1998")
	   (authors Doelfer Kepser)
	   (input   "liste - list of sublists"
		    "list-to-be-removed - a flat list")
	   (effect  "None")
	   (value   "The list build from deleting list-to-be-removed in liste"))
  (remove list-to-be-removed liste :test #'cpa=equal-words-p))

;; old Doelfer code
;;  (if (null liste)
;;      akku
;;      (if (cpa=equal-words-p (first liste) list-to-be-removed)
;;	  (append akku (rest liste))
;;	  (cpa=remove-list (rest liste)  list-to-be-removed (cons (first liste) akku)))))


(defun cpa=replace-word-for-char (char word l)
  (declare (edited  "10-FEB-1997" "18-SEP-1998")
           (authors Doelfer Kepser)
	   (input   "char - a keim term with length 1,"
		    "word - a list of char's (e.g. a word),"
		    "l    - a list of char's (e.g. a word),"
		    "akku - akkumulator (init with nil)")
	   (effect "None.")
	   (value  "Replaces every occurence of char in the list l"
		    "by the 'characters' of the word 'word' "))
  (a=subst-w char word l))

;;; Kepser: Funktion durch butlast ersetzt
;(defun cpa=almost-cdr (word akku)
;   (declare (edited  "05-MAR-1997")
;	   (authors Doelfer)
;	   (input   "a word over a")
;	   (effect  "None")
;	   (value   "all characters of word except for the last character"))
;   (if (rest word)
;       (cpa=almost-cdr (cdr word) (append akku (cons (first word) nil)))
;       akku))

	 

;;; +------------------------------------------------------------------------------------
;;;   Funktionen um mit den Constraints zu arbeiten
;;; +------------------------------------------------------------------------------------

(defun cpa=equal-ident-p (x y)
  (and (term~equal (unicp~id-left  x) (unicp~id-left  y))
       (term~equal (unicp~id-right x) (unicp~id-right y))))

(defun cpa=equal-conrest-p (x y)
  (if cpa*debug (format *terminal-io* "CPA=EQUAL-CONREST-P:: x - ~S y - ~S ~%" x y))
  (and (term~equal (unicp~cr-const x) (unicp~cr-const y))
       (cpa=equal-words-p (unicp~cr-varlist x) (unicp~cr-varlist y))))

(defun cpa=make-const-rest (w x)
  (declare (edited  "20-FEB-1997")
	   (authors Doelfer)
	   (input   "w - (w1 ... wn) is a word over A,"
                    "x - a 'character'")
	   (effect  "None")
	   (value   "The constraint w <= x"))
  (unicp~cr-create x w))

(defun cpa=make-discr (w x)
  (declare (edited  "04-MAR-1997")
	   (authors Doelfer)
	   (input   "a word w and a character x")
	   (effect  "None")
	   (value   "The constraint w != x"))
  (mapcar #'(lambda (y) (unicp~id-create y x)) w))

(defun cpa=are-identified-p (var1 var2 ident)
  (declare (edited  "04-MAR-1997")
	   (authors Doelfer)
	   (input   "Two variables var1 and var2, an identification constraint ident")
	   (effect  "none")
	   (value   "true, if var1=var2 is in the constraint ident"))
  (or
   (unicp~vars-identified-p var1 var2) 
   (some #'(lambda (x) (or (and (term~equal var1 (unicp~id-left x))
				(term~equal var2 (unicp~id-right x)))
			   (and (term~equal var2 (unicp~id-left x))
				(term~equal var1 (unicp~id-right x)))))
	 ident)))

(defun cpa=are-discr-p (var1 var2 discr)
  (or
   (member var2 (unicp~vars-discriminated var1))
   (some #'(lambda (x) (or (and (term~equal var1 (unicp~id-left x))
				(term~equal var2 (unicp~id-right x)))
			   (and (term~equal var2 (unicp~id-left x))
				(term~equal var1 (unicp~id-right x)))))
	 discr)))

(defun cpa=are-rest-p (var1 var2 const-rest)
  (or
   (unicp~vars-less-p var1 var2) 
   (some #'(lambda (x) (and (term~equal var1 (unicp~cr-const x))
			    (member var2 (unicp~cr-varlist x) :test #'term~equal)))
	 const-rest)))

(defun cpa=is-constant-p (var const)
   (declare (edited  "04-MAR-1997")
	    (authors Doelfer)
	    (input   "A variable var and a const constraint")
	    (effect  "none")
	    (value   "true, if var is a constant"))
   (or (unicp~constant-p var *cpa-th*)
       (member var const :test #'term~equal)))



;;; ---------------------------------
;;;   Code for debugging purpose
;;; ---------------------------------

(defun cpa=print-new-constraints-dbg (up ident discr ass const con-rest)
  (format *terminal-io* "cpa=find-new-constraints::~% ~
Gleichungen links: ~S ~% ~
Gleichungen rechts: ~S ~% ~
Identifications: ~S ~% ~
Discriminations: ~S ~% ~
Variables: ~S ~% ~
Constants: ~S ~% ~
Constant Restriction: ~S ~%" (cpa=words-left up) (cpa=words-right up)
ident discr ass const con-rest))

(defun cpa=print-new-rules-dbg (up ident discr const con-rest)
  (format *terminal-io* "cpa=more-rules-applicable::~% ~
Gleichungen links: ~S ~% ~
Gleichungen rechts: ~S ~% ~
Identifications: ~S ~% ~
Discriminations: ~S ~% ~
Constants: ~S ~% ~
Constant Restriction: ~S ~%" (cpa=words-left up) (cpa=words-right up)
ident discr const con-rest))

(defun cpa=print-state (up ident discr ass const con-rest)
  (format *terminal-io* "Aktueller Stand::~% ~
Gleichungen links: ~S ~% ~
Gleichungen rechts: ~S ~% ~
Identifications: ~S ~% ~
Discriminations: ~S ~% ~
ass. Vars: ~S ~% ~
Constants: ~S ~% ~
Constant Restriction: ~S ~%" (cpa=words-left up) (cpa=words-right up)
ident discr ass const con-rest))

(defun cpa=print-equ (left right)
  (format t "Creating a UP of: ~% ~
  Left:  ~S ~% ~
  Right: ~S ~%" left right))

(defun cpa=print-solveable (wl wr equ const rest result)
  (format *terminal-io* "A=DECIDE-UNIFY called with following Args: ~% ~
  Words-left:  ~S ~% ~
  Words-right: ~S ~% ~
  Equations: ~S ~% ~
  Constants:   ~S ~% ~
  Restr.       ~S ~% ~
  Answer is: ~S ~%" wl wr equ const rest result))

(defun cpa=print-rest-list (v a r)
  (format *terminal-io* "CPA=RETURN REST:: var - ~S akku - ~S Restr. von ~S - ~S ~%"                    v a (first v) r)) 

(defun cpa=print-new-constr (oi od ocr ni nd na ncr)
  (format *terminal-io* "CPA=new-constraints:: ~%
Old ident - ~% ~S ~%
New ident - ~% ~S ~%
Old discr - ~% ~S ~%
New discr - ~% ~S ~%
Old crest - ~% ~S ~%
New crest - ~% ~S ~%
Ass - ~% ~S ~%"
	  oi ni od nd ocr ncr na))
