
(in-package "KEIM")

;;; basis=give-fsym
;;;
;;; diese Funktion liefert das erste gefundene Funktionssymbol; da der Unifikationsalgorithmus
;;; stark vereinfacht wurde, wird angenommen, dass in den Eingabe- Termen nur
;;; ein einziges AC-Fkt-Symbol vorkommt

(defun basis=give-fsym (tl1 tl2)
  (declare (edited  "13-DEC-1994")
	   (authors Neskakis)
	   (input   "two termlists")
	   (effect  "None")
	   (value   "the first function-symbol found in one of the lists or NIL"))
  (if (endp tl1)
      (if (endp tl2)
	  NIL
	  (basis=give-fsym tl2 NIL))
      (if (appl~p (car tl1))
	  (appl~function (car tl1))
	  (basis=give-fsym (cdr tl1) tl2))))



;;; basis=tl-to-argl
;;;
;;; da die Terme (nach Voraussetzung) nur aus Variablen, Konstanten und einem 2-stelligem
;;; AC-Fkt-Symbol "+" bestehen kann man sie als Liste ihrer Argumente wiedergeben .
;;; Bsp: (g a (g y b)) -> (a y b) 

(defun basis=tl-to-argl (termliste)
  (declare (edited  "29-MAR-1995")
	   (authors Neskakis)
	   (input   "a list of TERMS")
	   (effect  "None")
	   (value   "a list of lists, each list contains the arguments of the terms"
		    "example: ((g a x) (g (g a x) (g a y))) -> ((a x) (a x a y))   "))
  (mapcar #'basis=tl termliste))

(defun basis=tl (term)
  (if (appl~p term)
      (mapcan #'basis=tl (appl~arguments term))
      (list term)))


;;; basis=args
;;;
;;; diese Funktion liefert fuer die Listen der Argument-Termlisten eine Liste, in der
;;; die Argumente (Var&Konst) der Terme ohne Wiederholungen stehen (bsp s. u.)


(defun basis=args (argliste)
  (declare (edited  "28-JUN-1995")
	   (authors Neskakis)
	   (input   "a list of list of args")
	   (effect  "None")
	   (value   "a list of all args without duplicates"))
  (remove-duplicates
   (do ((i (1- (length argliste)) (1- i)) (newliste NIL (append (elt argliste i) newliste)))
       ((< i 0) newliste))
   :test #'term~equal))




;;; basis~labels-reset!


(defun basis~labels-reset! (Argliste)
  (mapc #'(lambda(a) (sym~set-label! a NIL)) Argliste))



;;; basis=solve-h
;;;
;;; diese Funktion ist die "Schnittstelle" zu basis=diol-matrix 
;;; aus den eingegebenen Argument-termlisten (t1...tn),(s1...sn) und der Argumentenliste
;;; baut sie zuerst die "Gleichungen" "t1-s1","t2-s2","t3-s3" [...] mit diol-matrix auf und
;;; loescht dann redundante NULL-Vektoren bzw NULL-Spalten
;;; Bsp fuer diol-matrix: t1: (g a x)  s1: (g a (g y a)) Argliste: (a x y) --> (a x y) (-1 1 -1)


#+alt(defun basis=solve-h (tl1 tl2 Args &optional (Allargs Args))
  (declare (edited  "07-JUL-1995")
	   (authors Neskakis)
	   (input   "2 list of Terms represented by their arguments"
		    "and the list of all arguments used in these terms")
	   (effect  "None")
	   (value   "a list of all input-vectors for >dioaed-solve.equations< "
		    "minus redundant vectors like NULL-vectors etc. "))
  (let ((matrix (basis=diol-matrix tl1 tl2 Args Allargs)))
    (progn
      (delete-if #'(lambda (vec)
		     (do ((i (1- (length vec)) (1- i)) (Sum 0 (+ (abs (elt Vec i)) Sum)))
			 ((< i 0) (= Sum 0))))
		 (cdr matrix))
      (do ((i (1- (length (car matrix))) (1- i)))
	  ((< i 0))
	  (do ((j (length (cdr matrix)) (1- j)) (Sum 0 (+ (abs (elt (elt matrix j) i)) Sum))) 
	      ((< j 1) (if (= Sum 0)
			   (mapc #'(lambda(vec) (delete-if #'atom vec :start i :end (1+ i)))
				 matrix)))))
      matrix)))

(defun basis=solve-h (tl1 tl2 Args &optional (Allargs Args))
  (declare (edited  "18-JUL-1995" "07-JUL-1995")
	   (authors Richts Neskakis)
	   (input   "2 list of Terms represented by their arguments"
		    "and the list of all arguments used in these terms")
	   (effect  "None")
	   (value   "a list of all input-vectors for >dioaed-solve.equations< "
		    "minus redundant vectors like NULL-vectors etc. "))
  (let* ((matrix (basis=diol-matrix tl1 tl2 Args Allargs))
	 (row-length (length (car matrix))))
    (setf (cdr matrix)
	  (delete-if #'(lambda (vec)
			 (do ((i (1- row-length) (1- i))
			      (Sum 0 (+ (abs (elt Vec i)) Sum)))
			     ((or (< i 0) (> sum 0)) (= Sum 0))))
		     (cdr matrix)))
    (let ((zero-columns
	   (do ((i 0 (1+ i))
		(rows (length (cdr matrix)))
		(zero-columns
		 nil
		 (if (do* ((j rows (1- j))
			   (Sum 0 (+ (abs (elt (elt (cdr matrix) j) i)) Sum))
			   #+weg(ddd (format t "~%~A ~A ~A" i j sum)
				     (format t "~%~A ~A ~A" i j sum)))
			 ((or (<= j 0) (> sum 0)) (= Sum 0)))
		     (cons i zero-columns)
		   zero-columns)))
	       ((>= i row-length) zero-columns))))
      ;(format t "~%~A" zero-columns)
      (mapc #'(lambda (zero-column)
		(mapl #'(lambda (matrix-tail)
			  (setf (car matrix-tail)
				(delete-if #'atom (car matrix-tail)
					   :start zero-column
					   :end (1+ zero-column) :count 1)))
		      matrix))
	    zero-columns)
      matrix)))




;;; diol=matrix2 NEU !
;;;
;;; Diese Funktion baut aus den Argumentenlisten der Termlisten die Eingabevektoren
;;; fuer die Funktion  dioaed-solve.equations  auf.
;;; d.h. sie erhaelt als eingabe (t1 t2 t3 ...) (s1 s2 s3 ...) , wobei ti und si
;;; wiederrum die Listen der Argumente sind . Nun arbeitet sie die jeweils
;;; korrespondierenden Listen tj und sj durch und zaehlt in beiden die Argumente
;;; durch; und zwar in der Reihenfolge in der sie in Args stehen
;;; Bsp: (x a b z y y) (a a b y) ergibt (Args: (x a b z y)) : (1 -1 0 1 1)
;;; und liefert die so erhaltenen Listen wieder zurueck

(defun basis=diol-matrix (tl1 tl2 Args &optional (Allargs Args))
  (declare (edited  "29-JUN-1995")
	   (authors Neskakis)
	   (input   "two lists of lists of arguments and a list of all arguments occuring")
	   (effect  "None")
	   (value   "a list of all input-vectors for >dioaed-solve.equations< "))
  (if (= (length tl1) (length tl2))
      (unwind-protect
	  (cons Args (mapcar #'(lambda(t1 t2) (basis=diol-count t1 t2 Args Allargs)) tl1 tl2))
	(basis~labels-reset! Allargs))
    (error "KEIM(diol) Termlisten ungleich lang")))

(defun basis=diol-count (t1 t2 Args &optional (Allargs Args))
  (progn
    (basis=diol-labels-reset! Allargs)
    (mapc #'(lambda(a) (sym~set-label! a (1+ (sym~label a)))) t1)
    (mapc #'(lambda(a) (sym~set-label! a (1- (sym~label a)))) t2)
    (mapcar #'sym~label Args)))
  

  
(defun basis=diol-labels-reset! (Args)
  (mapc #'(lambda (a) (sym~set-label! a 0)) Args))
		    



