(in-package "KEIM")

;;; ac~unify
;;;
;;; der "main"-algorithmus, diese Funktion sollte von "aussen" aufgerufen werden
;;; Eingabeparameter: - zwei Listen von Termen in POST-Syntax: (t1...tn) und (s1...sn)
;;;                     wobei t1==s1 t2==s2 usw unifiziert werden soll
;;;                   - Liste von Konstantenrestriktionen (optional)
;;;                     in folgendem Format: ((a x y z) (b x))
;;;                   - Liste von Variablen, die wie Konstanten behandelt werden
;;;                     sollen
;;;                   - evorvm - EntscheidungsVerfahrenORVollstaendigeMenge
;;;                     soll das Programm alle Substitutionen liefern -> NIL
;;;                     soll es nur entscheiden und eine Substitution liefern -> TRUE
;;; Ausgabe: Der Algorithmus sollte die vollstaendige Menge der Unifikatoren liefern

(defvar ac*one-unifier nil)

(defmethod th~unify ((th th+ac) tl1 tl2 restr const-vars)
  (ac~unify tl1 tl2 restr const-vars nil))

(defmethod th~decide-unify ((th th+ac) tl1 tl2 restr const-vars)
  (ac~unify tl1 tl2 restr const-vars t))

(defun ac~unify (tl1 tl2 &optional (restr NIL) (var_as_const NIL) (evorvm ac*one-unifier))
  (declare (edited  "28-JUN-1995")
	   (authors Neskakis)
	   (input   "two list of Terms which shall be ac-unified"
		    "optional: a list of constant restrictions"
		    "example: ((a z y) (b y))")
	   (effect  "None")
	   (value   "the correct Substitutions"))
  (let* ((fsym   (basis=give-fsym tl1 tl2))
	 (argl1  (basis=tl-to-argl tl1))            
	 (argl2  (basis=tl-to-argl tl2))            
	 (args   (basis=args (append argl1 argl2)))
	 (matrix (basis=solve-h argl1 argl2 args))) ; Vorsicht! Hier wird der sym-label-slot benutzt!
    (cond ((null fsym)
	   (error "ac~~unify : Kein AC-Symbol gefunden ! "))
	  ((null (car matrix)) (list (subst~create nil nil)))
	  (t
	   (unwind-protect
	       (progn
		 (ac=mark-var-as-const var_as_const) ;
		 (ac~Subs		;
		  fsym			;
		  (ac=make-vecs		;
		   fsym			;
		   (ac~hullot		;
		    evorvm		;
		    (ac=const-restr	; > sym-label-slot durch ac=mark-var-as-const belegt
		     restr		;  
		     (ac=erwvek		;
		      (ac=clear-before	;
		       (ac=solve matrix))))))))	;
	     (basis~labels-reset! var_as_const))))))
	




;;; ac=constant-p
;;;
;;; Ueberprueft ob das Symbol eine Konstante ist oder als Konstante behandelt werden soll

(defun ac=constant-p (object)
  (declare (edited  "09-MAY-1995")
	   (authors Neskakis)
	   (input   "an object")
	   (effect  "None")
	   (value   "True if object is a constant or is a variable treated as aconstant, else false"))
  (or (sym~constant-p object)
      (sym~label object)))




;;; ac=mark-var-as-const
;;;
;;; Markiert gewisse Variablen als Konstanten

(defun ac=mark-var-as-const (varliste)
  (declare (edited  "09-MAY-1995")
	   (authors Neskakis)
	   (input   "a list of variables which shall be treated as constants")
	   (effect  "the sym-label-slot of the variables is set to true")
	   (value   "None"))
  (mapc #'(lambda(v) (sym~set-label! v T)) varliste))




;;; ac=solve
;;;
;;; Diese Funktion erzeugt aus den eingegebenen Termlisten und der Argumentenliste eine
;;; Matrix der Form: Liste der Argumente + Loesungsvektoren aus dioaed-solve.equ
;;;
;;; Vorgehensweise:
;;;    - basis=solve-h wird aufgerufen und liefert zu den eingegebenen Termlisten (t1...tn),
;;;      (s1...sn) und der Argumentenliste die "Gleichungsvektoren" "t1-s1","t2-s2","t3-s3" [...]
;;;      NULL-Vektoren und NULL-Spalten (somit auch Eintragungen in der "Args"-Liste)
;;;      werden geloescht.
;;;    - mit diesen "Gleichungsvektoren" wird dioaed-solve.equ augerufen und liefert die
;;;      dazugehoerigen Loesungsvektoren.

(defun ac=solve (matrix)
  (declare (edited  "07-JUL-1995")
	   (authors Neskakis)
	   (input   "a list , 1st element: a list of the Variables and Constants "
		    "Rest: the matrix corresponding to the terms to be unified"
		    "       (a list of lists of numbers).")
	   (effect  "None")
	   (value   "a list, 1st element: All Variables used in the terms"
		    "        Rest: The Solution-Vectors of the "
		    "              corresponding diophantine matrix "))
  (let ((constant-rows
	 (do ((symbols (car matrix) (rest symbols))
	      (i 0 (1+ i))
	      (rows nil (if (ac=constant-p (first symbols))
			    (cons i rows)
			  rows)))
	     ((null symbols) rows))))
    (cons (car matrix)
	  (if (cdr matrix)
	      (dioaed-solve.equations (cdr matrix)
				      constant-rows)))))




;;; ac=clear-before
;;;
;;; Die Loesungsvektoren der diophantischen Gleichungen enthalten einige
;;; Vektoren , die nicht zu einem Unifikator fuehren koennen, diese entfernt
;;; diese Funktion vor der Potenzmengenmengen-bildung 

(defun ac=clear-before (matrix)
  (declare (edited  "23-JAN-1995")
	   (authors Neskakis)
	   (input   "a list , car:a list of the Variables and Constants "
		    "Rest: some Vectors (the solution of the diophant equ)")
	   (effect  "None")
	   (value   "a list , car:as above Rest:as above minus some"
		    "redundant vectors"))
  (setf (cdr matrix)
	(remove-if #'(lambda (vec) (let ((Eintrag 0))
				    (do ((i (1- (length (car matrix))) (1- i)))
				    ((< i 0) (> Eintrag 1))
				    (if (ac=constant-p (elt (car matrix)  i)) (incf Eintrag (elt Vec i))))))
		     (cdr matrix)))
  matrix)




;;; ac=erwvek
;;;
;;; Diese Funktion erweitert die Vektoren: entweder mit der Konstante auf die sie
;;; abgebildet werden (Bsp: Matrix: ((a x y) [1 1 0] ...) -> ((a x y) (a . [1 1 0]) ...))
;;; oder mit einer Variablen, die neue erzeugt wird, (Bsp: ((a x y) [0 1 1] ...)
;;; -> ((a x y) (x-1 . [0 1 1]) ...))


(defun ac=erwvek (matrix)
  (declare (edited  "13-DEC-1994")
	   (authors Neskakis)
	   (input   "the matrix : a list :"
		    "car: a list of the variables and constants"
		    "cdr: the solution-vectors of the diophantic equations")
	   (effect  "None")
	   (value   "a matrix : car unchanged "
		    "cdr: cons-ed in front of any vector ist either a variable or a"
		    "constant" ))
  (cons (car matrix) (mapcar #'(lambda(vec) (ac=erwvek-cv (car matrix) vec))
			     (cdr matrix))))

(defun ac=erwvek-cv (Var Vec &optional (Stelle 0))
  (if (= Stelle (length Var))
      (cons (fo~variable-create) Vec)
      (if (and (ac=constant-p (elt Var Stelle)) (= (elt Vec Stelle) 1))
	  (cons (elt Var Stelle) Vec)
	  (ac=erwvek-cv Var Vec (1+ Stelle)))))






;;; ac=const-restr
;;;
;;; man kann (optional) Konstantenrestriktionen angeben. Dadurch ueberfluessig werdende
;;; Loesungsvektoren werden hier geloescht

(defun ac=const-restr (crestr matrix)
  (declare (edited  "24-MAR-1995")
	   (authors Neskakis)
	   (input   "a list of constant-restrictions , and a matrix (args + vectors) ")
	   (effect  "None")
	   (value   "new matrix: without the restricted vectors" ))
  (if (endp crestr)
      matrix
      (cons (car matrix) (ac=const-restr-2 (car matrix) (cdr matrix) crestr))))



(defun ac=const-restr-2 (Var Vecs restr)
  (delete-if #'(lambda(vec) 
		 (let ((restr-var (cdar (member (car Vec) restr :test #'keim~equal :key #'car)))
		       (Eintrag NIL))
		   (if (null restr-var)
		       NIL
		     (do ((i (1- (length Var)) (1- i)))
			 ((< i 0) Eintrag)
			 (if (and (> (elt (cdr Vec) i) 0) (member (elt Var i) restr-var :test #'keim~equal))
			     (setf Eintrag T))))))
	     Vecs))
  







;;; ac~hullot ...
;;;
;;; Dieser Teil des Programms liefert die Loesungen 


(defun ac~hullot (evorvm matrix)
  (declare (edited  "06-JUL-1995")
	   (authors Neskakis)
	   (input   "matrix (to solve) and evorvm-Flag")
	   (effect  "None")
	   (value   "calls ac~hullot-p if evorm-Flag is TRUE , else ac~hullot-vm"))
  (if evorvm
      (ac~hullot-p matrix)
      (ac~hullot-vm matrix)))
      

(defun ac~hullot-p (matrix)
  (declare (edited  "06-JUL-1995")
	   (authors Neskakis)
	   (input   "a matrix (a list of two lists: 1.list of arguments 2.list of vectors)")
	   (effect  "None")
	   (value   "T if there exists a correct substitution"))
  (multiple-value-bind (argnums constnums)
		       (ac=make-number (car matrix) (cdr matrix))
		       (list
			(car matrix)
			(cdr matrix)
			(ac=hullot-h-p argnums constnums
				       (1- (expt 2 (length (cdr matrix))))))))


			

(defun ac=hullot-h-p (argnums constnums zahl &optional (ug 0) (og zahl) (lefttree t))
  (if (if lefttree
	  (ac=subset-big-enough argnums zahl)
	  (ac=subset-small-enough constnums zahl))
      (if (= 0 (- og ug))
	  (progn
	    ;(format t "~%Zahl: ~A" zahl)
	    (list zahl))
	  (let ((newnum (floor (+ ug og) 2)))
	       (or (ac=hullot-h-p argnums constnums newnum ug newnum t)
		   (ac=hullot-h-p argnums constnums (1+ newnum) (1+ newnum) og NIL))))
    NIL))


(defun ac~hullot-vm (matrix)
  (declare (edited  "29-JUN-1995")
	   (authors Neskakis)
	   (input   "a matrix (a list of two lists: 1.list of arguments 2.list of vectors")
	   (effect  "None")
	   (value   "a new matrix(3 lists: 1&2 as above + list of numbers which are"
		    "correct substitutions"))
  (multiple-value-bind (argnums constnums)
		       (ac=make-number (car matrix) (cdr matrix))
		       (list  (car matrix)
			      (cdr matrix)
			      (ac=hullot-h argnums constnums
				       (1- (expt 2 (length (cdr matrix))))))))

(defun ac=hullot-h (argnums constnums zahl &optional (ug 0) (og zahl) (lefttree t))
  (if (if lefttree
	  (ac=subset-big-enough argnums zahl)
	  (ac=subset-small-enough constnums zahl))
      (if (= 0 (- og ug))
	  (progn
	    ;(format t "~%Zahl: ~A" zahl)
	    (list zahl))
	  (let ((newnum (floor (+ ug og) 2)))
	       (append (ac=hullot-h argnums constnums newnum ug newnum t)
		       (ac=hullot-h argnums constnums (1+ newnum) (1+ newnum) og NIL))))
	   
      NIL))
	  


(defun ac=make-number (args vectors &optional (argnums NIL) (constnums NIL))
  (declare (edited  "11-MAY-1995")
	   (authors Neskakis)
	   (input   "2 lists: arguments and vectors")
	   (effect  "None")
	   (value   "two lists of integer-numbers for the arguments and the constants"
		    "example: 11 <-> (1011) means the argument occurs in the vectors"
		    "1,3 and 4 "))
  (do ((i (1- (length args)) (1- i)))
      ((< i 0) (values argnums constnums))
      (let ((newnum (ac=make-number-2 vectors i)))
	(if (ac=constant-p (elt args i))
	    (push newnum constnums))
	(push newnum argnums))))

(defun ac=make-number-2 (vectors Stelle &optional (newnum 0) (pot 1))
  (if (endp vectors)
      newnum
      (ac=make-number-2 (cdr vectors)
		      Stelle
		      (+ (* pot (signum (elt (cdar vectors) Stelle))) newnum)
		      (* 2 pot))))


(defun ac=subset-big-enough (argnums subset)
  (declare (edited  "11-MAY-1995")
	   (authors Neskakis)
	   (input   "the set of argnums and the subset-number")
	   (effect  "None")
	   (value   "t if the subset is big enough "))
  (if (endp argnums)
      t
      (if (= 0 (logand (car argnums) subset))
	  NIL
	  (ac=subset-big-enough (cdr argnums) subset))))

(defun ac=subset-small-enough (constnums subset)
  (declare (edited  "11-MAY-1995")
	   (authors Neskakis)
	   (input   "the set of constnums and the subset-number")
	   (effect  "None")
	   (value   "t if the subset is small enough"))
  (if (endp constnums)
      t
      (if (= 0 (logand (logand subset (car constnums))
		       (1- (logand subset (car constnums)))))
	  (ac=subset-small-enough (cdr constnums) subset)
	  NIL)))


;;;  ac=make-vecs
;;;
;;; Diese Funktionen bereiten die Substitutionen vor. D.h. sie basteln aus den
;;; "Vektoren" (x . [1 0 1 4 0 3) neue Vektoren (x NIL x (ac (ac x x) (ac x x)) (ac x (ac x x)))
;;; die durch ac~Subs jeweils nur neu "kombiniert werden muessen, d.h. Unterterme, die
;;; in den Substitutionen auftauchen muessen nicht jedesmal neu erzeugt werden.


(defun ac=make-vecs (fsym matrix)
  (declare (edited  "11-MAY-1995")
	   (authors Neskakis)
	   (input   "the ac-symbol, a matrix (a list of 3 lists)"
		    "1.the arguments 2.the vectors 3.the subsets which represent"
		    "correct substitutions")
	   (effect  "None")
	   (value   "a new matrix: 1.as above 3.as above"
		    "2. the vectors are changed from (x-3 . [1 0 2 1]) to (x-3 NIL (ac x-3 x-3) x-3)"))
  (list (car matrix)
	(mapcar #'(lambda(x) (ac=make-vecs-2 (car x) (cdr x) fsym)) (cadr matrix))
	(caddr matrix)))

(defun ac=make-vecs-2 (arg vec fsym &optional (newvec NIL))
  (do ((i (1- (length vec)) (1- i)))
      ((< i 0) newvec)
      (case (elt vec i)
	    (0 (push NIL newvec))
	    (1 (push arg newvec))
	    (2 (push (appl~create fsym (list arg arg)) newvec))
	    (t (push (ac=make-vec fsym arg (- (elt vec i) 2) (appl~create fsym (list arg arg)))
		     newvec)))))

(defun ac=make-vec (fsym arg zahl term)
  (do ((i zahl (1- i)))
      ((<= i 0) term)
      (setf term (appl~create fsym (list arg term)))))


;;; ac~Subs
;;;
;;; Diese Funktionen bauen die Substitutionen auf


(defun ac~Subs (fsym matrix)
  (declare (edited  "11-MAY-1995")
	   (authors Neskakis)
	   (input   "the ac-symbol , a matrix (list of three lists)"
		    "1. the list of arguments  2. the vectors (build by ac=make-vecs"
		    "3. the list of numbers which represent correct substitutions ")
	   (effect  "None")
	   (value   "the substitutions"))
  (ac=Subs (car matrix)
	   (cadr matrix)
	   (caddr matrix)
	   fsym))



(defun ac=Subs (argliste vecliste mengenliste fsym)
  (let ((argl (remove-if #'ac=constant-p argliste)))
    (mapcar #'(lambda (m)
	      (subst~create argl
			    (ac=Subs-Termliste-create argliste vecliste m fsym (length argliste))
			     ))
	  mengenliste)))

(defun ac=Subs-Termliste-create (argliste vecliste menge fsym lang &optional (termliste NIL))
  (do ((i (1- lang) (1- i)))
      ((< i 0) termliste)
      (if (not (ac=constant-p (elt argliste i)))
      (setf termliste
	    (cons
	     (ac=Subs-term-create vecliste menge i fsym)
	     termliste)))))

(defun ac=Subs-term-create (vecliste menge Stelle fsym &optional (term NIL))
  (multiple-value-bind (divi rest) (floor menge 2)
     (if (= rest 1)
	     (let ((VecStelle (elt (car vecliste) Stelle)))
	          (if (null VecStelle)
		      (ac=Subs-term-create (cdr vecliste) divi Stelle fsym term)
		      (if (null term)
			  (ac=Subs-term-create (cdr vecliste) divi Stelle fsym VecStelle)
			  (ac=Subs-term-create (cdr vecliste) divi Stelle fsym
					       (appl~create fsym (list VecStelle term))))))
         (if (= divi 0)
	     term
	     (ac=Subs-term-create (cdr vecliste) divi Stelle fsym term)))))

	


;;; print-object
;;;
;;; damit die Vektoren vernuenftig aussehen


(defmethod print-object :around ((object vector) stream)
  (if (or (eq (class-of object) (find-class 'vector))
	  (eq (class-of object) (find-class 'bit-vector)))
      (format stream "[~{~A ~}]" (coerce object 'list))
    (call-next-method)))











