
(in-package "KEIM")

;+----------------------------------------------------------------------------+
;| Hilfsfunktionen - groesstenteils spezialisierte unicp~ - Fkt               |
;+----------------------------------------------------------------------------+

(defvar cpac*debug nil)

;;;
;;; Aenderung (x)
;;;

(defun cpac=variable-p (object)
  (declare (edited  "21-OCT-1996")
	   (authors Felix)
	   (input   "An object.")
	   (effect  "None.")
	   (value   "True iff object is a variable for AC.")
	   (special cpac*theory))
  (and
   (sym~variable-p object)
   (not (unicp~constant-p object cpac*theory))))

(defun cpac=constant-p (object)
  (declare (edited  "21-OCT-1996")
	   (authors Felix)
	   (input   "An object.")
	   (effect  "None.")
	   (value   "True iff object is a constant for AC.")
	   (special cpac*theory))
  (and
   (sym~variable-p object)
   (unicp~constant-p object cpac*theory)))

#|

(defun cpac=variable-p (object &optional (varlist NIL))
  (declare (edited  "23-JAN-1996")
	   (authors Neskakis)
	   (input   "an object and a list of variables(optional)")
	   (effect  " True iff object is a variable and"
		    "    1. is not owned by a theory  "
		    " or 2. is owned by cpac*theory   "
		    " or 3. is member of the list of variables ")
	   (value   "None.")
	   (special cpac*theory))
  (and
   (sym~variable-p object)
   (or (null (unicp~var-label-theory (unicp~var-label-repres object)))
       (keim~equal (unicp~var-label-theory (unicp~var-label-repres object)) cpac*theory)
       (member object varlist :test 'keim~equal))))
 

(defun cpac=constant-p (object &optional (constlist NIL))
  (declare (edited  "23-JAN-1996")
	   (authors Neskakis)
	   (input   "an object and a list of constants(optional)")
	   (effect  " True iff object is a sym~constant or "
		    " not owned by cpac*theory or member of"
		    " the list of constants")
	   (value   "None")
	   (special cpac*theory))
  (or (sym~constant-p object)
      (not (or (keim~equal (unicp~var-label-theory (unicp~var-label-repres object)) cpac*theory)
	       (null (unicp~var-label-theory (unicp~var-label-repres object)))))
      (member object constlist :test 'keim~equal)))

|#

(defun cpac=vars-discriminated-p (arg1 arg2 discriminated)
  (declare (edited  "23-JAN-1996")
	   (authors Neskakis)
	   (input   "two arguments and a list of discriminated pairs of arguments")
	   (effect  " True iff the arguments are discriminated or"
		    " the pair of the two arguments is member of the"
		    " list of dicriminated pairs")
	   (value   "None")
	   (special cpac*theory))
  (or (unicp~vars-discriminated-p arg1 arg2)
      (member (list arg1 arg2) discriminated :test 'keim~equal)
      (member (list arg2 arg1) discriminated :test 'keim~equal)))

(defun cpac=vars-lesseq-p (arg1 arg2 const-restr)
  (declare (edited  "23-JAN-1996")
	   (authors Neskakis)
	   (input   "two arguments and a list of constant-restrictions")
	   (effect  "True iff arg1<=arg2 or a list (arg2 ... arg1 ...) is"
		    "member of the list of constant restrictions")
	   (value   "None")
	   (special cpac*theory))
  (or (unicp~vars-lesseq-p arg1 arg2)
      (member arg1 (cdar (member arg2 const-restr :key 'car :test 'keim~equal)) :test 'keim~equal)))

(defun cpac=vars-identified-p (arg1 arg2 identified)
  (declare (edited  "23-JAN-1996")
	   (authors Neskakis)
	   (input   "two arguments and a list of identified pairs of arguments")
	   (effect  "True iff the arguments are identified or the pair of the"
		    "arguments is member of the list of identified pairs")
	   (value   "None")
	   (special cpac*theory))
  (or (unicp~vars-identified-p arg1 arg2)
      (member (list arg1 arg2) identified :test 'keim~equal)
      (member (list arg2 arg1) identified :test 'keim~equal)))

(defun cpac=owned-var-p (arg ownedvar)
  (declare (edited  "23-JAN-1996")
	   (authors Neskakis)
	   (input   "an argument and a list of variables")
	   (effect  "True iff the argument is member of the list")
	   (value   "None")
	   (special cpac*theory))
  (member arg ownedvar :test 'keim~equal))


(defun cpac=var-label-less (arg)
  (declare (edited  "23-JAN-1996")
	   (authors Neskakis)
	   (input   "an argument")
	   (effect  "intersection of the list of the variables in the unicp~var-label-less-slot of the arg"
		    "and the list of arguments in the cpac=arguments-slot")
	   (value   "None")
	   (special cpac*theory))
  (remove-if-not #'cpac=position (unicp~var-label-less arg)))

(defun cpac=var-label-greater (arg)
  (declare (edited  "23-JAN-1996")
	   (authors Neskakis)
	   (input   "an argument")
	   (effect  "intersection of the list of the variables in the unicp~var-label-greater-slot of the arg"
		    "and the list of arguments in the cpac=arguments-slot")
	   (value   "None")
	   (special cpac*theory))
  (remove-if-not #'cpac=position (unicp~var-label-greater arg)))



;+----------------------------------------------------------------------------+
;| Definition der Unterklasse  cpac+up                                        |
;| sowie Hilfsfunktionen, die direkt die Slots der Klasse veraendern          |
;+----------------------------------------------------------------------------+


(defclass cpac+up (unicp+elementary-up)
  ((choice-points :initform NIL :accessor cpac=choice-points)
   (arguments :initform NIL :accessor cpac=arguments)
   (solutions :initform NIL :accessor cpac=solutions)
   (permanent-solutions :initform nil :accessor cpac=permanent-solutions)
   (solutions-p :initform NIL :accessor cpac=solutions-p))
  (:documentation
   "This is the class of ."
   "vectors added to permanent-solutions are deleted in solutions!!!"
   "this important for action5 in cpac=action: cpac=twovectors-oneconstant! !!!"))


(defun cpac=position (object)
  (declare (edited  "23-JAN-1996")
	   (authors Neskakis)
	   (input   "an object")
	   (effect  "None")
	   (value   "the position of the object in the list of arguments(cpac=arguments)")
	   (special cpac*theory))
  (unicp~var-label object cpac*theory))

(defun cpac=set-position! (object value)
  (declare (edited  "23-JAN-1996")
	   (authors Neskakis)
	   (input   "an object and a value")
	   (effect  "the unicp~var-label-slot of the object is set to value")
	   (value   "value")
	   (special cpac*theory))
  (setf (unicp~var-label object cpac*theory) value))




;+-------------------------------------------------------------------------------+
;| Alles fuer cpac=new-solutions!                                                |
;| mittels dieser Funktionen werden einmalig (zu Beginn) die Loesungsvektoren    |
;| der diophantischen Gleichungen berechnet, sowie die Slots unserer Klasse      |
;| und der Argumente mit Werten belegt.                                          |
;| Hier wird die Funktion  dioaed-solve.equations  aus  adv-domen.lisp aufgerufen|
;+-------------------------------------------------------------------------------+

(defun cpac=term-arglist (term)
  (declare (edited  "31-DEC-1995")
	   (authors Richts)
	   (input   "A term, consisting of an AC-symbol and variables.")
	   (effect  "None.")
	   (value "A list of the AC-arguments of TERM, i.e. a multiset of the"
		  "variables in TERM.  Variable identifications are considered, i.e. the"
		  "representative of each variable is taken."))
  (if (appl~p term)
      (mapcan #'cpac=term-arglist (appl~arguments term))
      (list (unicp~var-label-repres term))))



(defun cpac=args (argl-1 argl-2)
  (declare (edited  "14-AUG-1995")
	   (authors Neskakis)
	   (input   "two lists of lists of arguments")
	   (effect  "None")
	   (value   "a list of all args without duplicates"))
  (let ((argliste (append argl-1 argl-2)))
  (remove-duplicates
   (do ((i (1- (length argliste)) (1- i)) (newliste NIL (append (elt argliste i) newliste)))
       ((< i 0) newliste))
   :test #'keim~equal)))


(defun cpac=clear-positions! (args)
  (declare (edited  "26-JAN-1996")
	   (authors Neskakis)
	   (input   "list of arguments")
	   (effect  "the unicp~var-label-slot of the arguments is set to NIL")
	   (value   "list of arguments")
	   (special cpac*theory))
  (mapc #'(lambda(arg) (cpac=set-position! arg NIL)) args))


(defun cpac=build-diophant-equations (argl-1 argl-2 args)
  (declare (edited  "30-JAN-1996 ")
	   (authors Neskakis)
	   (input   "two lists of lists of arguments,each argument-list represents a term"
		    "and a list of arguments")
	   (effect  "None")
	   (value   "a list of all input-vectors for >dioaed-solve.equations< "))
  (unwind-protect
      (mapcar #'(lambda(t1 t2) (cpac=bde-count t1 t2 Args)) argl-1 argl-2)
    (cpac=labels-reset! args NIL)))

(defun cpac=bde-count (t1 t2 Args)
  (progn
    (cpac=labels-reset! args 0)
    (mapc #'(lambda(var) (incf (unicp~var-label-mark var))) t1)
    (mapc #'(lambda(var) (decf (unicp~var-label-mark var))) t2)
    (mapcar #'unicp~var-label-mark Args)))

(defun cpac=labels-reset! (Args Val)  (mapc #'(lambda (var) (setf (unicp~var-label-mark var) Val)) Args))


(defun cpac=redundant-rows&vectors (matrix row-length)
  (declare (edited  "15-AUG-1995")
	   (authors Neskakis)
	   (input   "matrix:cons of args and dioeq , length of the dioeq-vectors")
	   (effect  "None")
	   (value   "a list of all input-vectors for >dioaed-solve.equations< "
		    "minus redundant vectors like NULL-vectors etc. "))
  (progn
    
    ;erst mal die NULL-Vektoren
    (setf (cdr matrix)
	  (delete-if #'(lambda (vec)
			 (do ((i (1- row-length) (1- i))
			      (Sum 0 (+ (abs (elt Vec i)) Sum)))
			     ((or (< i 0) (> sum 0)) (= Sum 0))))
		     (cdr matrix)))
    
    ;dann die NULL-Spalten
    (let ((zero-columns
	   (do ((i 0 (1+ i))
		(rows (length (cdr matrix)))
		(zero-columns
		 nil
		 (if (do* ((j rows (1- j)) (Sum 0 (+ (abs (elt (elt (cdr matrix) j) i)) Sum)))
			  ((or (<= j 0) (> sum 0)) (= Sum 0)))
		     (cons i zero-columns)
		   zero-columns)))
	       ((>= i row-length) zero-columns))))
      (mapc #'(lambda (zero-column)
		(mapl #'(lambda (matrix-tail)
			  (setf (car matrix-tail)
				(delete-if #'atom (car matrix-tail)
					   :start zero-column
					   :end (1+ zero-column) :count 1)))
		      matrix))
	    zero-columns) 
      matrix)
    ))



(defun cpac=set-positions! (args)
  (declare (special cpac*theory))
  (do ((i (1- (length args)) (1- i)))
      ((< i 0) args)
      (cpac=set-position! (elt args i) i)))


(defun cpac=constant-rows (args)
  (declare (edited  "15-AUG-1995")
	   (authors Neskakis)
	   (input   "list of the arguments: variables&constants")
	   (effect  "None")
	   (value   "a list of the numbers of the constant-rows")
	   (special cpac*theory))
  (do ((symbols args (rest symbols))
	      (i 0 (1+ i))
	      (rows nil (if (cpac=variable-p (first symbols))
			    rows
			  (cons i rows))))
	     ((null symbols) rows)))




(defun cpac=new-equations! (up)
  (declare (edited  "22-JAN-1996" "21-AUG-1995")
	   (authors Richts Neskakis)
	   (input   "unification-problem [cpac+up]")
	   (effect  "None")
	   (value   "a list, first element: a list of the occuring variables and constants"
		    "        rest: lists representing solution-vectors of the diophantine equations")
	   (special cpac*theory))
  (let* ((argl-1    (mapcar #'cpac=term-arglist (car (unicp~equations up))))
	 (argl-2    (mapcar #'cpac=term-arglist (cadr (unicp~equations up)))))
    (setf (cpac=solutions-p up) nil)
    (if (some #'(lambda (args1 args2)
		  (and (or (null args1) (null args2))
		       (or args1 args2)))
	      argl-1 argl-2)
	nil
	(values up
		(mapcan #'(lambda (args1 args2)
			    (unless (or (rest args1) (rest args2))
			      (unicp~id-create (first args1) (first args2))))
			argl-1 argl-2)
		(mapcan #'(lambda (args1 args2)
			    (cond ((and (null (rest args1)) (null (rest args2)))
				   nil)
				  ((null (rest args1))
				   (mapcar #'(lambda (var)
					       (unicp~id-create (first args1) var))
					   args2))
				  ((null (rest args2))
				   (mapcar #'(lambda (var)
					       (unicp~id-create (first args2) var))
					   args1))
				  (t nil)))
			argl-1 argl-2)
		(mapcan #'(lambda (args1 args2)
			    (cond ((and (null (rest args1)) (null (rest args2)))
				   nil)
				  ((null (rest args1))
				   (list (first args1)))
				  ((null (rest args2))
				   (list (first args2)))
				  (t nil)))
			argl-1 argl-2)
		nil
		(mapcan #'(lambda (args1 args2)
			    (cond ((and (null (rest args1)) (null (rest args2)))
				   nil)
				  ((null (rest args1))
				   (list (unicp~cr-create (first args1) args2)))
				  ((null (rest args2))
				   (list (unicp~cr-create (first args2) args1)))
				  (t nil)))
			argl-1 argl-2)
		nil))))


(defun cpac=new-solutions! (up)
  (declare (special cpac*theory)
           (edited  "21-AUG-1995")
	   (authors Neskakis)
	   (input   "unification-problem [cpac+up]")
	   (effect  "None")
	   (value   "a list, first element: a list of the occuring variables and constants"
		    "        rest: lists representing solution-vectors of the diophantine equations"))
  (let* ((argl-1    (mapcar #'cpac=term-arglist (car (unicp~equations up))))
	 (argl-2    (mapcar #'cpac=term-arglist (cadr (unicp~equations up))))
	 (argl-3    (cpac=args argl-1 argl-2))
	 (args      (progn (cpac=clear-positions! (unicp~variables up))
			   argl-3))
	 (dioeq     (cpac=build-diophant-equations argl-1 argl-2 args))
	 (matrix    (cpac=redundant-rows&vectors (cons args dioeq) (length args)))
	 (newdioeq  (cdr matrix))
	 (newargs   (cpac=set-positions! (car matrix)))
	 (constrows (cpac=constant-rows newargs))
	 (vectors   (dioaed-solve.equations newdioeq NIL))) ;constrows)))
    (setf (cpac=solutions-p up) t
          (cpac=solutions up) vectors
	  (cpac=arguments up) newargs
	  (cpac=permanent-solutions up) NIL)))


	 


;+-------------------------------------------------------------------------------+
;| Alles fuer cpac=action I & II                                                 |
;| das erste Drittel enthaelt Funktionen, die Vektoren loeschen bzw. von den     |
;| Loesungen zu den permanenten Loesungen verschieben koennen. D.h. also Slots   |
;| der Klasse direkt veraendern.                                                 |
;| die zweite Drittel enthaelt Funktionen, die neue Informationen produzieren    |
;| koennen,wie: x1 < x2  oder  x1 /= x2                                          |
;| im letzten Drittel ist die Funktion cpac=action                               |
;+-------------------------------------------------------------------------------+


(defun cpac=constant-gt-one! (up constants)
  (declare (edited  "22-NOV-1995")
	   (authors Neskakis)
	   (input   "up: unification-problem [cpac+up]"
		    "constants: a list of former variables which are now occupied by other theories")
	   (effect  "- sets >not*solvable< to >T< iff the unif.prob. is no longer solvable"
		    "- deletes vectors in cpac=solutions if they are not suitable for a"
		    "  correct subset"
		    "  -test1:delete $\vec{s}$ in $S$, if $\vec{s}(x)>1$, $x\notin AC$.")
	   (value   "None")
	   (special cpac*theory not*solvable))
  (progn
  (mapc #'(lambda(vec)
	    (do ((symbols constants (cdr symbols)))
		((or (null symbols) not*solvable))
		(if (> (elt vec (cpac=position (car symbols))) 1)
		    (setf not*solvable T))))
	    (cpac=permanent-solutions up))
  (if (null not*solvable)
      (setf (cpac=solutions up)
	    (remove-if #'(lambda(vec)
			   (do ((symbols constants (cdr symbols)))
			       ((or (null symbols)
				    (> (elt vec (cpac=position (car symbols))) 1))
				(if symbols T NIL))))
		       (cpac=solutions up)))) 
  NIL))


(defun cpac=const-restr-vector-p (cr vector olddis)
  (declare (edited  "12-DEC-1995")
	   (authors Neskakis)
	   (input   "cr: a constant-restriction"
		    "vector: a solution-vector ")
	   (effect  "None")
	   (value   "T iff vector ist not suitable for a correct subset")
           (special cpac*theory))
    (if (and (cpac=constant-p (unicp~cr-const cr))
	     (> (elt vector (cpac=position (unicp~cr-const cr))) 0))
	(if (do ((xliste (unicp~cr-varlist cr) (cdr xliste)) (Eintrag NIL))
		((or (null xliste) Eintrag) Eintrag)
		(if (and
		     (cpac=vars-discriminated-p (unicp~cr-const cr) (car xliste) olddis) ;HIER not ???
		     (cpac=variable-p (car xliste))
		     (> (elt vector (cpac=position (car xliste))) 0))
		    (setf Eintrag T)))
	    T
	  NIL)
      NIL))






(defun cpac=const-restr-newcr! (up const-restr olddis)
  (declare (edited  "23-NOV-1995")
	   (authors Neskakis)
	   (input   "up: unification-problem [cpac+up]"
		    "const-restr: a list of constant-restrictions")
	   (effect  "- sets >not*solvable< to >T< iff the unif.prob. is no longer solvable"
		    "- deletes vectors in cpac=solutions if they are not suitable for a"
		    "  correct subset"
		    "  -test2:delete $\vec{s}$ in $S$, if $\vec{s}(x)>0$, $\vec{s}(y)>0$, $x\dotls y$"
		    "         $y\notin AC$, $x\in AC$"
		    "  -performs test2 for evey constant-restriction in the list")
	   (value   "None")
	   (special cpac*theory not*solvable))
  (do ((constr const-restr (cdr constr)))
      ((or (null constr) not*solvable))
      (if (cpac=constant-p (unicp~cr-const (first constr)))
	  (progn
	    (do ((pervec (cpac=permanent-solutions up) (cdr pervec)))
		((or (null pervec) not*solvable))
	        (if (cpac=const-restr-vector-p  (first constr) (car pervec) olddis)
		    (setf not*solvable T)))
	    (if (null not*solvable)
		(setf (cpac=solutions up)
		      (remove-if #'(lambda(vec)
				     (cpac=const-restr-vector-p (first constr) vec olddis))
				 (cpac=solutions up))))))))







(defun cpac=const-restr-newconst! (up constants olddis)
  (declare (edited  "23-NOV-1995")
	   (authors Neskakis)
	   (input   "up: unification-problem [cpac+up]"
		    "constants: a list of former variables which are now occupied by other theories"
		    "olddis: a list of former variables which are now occupied by other theories"
		    "        but their theory-slot is still set to NIL")
	   (effect  "- sets >not*solvable< to >T< iff the unif.prob. is no longer solvable"
		    "- deletes vectors in cpac=solutions if they are not suitable for a"
		    "  correct subset"
		    "  -test2:delete $\vec{s}$ in $S$, if $\vec{s}(x)>0$, $\vec{s}(y)>0$, $x\dotls y$"
		    "         $y\notin AC$, $x\in AC$"
		    "  -performs test2 for evey constant in the list")
	   (value   "None")
	   (special cpac*theory not*solvable))
  (do ((c constants (cdr c)))
      ((or (null c) not*solvable))
      (if (cpac=var-label-less (car c))
	  (let ((constr (cons (car c) (cpac=var-label-less (car c)))))
	    (do ((pervec (cpac=permanent-solutions up) (cdr pervec)))
		((or (null pervec) not*solvable))
	        (if (cpac=const-restr-vector-p constr (car pervec) olddis)
		    (setf not*solvable T)))
	    (if (null not*solvable)
		(setf (cpac=solutions up)
		      (remove-if #'(lambda(vec)
				     (cpac=const-restr-vector-p constr vec olddis))
				 (cpac=solutions up))))))))




(defun cpac=const-restr-newvar! (up variables olddis)
  (declare (edited  "23-NOV-1995")
	   (authors Neskakis)
	   (input   "up: unification-problem [cpac+up]"
		    "variables: a list of variables which are now occupied by us"
		    "olddis: a list of former variables which are now occupied by other theories"
		    "        but their theory-slot is still set to NIL")
	   (effect  "- sets >not*solvable< to >T< iff the unif.prob. is no longer solvable"
		    "- deletes vectors in cpac=solutions if they are not suitable for a"
		    "  correct subset"
		    "  -test2:delete $\vec{s}$ in $S$, if $\vec{s}(x)>0$, $\vec{s}(y)>0$, $x\dotls y$"
		    "         $y\notin AC$, $x\in AC$"
		    "  -performs test2 for evey variable in the list")
	   (value   "None")
	   (special cpac*theory not*solvable))
  (do ((v variables (cdr v)))
      ((or (null v) not*solvable)) 
      (do ((c (cpac=var-label-greater (car v)) (cdr c)))
	  ((or (null c) not*solvable))
	  (let ((const-restr (unicp~cr-create (car c) (car v))))
	    (do ((pervec (cpac=permanent-solutions up) (cdr pervec)))
		((or (null pervec) not*solvable))
	        (if (cpac=const-restr-vector-p const-restr (car pervec) olddis)
		    (setf not*solvable T)))
	    (if (null not*solvable)
		(setf (cpac=solutions up)
		      (remove-if #'(lambda(vec)
				     (cpac=const-restr-vector-p const-restr vec olddis))
				 (cpac=solutions up))))))))


(defun cpac=const-restr-newdis! (up newdis const-restr)
  (declare (edited  "30-JAN-1996")
	   (authors Neskakis)
	   (input   "up: unification-problem [cpac+up]"
		    "newdis: A list of pairs of variables which must not be unified."
		    "const-restr: a list of constant-restrictions")
	   (effect  "- sets >not*solvable< to >T< iff the unif.prob. is no longer solvable"
		    "- deletes vectors in cpac=solutions if they are not suitable for a"
		    "  correct subset"
		    "  -test2:delete $\vec{s}$ in $S$, if $\vec{s}(x)>0$, $\vec{s}(y)>0$, $x\dotls y$"
		    "         $y\notin AC$, $x\in AC$"
		    "  -performs test2 for every pair in the list")
	   (value   "None")
	   (special cpac*theory not*solvable))
  (progn
    (mapc #'(lambda(vec)
	      (mapc #'(lambda(dis)
			(if (and
			     (or (and (cpac=vars-lesseq-p (first dis) (second dis) const-restr)
				      (cpac=constant-p (second dis))
				      (cpac=variable-p (first dis)))
				 (and (cpac=vars-lesseq-p (second dis) (first dis) const-restr)
				      (cpac=constant-p (first dis))
				      (cpac=variable-p (second dis))))
			     (> (elt vec (cpac=position (first dis))) 0)
			     (> (elt vec (cpac=position (second dis))) 0))
			    (setf not*solvable T)))
		    newdis))
	  (cpac=permanent-solutions up))
    (setf (cpac=solutions up)
	  (remove-if #'(lambda(vec)
			 (some #'(lambda(dis)
				   (and
				    (or (and (cpac=vars-lesseq-p (first dis) (second dis)
								 const-restr)
					     (cpac=constant-p (second dis))
					     (cpac=variable-p (first dis)))
					(and (cpac=vars-lesseq-p (second dis) (first dis)
								 const-restr)
					     (cpac=constant-p (first dis))
					     (cpac=variable-p (second dis))))
				    (> (elt vec (cpac=position (first dis))) 0)
				    (> (elt vec (cpac=position (second dis))) 0)))
			       newdis))
		     (cpac=solutions up)))))

    






(defun cpac=twoconst-gt-zero-co! (up constants olddis)
  (declare (edited  "02-FEB-1996")
	   (authors Neskakis)
	   (input   "up: unification-problem [cpac+up]"
		    "a list of former variables which are now occupied by other theories"
		    "olddis: a list of former variables which are now occupied by other theories"
		    "        but their theory-slot is still set to NIL")
	   (effect  "- sets >not*solvable< to >T< iff the unif.prob. is no longer solvable"
		    "- deletes vectors in cpac=solutions if they are not suitable for a"
		    "  correct subset"
		    "  -test3:delete $\vec{s}$ in $S$, if $\vec{s}(x)>0$, $\vec{s}(y)>0$, $x\not\doteqv"
		    "        y$, $x\notin AC$, $y\notin AC$."
		    "  -performs test3 for evey constant in the list")
	   (value   "None")
	   (special cpac*theory not*solvable))
  (let ((constants-2 (remove-if-not  #'cpac=constant-p (cpac=arguments up))))
    (do ((constant constants (cdr constant)))
	((or (null constant) not*solvable))
        (do ((pervecs (cpac=permanent-solutions up) (cdr pervecs)))
	    ((or (null pervecs) not*solvable))
	    (if (> (elt (car pervecs) (cpac=position (car constant))) 0)
		(do ((constant-2 constants-2 (cdr constant-2)))
		    ((or (null constant-2) not*solvable))
		    (if (and (> (elt (car pervecs) (cpac=position (car constant-2))) 0)
			     (cpac=vars-discriminated-p (car constant) (car constant-2) olddis))
			(setf not*solvable T))))))
    (if (null not*solvable)
	(setf (cpac=solutions up)
	      (remove-if #'(lambda(vec)
			     (do ((constant constants (cdr constant)) (Flag NIL))
				 ((or (null constant) flag) flag)
			         (if (> (elt vec (cpac=position (car constant))) 0)
				     (do ((constant-2 constants-2 (cdr constant-2)))
					 ((or (null constant-2) flag))
				         (if (and (> (elt vec (cpac=position (car constant-2))) 0)
						  (cpac=vars-discriminated-p (car constant) (car constant-2) olddis))
					     (setf flag T))))))
			 (cpac=solutions up))))
      not*solvable))
      



  


(defun cpac=twoconst-gt-zero-dis! (up newdis)
  (declare (edited  "02-FEB-1996")
	   (authors Neskakis)
	   (input   "up: unification-problem [cpac+up]"
	            "newdis: A list of pairs of variables which must not be unified.")
	   (effect  "- sets >not*solvable< to >T< if the unif.prob. is no longer solvable"
		    "- deletes vectors in cpac=solutions if they are not suitable for a"
		    "  correct subset"
		    "  -test3:delete $\vec{s}$ in $S$, if $\vec{s}(x)>0$, $\vec{s}(y)>0$, $x\not\doteqv"
		    "        y$, $x\notin AC$, $y\notin AC$."
		    "  -performs test3 for every pair in the list")
	   (value   "None")
	   (special cpac*theory not*solvable))
  (progn
    (mapc #'(lambda(ndis)
	      (if (and (cpac=constant-p (first ndis))
		       (cpac=constant-p (second ndis)))
		  (mapc #'(lambda(vec)
			    (if (and (> (elt vec (cpac=position (first ndis))) 0)
				     (> (elt vec (cpac=position (second ndis))) 0))
				(setf not*solvable T)))
			(cpac=permanent-solutions up))))
	  newdis)
    (if (null not*solvable)
	(setf (cpac=solutions up)
	      (remove-if #'(lambda(vec)
			     (some #'(lambda(ndis)
				       (and (cpac=constant-p (first ndis))
					    (cpac=constant-p (second ndis))
					    (> (elt vec (cpac=position (first ndis))) 0)
					    (> (elt vec (cpac=position (second ndis))) 0)))
				   newdis))
			 (cpac=solutions up))))))
					    
	







	       



(defun cpac=neq-constants-newid! (up identified)
  (declare (edited  "02-FEB-1996")
	   (authors Neskakis)
	   (input   "unification-problem [cpac+up]"
		    "identified: list of pairs of variables which are identical")
	   (effect  "- sets >not*solvable< to >T< if the unif.prob. is no longer solvable"
		    "- deletes vectors in cpac=solutions if they are not suitable for a"
		    "  correct subset"
		    "  -test4: delete $\vec{s}$ in $S$, if $\vec{s}(x)\not=\vec{s}(y)$, $x\doteqv y$,"
		    "          ;$x\notin AC$, $y\notin AC$."
		    "  -performs test4 for every pair in the list")
	   (value   "None")
	   (special cpac*theory not*solvable))
  (progn
    (mapc #'(lambda(id)
	      (if (and (cpac=constant-p (unicp~id-left id))
		       (cpac=constant-p (unicp~id-right id)))
		  (mapc #'(lambda(vec)
			    (if (/= (elt vec (cpac=position (unicp~id-left id)))
				    (elt vec (cpac=position (unicp~id-right id))))
				(setf not*solvable T)))
			(cpac=permanent-solutions up))))
	  identified)
    (if (null not*solvable)
	(setf (cpac=solutions up)
	      (remove-if #'(lambda(vec)
			     (some #'(lambda(id)
				       (and (/= (elt vec (cpac=position (unicp~id-left id)))
						(elt vec (cpac=position (unicp~id-right id))))
					    (cpac=constant-p (unicp~id-left id))
					    (cpac=constant-p (unicp~id-right id))))
				   identified))
			 (cpac=solutions up))))))

(defun cpac=neq-constants-newco! (up constants identified)
  (declare (edited  "02-FEB-1996")
	   (authors Neskakis)
	   (input   "unification-problem [cpac+up]"
		    "constants: a list of former variables which are now occupied by other theories"
		    "identified: list of pairs of variables which are identical")
	   (effect  "- sets >not*solvable< to >T< if the unif.prob. is no longer solvable"
		    "- deletes vectors in cpac=solutions if they are not suitable for a"
		    "  correct subset"
		    "  -test4: delete $\vec{s}$ in $S$, if $\vec{s}(x)\not=\vec{s}(y)$, $x\doteqv y$,"
		    "          ;$x\notin AC$, $y\notin AC$."
		    "  -performs test4 for every constant in the list")
	   (value   "None")
	   (special cpac*theory not*solvable))
  (let ((constants2 (remove-if-not #'cpac=constant-p (cpac=arguments up))))
    (mapc #'(lambda(co)
	      (mapc #'(lambda(co2)
			(mapc #'(lambda(vec)
				  (if (and (/= (elt vec (cpac=position co))
					       (elt vec (cpac=position co2)))
					   (cpac=vars-identified-p co co2 identified))
				      (setf not*solvable T)))
			      (cpac=permanent-solutions up)))
		    constants2))
	  constants)
    (if (null not*solvable)
	(setf (cpac=solutions up)
	      (remove-if #'(lambda(vec)
			     (do ((co constants (cdr co)) (Eintrag NIL))
				 ((or (null co) Eintrag) Eintrag)
			         (do ((co2 constants2 (cdr co2)))
				     ((or (null co2) Eintrag))
				     (if (and (cpac=vars-identified-p (first co) (first co2) identified)
					      (/= (elt vec (cpac=position (car co)))
						  (elt vec (cpac=position (car co2)))))
					 (setf Eintrag T)))))
			 (cpac=solutions up))))))
    
     

(defun cpac=2vectors-oneconst-newconstants! (up constants)
  (declare (edited  "02-FEB-1996")
	   (authors Neskakis)
	   (input   "unification-problem [cpac+up]"
		    "a list of former variables which are now occupied by other theories")
	   (effect  "- sets >not*solvable< to >T< iff the unif.prob. is no longer solvable"
		    "- deletes vectors in cpac=solutions if they are not suitable for a"
		    "  correct subset"
		    "  -test5:  delete $\vec{s}$ in $S$, if $\exists\vec{t}\in T$, $\vec{s}\not=\vec{t}$,
		    "		;$\vec{t}(x)>0$, $\vec{s}(x)>0$, $x\notin AC$."
		    "  -performs test5 for every constant in the list")
	   (value   "None")
	   (special cpac*theory not*solvable))
  (progn
    (setf not*solvable
	  (some #'(lambda (const)
		    (do ((sol (cpac=permanent-solutions up) (cdr sol)) (sum 0))
			((or (> sum 1) (null sol))
			 (> sum 1))
		      (incf sum (elt (car sol) (cpac=position const)))))
		constants))
    (if (null not*solvable)
	(setf (cpac=solutions up)
	      (remove-if #'(lambda(vec)
			     (let ((flag NIL))
			       (mapc #'(lambda(const)
					 (if (< 0 (elt vec (cpac=position const)))
					     (mapc #'(lambda(npv)
						       (if (< 0 (elt npv (cpac=position const)))
							   (setf flag T)))
						   (cpac=permanent-solutions up))))
				     constants)
			       flag))
			 (cpac=solutions up))))))


(defun cpac=2vectors-oneconst-newpervecs! (up newpervecs)
  (declare (edited  "02-FEB-1996")
	   (authors Neskakis)
	   (input   "up: unification-problem [cpac+up]"
		    "vectors which are now in the set of permanent solutions")
	   (effect  "- sets >not*solvable< to >T< if the unif.prob. is no longer solvable"
		    "- deletes vectors in cpac=solutions if they are not suitable for a"
		    "  correct subset"
		    "  -test5:  delete $\vec{s}$ in $S$, if $\exists\vec{t}\in T$, $\vec{s}\not=\vec{t}$,
		    "		;$\vec{t}(x)>0$, $\vec{s}(x)>0$, $x\notin AC$."
		    "  -performs test5 for every vector in the list")
	   (value   "T iff some vector is deleted and problem is still solvable")
	   (special cpac*theory not*solvable))
  (let ((constants (remove-if-not #'cpac=constant-p (cpac=arguments up)))
	(start-over NIL))
    (setf not*solvable
	  (some #'(lambda (const)
		    (do ((sol (cpac=permanent-solutions up) (cdr sol)) (sum 0))
			((or (> sum 1) (null sol))
			 (> sum 1))
		      (incf sum (elt (car sol) (cpac=position const)))))
		constants))
    (if (null not*solvable)
	(progn
	  (setf (cpac=solutions up)
		(remove-if #'(lambda(vec)
			       (let ((flag NIL))
				 (mapc #'(lambda(const)
					   (if (< 0 (elt vec (cpac=position const)))
					       (mapc #'(lambda(npv)
							 (if (< 0 (elt npv (cpac=position const)))
							     (setf flag T
								   start-over T)))
						     newpervecs)))
				       constants)
				 flag))
			   (cpac=solutions up)))
	  start-over)
	NIL)))
  
 


(defun cpac=add-to-permanents! (up)
  (declare (edited  "02-FEB-1996")
	   (authors Neskakis)
	   (input   "up: unification-problem [cpac+up]")
	   (effect  "vectors that must be in the set of solutions are moved"
		    "form SOLUTIONS to PERMANENT-SOLUTIONS of UP")
	   (Value   "True, iff some vector has been moved from SOLUTIONS to"
		    "PERMANENT-SOLUTIONS of UP. and problem is still solvable")
	   (special cpac*theory not*solvable))

  (let ((vecs (cpac=solutions up))
	(vecs-to-pervecs NIL)
	(new-pervecs NIL))

    (do ((args (cpac=arguments up) (cdr args)))
	((null args) vecs-to-pervecs)
	(do ((i (1- (length vecs)) (1- i))
	     (found1 NIL)
	     (found2 (some #'(lambda (vec)
			       (< 0 (elt vec (cpac=position (car args)))))
			   (cpac=permanent-solutions up))))
	    ((or (< i 0) found2)
	     (if (and found1 (not found2))
		 (push found1 vecs-to-pervecs)))
	    (if (< 0 (elt (elt vecs i) (cpac=position (car args))))
		(if found1
		    (setf found2 T)
		    (setf found1 i)))))

    (remove-duplicates vecs-to-pervecs)

    (setf (cpac=solutions up) NIL)

    (do ((i (1- (length vecs)) (1- i)) )
	((< i 0))
        (if (member i vecs-to-pervecs)
	    (progn
	      (push (elt vecs i) (cpac=permanent-solutions up))
	      (push (elt vecs i) new-pervecs))
	    (push (elt vecs i) (cpac=solutions up))))

    (if (> (length vecs-to-pervecs) 0)   ;werden Vektoren zu den permanent-solutions hinzugefuegt so
	(if (cpac=2vectors-oneconst-newpervecs! up new-pervecs) ; kann action5 nochmal greifen!
	    (cpac=add-to-permanents! up))) ;woraufhin add-to-permanents wieder greifen koennte

    (> (length vecs-to-pervecs) 0)))
	       
 
       



    

;+++ zweites Drittel von cpac=action +++


;;;
;;; Aenderung
;;;




(defun cpac=variable-to-ac (up ownedvar)
  (declare (edited  "02-FEB-1996")
	   (authors Neskakis)
	   (input   "unification-problem [cpac+up]"
		    "ownedvar:  A list of variables which need not to be constants")
	   (effect  "None")
	   (value   "performs information-test1: $x\in AC$, if $\sum_{\vec{s}\in T} \vec{s}(x)>1$."
	            "and returns a list of variables which need not to be constants")
	   (special cpac*theory))
  (let ((newownedvar NIL))
      (do ((args (cpac=arguments up) (cdr args)) )
	  ((null args) newownedvar)
	  (if (and (null (unicp~var-label-theory (car args)))
		   (unicp~shared-variable-p (car args)))
	      (do ((sol (cpac=permanent-solutions up) (cdr sol)) (sum 0))
		  ((or (> sum 1) (null sol))
		   (if (> sum 1) (setf newownedvar (cons (car args) newownedvar))))
		  (incf sum (elt (car sol) (cpac=position (car args)))))))
      (remove-if #'(lambda(var) (member var ownedvar :test 'keim~equal)) newownedvar)))
  






(defun cpac=new-identified-1 (up identified)
  (declare (edited  "02-FEB-1996")
	   (authors Neskakis)
	   (input   "unification-problem [cpac+up]"
		    "identified: list of pairs of variables which are identical")
	   (effect  "None")
	   (value   "performs information-test3: $x\doteqv y$, if $\exists\vec{s}\in T$, $\vec{s}(x)=1$, $\vec{s}(y)=1$,"
                    "                            $x\notin AC$, $y\notin AC$."
	            "which returns a list of new identifications")
	   (special cpac*theory))
  (let* ((args-1 (cpac=arguments up))
	 (constants-1  (remove-if #'cpac=variable-p args-1))
	 (constants    (remove-if-not #'unicp~shared-variable-p constants-1))
	 (newidentified NIL))
    (do ((c constants (cdr c)))
	((null c) newidentified)
        (do ((v (cpac=permanent-solutions up) (cdr v)))
	    ((null v))
	    (if (= (elt (car v) (cpac=position (car c))) 1)
		(do ((a constants (cdr a)))
		    ((null a))
		    (if (and (= (elt (car v) (cpac=position (car a))) 1)
			     (not (cpac=vars-identified-p (car a) (car c) identified)))
			(setf newidentified (cons (unicp~id-create (car c) (car a)) newidentified)))))))
    newidentified))
		


(defun cpac=subset-information (up cr id dis ov)
  (declare (edited  "02-FEB-1996")
	   (authors Neskakis)
	   (input   "up: unification-problem [cpac+up]"
		    "cr: a list of constant-restrictions"
		    "id: list of identification of variables"
		    "dis: list of pairs of variables which must not be unified "
		    "ov: list of variables which must not be constants")
	   (effect  "None")
	   (value   "performs information-test4: $x\dotle y$, if $c_S(x)\subseteq c_S(y)$"
                    "                            ($\forall\vec{s}\in S$, $\vec{s}(x)\le\vec{s}(y)$)"
	            "     and information-test5: $x\dotls y$, if $c_S(x)\subset c_S(y)$ ($\forall\vec{s}\in S$"
		    "                            $\vec{s}(x)\le\vec{s}(y)$, $\exists\vec{s}\in T$,"
                    "                            $\vec{s}(x)\not=\vec{s}(y)$)."
		    "     and information-test7: $x\in AC$, $y\dotls x$, $z\dotls x$, if $c_S(y)\subset c_S(x)$,"
                    "                            $c_S(z)\subset c_S(x)$, $y\not=z$, $c_S(y)\cap c_S(z)=\emptyset$"
		    "and returns a list of new information consisting of:"
		    "     - a list of constant-restrictions"
		    "     - a list pairs of variables which must not be unified "
		    "     - a list of variables which need not to be variables")
	   (special cpac*theory))
  (let* ((ncr  (cpac=new-constr-2 up cr id))
	 (hlp  (cpac=not-identified up ncr dis))
	 (ndis (car hlp))
	 (hlp2 (cadr hlp))
	 (now  (cpac=variable-to-ac-2 up hlp2 ov)))
    (list ncr ndis now))) 






  

(defun cpac=new-constr-2 (up up-const-restr up-identified)
  ;this function builds information4: $x\dotle y$, if $c_S(x)\subseteq c_S(y)$
  ;($\forall\vec{s}\in S$, $\vec{s}(x)\le\vec{s}(y)$)
  (declare (edited  "05-DEC-1995")
	   (authors Neskakis)
	   (input   "unification-problem [cpac+up]")
	   (effect  "None")
	   (value   "a list of new constant-restrictions")
	   (special cpac*theory))
  (let ((Solution-vecs  (append (cpac=solutions up) (cpac=permanent-solutions up)))
	(newcr NIL))
    (do ((y-args (cpac=arguments up) (cdr y-args))) 
	((null y-args))
        (if (unicp~shared-variable-p (car y-args))
	    (do ((x-args (cpac=arguments up) (cdr x-args)) (newcr-for-xarg NIL))
		((null x-args) (if newcr-for-xarg
				   (setf newcr (cons (unicp~cr-create (car y-args) newcr-for-xarg)
						     newcr))))
	        (if (and (unicp~shared-variable-p (car x-args))
			 (not (or (cpac=vars-lesseq-p (car y-args) (car x-args) up-const-restr)
				  (cpac=vars-lesseq-p (car x-args) (car y-args) up-const-restr)
				  (cpac=vars-identified-p (car x-args) (car y-args) up-identified))))
		    (do ((vecs Solution-vecs (cdr vecs)) (no-cr-flag NIL))
			((or no-cr-flag (null vecs)) (if (not no-cr-flag)
							 (setf newcr-for-xarg (cons (car x-args) newcr-for-xarg))))
		      (if (> (elt (car vecs) (cpac=position (car x-args)))
			     (elt (car vecs) (cpac=position (car y-args))))
			  (setf no-cr-flag T)))))))
    newcr))


(defun cpac=not-identified (up newcr discriminated)
  ;this function builds information5: $x\dotls y$, if $c_S(x)\subset c_S(y)$ ($\forall\vec{s}\in S$,
  (declare (edited  "05-DEC-1995")
	   (authors Neskakis)
	   (input   "unification-problem [cpac+up]"
		    "newconst-restr: a list of new constant-restrictions")
	   (effect  "None")
	   (value   "new-discriminated")
	   (special cpac*theory))
  (let ((new-discriminated NIL)
	(newcrlessnoteq NIL))
    (do ((ncr newcr (cdr ncr)))
	((null ncr))
        (do ((ncrvl (cdar ncr) (cdr ncrvl)))
	    ((null ncrvl))
	    (do ((pervecs (cpac=permanent-solutions up) (cdr pervecs)) (ncrlneq NIL))
		((null pervecs) (if (> (length ncrlneq) 1)
				    (setf newcrlessnoteq
					  (cons (unicp~cr-create (car ncr) ncrlneq)
						newcrlessnoteq))))
	        (if (/= (elt (car pervecs) (cpac=position (caar ncr)))
			(elt (car pervecs) (cpac=position (car ncrvl))))
		    (if (not (cpac=vars-discriminated-p (caar ncr) (car ncrvl) discriminated))
			(setf new-discriminated (cons (list (caar ncr) (car ncrvl)) new-discriminated)))
		    (setf ncrlneq (cons (car ncrvl) ncrlneq))))))
    (list new-discriminated newcrlessnoteq)))


(defun cpac=variable-to-ac-2 (up newcrlessnoteq ownedvar)
  (declare (edited  "05-DEC-1995")
	   (authors Neskakis)
	   (input   "unification-problem [cpac+up]"
		    "newcrlessnoteq: newconst-restr without discriminated variables in the variablelist")
	   (effect  "None")
	   (value   "A list of variables which need not to be constants")
	   (special cpac*theory))
  (let ((newownedvar NIL))
    (do ((ncrlneq newcrlessnoteq (cdr ncrlneq)))
	((null ncrlneq))
        (if (not (cpac=owned-var-p (unicp~cr-const (first ncrlneq)) ownedvar))
	    (do ((liste1 (unicp~cr-varlist (first ncrlneq)) (cdr liste1))
		 (x (unicp~cr-const (first ncrlneq))) (ready-flag NIL))
		((or (null liste1) ready-flag))
	        (do ((liste2 (unicp~cr-varlist (first ncrlneq)) (cdr liste2)))
		    ((or (null liste2) ready-flag))
		    (do ((vecs (append (cpac=solutions up) (cpac=permanent-solutions up)) (cdr vecs)) (fail-flag NIL))
			((or (null vecs) fail-flag) (if (not fail-flag)
							(progn
							  (setf newownedvar (cons x newownedvar))
							  (setf ready-flag T))))
		        (if (and (> (elt (car vecs) (cpac=position (car liste1))) 0)
				 (> (elt (car vecs) (cpac=position (car liste2))) 0))
			    (setf fail-flag T)))))))
    newownedvar))









(defun cpac=spalte (up)
  (declare (edited  "02-FEB-1996")
	   (authors Neskakis)
	   (input   "unification-problem [cpac+up]")
	   (effect  "not*solvable is set to TRUE iff the unificationproblem is no longer solvable")
	   (value   "None")
	   (special cpac*theory not*solvable))
  (do ((args (cpac=arguments up) (cdr args)))
      ((null args))
      (do ((vecs (append (cpac=permanent-solutions up) (cpac=solutions up)) (cdr vecs)) (sum 0))
	  ((or (null vecs) (> sum 0)) (if (= 0 sum) (setf not*solvable T)))
	  (incf sum (elt (car vecs) (cpac=position (car args)))))))


;;;
;;; Aenderung
;;;


(defun cpac=newov-co (up constants newownedvar)
  (declare (edited  "02-FEB-1996")
	   (authors Neskakis)
	   (input   "unification-problem [cpac+up]"
		    "constants: a list of former variables which are now occupied by other theories"
		    "newownedvar:  A list of variables which need not to be constants")
	   (effect  "None")
	   (value   "performs information-test9: $\ind(x)=AC$, if \sum_{\vec{s}\in T} \vec{s}(x) > 1, & \oor\\[3ex]"
                    "                            \vec{s}(x) > 0 & \fora \vec{s}\in T\ \with \vec{s}(y)>0, \ind(y)=E'."
		    "for every constant in the argument-list of constants"
		    "and returns a list of variables which must not be constants")
	   (special cpac*theory))
  (let ((nov NIL))
    (mapc #'(lambda(co)
	      (mapc #'(lambda(id-class)
			(let ((codis (unicp~var-label-repres (first id-class))))
			  (if (and (not (unicp~var-label-theory codis))
				   (member codis newownedvar :test 'keim~equal)
				   (unicp~shared-variable-p codis))
			      (if (some #'(lambda(vec)
					    (and (> (elt vec (cpac=position co)) 0)
						 (> (elt vec (cpac=position codis)) 0)))
					(cpac=permanent-solutions up))
				  (push codis nov)))))
		    (unicp~var-label-discrim-lists co)))
	  constants)
    nov))


;;;
;;; Aenderung
;;;


(defun cpac=newov-dis (up newdiscriminated newownedvar)
  (declare (edited  "02-FEB-1996")
	   (authors Neskakis)
	   (input   "unification-problem [cpac+up]"
		    "A list of pairs of variables which must not be unified."
		    "newownedvar:  A list of variables which need not to be constants")
	   (effect  "None")
	   (value   "performs information-test9: $\ind(x)=AC$, if \sum_{\vec{s}\in T} \vec{s}(x) > 1, & \oor\\[3ex]"
                    "                            \vec{s}(x) > 0 & \fora \vec{s}\in T\ \with \vec{s}(y)>0, \ind(y)=E'."
		    "for every pair in the list of pairs of discriminated variables"
		    "and returns a list of variables which must not be constants")
	   (special cpac*theory))
  (let ((nov NIL))
    (mapc #'(lambda(dis)
	      (if (and (unicp~shared-variable-p (first dis))
		       (unicp~shared-variable-p (second dis)))
		  (if (and (cpac=constant-p (first dis))
			   (not (unicp~var-label-theory (second dis)))
			   (cpac=newov-dis-hlp-p up (first dis) (second dis) newownedvar))
		      (push (second dis) nov)
		      (if (and (cpac=constant-p (second dis))
			       (not (unicp~var-label-theory (first dis)))
			       (cpac=newov-dis-hlp-p up (second dis) (first dis) newownedvar))
			  (push (first dis) nov)))))
	  newdiscriminated)
    nov))
 
(defun cpac=newov-dis-hlp-p (up y x newownedvar)
  (declare (special cpac*theory))
  (if (not (member x newownedvar :test 'keim~equal))
      (some #'(lambda(vec)
		(and (> (elt vec (cpac=position x)) 0)
		     (> (elt vec (cpac=position y)) 0)))
	    (cpac=permanent-solutions up))
      NIL))
		    
		      






  


;+-----------------------------------------------------------------------------+
;| cpac=action                                                                 |
;| dies ist sozusagen die main-Fkt zur Berechnung neuer Informationen.Sie ruft |
;| die Hilfsfunktionen aus den ersten zwei Haelften in der richtigen Reihen-   |
;| folge auf, merkt sich die Informationen, die nach aussen gegeben werden und |
;| ruft sich selbst gegebenenfalls mit seinen eigenen neuen Informationen      |
;| erneut selber auf.                                                          |
;+-----------------------------------------------------------------------------+


(defun cpac=actions-main (up input-id input-dis input-co input-cr input-ov
			     &optional (output-id NIL) (output-ov NIL) (output-cr NIL) (output-dis NIL))
  (declare (edited  "12-JAN-1996")
	   (authors Neskakis)
	   (input   "unification-problem [cpac+up]"
		    "the input-XX-arguments can lead to new information"
		    "input-id: a list of identification of variables (i.e. sets of variables)"
		    "input-dis: a list of pairs of variables which must not be unified."
		    "input-co: a list of former variables which are now occupied by other theories"
		    "input-cr: a list of constant-restrictions"
		    "input-ov: a list of variables which must not be constants."
		    "the output-XX-arguments store the information produced through the recursion"
		    "of cpac=actions-main; at the end of the recursion this information is returned"
		    "output-id: a list of identification of variables (i.e. sets of variables)"
		    "output-ov: a list of variables which must not be constants."
		    "output-cr: a list of constant-restrictions"
		    "output-dis: a list of pairs of variables which must not be unified.")
	   (effect  "- sets >not*solvable< to >T< iff the unif.prob. is no longer solvable"
		    "- deletes vectors in cpac=solutions if they are not suitable for a"
		    "  correct subset (action1 - action5)"
		    "- adds vectors to cpac=permanent-solutions which have to be in the minimal"
		    "  subset of solutions")
	   (value   "see unicp~new-constr")
	   (special cpac*theory not*solvable))

  (let ((newid  NIL)
	(newov  NIL)
	(newcr  NIL)
	(newdis NIL))
    
    (cpac=actions up input-id input-dis input-co input-cr input-ov output-cr output-dis)
     
    (if (null not*solvable)
	(multiple-value-setq (newid newov newcr newdis)
	  (cpac=newinformations up input-dis input-co input-ov
				output-id output-ov output-cr output-dis)))

    (if (null not*solvable)
	  (if (or newid newov newdis)
	      (cpac=actions-main up newid newdis NIL NIL newov
				 (append newid  output-id)
				 (append newov  output-ov)
				 (append newcr  output-cr)
				 (append newdis output-dis))
	      (values up
		      output-id
		      output-dis
		      output-ov
		      NIL
		      (append newcr output-cr)
		      NIL))
	  NIL)
    ))





(defun cpac=actions (up input-id input-dis input-co input-cr input-ov output-cr output-dis)

  (declare (special cpac*theory not*solvable))

  (progn
      
    ;action1: delete $\vec{s}$ in $S$, if $\vec{s}(x)>1$, $x\notin AC$.
    (if input-co
	(cpac=constant-gt-one! up input-co))
    (when cpac*debug (format cpac*debug "act1 not*solvable: ~A~%" not*solvable))

      
    ;action2: delete $\vec{s}$ in $S$, if $\vec{s}(x)>0$, $\vec{s}(y)>0$, $x\dotls y$,
    ;$y\notin AC$, $x\in AC$.
    ;action2 is split up into three functions: cpac=const-restr-newcr!
    ;cpac=const-restr-newconst! cpac=const-restr-newvar!
    (if (and (null not*solvable) input-co)
	(cpac=const-restr-newconst! up input-co output-dis))
    (when cpac*debug (format cpac*debug "act2a not*solvable: ~A~%" not*solvable))
    (if (and (null not*solvable) input-cr)
	(cpac=const-restr-newcr! up input-cr output-dis))
    (when cpac*debug (format cpac*debug "act2b not*solvable: ~A~%" not*solvable))
    (if (and (null not*solvable) input-ov)
	(cpac=const-restr-newvar! up input-ov output-dis))
    (when cpac*debug (format cpac*debug "act2c not*solvable: ~A~%" not*solvable))
    (if (and (null not*solvable) input-dis)
	(cpac=const-restr-newdis! up input-dis output-cr))
    (when cpac*debug (format cpac*debug "act2d not*solvable: ~A~%" not*solvable))
 

    ;action3: delete $\vec{s}$ in $S$, if $\vec{s}(x)>0$, $\vec{s}(y)>0$, $x\not\doteqv
    ;y$, $x\notin AC$, $y\notin AC$.
    (if (and (null not*solvable) input-co)
	(cpac=twoconst-gt-zero-co! up input-co output-dis))
    (when cpac*debug (format cpac*debug "act3a not*solvable: ~A~%" not*solvable))
    (if (and (null not*solvable) input-dis)
	(cpac=twoconst-gt-zero-dis! up input-dis))
    (when cpac*debug (format cpac*debug "act3b not*solvable: ~A~%" not*solvable))


    
    #|
    ;action4: delete $\vec{s}$ in $S$, if $\vec{s}(x)\not=\vec{s}(y)$, $x\doteqv y$,
    ;$x\notin AC$, $y\notin AC$.
    (if (and (null not*solvable) input-id)
	(cpac=neq-constants-newid! up input-id))
    (if (and (null not*solvable) input-co)
	(cpac=neq-constants-newco! up input-co input-id))
    (when cpac*debug (format cpac*debug "act4 not*solvable: ~A~%" not*solvable))
    |#
    



    ;action5: delete $\vec{s}$ in $S$, if $\exists\vec{t}\in T$, $\vec{s}\not=\vec{t}$,
    ;$\vec{t}(x)>0$, $\vec{s}(x)>0$, $x\notin AC$.
    (if (and (null not*solvable) input-co)
	(cpac=2vectors-oneconst-newconstants! up input-co))


    
    ;action6:  add $\vec{s}$ to $T$, if $\vec{s}(x)>0$, $\forall\vec{t}\in S$,
    ;$\vec{s}\not=\vec{t} \leadsto \vec{t}(x)=0$.
    (if (null not*solvable)
	(cpac=add-to-permanents! up))


          
    (when cpac*debug (format cpac*debug "Solutions: ~A~%" (cpac=solutions up)))
    (when cpac*debug (format cpac*debug "Permanent solutions: ~A~%" (cpac=permanent-solutions up)))

    
    ))




(defun cpac=newinformations (up input-dis input-co input-ov
			     &optional (output-id NIL) (output-ov NIL) (output-cr NIL) (output-dis NIL))

  (declare (special cpac*theory not*solvable))

  (let ((newid  NIL)
	(newov  NIL)
	(newcr  NIL)
	(newdis NIL))
    
      ;newinformation1: $x\in AC$, if $\sum_{\vec{s}\in T} \vec{s}(x)>1$.
      (if (null not*solvable)
	  (setf newov (append (cpac=variable-to-ac up output-ov) newov)))

      ;newinformation3: $x\doteqv y$, if $\exists\vec{s}\in T$, $\vec{s}(x)=1$, $\vec{s}(y)=1$,
      ;$x\notin AC$, $y\notin AC$.
      (if (null not*solvable)
	  (setf newid (append (cpac=new-identified-1 up output-id) newid)))

      ;newinformation4:  $x\dotle y$, if $c_S(x)\subseteq c_S(y)$ ($\forall\vec{s}\in S$, $\vec{s}(x)\le\vec{s}(y)$)
      ;+ newinformation5:  $x\dotls y$, if $c_S(x)\subset c_S(y)$ ($\forall\vec{s}\in S$,
      ;  $\vec{s}(x)\le\vec{s}(y)$, $\exists\vec{s}\in T$,
      ;  $\vec{s}(x)\not=\vec{s}(y)$).
      ;+ newinformation7:  $x\in AC$, $y\dotls x$, $z\dotls x$, if $c_S(y)\subset c_S(x)$,
      ;  $c_S(z)\subset c_S(x)$, $y\not=z$, $c_S(y)\cap c_S(z)=\emptyset$.
      (if (null not*solvable)
	  (let ((inf (cpac=subset-information up output-cr output-id output-dis
					      input-ov)))
	      (setf newcr (append (first inf) newcr))
	      (setf newdis (append (second inf) newdis))
	      (setf newov (append (third inf)  newov))))

 
      ;newinformation6: {\bf not solvable}, if $\sum_{\vec{s}\in S} \vec{s}(x)=0$.
      (if (null not*solvable)
	  (cpac=spalte up))

      (when cpac*debug (format cpac*debug "act6 not*solvable: ~A~%" not*solvable))
      
      ;newinformation9:  $\ind(x)=AC$, if \sum_{\vec{s}\in T} \vec{s}(x) > 1, & \oor\\[3ex]
      ;\vec{s}(x) > 0 & \fora \vec{s}\in T\ \with \vec{s}(y)>0, \ind(y)=E'.
      (if (and (null not*solvable) input-co)
	  (setf newov (append (cpac=newov-co up input-co output-ov) newov)))

      (if (and (null not*solvable) input-dis)
	  (setf newov (append (cpac=newov-dis up input-dis output-ov) newov)))      

      (values newid newov newcr newdis)
      
      ))





;+-----------------------------------------------------------------------------------+
;| Hilfsfunktionen, die dazu dienen die Informationen, die von aussen kommen         |
;| aufzubereiten                                                                     |
;+-----------------------------------------------------------------------------------+

(defun cpac=check-identified (identified)
  (remove-if
   #'(lambda(elem)
       (notany #'cpac=position elem))
   identified))

(defun cpac=check-discriminated (discriminated)
  (remove-if
   #'(lambda(elem)
       (notevery #'cpac=position elem))
   discriminated))

(defun cpac=check-constants (constants)
  (remove-if-not #'cpac=position constants))

(defun cpac=check-const-restr (const-restr)
  (let ((newcr NIL))
  (mapc #'(lambda(cr)
	    (if (cpac=position (car cr))
		(let ((hlp (remove-if-not #'cpac=position cr)))
		  (if (> (length hlp) 1)
		      (push hlp newcr)))))
	    const-restr)
  newcr))
		




;+-----------------------------------------------------------------------------------+
;| Methoden mittels derer die AC-Unifikation von aussen gestartet und neu aufgerufen |
;| wird                                                                              |
;+-----------------------------------------------------------------------------------+


(defmethod unicp~up-save-choice-point ((up cpac+up))
  (push (list (cpac=solutions up)
	      (cpac=permanent-solutions up)
	      (cpac=arguments up)
	      (cpac=solutions-p up))
	(cpac=choice-points up)))

(defmethod unicp~up-restore-choice-point ((up cpac+up))
  (setf (cpac=solutions up)
	(first (first (cpac=choice-points up)))
	(cpac=permanent-solutions up)
	(second (first (cpac=choice-points up)))
	(cpac=arguments up)
	(third (first (cpac=choice-points up)))
	(cpac=solutions-p up)
	(fourth (first (cpac=choice-points up))))
  (pop  (cpac=choice-points up))
  up)

(defmethod unicp~theo-init ((cpac*theory th+ac) general-up equations
			    identified discriminated constants const-restr)
  (declare (edited  "13-DEC-1995")
	   (authors Neskakis)
	   (input   "cpac*theory: the theoryindex of the unification problem"
		    "general-up: "
		    "equations:  a list of pure equations (i.e. pairs of terms)"
		    "identified: list of pairs of variables which are identical"
		    "constants: a list of former variables which are now occupied by other theories"
		    "const-restr: a list of constant-restrictions")
	   (effect  "a unification problem [cpac+up] is created"
	            "+ effect as in unicp~new-constr")
	   (value   "value as in unicp~new-constr")
	   (special cpac*theory))
  (let ((up (make-instance 'cpac+up
			   :general-up general-up
			   :equations equations
			   :theory cpac*theory)))
      (cpac=new-equations! up)
      #+weg(cpac=new-solutions! up)
      #+weg(cpac=new-constr up
			(cpac=check-identified identified)
			(cpac=check-discriminated discriminated)
			(cpac=check-constants constants)
			(cpac=check-const-restr const-restr))))




(defmethod unicp~new-constr ((up cpac+up) identified discriminated constants const-restr)
  (declare (edited  "13-DEC-1995")
	   (authors neskakis)
	   (input   "unification-problem [cpac+up]"
		    "identified: list of pairs of variables which are identical"
		    "constants: a list of former variables which are now occupied by other theories"
		    "const-restr: a list of constant-restrictions")
	   (effect  "")
	   (value   "The elementary algorithm for the theory ELEMENTARY-UP"
		    "belongs to is called. It returns NIL if the unification"
		    "problem is not solvable with these new restrictions."
		    "Otherwise a multiple value is returned that contains"
		    "some new constraints wich must be met if the unification"
		    "problem should be solvable:"
		    "1.*UP, indicating that the unification problem is still"
		    "   solvable."
		    "2.*A list of new identification of variables (i.e. sets"
		    "   of variables)."
		    "3.*A list of pairs of variables which must not be"
		    "   unified."
		    "4.*A list of variables which must not be constants."
		    "5. A list of variables which need not to be variables."
		    "6.*A list of variables $(x y_1 ... y_n)$ where"
		    "   $\sigma(y_i)$ is a subterm of $\sigma(x)$ for a"
		    "   unifier $\sigma$."
		    "7. (A list of pairs $(x . y)$ of variables where no"
		    "   solution is lost if $y$ is in the constant"
		    "   restrictions of $x$)."))
  (let ((cpac*theory (unicp~theory up)))
    (declare (special cpac*theory))
    (let ((id  (cpac=check-identified identified))
	  (flag (cpac=solutions-p up)))
      (cond (id (cpac=new-equations! up))
	    (t
	     (unless flag
	       (cpac=new-solutions! up))
	     (let ((dis (cpac=check-discriminated discriminated))
		   (co  (if flag
			    (cpac=check-constants constants)
			    (remove-if-not #'(lambda (var)
					       (and (cpac=constant-p var)
						    (cpac=position var)))
					   (cpac=arguments up))))
		   (cr  (cpac=check-const-restr const-restr)))
	       (cpac=new-constr up id dis co cr)))))
    ))

(defmethod unicp~solvable-p ((up cpac+up))
  t)

(defmethod unicp~more-information-p ((up cpac+up))
  (not (cpac=solutions-p up)))

(defun cpac=new-constr (up identified discriminated constants const-restr)
  (declare (edited  "13-DEC-1995")
	   (authors Neskakis)
	   (input   "unification-problem [cpac+up]"
		    "identified: list of pairs of variables which are identical"
		    "constants: a list of former variables which are now occupied by other theories"
		    "const-restr: a list of constant-restrictions")
	   (effect  "see unicp~new-constr")
	   (value   "see unicp~new-constr")
	   (special cpac*theory))
  (let ((not*solvable NIL))
    (declare (special not*solvable))
    (cpac=actions-main up identified discriminated constants const-restr NIL)))









