
(in-package "KEIM")



(defun evac=sortiere (args &optional (var_as_const NIL))
  (unwind-protect
      (progn
	(ac=mark-var-as-const var_as_const)
	(let ((var   NIL)
	      (const NIL))
	  (progn
	    (mapc #'(lambda(arg)
		      (if (ac=constant-p arg)
			  (push arg const)
			  (push arg var)))
		  args)
	    (list var const))))
    (basis~labels-reset! args))) 



(defun evac=eingabe (tl1 tl2 var_as_const)
  (let*
      ((argl1   (basis=tl-to-argl tl1))
       (argl2   (basis=tl-to-argl tl2))
       (allargs (basis=args (append argl1 argl2)))
       (args    (evac=sortiere allargs var_as_const))
       (var     (car args))
       (const   (cadr args))
       (vecv    (cdr (basis=diol-matrix argl1 argl2 var allargs)))
       (vecc    (cdr (basis=diol-matrix argl1 argl2 const allargs)))
       )
    (values var vecv const vecc)))

(defun evac=baugleichung (vec varstelle varwert stream)
  (do ((i (1- (length vec)) (1- i)))
      ((< i 0) (format stream " = ~A;~%" varwert))
      (let*  ((wert    (elt vec i))
	      (zeichen (evac=plus-minus wert))
	      (abswert (abs wert)))
	(case abswert
	      (0 NIL) 
	      (1 (format stream " ~A x_~A_~A" zeichen i varstelle))
	      (t (format stream " ~A ~A x_~A_~A" zeichen abswert i varstelle))))))


(defun evac=plus-minus (zahl)
  (if (>= zahl 0) "+" "-"))


(defun evac=constrestr (var const crestr stream)
  (unwind-protect
      (progn
	(do ((i (1- (length var)) (1- i)))
	    ((< i 0))
	    (sym~set-label! (elt var i) i))
	(do ((i (1- (length const)) (1- i)))
	    ((< i 0))
	    (sym~set-label! (elt const i) (1+ i)))
	;(mapc #'(lambda(l) (format stream " ~A:~A ~%" l (sym~label l))) (append var const))
	(do ((i (1- (length crestr)) (1- i)))
	    ((< i 0))
	    (do ((j (1- (length (elt crestr i))) (1- j)))
		((< j 1))
		(format stream " x_~A_~A <= 0;~%" (sym~label (elt (elt crestr i) j)) (sym~label (elt (elt crestr i) 0)))))
	)
    (mapc #'(lambda(l) (sym~set-label! l NIL)) (append var const))))


(defun evac=patch-lpsolve! (tl1 tl2 crestr var_as_const  &optional (inner-subst NIL))
  (let ((subst (some #'(lambda(t1 t2)
			 (if (and (sym~variable-p t1)
				  (not (member t1 var_as_const :test 'keim~equal))
				  (not (member t1 (top~free-variables t2) :test 'keim~equal)))
			     (subst~create (list t1) (list t2))
			     (if (and (sym~variable-p t2)
				      (not (member t2 var_as_const :test 'keim~equal))
				      (not (member t2 (top~free-variables t1))))
				 (subst~create (list t2) (list t1))
				 NIL)))
		     tl1 tl2))
	(tl1new NIL)
	(tl2new NIL)
	(crestrnew NIL))
    (if subst
	(progn
	  (setf tl1new (mapcar #'(lambda(term) (subst~apply subst term)) tl1))
	  (setf tl2new (mapcar #'(lambda(term) (subst~apply subst term)) tl2))
	  (setf crestrnew (mapcar #'(lambda(cr)
				      (mapcar #'(lambda(arg) (subst~apply subst arg)) cr))
				  crestr))
	  (let ((tl1-help NIL)(tl2-help NIL))
	    (mapc #'(lambda(t1 t2) (if (not (keim~equal t1 t2))
				       (progn
					 (push t1 tl1-help)
					 (push t2 tl2-help))))
		  tl1new tl2new)
	    (setf tl1new tl1-help
		  tl2new tl2-help))
	  (if inner-subst
	      (setf inner-subst (subst~compose-substitution subst inner-subst))
	      (setf inner-subst subst))
	  (evac=patch-lpsolve! tl1new tl2new crestrnew var_as_const inner-subst))
	(values tl1 tl2 crestr inner-subst))))
    
	  
		     
	      
			     



(defun evac~baueingabe (terml1 terml2 constrestr var_as_const stream)
  (declare (edited  "24-JUL-1995" "16-JAN-1996")
	   (authors Richts Neskakis)
	   (input  "Two lists of terms, a list of constant restrictions, a list"
		   "of constant variables, and an output stream")
	   (effect "Prints input for the program lpsolve onto STREAM")
	   (value  "The lists of the variables and of the constants in TL1 and TL2,"
		   "or NIL if for one equation the gcd of the variable"
		   "koefficients is not a divisor of one constant koefficient (i.e. the problem is"
		   "unsolvable).  This is necessary because lp_solve would"
		   "result in a segmentation fault in this case.")
	   (special first*subst))
  (multiple-value-bind (tl1 tl2 crestr subst)
      (evac=patch-lpsolve! terml1 terml2 constrestr var_as_const)

    (setf first*subst subst)
    
    (multiple-value-bind (var vecv const vecc)
	(evac=eingabe tl1 tl2 var_as_const)
      
      (if (or (some #'(lambda (var-vector const-vector)
			(let ((divisor (apply #'gcd var-vector)))
			  (some #'(lambda (const-koeff)
				    (if (zerop divisor)
					(not (zerop const-koeff))
					(not (zerop (mod const-koeff divisor)))))
				const-vector)))
		    vecv vecc)
	      (some #'(lambda (var-vector const-vector)
			(and (every #'zerop var-vector)
			     (notevery #'zerop const-vector)))
		    vecv vecc))
					;then
	  (values nil nil)
	  
					;else
	  (progn
	    
	    (format stream " min: x;~%~%")
	    
	    (let (c)
	      (mapc #'(lambda (var-vector const-vector)
			(when (notevery #'zerop var-vector)
			  (evac=baugleichung var-vector (setf c 0) 0 stream)
			  (mapc #'(lambda (const-koeff)
				    (evac=baugleichung var-vector (incf c) (* -1 const-koeff) stream))
				const-vector)))
		    vecv vecc))
	    
	    (do ((v (1- (length var)) (1- v)))
		((< v 0) (format stream "~%"))
	      (do ((c (length const) (1- c)))
		  ((< c 0) (format stream " >= 1;~%"))
		(format stream " + x_~A_~A" v c)))
	    
	    (evac=constrestr var const crestr stream)
	    
	    (format stream "~%")
	    
	    (format stream " int")
	    
	    (do ((v (1- (length var)) (1- v)))
		((< v 0) (format stream ";~%"))
	      (do ((c (length const) (1- c)))
		  ((< c 0) 1)
		(format stream " x_~A_~A" v c)))
	    
	    (values var const)))
      )))







;(defmethod th~decide-unify ((th th+ac) tl1 tl2 restr const-vars)
;  (evac~lpsolve tl1 tl2 restr const-vars))






(defun evac~lpsolve (tl1 tl2 &optional (crestr NIL) (var_as_const NIL))
  (declare (edited  "24-JUL-1995" "21-APR-1995")
	   (authors Richts Neskakis)
	   (input   "2 lists of terms, optional list of restricted constants")
	   (effect  "calls lpsolve")
	   (value   "returns a substitution if the unification-problem is solvable, else NIL"))
  (let ((lp-solve (run-program "/usr/local/keim/bin/lp_solve"
			       :input :stream
			       :output :stream
			       :error-output :output
			       :wait nil))
	(lines nil) (var nil) (const nil)
	(first*subst NIL))
    (declare (special first*subst))
    (unwind-protect
	 (progn
	   (unwind-protect
		(multiple-value-setq (var const)
		  (evac~baueingabe tl1 tl2 crestr var_as_const lp-solve))
	     (close (underlying-stream lp-solve :output)))
	   
	   (cond ((and (null var) (null const))  ; evac~baueingabe detected that 
		  nil)                           ; the problem is unsolvable
		 ((null var) (subst~create nil nil))
		 (t
		  (do ((line (read-line lp-solve nil lp-solve)
			     (read-line lp-solve nil lp-solve)))
		      ((eq line lp-solve) (setf lines (nreverse lines)))
		      (push line lines))
		  (evac~eval lines var const (basis=give-fsym tl1 tl2) first*subst))))
      (close (underlying-stream lp-solve :input)))
    ))



(defun evac~eval (lines varliste constliste fsym subst)
  (cond ((endp lines)
	 (error " KEIM(evac) : Unknown output of lp-solve "))
	((or (string-equal "(M)ILP problem has no solution, problem is infeasible" (car lines))
	     (search "contradicts earlier bounds, exiting" (car lines))
	     (search "variables must always be non-negative" (car lines)))
	 NIL)
	((or (string-equal "Error, you are trying to solve a problem without real constraints"
			   (car lines)))
	 (subst~create nil nil))
	((string-equal "Value of objective function:" (car lines)
		       :end2 (min 28 (length (car lines))))
	 (unwind-protect
	     (evac~buildsub (cddr lines) varliste constliste fsym subst)
	   (basis~labels-reset! varliste)))
	(t
	 (evac~eval (cdr lines) varliste constliste fsym subst))))

 
(defun evac~buildsub (lines varliste constliste fsym subst)
  (progn
    (basis~labels-reset! varliste)
    (push (fo~variable-create) constliste)
    ;(pprint lines)
    (mapc #'(lambda(l) (evac=parseline l varliste constliste)) lines)
    ;(mapc #'(lambda(v) (format T "~% ~A --> ~A" v (sym~label v))) varliste)
    (if subst
	(subst~compose-substitution
	 (subst~create varliste
		       (mapcar #'(lambda(var) (evac=build-rec (sym~label var) fsym)) varliste))
	 subst)
	(subst~create varliste
		       (mapcar #'(lambda(var) (evac=build-rec (sym~label var) fsym)) varliste)))
    ))
 
(defun evac=build-rec (subliste fsym)
  (if (= 1 (length subliste))
      (car subliste)
     (evac=build-rec (cons
		       (appl~create fsym (list (car subliste) (cadr subliste)))
		       (cddr subliste))
		      fsym)))
      



      

(defun evac=parseline (line vliste cliste)
  (let* ((pos1  (1+ (position #\_ line)))
	 (pos2  (1+ (position #\_ line :start pos1)))
	 (pos3  (position #\space line))
	 (pos4  (1+ (position #\space line :from-end T)))
	 (pos5  (length line))
	 (var   (evac=string-int line pos1 (1- pos2)))
	 (const (evac=string-int line pos2 pos3))
	 (times (evac=string-int line pos4 pos5)))
    ;(format T "~% ~A ~A ~A" var const times)
    (do ((i 1 (1+ i))) ((> i times))
	(sym~set-label! (elt vliste var) (cons (elt cliste const) (sym~label (elt vliste var)))))
    ))

(defun evac=string-int (line pos1 pos2)
  (declare (edited  "21-APR-1995")
	   (authors Neskakis)
	   (input   "a string and two positions")
	   (effect  NONE)
	   (value   "transforms the char from pos1 up to pos2 into numbers and adds them together"))
    (do ((i pos1 (1+ i)) (j (- (char-int (elt line pos1)) 48) (+ j (- (char-int (elt line (1+ i))) 48)))) 
        ((>= i (1- pos2)) j)))














