
;******************************************************************************
;* Test                                                                       *
;******************************************************************************

(in-package :keim)

;+----------------------------------------------------------------------------+
;| universe                                                                   |
;+----------------------------------------------------------------------------+

(defvar env (env~create))      ; universe

(post~read-object-list '((type-constants i o)
			 (constants (f (i i i))
				    (a i) (b i) (c i) (d i)
				    (c1 i) (c2 i))
			 (variables (u i) (v i) (x i) (y i) (z i)
				    (v1 i) (v2 i)))
		       env)
                               ; universe filled with individuals


;+----------------------------------------------------------------------------+
;| terms                                                                      |
;+----------------------------------------------------------------------------+

(defvar a (term~read 'a env))          ; constants 
(defvar b (term~read 'b env))
(defvar c (term~read 'c env)) 
(defvar c1 (term~read 'c1 env))
(defvar c2 (term~read 'c2 env))
(defvar d (term~read 'd env))
(defvar u (term~read 'u env))          ; variables
(defvar v (term~read 'v env))
(defvar v1 (term~read 'v1 env))
(defvar v2 (term~read 'v2 env))
(defvar x (term~read 'x env))
(defvar y (term~read 'y env))
(defvar z (term~read 'z env))

(defvar ab (term~read '(f a b) env))            ; applications
(defvar abc (term~read '(f a (f b c)) env))
(defvar ax (term~read '(f a x) env))
(defvar axb (term~read '(f a (f x b)) env))
(defvar axy (term~read '(f a (f x y)) env))
(defvar ayxz (term~read '(f (f a y) (f x z)) env))
(defvar az (term~read '(f a z) env))
(defvar bc (term~read '(f b c) env))
(defvar by (term~read '(f b y) env))
(defvar cd (term~read '(f c d) env))
(defvar c1c2u (term~read '(f c1 (f c2 u)) env))
(defvar uv (term~read '(f u v) env))
(defvar uva (term~read '(f u (f v a)) env))
(defvar uvcd (term~read '(f (f u v) (f c d)) env))
(defvar uv1cd (term~read '(f (f u v1) (f c d)) env))
(defvar v2bc (term~read '(f v2 (f b c)) env))
(defvar xay (term~read '(f x (f a y)) env))
(defvar xuv (term~read '(f x (f u v)) env))
(defvar xuv1 (term~read '(f x (f u v1)) env))
(defvar xyv2b (term~read '(f (f x y) (f v2 b)) env))
(defvar xy (term~read '(f x y) env))
(defvar xyab (term~read '(f (f x y) (f a b)) env))
(defvar xyb (term~read '(f x (f y b)) env))
(defvar xz (term~read '(f x z) env))
(defvar yuac (term~read '(f (f y u) (f a c)) env))
(defvar yuv2c (term~read '(f (f y u) (f v2 c)) env))
(defvar yy (term~read '(f y y) env))
(defvar zv1 (term~read '(f z v1) env))
(defvar zv2 (term~read '(f z v2) env))
(defvar zxv (term~read '(f z (f x v)) env))
(defvar zxv1 (term~read '(f z (f x v1)) env))

;+----------------------------------------------------------------------------+
;| constant - restrictions                                                    |
;+----------------------------------------------------------------------------+

(defvar ra-u   (list a a u))     
(defvar ra-xy  (list a x y))
(defvar rb-u   (list b b u))
(defvar rb-x   (list b x))
(defvar rb-xy  (list b x y))
(defvar rb-y   (list b y))
(defvar rb-yz  (list b y z))
(defvar rd-z   (list d d z))

;+----------------------------------------------------------------------------+
;| test - functions                                                           |
;+----------------------------------------------------------------------------+

(defun test~test (key)
  (case key
   ((1) (aci~unif (list (cons xay az) (cons ayxz y) (cons az b)) nil nil))  
   ((2) (aci~unif (list (cons xay az) (cons ayxz y) (cons yy u)) nil nil))  
   ((3) (aci~unif (list (cons b a)) nil nil))
   ((4) (aci~unif (list (cons a xyb)) nil nil))
   ((5) (aci~unif (list (cons ab xy) (cons cd xz)) nil nil))
   ((6) (aci~unif (list (cons ab xy) (cons xy zv1)
			(cons uv zv2) (cons cd uv)) nil nil))
   ((7) (aci~unif (list (cons ab xy) (cons xy zv1)
			(cons uva zv2) (cons cd uv)) nil nil))
   
   ((8) (aci~unif (list (cons axy axb)) (list ra-xy) nil))
   ((9) (aci~unif (list (cons axy axb)) (list rb-y) nil))
   ((10) (aci~unif (list (cons axy axb)) (list rb-x) nil))
   ((11) (aci~unif (list (cons axy axb)) (list rb-xy) nil))

   ((12) (aci~unif (list (cons ab xy) (cons bc xz)) (list rb-yz) nil))
   ((13) (aci~unif (list (cons ab xy) (cons bc xz)) (list rb-x) nil))

   ((14) (aci~unif (list (cons ax by)) (list rb-x) nil))
   ((15) (aci~unif (list (cons ax by)) nil nil))
   
   ((16) (aci~unif (list (cons xyab uvcd)
			 (cons xuv abc)
			 (cons zxv yuac)) nil nil))
   ((17) (aci~unif (list (cons xyab uvcd)
			 (cons xuv abc)
			 (cons zxv yuac)) (list rd-z) nil))
   ((18) (aci~unif (list (cons xyab uvcd)
			 (cons xuv abc)
			 (cons zxv yuac)) (list ra-u rb-u) nil))
   
   ((19) (aci~unif (list (cons xyab uvcd) (cons xuv abc)
			 (cons zxv yuac)  (cons xy c1c2u)) nil nil))

   ((20) (aci~unif (list (cons ab xy)
			 (cons xy uv)
			 (cons uv cd)) nil nil))
   ((21) (aci~unif (list (cons xyv2b uv1cd)
			 (cons xuv1 v2bc)
			 (cons zxv1 yuv2c)) nil (list v2)))
   ))

(defvar test**max 22)                      ; maximum of test-functions + 1
(defvar test**result (make-array (list test**max)))
					   ; array of exspected results

(setf (aref test**result 1) nil)           ; exspected results
(setf (aref test**result 2) T)
(setf (aref test**result 3) nil)
(setf (aref test**result 4) nil)
(setf (aref test**result 5) nil)
(setf (aref test**result 6) nil)
(setf (aref test**result 7) T)
(setf (aref test**result 8) T)
(setf (aref test**result 9) T)
(setf (aref test**result 10) T)
(setf (aref test**result 11) nil)
(setf (aref test**result 12) T)
(setf (aref test**result 13) nil)
(setf (aref test**result 14) nil)
(setf (aref test**result 15) T)
(setf (aref test**result 16) T)
(setf (aref test**result 17) nil)
(setf (aref test**result 18) nil)
(setf (aref test**result 19) nil)
(setf (aref test**result 20) nil)
(setf (aref test**result 21) T)
 
(defun test~all ()                         ; compares results with exspected
  (format T "~%")                          ; results
  (loop for i from 1 to 21 do
	(format T "~A" (equal (test~test i) (aref test**result i))))
  (values))
  
;-----------------------------------------------------------------------------

(defun test~label-check ()     ; list of labels of atomic terms
  (mapcar #' sym~label (list a b c c1 c2 d u v v1 v2 x y z)))

(defun test~label-reset! ()    ; reset of labels of atomic terms
  (mapc #' (lambda (term) (sym~set-label! term nil))
	   (list a b c c1 c2 d u v v1 v2 x y z))
  (values))


(defun test~binding-check ()     ; list of labels of atomic terms
  (mapcar #' sym~symbol-binding (list a b c c1 c2 d u v v1 v2 x y z)))

(defun test~binding-reset! ()    ; reset of labels of atomic terms
  (mapc #' (lambda (term) (top~bind-term! term nil))
	   (list a b c c1 c2 d u v v1 v2 x y z))
  (values))
;-----------------------------------------------------------------------------


(defun test~unifier-correct (el var-constlist unifier sel)

  ; applys unifier to equation list
  ; prints equation list after substitution
  ; computes new symset to prove correctness of unifier
  
  (mapc #' (lambda (eq)
	     (setf (car eq) (subst~apply unifier (car eq)))
	     (setf (cdr eq) (subst~apply unifier (cdr eq))))
	   el)
  (test=el-print el)
  (symset~slot-reset! sel)
  (symset~create el var-constlist))
	     
(defun test=el-print (el)
  (mapc #' (lambda (eq)
	     (format T "~%~A = ~A" (car eq) (cdr eq)))
	   el))
	     
	     









 
