(in-package "KEIM")

     ;;; this file contains three modules
     ;;;  i)    the original algorithm (all parts named bool=alg... )
     ;;;  ii)   the constant elimination algorithm (named bool=ce... )
     ;;;  iii)  the normalization algorithm (named bool=norm... )
     ;;; 
     ;;;
     ;;; The main function is bool~unify
     ;;; This function needs three arguments:
     ;;;   1) a term or a list of terms (where the terms are unified pairwise
     ;;;                                  (even number of terms!!!))
     ;;;   2) a list of variables, which are used like constants
     ;;;   3) a list of restrictions
     ;;;  


     ;;;  Part I: the boolean unification algorithm
     ;;;
     ;;;  unify(f(x~)) =
     ;;;       if x~ = ()
     ;;;       then if f(x~) = 0 then () else fail
     ;;;       else let G = unify(f(0,y~)*f(1,y~))
     ;;;            in ((f(0,G(y~))+f(1,G(y~))+1)*x1 + f(0,G(y~)),G(y~))


(defmethod th~decide-unify ((th th+br) tl1 tl2 restr const-vars)
  (let ((unifier (bool~unify (mapcan #'list tl1 tl2) const-vars restr)))
    (if unifier
	t
	nil)))


(defmethod th~unify ((th th+br) tl1 tl2 restr const-vars)
  (let ((unifier (bool~unify (mapcan #'list tl1 tl2) const-vars restr)))
    (if unifier
	(list unifier)
	nil)))


(defun bool~unify (term-list var-as-const restrictions)
  (declare (edited  "18-JUL-1997")
	   (authors Hendrik Loehn)
	   (input   "a list of terms.")
	   (effect  "none.")
	   (value   "a substitution."))
  (let* ((theory (if (eq (class-of (th~term-theory (car term-list))) (th~theory-class 'br))
		     (th~term-theory (car term-list))
		     (th~term-theory (cadr term-list))))
	 (bool*add (th~symbol theory :plus))
	 (bool*mul (th~symbol theory :mal))
	 (bool*one (th~symbol theory 1))
	 (bool*zero (th~symbol theory 0))
	 (bool*vac var-as-const))
    (declare (special bool*add))
    (declare (special bool*mul))
    (declare (special bool*one))
    (declare (special bool*zero))
    (declare (special bool*vac))
    (let ((subst (cond ((null (cdr term-list)) (bool=alg (bool=alg-simplify (bool=norm (car term-list)))))
		       ((null (cddr term-list)) (bool=alg
				    (bool=norm
				     (appl~create bool*add (cons (first term-list)
								   (list (second term-list)))))))
		       (t (let* ((norm-terms (bool=order-norm-and-conc-terms term-list))
				 (var-lists (bool=order-create-var-list norm-terms))
				 (new-term-list (bool=order norm-terms var-lists)))
			    (bool=unify-prepare new-term-list))))))
      (if (null restrictions) subst
	  (bool=const-elimination subst restrictions)))))


(defun bool=unify-prepare (term-list-list)
  (if (null term-list-list) (subst~create nil nil)
      (if (null (car (car  term-list-list))) (bool=unify-prepare (cdr term-list-list))
	  (let ((subst1 (bool=unify-prep1 (car term-list-list))))
	    (if (null subst1) nil
		(let ((subst2 (bool=unify-prepare (cdr term-list-list))))
		  (if (null subst2) nil
		      (subst~compose-substitution subst2 subst1))))))))


   (defun bool=unify-prep1 (term-list)
     (if (null (cdr term-list)) (bool=alg (car term-list))
	   (bool=unify-prep1 (cons (bool=alg-simplify (append
		   			        	  (append (first term-list)
			   			        	  (second term-list))
				   		          (bool=alg-new-mult (first term-list)
							          (second term-list))))
			         	  (cddr term-list)))))



     ;;; this part of the module is written for reuse
     ;;; it contains a function (bool=order), which is called with a list of variable-lists
     ;;; and a list of objects (no specific type) (both lists with the same length)
     ;;; the function creates a list of groups of objects, such that the variable-lists
     ;;; in each group are disjunct to the lists in other groups


(defun bool=order (term-list var-lists)
  (declare (edited  "31-JUL-1997")
	   (authors Hendrik Loehn)
	   (input   "a list of normalized terms, represented by lists.")
	   (effect  "none.")
	   (value   "a list of disjunct groups of terms in respect to their variables."))
  (mapcar #'cdr (bool=order1 term-list var-lists)))


   (defun bool=order1 (term-list var-lists)
     (if (null term-list) nil
         (let ((new-list (bool=order1 (cdr term-list) (cdr var-lists))))
	   (bool=order-insert-and-merge new-list (car term-list) (car var-lists)))))


(defun bool=order-insert-and-merge (work-list term var-list)
  (if (null work-list) (list (cons var-list (list term)))
      (let ((new-list (bool=order-insert-and-merge (cdr work-list) term var-list)))
	(if (bool=order-member-p (car (car work-list)) (car (car new-list)))
	    (cons (cons (bool=order-merge-var-list (car (car work-list)) (car (car new-list)))
		 	(append (cdr (car work-list))
				(cdr (car new-list))))
	         (cdr new-list))
	    (append new-list (list (car work-list)))))))


(defun bool=order-member-p (var-list-1 var-list-2)
  (if (null var-list-1) nil
      (if (member (car var-list-1) var-list-2 :test #'term~equal) T
	  (bool=order-member-p (cdr var-list-1) var-list-2))))


(defun bool=order-merge-var-list (var-list-1 var-list-2)
  (if (null var-list-1) nil
      (let ((new-list (or (bool=order-merge-var-list (cdr var-list-1) var-list-2)
			  var-list-2)))
	  (if (member (car var-list-1) new-list :test #'term~equal) new-list
	    (cons (car var-list-1) new-list)))))


(defun bool=order-norm-and-conc-terms (term-list)
  (if (null term-list) nil
      (cons (append
	       (bool=norm (car term-list))
	       (bool=norm (cadr term-list)))
       (bool=order-norm-and-conc-terms (cddr term-list)))))


(defun bool=order-create-var-list (term-list)
  (if (null term-list) nil
      (cons (bool=order-find-all-vars (car term-list))
	    (bool=order-create-var-list (cdr term-list)))))


(defun bool=order-find-all-vars (norm-term)
  (if (null norm-term) nil
      (let ((var-list (bool=order-find-all-vars (cdr norm-term))))
	(bool=order-find-new-vars var-list (car norm-term)))))


(defun bool=order-find-new-vars (var-list list)
  (declare (special bool*vac))
  (if (null list) var-list
      (if (and
	   (sym~variable-p (car list))
	   (and (not (member (car list) var-list :test #'term~equal))
		(not (member (car list) bool*vac :test #'term~equal))))
	  (bool=order-find-new-vars (cons (car list) var-list) (cdr list))
	  (bool=order-find-new-vars var-list (cdr list)))))

     ;;; end of bool=order
		

(defun bool=term-equal (list-1 list-2)
  (declare (edited  "11-JUL-1997")
	   (authors Hendrik Loehn)
	   (input  "Two lists representing a term." )
	   (effect  "none.")
	   (value   "T if lists are equal in respekt to "
		    "commutativity and associativity."))
  (if (null list-1)
      (if (null list-2) T
	  Nil)
      (if (member (car list-1) list-2 :test #'term~equal)
	  (bool=term-equal (cdr list-1) (bool=delete-term (car list-1) list-2))
	  Nil)))


(defun bool=delete-term (element list)
  (if (term~equal element (car list)) (cdr list)
      (cons (car list) (bool=delete-term element (cdr list)))))


(defun bool=alg (list)
  (declare (edited  "11-JUL-1997")
	   (authors Hendrik)
	   (input   "a list, representing a normalized term.")
	   (effect  "none.")
	   (value   "a substitution if possible, nil else."))
  (declare (special bool*zero))
  (declare (special bool*one))
  (let ((var (bool=alg-get-next-variable list)))
    (if (equal var nil) (if (term~equal (car (car list)) bool*zero) (subst~create nil nil)
			    nil)
	(let ((f0 (bool=alg-simplify (bool=alg-replace-var bool*zero var list)))
	      (f1 (bool=alg-simplify (bool=alg-replace-var bool*one var list))))
	  (let ((subst (bool=alg (bool=alg-simplify (bool=alg-new-mult f0 f1)))))
	    (if (equal subst nil) nil
		(subst~add-component var (bool=alg-create-subst-term f0 f1 var subst) subst)))))))


(defun bool=alg-simplify (list)
  (declare (edited  "11-JUL-1997")
	   (authors Hendrik Loehn)
	   (input   "a list, representing a normalized term.")
	   (effect  "none.")
	   (value   "a list, representing the simplified term."))
  (bool=alg-simpl-add (bool=alg-simpl-mul list)))


(defun bool=alg-simpl-mul (list)
  (declare (edited  "11-JUL-1997")
	   (authors Hendrik Loehn)
	   (input   "a list, representing a normalized term.")
	   (effect  "none.")
	   (value   "a list, representing the term after simplifying the multiplications"))
  (if (null list) list
      (cons (bool=alg-del-ones (remove-duplicates (bool=alg-get-zero (car list))
						      :test #'term~equal))
	    (bool=alg-simpl-mul (cdr list)))))


(defun bool=alg-get-zero (list)
  (declare (special bool*zero))
  (if (member bool*zero list :test #'term~equal) (list bool*zero)
      list))


(defun bool=alg-del-ones (list)
  (declare (special bool*one))
  (if (and (null (cdr list)) (term~equal bool*one (car list))) list
      (remove bool*one list :test #'term~equal)))


(defun bool=alg-simpl-add (list)
  (declare (edited  "11-JUL-1997")
	   (authors Hendrik Loehn)
	   (input   "a list, representing a normalized term.")
	   (effect  "none.")
	   (value   "a list, representing the term after simplifying the additions."))
  (bool=alg-get-non-zero (bool=alg-del-inverse list)))


(defun bool=alg-del-inverse (list)
  (declare (special bool*zero))
  (if (null (cdr list)) list
      (if (member (car list) (cdr list) :test #'bool=term-equal)
          (cons (list bool*zero)
	        (bool=alg-del-inverse (bool=alg-delete-member (car list) (cdr list))))
          (cons (car list) (bool=alg-del-inverse (cdr list))))))


(defun bool=alg-delete-member (element list)
  (if (bool=term-equal element (car list)) (cdr list)
      (cons (car list) (bool=alg-delete-member element (cdr list)))))


(defun bool=alg-get-non-zero (list)
  (declare (special bool*zero))
  (if (null (cdr list)) list
      (cond ((bool=term-equal (list bool*zero) (car list))
	     (bool=alg-get-non-zero (cdr list)))
	    ((bool=term-equal (list bool*zero) (cadr list))
	     (bool=alg-get-non-zero (cons (car list) (cddr list))))
	  (t (cons (car list) (bool=alg-get-non-zero (cdr list)))))))


(defun bool=alg-get-next-variable (list)
  (declare (edited  "11-JUL-1997")
	   (authors Hendrik Loehn)
	   (input   "a list, representing a normalized term.")
	   (effect  "none.")
	   (value   "the next free variable with the least occurencies in the term."))
  (car (bool=alg-gnv-1 (bool=alg-gnv-count list nil))))


(defun bool=alg-gnv-count (list var-list)
  (if (null list) var-list
      (bool=alg-gnv-count (cdr list) (bool=alg-gnv-count1 (car list) var-list))))


   (defun bool=alg-gnv-count1 (list var-list)
     (declare (special bool*vac))
     (if (null list) var-list
         (if (and (sym~variable-p (car list))
	          (not (member (car list) bool*vac :test #'term~equal)))
	     (bool=alg-gnv-count1 (cdr list) (bool=alg-gnv-enter (car list) var-list))
	     (bool=alg-gnv-count1 (cdr list) var-list))))


(defun bool=alg-gnv-enter (var var-list)
  (if (null var-list) (list (cons var (list '1)))
      (if (term~equal (car (car var-list)) var)
	  (cons (cons var (list (+ 1 (cadr (car var-list))))) (cdr var-list))
	  (cons (car var-list) (bool=alg-gnv-enter var (cdr var-list))))))


   (defun bool=alg-gnv-1 (var-list)
     (if (null var-list) nil
         (let ((lowest (car var-list)))
	   (bool=alg-gnv-search-lowest lowest (cdr var-list)))))


(defun bool=alg-gnv-search-lowest (p-lowest var-list)
  (if (null var-list) p-lowest
      (if (< (cadr (car var-list)) (cadr p-lowest))
	  (bool=alg-gnv-search-lowest (car var-list) (cdr var-list))
	  (bool=alg-gnv-search-lowest p-lowest (cdr var-list)))))
    

(defun bool=alg-replace-var (const var list)
  (declare (edited  "11-JUL-1997")
	   (authors Hendrik Loehn)
	   (input   "a constant, a variable and a list, representing a normalized term.")
	   (effect  "none")
	   (value   "a list, representing a normalized term in which all occurrencies"
		    "of the specified variable are replaced by the constant."))
  (if (null list) list
      (cons (bool=alg-repl-var1 const var (car list))
	    (bool=alg-replace-var const var (cdr list)))))

   (defun bool=alg-repl-var1 (const var list)
     (if (null list) list
         (if (term~equal var (car list))
	     (cons const (bool=alg-repl-var1 const var (cdr list)))
	     (cons (car list) (bool=alg-repl-var1 const var (cdr list))))))


(defun bool=alg-new-mult (list-1 list-2)
  (declare (edited  "11-JUL-1997")
	   (authors Hendrik Loehn)
	   (input   "two lists, representing normalized terms")
	   (effect  "none.")
	   (value   "a list, representing a normalized term, created by multiplying"
		    "the two original terms."))
  (if (or (null list-1) (null list-2)) list-1
      (append (bool=alg-new-mult1 (car list-1) list-2)
	      (bool=alg-new-mult (cdr list-1) list-2))))


   (defun bool=alg-new-mult1 (element list-2)
     (if (null list-2) list-2
         (cons (append element (car list-2))
	         (bool=alg-new-mult1 element (cdr list-2)))))


(defun bool=alg-create-new-term (list)
  (declare (edited  "11-JUL-1997")
	   (authors Hendrik Loehn)
	   (input   "a list, representing a normalized term.")
	   (effect  "none.")
	   (value   "the term, represented by a term+term"))
  (bool=alg-create-add list))


(defun bool=alg-create-add (list)
  (declare (special bool*add))
  (if (null (cddr list))
      (if (null (cdr list)) (bool=alg-create-mult (car list))
	  (appl~create bool*add (cons (bool=alg-create-mult (car list))
				      (list (bool=alg-create-mult (cadr list))))))
      (appl~create bool*add (cons (bool=alg-create-mult (car list))
				  (list (bool=alg-create-add (cdr list)))))))


(defun bool=alg-create-mult (list)
  (declare (special bool*mul))
  (if (null (cddr list))
      (if (null (cdr list)) (car list)
	  (appl~create bool*mul (cons (car list)
				      (list (cadr list)))))
      (appl~create bool*mul (cons (car list)
			          (list (bool=alg-create-mult (cdr list)))))))


(defun bool=alg-create-subst-term (f0 f1 var subst)
  (declare (edited  "11-JUL-1997")
	   (authors Hendrik Loehn)
	   (input   "two terms, a variable and a substitution.")
	   (effect  "none.")
	   (value   "a term+term, created as (f0+f1+1)*var+f0)."))
  (bool=alg-create-new-term
   (bool=alg-simplify
    (append
     (append (bool=alg-new-mult (list (list var)) (bool=alg-apply-subst f0 subst))
	     (bool=alg-new-mult (list (list var)) (bool=alg-apply-subst f1 subst)))
     (cons (list var)
	   (bool=alg-apply-subst f0 subst))))))


(defun bool=alg-apply-subst (list subst)
  (declare (edited  "11-JUL-1997")
	   (authors Hendrik Loehn)
	   (input   "a list, representing a normalized term and a substitution.")
	   (effect  "none.")
	   (value   "a list, representing a new term, generated by applying"
		    "the substitution to the old term."))
  (if (null list) nil
      (append (bool=alg-app-sub1 (car list) subst) (bool=alg-apply-subst (cdr list) subst))))

   (defun bool=alg-app-sub1 (list subst)
     (if (null list) nil
         (bool=alg-new-mult (let ((zwischen (subst~apply subst (car list))))
		          (if (term~p zwischen) (bool=norm zwischen)
			      zwischen))
		        (bool=alg-app-sub1 (cdr list) subst))))





     ;;;  Part II : The constant elimination algorithm
     ;;;
     ;;;  c0 = set of all constants to be eliminated
     ;;;  v0 = set of all variables occuring in the terms of the problems
     ;;;  D  = set of all possible products of elements in c0 including "1"


(defun bool=const-elimination (subst restrictions)
  (declare (edited  "17-MAR-1998")
	   (authors Hendrik)
	   (input   "A substitution and a list of constant restrictions.")
	   (effect  "none.")
	   (value   "a new substitution, in which all restrictions are ?...?"
		    "if one can be found, nil else."))
  (if (null subst) nil
      (let* ((old-ti (bool=ce-get-ti restrictions))
	     (ci (bool=ce-create-ci old-ti restrictions))
	     (ti (mapcar #'(lambda (ti) (subst~apply subst ti)) old-ti))
	     (ti-norm (mapcar #'bool=norm ti))
	     (v0 (bool=ce-create-v0 ti-norm))
	     (c0 (bool=ce-create-c0 restrictions))
	     (d (bool=ce-create-d c0))
	     (new-subst (bool=ce-gen-subst v0 d))) 
	(let* ((norm-terms (mapcan #'(lambda (list c)
				       (bool=ce-get-only-ci (bool=ce-create-tijd list d) c d))
				   (mapcar #'bool=norm
					   (mapcar #'(lambda (list)
						       (subst~apply new-subst list)) ti))
				   ci))
		    (var-lists (bool=order-create-var-list norm-terms))
		    (new-term-list (bool=order norm-terms var-lists)))
	       (let ((elimination (bool=unify-prepare new-term-list)))
		 (if (null elimination) nil
		     (let ((final-subst (subst~compose-substitution
					 (subst~apply elimination new-subst)
					 subst)))
		       (bool=ce-make-new-subst (subst~domain final-subst)
					       (subst~codomain final-subst)))))))))


(defun bool=ce-create-ci(ti restrictions)
  (declare (edited  "17-MAR-1998")
	   (authors Hendrik)
	   (input   "a list of terms in which some constants are restricted (mostly variables only).")
	   (effect  "none.")
	   (value   "a list of constants, which are restricted."))
  (mapcar #'(lambda (var1)
	      (mapcan #'(lambda (var2)
			  (if (member var1 var2 :test #'term~equal)
			      (list (car var2))
			      nil))
		      restrictions))
	  ti))


(defun bool=ce-get-only-ci (list c d)
  (if (null d) nil
      (if (bool=ce-get-only-ci2 c (car d))
	  (cons (car list) (bool=ce-get-only-ci (cdr list) c (cdr d)))
	  (bool=ce-get-only-ci (cdr list) c (cdr d)))))


   (defun bool=ce-get-only-ci2 (c prefix)
     (if (null c) nil
         (if (member (car c) prefix :test #'term~equal) t
	     (bool=ce-get-only-ci2 (cdr c) prefix))))


(defun bool=ce-make-new-subst (domain co-domain)
  (if (null co-domain) (subst~create nil nil)
      (subst~insert-component (car domain)
			      (bool=alg-create-new-term
			       (bool=alg-simplify
				(bool=norm (car co-domain))))
			      (bool=ce-make-new-subst (cdr domain) (cdr co-domain)))))


(defun bool=ce-gen-subst (v0 d)
  (if (null v0) (subst~create nil nil)
      (subst~insert-component (car v0)
			      (bool=alg-create-new-term
			          (mapcan #'(lambda (var)
					     (list (append var
							   (list (sym~rename-var (car v0))))))
					  d))
			      (bool=ce-gen-subst (cdr v0) d))))


(defun bool=ce-create-tijd (list d)
  (bool=ce-check-for-nil (bool=ce-create-tijd1 (bool=alg-simplify list) d (bool=ce-prepare-list d))))


(defun bool=ce-check-for-nil (list)
  (declare (special bool*zero))
  (if (null list) nil
      (if (null (car list)) (cons (list (list bool*zero))
				    (bool=ce-check-for-nil (cdr list)))
	  (cons (car list) (bool=ce-check-for-nil (cdr list))))))


   (defun bool=ce-create-tijd1 (list d new-list)
     (if (null list) new-list
         (bool=ce-create-tijd1 (cdr list) d (bool=ce-create-tijd2 (car list) d new-list))))


   (defun bool=ce-create-tijd2 (list d new-list)
     (if (null (cdr d)) (list (cons list (car new-list)))
         (let ((term-part (bool=ce-is-ci-in-tijd-p list (car d))))
	   (if term-part (cons (append (list term-part) (car new-list))
			       (cdr new-list))
	       (cons (car new-list)
		     (bool=ce-create-tijd2 list (cdr d) (cdr new-list)))))))


(defun bool=ce-prepare-list (d)
  (if (null d) nil
      (cons nil (bool=ce-prepare-list (cdr d)))))


(defun bool=ce-is-ci-in-tijd-p (list const-list)
  (if (null const-list) list
      (let ((list-part (bool=ce-is-ci-in-tijd2-p list (car const-list))))
	(if list-part (bool=ce-is-ci-in-tijd-p list-part (cdr const-list))
	    nil))))


   (defun bool=ce-is-ci-in-tijd2-p (list const)
     (declare (special bool*one))
     (if (null list) nil
         (if (term~equal (car list) const)
	     (if (null (cdr list)) (list bool*one)
	         (cdr list))
             (let ((test (bool=ce-is-ci-in-tijd2-p (cdr list) const)))
	       (if test (cons (car list) test)
		   nil)))))


(defun bool=ce-create-v0 (term-list)
  (let ((new-list (if (atom (car (car term-list))) term-list
		      (mapcan #'(lambda (elem) elem) term-list))))
    (bool=ce-c-v0-get-vars new-list nil)))


(defun bool=ce-c-v0-get-vars (term vars)
  (declare (special bool*vac))
  (if (null term) vars
      (bool=ce-c-v0-get-vars
       (cdr term)
       (append vars
	       (remove-duplicates
		       (mapcan #' (lambda (atom)
			         (if (and (and (sym~variable-p atom)
					  (not (member atom bool*vac :test #'term~equal)))
				          (not (member atom vars :test #'term~equal)))
				     (list atom)
				     nil))
				  (car term))
		       :test #'term~equal)))))
				  

(defun bool=ce-get-ti (restrictions)
  (if (null restrictions) nil
      (bool=ce-new-ti (cdr (car restrictions)) (bool=ce-get-ti (cdr restrictions)))))


(defun bool=ce-new-ti (ti old-ti)
  (if (null ti) old-ti
      (if (member (car ti) old-ti :test #'term~equal)
	  (bool=ce-new-ti (cdr ti) old-ti)
	  (bool=ce-new-ti (cdr ti) (cons (car ti) old-ti)))))


(defun bool=ce-create-c0 (restrictions)
  (if (null restrictions) nil
      (cons (car (car restrictions))
	    (bool=ce-create-c0 (cdr restrictions)))))


(defun bool=ce-create-d (c0)
  (declare (special bool*one))
  (let ((n (list-length c0)))
    (append (bool=ce-potenz c0 n n) (list (list bool*one)))))


(defun bool=ce-potenz (c0 n m)
  (if (equal m 1) (bool=ce-pot1 c0 n 1)
      (append (bool=ce-pot1 c0 n m)
	      (bool=ce-potenz c0 n (- m 1)))))


   (defun bool=ce-pot1 (C0 n m)
     (if (equal m 1) (mapcar #'list c0)
         (bool=ce-pot2 C0 n (- m 1))))


   (defun bool=ce-pot2 (C0 n m)
      (if (equal 1 (- n m))
         (list c0)
         (append (bool=ce-pot-mult (car c0) (bool=ce-pot1 (cdr c0) (- n 1) m))
	         (bool=ce-pot2 (cdr c0) (- n 1) m))))


(defun bool=ce-pot-mult (elem list)
  (if (null list) nil
      (cons (bool=ce-pot-add elem (car list))
	    (bool=ce-pot-mult elem (cdr list)))))


(defun bool=ce-pot-add (elem1 elem2)
  (if (atom elem2) (cons elem1 (list elem2))
      (cons elem1 elem2)))




     ;;;  Part III : The normalization algorithm
     ;;;
     ;;;  the original term is normalized. The new term has the form of lists.


(defun bool=norm (term)
  (declare (edited  "17-MAR-1998")
	   (authors Hendrik)
	   (input   "a term*term.")
	   (effect  "none.")
	   (value   "a list of lists of terms, representing the term in its normalized form"
		    "i.e. a disjunction of conjunctios."))
  (declare (special bool*add))
  (declare (special bool*mul))
  (cond ((sym~constant-p Term) (list (list Term)))
	((sym~variable-p Term) (list (list Term)))
	((term~equal (term~top Term) bool*add)
	  (append (bool=norm (car (appl~arguments Term)))
		  (bool=norm (cadr (appl~arguments Term)))))
	((term~equal (term~top Term) bool*mul)
	  (bool=norm-mult-new (bool=norm (car (appl~arguments Term)))
			 (bool=norm (cadr (appl~arguments Term)))))
	)
  )

(defun bool=norm-mult-new (list-1 list-2)
  (if (null list-1) list-1
      (append (bool=norm-mult-each-new (car list-1) list-2)
	      (bool=norm-mult-new (cdr list-1) list-2))))

(defun bool=norm-mult-each-new (elem list)
  (if (null list) list
      (cons (append elem (car list))
	    (bool=norm-mult-each-new elem (cdr list)))))
