;;;****************************************************************************
;;;* Boolean Unification                                                      *
;;;****************************************************************************

(in-package "KEIM")

(setf br*env (env~create))

(post~read-object-list '((type-constants i)
			 (constants (a i) (b i) (a0 i) (b0 i) (a1 i) (b1 i)
			            (a_q i) (a_k i) (a_l i) (a_e i) (a_f i)
			            (a_A i) (a_B i) (a_0 i) (a_1 i)
			            (inv (i i)) (n (i i)) (p (i i))
			            (s (i i i)) (dot (i i i)))
			 (variables (x i) (y i) (z i) (u i) (v i) (w i)
				    (x1 i) (x2 i) (x3 i) (x4 i) (x5 i) (x6 i)
				    (y1 i) (y2 i) (y3 i) (y4 i) (y5 i) (y6 i)
				    (v_0 i) (v_1 i) (v_2 i) (v_3 i))
			 (signatures (br (+ (i i i)) (* (i i i)) (o i) (e i))))
		       br*env)


(setf cpbr-theory (th~term-theory (term~read '+ br*env)))
(setf a1 (term~read '(* (+ a b) (* (+ x y) z)) br*env))
(setf a2 (term~read
	  '(+
	    (* (+ (* a e) b) (* (+ x y) z))
	    (+ (* o a) (* (+ (* a (* a e)) b) (* (+ y x) z))))
	  br*env))
;;;+--------------------------------------------------------------------------+
;;;| Normalization                                                            |
;;;+--------------------------------------------------------------------------+

;;;  This function transforms a br-term in the shortest disjunctive normalform 
;;;  by applying the following rewrite rules:
;;;
;;;  (1) term -> DNF                       ||   order of use
;;;  (2) 0 * x -> 0                        ||
;;;  (3) 1 * x -> x                        ||
;;;  (4) (0 + x -> x) performed by (2)     || 
;;;  (5) x + x -> 0                        \/ 
;;; 
;;;  Afterwards the term has the following representation:
;;;  
;;;  e.g. ((A B) (A) (C D)) means A*B + A + C*D
;;;


(defun cpbr=normalize (theory term)
  (let*
      ((cpbr*theory theory)
       (cpbr*plus (th~symbol cpbr*theory :plus))
       (cpbr*mal (th~symbol cpbr*theory :mal))
       (cpbr*o (th~symbol cpbr*theory 0))
       (cpbr*e (th~symbol cpbr*theory 1)))
    (declare (special cpbr*theory cpbr*plus cpbr*mal cpbr*o cpbr*e))
    (cpbr=x-plus-x
     (cpbr=x-mal-x
      (cpbr=mal-e
       (cpbr=mal-o
	(cpbr=dnf term)))))))

;;;+- x + x -> 0 -------------------------------------------------------------+

(defun cpbr=x-plus-x (dnf-term)
  (let
      ((len (length dnf-term)))
    (if (<= len 1)
	dnf-term
	(if (eq
	     (1- len)
	     (length (cpbr=del-disjunct! (car dnf-term) (cdr dnf-term))))
	    (cons (car dnf-term) (cpbr=x-plus-x (cdr dnf-term)))
	    (cpbr=x-plus-x (cdr dnf-term))))))

(defun cpbr=disjunct-equal-p (d1 d2)
  (not (set-exclusive-or d1 d2 :test #'term~equal)))

(defun cpbr=del-disjunct! (disjunct dnf-term)
  (cpbr=del-disjunct-aux! disjunct dnf-term nil))

(defun cpbr=del-disjunct-aux! (disjunct dnf-term buffer)
  (if (null dnf-term)
      buffer
      (if (cpbr=disjunct-equal-p disjunct (car dnf-term))
	  (cpbr=del-disjunct-aux!
	   disjunct
	   (cdr dnf-term)
	   buffer)
	  (cpbr=del-disjunct-aux! 
	   disjunct
	   (cdr dnf-term)
	   (append buffer (list (car dnf-term)))))))

;;;+- 0 + x -> x -------------------------------------------------------------+

;;;                            see 0 * x -> 0

;;;+- x * x -> x -------------------------------------------------------------+

(defun cpbr=x-mal-x (dnf-term)
  (mapcar
   #'(lambda (disjunct)
		   (remove-duplicates disjunct :test #'term~equal))
   dnf-term))

;;;+- 1 * x -> x -------------------------------------------------------------+

(defun cpbr=mal-e (dnf-term)
  (let
      ((temp (mapcan #'cpbr=e-factor dnf-term)))
    (if (null temp)
	dnf-term
	temp)))

(defun cpbr=e-factor (disjunct)
  (declare (special cpbr*e))
  (let
      ((temp (mapcan
	      #'(lambda (factor)
		  (if (term~equal factor cpbr*e)
		      nil
		      (list factor)))
	      disjunct)))
    (if temp
	(list temp)
	nil)))
		
;;;+- 0 * x -> 0 -------------------------------------------------------------+

(defun cpbr=mal-o (dnf-term)
  (let
      ((temp (mapcan #'cpbr=o-factor dnf-term)))
    (if (null temp)
	dnf-term
	temp)))

(defun cpbr=o-factor (disjunct)
  (declare (special cpbr*o))
  (if (some #'(lambda (factor) (term~equal cpbr*o factor)) disjunct)
      nil 
      (list disjunct)))

;;;+- DNF --------------------------------------------------------------------+

(defun cpbr=dnf (term)
  (cond
    ((cpbr=produkt-p term)
     (let*
	 ((flat-term (cpbr=produkt-flat term))
	  (factor-dnf-list (mapcar #'cpbr=dnf flat-term))
	  (dnf nil)
	  (temp nil))
       (mapc
	#'(lambda (factor-dnf)
	    (if (null dnf)
		(setf dnf factor-dnf)
		(mapc
		 #'(lambda (si)
		     (setf temp dnf)
		     (setf dnf nil)
		     (mapc
		      #'(lambda (sj)
			  (push (append si sj) dnf))
		      temp))
		 factor-dnf)))
	factor-dnf-list)
       dnf))
    ((cpbr=summe-p term)
     (let
	 ((flat-term (cpbr=summe-flat term)))
       (mapcan #'cpbr=dnf flat-term)))
    (T (list (list term)))))

(defun cpbr=summe-p (term)
  (declare (special cpbr*plus)) 
  (and (appl~p term)
       (term~equal (appl~function term) cpbr*plus)))

(defun cpbr=produkt-p (term)
  (declare (special cpbr*mal))
  (and (appl~p term)
       (term~equal (appl~function term) cpbr*mal)))

(defun cpbr=produkt-flat (term)
  (if (cpbr=produkt-p term)
      (append (cpbr=produkt-flat (first (appl~arguments term)))
	      (cpbr=produkt-flat (second (appl~arguments term))))
      (list term)))

(defun cpbr=summe-flat (term)
  (if (cpbr=summe-p term)
      (append (cpbr=summe-flat (first (appl~arguments term)))
	      (cpbr=summe-flat (second (appl~arguments term))))
      (list term)))





