(in-package 'keim)


;;; Additions to other modules

;;; for module Theories (th):


(defun th~term-theory (term)
  (declare (edited  "16-MAR-1995")
	   (authors Kepser Richts)
	   (input   "A term.")
	   (effect  "None.")
	   (value   "The theory instance TERM belongs to. (th*free if"
		    "it is free).")) 
  (if (appl~p term)
      (th~symbol-theory (appl~function term))
    (th~symbol-theory term)))


;;; for module Terms (term):

(defun term~vars-of-term (term &optional (vars-found NIL))
  (declare (edited  "22-MAR-1995")
	   (authors Kepser)
	   (input   "A term.")
	   (effect  "None.")
	   (value   "A set of all variables in the term")) 
  (cond ((not (term~p term))
	 (error "term~~vars: ~A is not a term!~%" term))
	((sym~variable-p term)
	 (adjoin term vars-found))
	((sym~constant-p term)
	 vars-found)
	((appl~p term)
	 (term~vars-of-termlist (appl~arguments term) vars-found))
	(T vars-found)))


(defun term~vars-of-termlist (term-list &optional (vars-found NIL))
  (declare (edited  "22-MAR-1995")
           (authors Kepser)
           (input   "A list of terms.")
           (effect  "None.")
           (value   "A set of all variables in the terms")) 
  (if (endp term-list)
      vars-found
    (term~vars-of-termlist (rest term-list)
			   (term~vars-of-term (first term-list)
                                                 vars-found))))



; ------------------------------------------------------------------------

;;; cb~create-new-var Environment
;;; create new variable of type I and place it
;;; in the environment Environment
;;; Variable names are X0, X1, X2, ...

(defun cb~create-new-var (Environment)
  (declare (edited  "30-MAY-1995")
	   (authors Kepser)
	   (input   "A KEIM environment.")
	   (effect  "The new variable is added to the environment.")
	   (value   "A new KEIM variable."))

  (let ((newvar (gentemp "X")))
    (sym~env-enter-variable newvar (type~i) Environment)
    (term~read newvar Environment)))



;;; remove equations x = y where both x and y is a variable

(defun cb~remove-vacuous-eqs (equations &optional (eqs-so-far NIL))
  (declare (edited  "30-MAY-1995")
	   (authors Kepser)
	   (input   "equations: a list of pairs (list) of KEIM terms.")
	   (effect  "None.")
	   (value   "A new list of equations where all equations of"
		    "type  var1 = var2  are removed."))

  (cond ((endp equations)
	 eqs-so-far)
	((and (sym~variable-p (first (first equations)))
	     (sym~variable-p (second (first equations))))
	 (let ((subst (subst~create `(,(first (first equations)))
				    `(,(second (first equations))))))
	   (cb~remove-vacuous-eqs (subst~apply subst (rest equations))
			       (subst~apply subst eqs-so-far))))
	(T ; non-vacuous eq
	 (cb~remove-vacuous-eqs (rest equations)
			     (cons (first equations) eqs-so-far)))))



;;; functions to purify equations, terms, etc.
;;; equartions t1 = t2 are supposed to be represented as
;;;  lists (t1 t2)

(defun cb=purify-const (const environment theory equations)
  (declare (edited  "30-MAY-1995")
	   (authors Kepser)
	   (input   "const: a KEIM constant; "
		    "environment: an environment; "
		    "theory: a KEIM theory instance to compare the"
		    "constant with; "
		    "equations: a list of equations still to purify.")
	   (effect  "None.")
	   (value   "If const is of theory  theory  ,then (const "
		    "equations), else a new variable and the old"
		    "equations plus the new (var const)."))
		    
  (if (eq theory (th~symbol-theory const))
      (values const equations)
    (let ((newvar (cb~create-new-var environment)))
      (values newvar (cons (list newvar const) equations)))))


(defun cb=purify-term (term environment theory equations)
  (declare (edited  "30-MAY-1995")
	   (authors Kepser)
	   (input   "term: a KEIM term; "
		    "environment: an environment; "
		    "theory: a KEIM theory instance to compare the"
		    "constant with; "
		    "equations: a list of equations still to purify.")
	   (effect  "None.")
	   (value   "The purified term and the - possibly extended - "
		    "list of equations still to purify."))
  (cond ((sym~constant-p term)
         (multiple-value-bind (const equations1)
	     (cb=purify-const term environment
			      theory equations)
	   (values const equations1)))
	((sym~variable-p term)
	 (values term equations))

        ; term must be complex
	((eq theory (th~term-theory term))
	 (multiple-value-bind (arglist equations1)
	     (cb=purify-subterms (appl~arguments term)
				 environment theory
				 equations)
	   (values (appl~create (appl~function
				 term) arglist)
		   equations1))) 

	(t  ; term is complex and of different theory
	 (let ((newvar (cb~create-new-var environment)))
	   (values newvar (cons (list newvar term)
				equations)))))) 


(defun cb=purify-subterms (subterms environment theory equations)
  (let ((new-subterms ()) (equations1 equations))
    (values (dolist (subterm subterms (reverse new-subterms))
	      (multiple-value-bind (new-subterm equations2)
		  (cb=purify-term subterm environment
			       theory equations1)
		(setf new-subterms (cons new-subterm
					 new-subterms))
		(setf equations1 equations2)))
	    equations1)))


(defun cb~purify-equation (equation pure-th1 pure-th2 environment
				 equations)
  (declare (edited  "30-MAY-1995")
	   (authors Kepser)
	   (input   "equation: a pair (list) of KEIM terms; "
		    "pure-th1: purified equations of first theory; "
		    "pure-th2: purified equations of second theory; "
		    "environment: an environment; "
		    "equations: a list of equations still to purify.")
	   (effect  "None.")
	   (value   "The new pair of purified equations and the"
		    "new equations still to purify."))

  ;;; we assume that variables have NO theory at all.
  
  (let* ((lhs (first equation))
	 (rhs (second equation))
	 (lhs-th (th~term-theory lhs))
	 (rhs-th (th~term-theory rhs)))
    (cond ((eq lhs-th rhs-th)
	   ;;both sides same theory (non variables)
	   (multiple-value-bind (new-lhs equations1)
	       (cb=purify-term lhs environment lhs-th equations)
	     (multiple-value-bind (new-rhs equations2)
		 (cb=purify-term rhs environment rhs-th equations1)
	       (if (eql lhs-th (first pure-th1))
		   (values (list lhs-th
				 (cons new-lhs (second pure-th1))
				 (cons new-rhs (third pure-th1)))
			   pure-th2 equations2)
		 (values pure-th1 (list (cons new-lhs (first pure-th2))
					(cons new-rhs (second pure-th2)))
			 equations2)))))

	  ((sym~variable-p lhs)
	   ;; left hand side variable
	   (multiple-value-bind (new-rhs equations1)
	       (cb=purify-term rhs environment rhs-th equations)
	     (if (eql rhs-th (first pure-th1))
		 (values (list rhs-th
			       (cons lhs (second pure-th1))
			       (cons new-rhs (third pure-th1)))
			 pure-th2 equations1)
	       (values pure-th1 (list (cons  lhs (first pure-th2))
				      (cons new-rhs (second pure-th2)))
		       equations1))))

	  ((sym~variable-p rhs)
	   ;; right hand side variable
	   (multiple-value-bind (new-lhs equations1)
	       (cb=purify-term lhs environment lhs-th equations)
	     (if (eql lhs-th (first pure-th1))
		 (values (list lhs-th
			       (cons rhs (second pure-th1))
			       (cons new-lhs (third pure-th1)))
			 pure-th2 equations1)
	       (values pure-th1 (list (cons rhs (first pure-th2))
				      (cons new-lhs (second pure-th2)))
		       equations1))))

	  (T  ;; sides have different theories, non-variable
	   (multiple-value-bind (new-lhs equations1)
	       (cb=purify-term lhs environment lhs-th equations)
	     (let ((newvar (cb~create-new-var environment)))
	       (if (eql lhs-th (first pure-th1))
		   (values (list lhs-th
				 (cons newvar (second pure-th1))
				 (cons new-lhs (third pure-th1)))
			   pure-th2
			   (cons (list newvar rhs) equations1))
		 (values pure-th1
			 (list (cons newvar (first pure-th2))
			       (cons new-lhs (second pure-th2)))
			 (cons (list newvar rhs) equations1)))))))))



(defun cb=aux-purify-equations (equations environment pure-th1 pure-th2)
  (declare (edited  "30-MAY-1995")
	   (authors Kepser)
	   (input   "equations: a list of equations to purify; "
		    "environment: an environment; "
		    "pure-th1: already purified equations of first"
		    "theory; "
		    "pure-th2: purified equations of second theory.")
	   (effect  "None.")
	   (value   "A list of purified equations for each of the two"
		    "theories."))
  (if (endp equations)
      (values pure-th1 pure-th2)
    (multiple-value-bind (new-pure-th1 new-pure-th2 equations1)
	(cb~purify-equation (first equations) pure-th1 pure-th2
			 environment (rest equations))
      (cb=aux-purify-equations equations1 environment new-pure-th1
			new-pure-th2))))


(defun cb~purify-equations (equations environment)
  (declare (edited  "30-MAY-1995")
	   (authors Kepser)
	   (input   "equations: a list of equations to purify; "
		    "environment: an environment. ")
	   (effect  "None.")
	   (value   "A list of purified equations for each of the"
		    "two theories. Format of equations is changed"
		    "inbetween. In  equations  each equation is a"
		    "pair consisting of the lhs and rhs. The purified"
		    "equations for one theory is a triple. First"
		    "element is the theory instance, second a list"
		    "of left hand sides, third a list of"
		    "corresponding right hand sides."))

  (if (endp equations)
      (values NIL NIL)
    (multiple-value-bind (pure-th1 pure-th2)
	(cb=aux-purify-equations
	 equations environment
	 (list (if (sym~variable-p (first (first equations)))
		   (th~term-theory (second (first equations)))
		 (th~term-theory (first (first equations))))
	       NIL NIL)
	 (list NIL NIL))
      (values pure-th1
	      (if (endp (first pure-th2))
		  (list th*free NIL NIL)
		(cons (th~term-theory (first (second pure-th2)))
		      pure-th2))))))











;; -----------------------------------------------------------------
; decision procedures



(defun cb~decide-unify (equations environment)
  (declare (edited  "30-MAY-1995")
	   (authors Kepser)
	   (input   "equations: a list of equations. Each equation is"
		    " a pair (list) of two KEIM terms. The function"
		    "symbols must be from exactly two signatures. "
		    "environment: an environment. ")
	   (effect  "None.")
	   (value   "NIL, if equations are not unifiable, "
		    "else non-NIL."))
  
  (multiple-value-bind (pure-th1 pure-th2)
      (cb~purify-equations (cb~remove-vacuous-eqs equations)
			environment)
    (let* ((vars-th1 (term~vars-of-termlist
		      (third pure-th1)
		      (term~vars-of-termlist (second pure-th1))))
	   (vars-th2 (term~vars-of-termlist
		      (third pure-th2)
		      (term~vars-of-termlist (second pure-th2))))
	   (common-vars (intersection vars-th1 vars-th2))
	   (number-of-vars (length common-vars))
	   (rgf (make-array number-of-vars))
	   (unifiable NIL))

      (if (endp common-vars)  ; NO common variables
	  (progn
	    (format T "This is not a combination problem!~%")
	    (values NIL))

	(progn
          ;;; initialise with last rgf
	  (dotimes (i number-of-vars T)
	    (setf (svref rgf i) (+ i 1)))
	  (loop while (and (not unifiable) rgf) do
		(multiple-value-bind (new-vars subst)
		    (cb=create-substitution rgf common-vars)
		  (format T "~A ~A ~A ~A~%"
			  (subst~apply subst pure-th1)
			  (subst~apply subst pure-th2)
			  subst new-vars)
		  (setf unifiable
			(cb=enumerate-indexes-decide
			 (subst~apply subst pure-th1)
			 (subst~apply subst pure-th2)
			 new-vars))
		  (setf rgf (cb=previous-rgf rgf))))
	  unifiable)))))






(defun cb=create-substitution (rgf variables)
  ;;; Input: rgf: restricted growth function. A vector of positive
  ;;;             integers of length variables
  ;;;        variables: A list of KEIM variables
  (declare (edited  "16-MAY-1995")
	   (authors Kepser)
	   (input   "A restricted growth function, i.e. a vector of"
		    "positive integers and a list of KEIM variables of"
		    "the same length as the vector.")
	   (effect  "None.")
	   (value   "An Array of KEIM Variables and a KEIM"
		    "substitution: All variables with the same"
		    "Index are substituted by one representative."))

  (let* ((domain NIL) (codomain NIL)
	 (alength (cb=max-vector rgf))
	 (reps (make-array alength :initial-element NIL))
	 (i 0))
    (dolist (var variables T)
      (let ((rep (svref reps (- (svref rgf i) 1))))
	(if rep
	    (setf domain (cons var domain)
		  codomain (cons rep codomain))
	  (setf (svref reps (- (svref rgf i) 1)) var)))
      (setf i (+ i 1)))
    (values reps (subst~create domain codomain))))



  
(defun cb=max-vector (vector &optional (l (array-dimension vector 0)))
  (let ((mx 0))
    (dotimes (i l mx)
      (when (> (svref vector i) mx)
	(setf mx (svref vector i))))))



(defun cb=previous-rgf (current-rgf)
  (let ((n (- (array-dimension current-rgf 0) 1)))
    (if (= n 0)
	NIL  ; only one variable
      (if (= (svref current-rgf n) 1)
	  (let ((j
		 (do ((i (- n 1) (- i 1)))
		     ((or (> (svref current-rgf i) 1) (= i 0)) i))))
	    (if (= j 0)
		NIL
	      (progn
		(setf (svref current-rgf j) (- (svref current-rgf j) 1))
		(do ((i (+ j 1) (+ i 1))
		     (mx (+ (cb=max-vector current-rgf j) 1) (+ mx 1)))
		    ((> i n))
		  (setf (svref current-rgf i) mx))
		current-rgf)))
	(progn
	  (setf (svref current-rgf n) (- (svref current-rgf n) 1))
	  current-rgf)))))


;----------------------------------------------------------------------

;;; cb=enumerate-orders-decide
;;;
;;; The function uses backtracking to enumerate all relevant
;;; orders. In fact, none of the orders is effectively calculated,
;;; rather the constant restrictions needed are created. But each
;;; time the base case is reached for calling the unification
;;; functions corresponds to the calculation of an order. As demanded,
;;; not all possible orders are created, rather equivalence classes of
;;; them in which all orders are regarded as equivalent which differ
;;; only in transposition of adjacent variables of the same
;;; theory. More precisely, it is not equivalence classes as such, but
;;; representatives that are generated.
;;; The backtracking is implemented by recursion without use of Lisp
;;; extensions or dynamic / global variables or side effects.
;;;
;;; Input:
;;; eqs-th1:         left hand side of the equations; a list of KEIM terms
;;; eqs-th2:         corresponding right hand side of the equations to be
;;;              unified; a list of KEIM terms
;;; variables1:  variables of theory 1 (regarded as constants of
;;;              theory 2); a list of KEIM variables
;;; variables2:  variables of theory 2; a list of KEIM variables
;;;
;;; (These 4 parameters are not altered but only passed through to the
;;; unification calls.)
;;;
;;; last-theory:
;;; vars11:      variables of theory 1 which must still be processed
;;;              after a theory change; a list of KEIM variables; a
;;;              subset of variables 1; initial value: variables1
;;; vars12:      variables of theory 1 which must still be processed
;;;              when no theory change takes place; a list of KEIM
;;;              variables; a subset of vars11; initial value:
;;;              variables1 
;;; vars21:      variables of theory 2 which must still be processed
;;;              after a theory change; a list of KEIM variables; a
;;;              subset of variables2; initial value: variables2
;;; vars22:      variables of theory 2 which must still be processed
;;;              when no theory change takes place; a list of KEIM
;;;              variables; a subset of vars21; initial value:
;;;              variables2
;;; restrictions1: constant restrictions for theory 1 calculated so
;;;                far; a list of lists of KEIM variables; initial
;;;                value: NIL
;;; restrictions2: constant restrictions for theory 2 calcualted so
;;;                far; a list of lists of KEIM variables; itnital
;;;                value: NIL
;;; rest-aux1:   variables of theory 1 used so far; a list of KEIM
;;;              variables; a subset of variables1;
;;;              initial value: NIL
;;; rest-aux2:   variables of theory 2 used so far; a liset of KEIM
;;;              variables; a subset of variables2;
;;;              initial value: NIL


(defun cb=enumerate-orders-decide
  ( eqs-th1 eqs-th2 variables1 variables2 last-theory
   vars11 vars12 vars21 vars22 restrictions1 restrictions2 rest-aux1
   rest-aux2)

  (declare (edited  "18-MAY-1995")
	   (authors Kepser)
	   (input   "Two unification functions, lhs and rhs of the"
		    "equations, variables of the first and second"
		    "theory, the number of the theory of the last"
		    "call of cb=enumerate-orders-decide, variables of theory1"
		    "still to be considered (for next the2 and the1),"
		    "variables of theory2 still to be considered"
		    "(for next the1 and the2), restrictions for"
		    "theories 1 and 2 calculated so far, variables"
		    "of theories 1 and 2 used so far.")
	   (effect  "None.")
	   (value   "The value of the unifications"))

  (cond	((endp vars11)
	 (let ((restr1 restrictions1))
	   (dolist (v vars21 T)
	     (setf restr1 (cons (cons v rest-aux1) restr1)))
	   (format T "U1: ~A ~A ~A~%" eqs-th1 restr1
		   variables2)
	   (format T "U2: ~A ~A ~A~%" eqs-th2 restrictions2
		   variables1)
	   (when (th~decide-unify (first eqs-th1) (second eqs-th1)
				  (third eqs-th1) restr1 variables2)
	     (th~decide-unify  (first eqs-th2) (second eqs-th2)
			       (third eqs-th2) restrictions2 variables1))))

	((endp vars21)
	 (let ((restr2 restrictions2))
	   (dolist (v vars11 T)
	     (setf restr2 (cons (cons v rest-aux2) restr2)))
	   (format T "U1: ~A ~A ~A~%" eqs-th1 restrictions1 variables2)
	   (format T "U2: ~A ~A ~A~%" eqs-th2 restr2 variables1)
	   (when (th~decide-unify (first eqs-th1) (second eqs-th1)
				  (third eqs-th1) restrictions1
				  variables2)
	     (th~decide-unify  (first eqs-th2) (second eqs-th2)
			       (third eqs-th2) restr2 variables1))))

	; vars11 <> nil  and  vars21 <> nil

	((eql last-theory 1)
	 (let ((unifiable NIL))
	   (do ((v1 (first vars12) (first vs1))
		(vs1 (rest vars12) (rest vs1)))
	       ((or unifiable (null v1)) unifiable)
	     (setf unifiable
		   (cb=enumerate-orders-decide
		    eqs-th1 eqs-th2 variables1 variables2 1
		    (remove v1 vars11) vs1 vars21 vars21
		    restrictions1 (cons (cons v1 rest-aux2) restrictions2)
		    (cons v1 rest-aux1) rest-aux2)))
	 
	   (do ((v2 (first vars21) (first vs2))
		(vs2 (rest vars21) (rest vs2)))
	       ((or unifiable (null v2)) unifiable)
	     (setf unifiable
		   (cb=enumerate-orders-decide
		    eqs-th1 eqs-th2 variables1 variables2 2
		    vars11 vars11  (remove v2 vars21) vs2
		    (cons (cons v2 rest-aux1) restrictions1)
		    restrictions2 rest-aux1 (cons v2 rest-aux2))))))
	
	(t ; last-theory = 2
	 (let ((unifiable NIL))
	   (do ((v2 (first vars22) (first vs2))
		(vs2 (rest vars22) (rest vs2)))
	       ((or unifiable (null v2)) unifiable)
	     (setf unifiable
		   (cb=enumerate-orders-decide
		    eqs-th1 eqs-th2 variables1 variables2 2
		    vars11 vars11 (remove v2 vars21) vs2
		    (cons (cons v2 rest-aux1) restrictions1)
		    restrictions2 rest-aux1 (cons v2 rest-aux2))))
	 
	   (do ((v1 (first vars11) (first vs1))
		(vs1 (rest vars11) (rest vs1)))
	       ((or unifiable (null v1)) unifiable)
	     (setf unifiable
		   (cb=enumerate-orders-decide
		    eqs-th1 eqs-th2 variables1 variables2 1
		    (remove v1 vars11) vs1 vars21 vars21
		    restrictions1 (cons (cons v1 rest-aux2) restrictions2)
		    (cons v1 rest-aux1) rest-aux2)))))))







(defun cb=enumerate-indexes-decide
  (eqs-th1 eqs-th2 variables)

  (declare (edited  "29-May-1995")
	   (authors Kepser)
	   (input   "eqs-th1: The pure equations of theory 1, a triple"
		    "consisting of the theory instance, a list of"
		    "left hand sides and a list of corresponding"
		    "right hand sides; "
		    "eqs-th2: The pure equations of theory 2, same"
		    "format as eqs-th1;"
		    "variables: a simple vector of KEIM variables, "
		    "the shared variables of the theories.")
	   (effect  "None.")
	   (value   "non-NIL, if equations are unifiable; NIL, else."))
  
  (let* ((n-vars (array-dimension variables 0))
	 (card-subset (if (= n-vars 1) 1 (floor n-vars 2)))
	 (subset (make-array card-subset))
	 (vars1) (vars2)
	 (unifiable NIL))

    ; initialise with first index
    (dotimes (i card-subset T)
      (setf (svref subset i) (+ i 1)))
    (loop while (and (not unifiable) subset) do
	  (setf card-subset (array-dimension subset 0)
		vars1 (cb=first-part subset card-subset variables)
		vars2 (cb=second-part vars1 variables n-vars))
	  (format T "V1: ~A   V2: ~A~%" vars1 vars2)
	  (setf unifiable 
		(cb=enumerate-orders-decide eqs-th1 eqs-th2 vars1
				  vars2 1 vars1 vars1 vars2 vars2 NIL NIL
				  NIL NIL ))
	  (format T "V1: ~A   V2: ~A~%" vars2 vars1)
	  (unless unifiable
	    (setf unifiable 
		  (cb=enumerate-orders-decide eqs-th1 eqs-th2 vars2
				    vars1 1 vars2 vars2 vars1 vars1 NIL NIL
				    NIL NIL)))
	  (format T "~% ")
	  (setf subset (cb=next-index subset card-subset n-vars)))
    unifiable))






(defun cb=next-index (subset card-subset n-vars)
  (declare (edited  "29-May-1995")
	   (authors Kepser)
	   (input   "subset: a simple vector of positive integers"
		    "<= n-vars;"
		    "card-subset: the length of the vector  subset; "
		    "n-vars: a positive integer.")
	   (effect  "None.")
	   (value   "The next index."))

  (if (< n-vars 2)
      NIL
    (let ((s1 (cb=next-subset subset n-vars)))
      (if s1
          ;;; check, if we are in the even case and start producing
          ;;; duplicates
	  (if (and (= n-vars (* card-subset 2))
		   (cb=start-doubles-p s1 card-subset))
	      (if (> card-subset 0)
		  (cb=first-subset (- card-subset 1))
		NIL)
	    s1)
	(if (> (decf card-subset) -1)
	    (cb=first-subset card-subset)
	  nil)))))



(defun cb=start-doubles-p (seq lang)
  ;;; T, iff seq = #(2 3 4 ... (n/2)+1)
  (declare (edited  "29-May-1995")
	   (authors Kepser)
	   (input   "seq: a simple vector of positive integers; "
		    "lang: its length.")
	   (effect  "None.")
	   (value   "T, iff seq = #(2 3 4 ... (n/2)+1)"))
  
  (dotimes (i lang T)
    (unless (= (+ i 2) (svref seq i))
      (return NIL))))


;;; List of k-elements subsets of an n-elements set.
;;; Again, we use lexicographic order

(defun cb=first-subset (k)
  (declare (edited  "29-May-1995")
	   (authors Kepser)
	   (input   "k: length of the subset to be constructed.")
	   (effect  "None.")
	   (value   "A simple vector of integers: #(1 2 3 ... k)"
		    "which is the first subset in lexicographical"
		    "order."))
  
  (let ((subset (make-array k)))
    (dotimes (i k subset)
      (setf (svref subset i) (+ i 1)))))


(defun cb=next-subset (subset n-vars)
  (declare (edited  "29-May-1995")
	   (authors Kepser)
	   (input   "subset: a simple vector of positive integers"
		    "each <= n-vars, and (length subset) <= n-vars; "
		    "n-vars: a positive integer.")
	   (effect  "None.")
	   (value   "Next subset in lexicographical order."))
  
  (let ((k (- (array-dimension subset 0) 1)))
    (if (< k 0)
	NIL
      (if (= (svref subset k) n-vars)
	  (let ((ind 
		 (do ((i k (- i 1))
		      (h n-vars (- h 1)))
		     ((or (< i 0)
			  (> h (svref subset i))) i))))
	    (if (= ind -1)
		NIL
	      (let ((nv (+ (svref subset ind) 1)))
		(setf (svref subset ind) nv)
		(do ((i ind (+ i 1))
		     (nv nv (+ nv 1)))
		    ((> i k) subset)
		  (setf (svref subset i) nv)))))
	(progn
	  (setf (svref subset k) (+ (svref subset k) 1))
	  subset)))))



(defun cb=first-part (indices n variables)
  (declare (edited  "29-May-1995")
	   (authors Kepser)
	   (input   "indices: a simple vector of positive integers"
		    "which can serve as indices (-1) to variables; "
		    "n: the length of the vector indices; "
		    "variables: a simple vector of KEIM variables.")
	   (effect  "None.")
	   (value   "A list of KEIM variables containing all those"
		    "to which indices pointed."))
  
  (let ((part NIL))
    (dotimes (i n part)
      (push (svref variables (- (svref indices i) 1)) part))))



(defun cb=second-part (first-part variables n)
  (declare (edited  "29-May-1995")
	   (authors Kepser)
	   (input   "first-part: a list of KEIM variables; "
		    "variables: a simple vector of KEIM variables; "
		    "n: the lenght of the vector variables.")
	   (effect  "None.")
	   (value   "A list of KEIM variables containing all those"
		    "from variables which are not in first-part."))
  
  (let ((part NIL) (elem))
    (dotimes (i n part)
      (unless (find (setf elem (svref variables i)) first-part)
	(push elem part)))))




; -------------------------------------------------------------------
; complete sets of unifiers



(defun cb~unify (equations environment)
  (declare (edited  "31-MAY-1995")
	   (authors Kepser)
	   (input   "equations: a list of equations. Each equation is"
		    " a pair (list) of two KEIM terms. The function"
		    "symbols must be from exactly two signatures. "
		    "environment: an environment. ")
	   (effect  "None.")
	   (value   "Complete set of unifiers for the combined problem."))
  
  (multiple-value-bind (pure-th1 pure-th2)
      (cb~purify-equations (cb~remove-vacuous-eqs equations)
			environment)
    (let* ((vars-th1 (term~vars-of-termlist
		      (third pure-th1)
		      (term~vars-of-termlist (second pure-th1))))
	   (vars-th2 (term~vars-of-termlist
		      (third pure-th2)
		      (term~vars-of-termlist (second pure-th2))))
	   (common-vars (intersection vars-th1 vars-th2))
	   (number-of-vars (length common-vars))
	   (rgf (make-array number-of-vars))
	   (unifiers NIL))

      (if (endp common-vars)  ; NO common variables
	  (progn
	    (format T "This is not a combination problem!~%")
	    (values NIL))

	(progn
          ;;; initialise with last rgf
	  (dotimes (i number-of-vars T)
	    (setf (svref rgf i) (+ i 1)))
	  (loop while rgf do
		(multiple-value-bind (new-vars subst)
		    (cb=create-substitution rgf common-vars)
		  (format T "~A ~A ~A ~A~%"
			  (subst~apply subst pure-th1)
			  (subst~apply subst pure-th2)
			  subst new-vars)
		  (let ((unif
			 (cb=enumerate-indexes-complete
			  (subst~apply subst pure-th1)
			  (subst~apply subst pure-th2)
			  new-vars)))
		    (if unif
			(push (cons subst unif)
			      unifiers))))
		(setf rgf (cb=previous-rgf rgf)))
	  unifiers)))))




(defun cb=enumerate-indexes-complete
  (eqs-th1 eqs-th2 variables)

  (declare (edited  "31-May-1995")
	   (authors Kepser)
	   (input   "eqs-th1: The pure equations of theory 1, a triple"
		    "consisting of the theory instance, a list of"
		    "left hand sides and a list of corresponding"
		    "right hand sides; "
		    "eqs-th2: The pure equations of theory 2, same"
		    "format as eqs-th1;"
		    "variables: a simple vector of KEIM variables, "
		    "the shared variables of the theories.")
	   (effect  "None.")
	   (value   "The set of unifiers."))
  
  (let* ((n-vars (array-dimension variables 0))
	 (card-subset (if (= n-vars 1) 1 (floor n-vars 2)))
	 (subset (make-array card-subset))
	 (vars1) (vars2)
	 (unifiers NIL))

    ; initialise with first index
    (dotimes (i card-subset T)
      (setf (svref subset i) (+ i 1)))
    (loop while subset do
	  (setf card-subset (array-dimension subset 0)
		vars1 (cb=first-part subset card-subset variables)
		vars2 (cb=second-part vars1 variables n-vars))
	  (format T "V1: ~A   V2: ~A~%" vars1 vars2)
	  (setf unifiers
		(nconc 
		 (cb=enumerate-orders-complete
		  eqs-th1 eqs-th2 vars1 vars2 1 vars1 vars1
		  vars2 vars2 NIL NIL NIL NIL)
		 unifiers))
	  (format T "V1: ~A   V2: ~A~%" vars2 vars1)
	  (setf unifiers
		(nconc
		 (cb=enumerate-orders-complete
		  eqs-th1 eqs-th2 vars2 vars1 1 vars2 vars2
		  vars1 vars1 NIL NIL NIL NIL)
		 unifiers))
	  (format T "~% ")
	  (setf subset (cb=next-index subset card-subset n-vars)))
    unifiers))




(defun cb=enumerate-orders-complete
  ( eqs-th1 eqs-th2 variables1 variables2 last-theory
   vars11 vars12 vars21 vars22 restrictions1 restrictions2 rest-aux1
   rest-aux2)

  (declare (edited  "31-MAY-1995")
	   (authors Kepser)
	   (input   "Two unification functions, lhs and rhs of the"
		    "equations, variables of the first and second"
		    "theory, the number of the theory of the last"
		    "call of cb=enumerate-orders-complete, variables"
		    "of theory1"
		    "still to be considered (for next the2 and the1),"
		    "variables of theory2 still to be considered"
		    "(for next the1 and the2), restrictions for"
		    "theories 1 and 2 calculated so far, variables"
		    "of theories 1 and 2 used so far.")
	   (effect  "None.")
	   (value   "The set of unifiers"))

  (cond	((endp vars11)
	 (let ((restr1 restrictions1))
	   (dolist (v vars21 T)
	     (setf restr1 (cons (cons v rest-aux1) restr1)))
	   (format T "U1: ~A ~A ~A~%" eqs-th1 restr1
		   variables2)
	   (format T "U2: ~A ~A ~A~%" eqs-th2 restrictions2
		   variables1)
	   (let ((unif1
		  (th~unify (first eqs-th1) (second eqs-th1)
				   (third eqs-th1) restr1 variables2)) 
		 (unif2 NIL))
	     (when unif1
	       (setf unif2
		     (th~unify  (first eqs-th2) (second eqs-th2)
				       (third eqs-th2) restrictions2
				       variables1))
	       (if unif2
		   (list (list unif1 unif2))
		 NIL)))))

	((endp vars21)
	 (let ((restr2 restrictions2))
	   (dolist (v vars11 T)
	     (setf restr2 (cons (cons v rest-aux2) restr2)))
	   (format T "U1: ~A ~A ~A~%" eqs-th1 restrictions1 variables2)
	   (format T "U2: ~A ~A ~A~%" eqs-th2 restr2 variables1)
	   (let ((unif1
		  (th~unify (first eqs-th1) (second eqs-th1)
				   (third eqs-th1) restrictions1
				   variables2))
		 (unif2 NIL))
	     (when unif1
	       (setf unif2
		     (th~unify  (first eqs-th2) (second eqs-th2)
				       (third eqs-th2) restr2
				       variables1))
	       (if unif2
		   (list (list unif1 unif2))
		 NIL)))))

	; vars11 <> nil  and  vars21 <> nil

	((eql last-theory 1)
	 (let ((unifiers NIL))
	   (do ((v1 (first vars12) (first vs1))
		(vs1 (rest vars12) (rest vs1)))
	       ((null v1) unifiers)
	     (setf unifiers
		   (nconc 
		    (cb=enumerate-orders-complete
		     eqs-th1 eqs-th2 variables1 variables2 1
		     (remove v1 vars11) vs1 vars21 vars21
		     restrictions1 (cons (cons v1 rest-aux2) restrictions2)
		     (cons v1 rest-aux1) rest-aux2)
		    unifiers)))
	 
	   (do ((v2 (first vars21) (first vs2))
		(vs2 (rest vars21) (rest vs2)))
	       ((null v2) unifiers)
	     (setf unifiers
		   (nconc
		    (cb=enumerate-orders-complete
		     eqs-th1 eqs-th2 variables1 variables2 2
		     vars11 vars11  (remove v2 vars21) vs2
		     (cons (cons v2 rest-aux1) restrictions1)
		     restrictions2 rest-aux1 (cons v2 rest-aux2))
		    unifiers)))))
	
	(t ; last-theory = 2
	 (let ((unifiers NIL))
	   (do ((v2 (first vars22) (first vs2))
		(vs2 (rest vars22) (rest vs2)))
	       ((null v2) unifiers)
	     (setf unifiers
		   (nconc
		    (cb=enumerate-orders-complete
		     eqs-th1 eqs-th2 variables1 variables2 2
		     vars11 vars11 (remove v2 vars21) vs2
		     (cons (cons v2 rest-aux1) restrictions1)
		     restrictions2 rest-aux1 (cons v2 rest-aux2))
		    unifiers)))
	 
	   (do ((v1 (first vars11) (first vs1))
		(vs1 (rest vars11) (rest vs1)))
	       ((null v1) unifiers)
	     (setf unifiers
		   (nconc 
		    (cb=enumerate-orders-complete
		     eqs-th1 eqs-th2 variables1 variables2 1
		     (remove v1 vars11) vs1 vars21 vars21
		     restrictions1 (cons (cons v1 rest-aux2) restrictions2)
		     (cons v1 rest-aux1) rest-aux2)
		    unifiers)))))))

