(in-package "KEIM")

(export '(cpcb~decide-unify))

(defvar cpcb*iterative t
  "Flag for testing the effect of the iterative method.")


(defun cpcb~decide-unify (terms1 terms2 &key return-up)
  (declare (edited  "29-SEP-1995" "30-MAY-1995")
	   (authors Richts Kepser)
	   (input   "Two lists of terms of equal length.")
	   (effect  "None.")
	   (value   "NIL, if the terms are not unifiable, "
		    "else non-NIL."))
  (cpcb=unify terms1 terms2 :decide t :return-up return-up))

(defun cpcb~unify (terms1 terms2)
  (declare (edited  "29-SEP-1995" "30-MAY-1995")
	   (authors Richts Kepser)
	   (input   "Two lists of terms of equal length.")
	   (effect  "None.")
	   (value   "NIL, if the terms are not unifiable, "
		    "else non-NIL."))
  (cpcb=unify terms1 terms2 :decide nil))

(defun cpcb~decide-unify-pred (eq-list &key return-up)
  (declare (edited  "13-AUG-1997")
	   (authors Doelfer)
	   (input   "List of equations (predicate notion!)")
	   (effect  "None.")
	   (value   "NIL, if the equations are not unifiable, "
		    "else non-NIL."))
  (cpcb=unify-pred eq-list :decide t :return-up return-up))

(defun cpcb~unify-pred (eq-list)
  (declare (edited  "13-AUG-1997")
	   (authors Doelfer)
	   (input   "List of equations (predicate notion!)")
	   (effect  "None.")
	   (value   "NIL, if the equations are not unifiable, "
		    "else non-NIL."))
  (cpcb=unify-pred eq-list :decide nil))

(defun cpcb=unify (terms1 terms2 &key decide return-up)
  (declare (edited  "22-AUG-1997")
	   (authors Doelfer)
	   (input   "Two lists of terms of equal length.")
	   (effect  "None.")
	   (value   "NIL, if the terms are not unifieable, "
		    "else non-NIL"))
  (let (general-up)
    (unwind-protect
	 (progn
	   (setf general-up (cpcb=initial-gup terms1 terms2))
	   (cond ((not (cpcb=unify-iteration general-up :decide decide))
		  nil)
		 (return-up general-up)
		 (t t)))
      (mapc #'(lambda (var)
		(cond (return-up
		       (keim~put var 'cpcb=label (unicp~orig-label var))
		       (unicp~set-orig-label! var nil))
		      (t (unicp~set-orig-label! var nil))))
	    (if (and general-up
		     (unicp~variables general-up))
		(unicp~variables general-up)
		(top~free-variables (list terms1 terms2)))))))

(defun cpcb=unify-pred (eq-list &key decide return-up)
  (declare (edited  "27-AUG-1997")
	   (authors Doelfer)
	   (input   "List of equations (predicate notion!)")
	   (effect  "None.")
	   (value   "Same as cpcb=unify." ))
  (let (general-up)
    (unwind-protect
	 (progn
	   (setf general-up (cpcb=new-initial-gup eq-list))
	   (cond ((not (cpcb=unify-iteration general-up :decide decide))
		  nil)
		 (return-up general-up)
		 (t nil)))
      (mapc #'(lambda (var)
		(unicp~set-orig-label! var nil))
	    (if (and general-up
		     (unicp~variables general-up))
		(unicp~variables general-up)
		(top~free-variables eq-list))))))

(defun cpcb=unify-iteration (general-up &key decide)
  (declare (edited  "15-AUG-1997" "29-SEP-1995" "30-MAY-1995")
	   (authors Doelfer Richts Kepser)
	   (input   "A general UP")
	   (effect  "None.")
	   (value   "NIL, if the terms are not unifiable, "
		    "else non-NIL."))
  (let (result-list)
    (do* ((result (cpcb=next-step general-up)
		  (cpcb=next-step general-up))
	  (ready-p (if (eq result :choice-point) (cpcb=ready-p general-up))
		   (if (eq result :choice-point) (cpcb=ready-p general-up)))
	  (solvable-p (if ready-p
			  (every #'unicp~solvable-p
				 (misc~queue-list (unicp~gup-up-queue general-up))))
		      (if ready-p
			  (every #'unicp~solvable-p
				 (misc~queue-list (unicp~gup-up-queue general-up))))))
	 ((or (and ready-p
		   (or (and solvable-p decide)
		       (null (unicp~gup-choice-points general-up))))
	      (and (null result)
		   (null (unicp~gup-choice-points general-up))))
	  (cond (decide solvable-p)
		(solvable-p (cons (cpcb=compute-result general-up) result-list))
		(t result-list)))
      (when solvable-p
	(push (cpcb=compute-result general-up) result-list))
      (cond ((and (eq result :choice-point) (not ready-p))
	     (cpcb=choice-point general-up))
	    ((or (null result) (eq result :choice-point))
	     (cpcb==trace-backtracking general-up result)
	     (cpcb=backtrack general-up))))))
;;;
;;; Aenderung: lqueue mit 5 Listen, misc! (x)
;;;


(defun cpcb=next-step (general-up)
  (declare (edited  "22-AUG-1995")
	   (authors Richts)
	   (input   "A unification problem")
	   (effect  "If the next elementary unification problem in GENERAL-UP"
		    "can be called in order to compute new constraints, then it"
		    "is called and popped from the up-queue, the new constraints"
		    "are inserted into the constraints-queue and the new UP"
		    "(which might be eq to the old one) is inserted at the end"
		    "of the up-queue.")
	   (value   "NIL iff the problem turns out to be unsolvable;"
		    ":CHOICE-POINT iff no further information can be computed"
		    "deterministically;"
		    "NIL else."))
  (let* ((lqueue (unicp~gup-constraint-lqueue general-up))
	 (up-queue (unicp~gup-up-queue general-up))
	 (up (misc~queue-front up-queue))
	 (constraint-lists (misc~lqueue-next-lists lqueue)))
    (cond ((or (cpcb=initial-up-p up)
	       (notevery #'null constraint-lists)
	       (unicp~more-information-p up))
	   (let ((new-up (cpcb=compute-up lqueue up constraint-lists)))
	     (when new-up
	       (misc~queue-delete! up-queue)
	       (misc~queue-insert! up-queue new-up))
	     new-up))
	  ((some #'unicp~more-information-p (misc~queue-list up-queue))
	   (misc~queue-delete! up-queue)
	   (misc~queue-insert! up-queue up)
	   (misc~lqueue-insert! lqueue nil nil nil nil nil)
	   (misc~lqueue-delete! lqueue)
	   (misc~lqueue-insert! lqueue nil nil nil nil nil)
	   (misc~lqueue-delete! lqueue)
	   t)
	  (t :choice-point))))

;;;
;;; Aenderung: lqueue mit 5 Listen, misc! (x)
;;;

(defun cpcb=initial-gup (terms1 terms2)
  (declare (edited  "13-FEB-1996")
	   (authors Richts)
	   (input   "Two lists of terms of equal length.")
	   (effect  "The variables of TERMS1 and TERMS2 are initalized.")
	   (value   "The equations of TERMS1 = TERMS2 are purified by variable"
		    "abstraction.  For each theory in TERMS1 and TERMS2 an"
		    "elementary unification problem with pure equations is"
		    "created.  A general unification problem, which contains"
		    "these elementary UPs, is returned"))
  (multiple-value-bind (theories equations-list)
      (cpcb=purify-equations terms1 terms2)
    ;(format t "~%~{~%~A~}" equations-list)
    (let* ((var-lists (mapcar #'top~free-variables equations-list))
	   (variables (top~free-variables var-lists)))
      (unicp~vars-label-initialize! variables)
      (mapc #'unicp~vars-identify!      ;; identify variables of variable equations
	    (unicp~eq-lhss (first equations-list))
	    (unicp~eq-rhss (first equations-list)))
      (mapc #'(lambda (theory var-list)
		(mapc #'(lambda (var)
			  (mapc #'(lambda (v)
				    (push theory (unicp~var-label-theories v)))
				(unicp~var-label-identified var)))
		      var-list))
	    (rest theories) (rest var-lists))
      (let* ((shared-variables
	      #+weg(print (remove-if-not #'unicp~shared-variable-p variables))
	      (sort
	       (copy-list (remove-if-not #'unicp~shared-variable-p variables))
	       #'string< :key #'(lambda (var) (reverse (sym~name var)))))
	     (general-up (unicp~general-up-create (unicp~eq-create terms1 terms2)
						  :variables variables
						  :shared-variables shared-variables))
	     (up-list (sort (mapcar #'(lambda (theory equations)
					(cpcb=initial-up-create theory general-up equations))
				    (rest theories) (rest equations-list))
			    #'(lambda (theory1 theory2)
				(and (typep theory1 'th+free)
				     (not (typep theory2 'th+free))))
			    :key #'unicp~theory)))
	(do ((up-tail up-list (rest up-tail))
	     (up-queue (misc~queue-create)
		       (misc~queue-insert! up-queue (first up-tail)))
	     (lqueue (misc~lqueue-create)
		     (misc~lqueue-insert!
		      (misc~lqueue-insert! lqueue nil nil nil nil nil)
		      nil nil nil nil nil)))
	    ((null up-tail)
	       (setf (unicp~gup-up-queue general-up) up-queue
		     (unicp~gup-up-list general-up) up-list
		     (unicp~gup-constraint-lqueue general-up) lqueue)))
	;(print (mapcar #'(lambda (up) (th~symbols (unicp~theory up))) up-list))
	general-up))))


(defun cpcb=new-initial-gup (eq-list)
  (declare (edited  "25-AUG-1996")
	   (authors Doelfer)
	   (input   "List of equations (Predicate notion).")
	   (effect  "The variables of EQ-LIST are initalized.")
	   (value   "The equations of EQ-LIST are purified by variable"
		    "abstraction.  For each theory in EQ-LIST an"
		    "elementary unification problem with pure equations is"
		    "created.  A general unification problem, which contains"
		    "these elementary UPs, is returned"))
  (multiple-value-bind (theories equations-list)
      (cpcb=purify-equations-pred eq-list)
    ;(format t "~%~{~%~A~}" equations-list)
    (let* ((var-lists (mapcar #'top~free-variables equations-list))
	   (variables (top~free-variables var-lists)))
      (unicp~vars-label-initialize! variables)
      (mapc #'unicp~vars-identify!      ;; identify variables of variable equations
	    (unicp~eq-lhss (first equations-list))
	    (unicp~eq-rhss (first equations-list)))
      (mapc #'(lambda (theory var-list)
		(mapc #'(lambda (var)
			  (mapc #'(lambda (v)
				    (push theory (unicp~var-label-theories v)))
				(unicp~var-label-identified var)))
		      var-list))
	    (rest theories) (rest var-lists))
      (let* ((shared-variables
	      #+weg(print (remove-if-not #'unicp~shared-variable-p variables))
	      (sort
	       (copy-list (remove-if-not #'unicp~shared-variable-p variables))
	       #'string< :key #'(lambda (var) (reverse (sym~name var)))))
	     (general-up (unicp~general-up-create eq-list
						  :variables variables
						  :shared-variables shared-variables))
	     (up-list (sort (mapcar #'(lambda (theory equations)
					(cpcb=initial-up-create theory general-up equations))
				    (rest theories) (rest equations-list))
			    #'(lambda (theory1 theory2)
				(and (typep theory1 'th+free)
				     (not (typep theory2 'th+free))))
			    :key #'unicp~theory)))
	(do ((up-tail up-list (rest up-tail))
	     (up-queue (misc~queue-create)
		       (misc~queue-insert! up-queue (first up-tail)))
	     (lqueue (misc~lqueue-create)
		     (misc~lqueue-insert!
		      (misc~lqueue-insert! lqueue nil nil nil nil nil)
		      nil nil nil nil nil)))
	    ((null up-tail)
	       (setf (unicp~gup-up-queue general-up) up-queue
		     (unicp~gup-up-list general-up) up-list
		     (unicp~gup-constraint-lqueue general-up) lqueue)))
	;(print (mapcar #'(lambda (up) (th~symbols (unicp~theory up))) up-list))
	general-up))))	     

;;;
;;; Aenderung: lqueue mit 5 Listen, misc! (x)
;;;

(defun cpcb=compute-up (queue up constraint-lists)
  (declare (edited  "22-AUG-1995")
	   (authors Richts)
	   (input   "A list-queue, an elementary unification problem and a list"
		    "of constraints.")
	   (effect  "The unicp~new-constr method for UP is called with"
		    "CONSTRAINT-LIST and returns new constraints.  Then"
		    "constraints which can be infered from these new constraints"
		    "are computed.  All new constraints are added to the"
		    "datastructures (i.e. the variables) and inserted in"
		    "QUEUE. Finally the oldest constraints are deleted in QUEUE.")
	   (value   "The new unification problem (probably equal to UP) or NIL"
		    "if it is unsolvabel."))
  (let ((vars (unicp~shared-variables (unicp~general-up up))))
    #+weg(format t "~%~%~A   ~A Variables ~% ~A variables with index~% ~A identifications; ~A discriminations"
	    up (length vars) (count-if #'unicp~var-label-theory vars)
	    (apply #'+ (mapcar #'(lambda (var) (length (unicp~var-label-identified var))) vars))
	    (apply #'+ (mapcar #'(lambda (var)
				   (apply #'+ (mapcar #'length
						      (unicp~var-label-discrim-lists var))))
			       vars))))
  (multiple-value-bind (up
			ident not-ident
			const-vars not-const-vars
			const-restr not-const-restr)
      (apply #'cpcb=new-constr up constraint-lists)
    (when up
      (let* ((theory (unicp~theory up))
	     (var-index (mapcar #'(lambda (var) (unicp~vi-create var theory))
				const-vars))
	     (not-var-index (mapcar #'(lambda (var) (unicp~vi-create var theory))
				    not-const-vars)))
	;(cpcb=vars-set! ident not-ident var-index not-var-index const-restr not-const-restr)
	(multiple-value-bind (consistent ident1 not-ident1 var-index1 not-var-index1 const-restr1
					 ident2 not-ident2 var-index2 not-var-index2 const-restr2)
	    (ord~order (unicp~general-up up) ident not-ident var-index not-var-index const-restr)
	  (when consistent
	    (misc~lqueue-insert! queue ident1 not-ident1 var-index1 not-var-index1 const-restr1)
	    (misc~lqueue-delete! queue)
	    (misc~lqueue-insert! queue ident2 not-ident2 var-index2 not-var-index2 const-restr2)
	    (misc~lqueue-delete! queue)
	    up))))))


;;; Purification

(defun cpcb=purify-equations (terms1 terms2)
  (declare (edited  "06-NOV-1995")
	   (authors Richts)
	   (input   "Two lists of terms of equal length.")
	   (effect  "None.")
	   (value   "Variable abstraction is performed on the equations"
		    "represented by TERMS1 and TERMS2, i.e. alien subterms"
		    "(subterms with another theory than the direct superterm)"
		    "are substituted by new variables and corresponding"
		    "equations are added.  Each of the resulting equations"
		    "contains symbols of only one theory.  The result is a"
		    "multiple value:"
		    "1. A list of theories (the first element is NIL)."
		    "2. A list of equations (the first element contains the"
		    "equations between variables)."))
  (cpcb=purify-equations-aux terms1 terms2 (list nil) (list (unicp~eq-create nil nil))))

(defun cpcb=purify-equations-pred (terms)
  (declare (edited  "22-AUG-1997")
	   (authors Doelfer)
	   (input   "A list of equations (in Predicate notion)")
	   (effect  "None.")
	   (value   "Same as cpcb=purify-equations."))
  (cpcb=purify-equations-aux-pred terms (list nil) (list (unicp~eq-create nil nil))))

(defun cpcb=purify-equations-aux (terms1 terms2 theory-list equations-list)
  (declare (edited  "06-NOV-1995")
	   (authors Richts)
	   (input   "Two lists of terms of equal length, a list of theories"
		    "and a list of equations.")
	   (effect  "THEORY-LIST and EQUATIONS-LIST are destructively changed.")
	   (value   "Multiple value:"
		    "1. A list of theories (the first element is NIL)."
		    "2. A list of equations (the first element contains the"
		    "   equations between variables)."))
  (if (or (endp terms1) (endp terms2))
      (if (and (endp terms1) (endp terms2))
	  (values theory-list equations-list)
	  (error "CPCB:  Lists of terms have not equal length."))
    (multiple-value-bind (theory-list equations-list)
	(cpcb=purify-equation (first terms1) (first terms2) theory-list equations-list)
      (cpcb=purify-equations-aux (rest terms1) (rest terms2)
				 theory-list equations-list))))

(defun cpcb=purify-equations-aux-pred (terms theory-list equations-list)
  (declare (edited  "27-AUG-1997")
	   (authors Doelfer)
	   (input   "A list of terms (predicate notion), a list of theories"
		    "and a list of equations.")
	   (effect  "THEORY-LIST and EQUATIONS_LIST are destructively changed")
	   (value   "Multiple value:"
		    "1. A list of theories (the first element ist nil)"
		    "2. A list of equations"))
  (if (endp terms)
      (values theory-list equations-list)
      (multiple-value-bind (theory-list equations-list)
	  (cpcb=purify-equation-pred (first terms) theory-list equations-list)
	(cpcb=purify-equations-aux-pred (rest terms) theory-list equations-list))))

(defun cpcb=purify-equation (term1 term2 theory-list equations-list)
  (declare (edited  "06-NOV-1995")
	   (authors Richts)
	   (input   "Two terms, a list of theories and a list of equations.")
	   (effect  "THEORY-LIST and EQUATIONS-LIST are changed destructively.")
	   (value   "Variable abstraction is performed on TERM1 = TERM2,"
		    "i.e. all new equations are added to EQUATIONS-LIST (if a"
		    "new theory appears it is added to THEORY-LIST). The result"
		    "is a multiple value:"
		    "1. The list of theories."
		    "2. The list of equations."))
  (let ((theory1 (th~term-theory term1))
	(theory2 (th~term-theory term2)))
    (multiple-value-bind (new-term1 theory-list equations-list)
	(cpcb=purify-term term1 theory1 theory-list equations-list)
      (multiple-value-bind (new-term2 theory-list equations-list)
	  (cpcb=purify-term term2 theory2 theory-list equations-list)
	(if (or (eq theory1 theory2)
		(sym~variable-p new-term1)
		(sym~variable-p new-term2))
	    ;; both sides have same theory or at least one is a variable
	    (cpcb=purify-insert-equation! new-term1 new-term2
					  (or theory1 theory2)
					  theory-list
					  equations-list)
	    ;; sides have different theories, and are non-variable
	    (let ((new-var (fo~variable-create)))
	      (multiple-value-bind (theory-list equations-list)
		  (cpcb=purify-insert-equation! new-var new-term1
						theory1
						theory-list
						equations-list)
		(cpcb=purify-insert-equation! new-var new-term2
					      theory2
					      theory-list
					      equations-list))))))))

(defun cpcb=purify-equation-pred (term theory-list equations-list)
  (declare (edited  "29-AUG-1997")
	   (authors Doelfer)
	   (input   "A term (predicate-notion) e.g. an equation,"
		    "a list of theories and a list of equations.")
	   (effect  "THEORY-LIST and EQUATIONS-LIST are changed destructively.")
	   (value   "Variable abstraction is performed on 'equation' term,"
		    "i.e. all new equations are added to EQUATIONS-LIST (if a"
		    "new theory appears it is added to THEORY-LIST). The result"
		    "is a multiple value:"
		    "1. The list of theories."
		    "2. The list of equations."))
  (if (string-equal (sym~name (appl~function term)) "=")
      (let ((terms (appl~arguments term)))
	(cpcb=purify-equation (first terms) (second terms) theory-list equations-list))
      (multiple-value-bind (new-term theory-list equations-list)
	  (cpcb=purify-term term (th~term-theory term) theory-list equations-list)
	(cpcb=purify-insert-pred! new-term (th~term-theory term) theory-list equations-list))))

(defun cpcb=purify-term (term theory theory-list equations-list)
  (declare (edited  "06-NOV-1995")
	   (authors Richts)
	   (input   "A term, the theory of the superterm, a list of theories and"
		    "a list of equations.")
	   (effect  "THEORY-LIST and EQUATIONS-LIST are changed destructively.")
	   (value   "If the top-symbol of TERM does not belong to THEORY, an"
		    "equation between TERM and a new variable is added to"
		    "EQUATIONS-LIST.  This operation is performed recursively on"
		    "all subterms of TERM (with the theory of TERM).  Thereby"
		    "all alien subterms of TERM are replaced with new variabels"
		    "and corresponding equations are added to EQUATIONS-LIST."
		    "The result is a multiple value:"
		    "1. A term corresponding to TERM where all alien subterms are"
		    "   replaced by new variables, or a new variable if TERM does not"
		    "   belong to THEORY."
		    "2. The list of theories."
		    "3. The list of equations."))
  (if (sym~variable-p term)
      (values term theory-list equations-list)
      (let ((new-theory (th~term-theory term)))
	(multiple-value-bind (new-arguments theory-list equations-list)
	    (cpcb=purify-arguments (fo~arguments term) new-theory
				   theory-list equations-list)
	  (let ((new-term (if new-arguments
			      (appl~create (appl~function term) new-arguments)
			      term)))
	    (if (eq theory new-theory)
		(values new-term theory-list equations-list)
		(let ((new-var (fo~variable-create)))
		  (multiple-value-bind (theory-list equations-list)
		      (cpcb=purify-insert-equation! new-var new-term
						    new-theory
						    theory-list
						    equations-list)
		    (values new-var theory-list equations-list)))))))))

(defun cpcb=purify-arguments (arguments theory theory-list equations-list)
  (declare (edited  "06-NOV-1995")
	   (authors Richts)
	   (input   "A list of terms, the theory of the superterm, a list of"
		    "theories and a list of equations.")
	   (effect  "THEORY-LIST and EQUATIONS-LIST are changed destructively.")
	   (value   "This function calls CPCB=PURIFY-TERM on each term in"
		    "ARGUMENTS. The result is like in this function, except that"
		    "the first value is a list of terms."))
  (if (null arguments)
      (values nil theory-list equations-list)
      (multiple-value-bind (new-argument theory-list equations-list)
	  (cpcb=purify-term (first arguments) theory
			    theory-list equations-list)
	(multiple-value-bind (new-arguments theory-list equations-list)
	    (cpcb=purify-arguments (rest arguments) theory
				   theory-list equations-list)
	  (values (cons new-argument new-arguments) theory-list equations-list)))))


(defun cpcb=purify-insert-equation! (term1 term2 theory theory-list equations-list)
  (declare (edited  "06-NOV-1995")
	   (authors Richts)
	   (input   "Two terms (which are pure), the theory (of all symbols in"
		    "both terms), a list of theories and a list of equations.")
	   (effect  "THEORY-LIST and EQUATIONS-LIST are changed destructively.")
	   (value   "The equation TERM1 = TERM2 is added to the equations in"
		    "EQUATIONS-LIST corresponding to THEORY.  If there are no"
		    "such equations (i.e. THEORY does not occur in THEORY-LIST),"
		    "THEORY is added to THEORY-LIST and a new equations-object"
		    "containing the new equation is added to EQUATIONS-LIST."))
  (let ((index (position theory theory-list)))
    (if index
	(let ((equations (elt equations-list index)))
	  (setf (elt equations-list index)
		(unicp~eq-create (cons term1 (unicp~eq-lhss equations))
				 (cons term2 (unicp~eq-rhss equations))))
	  (values theory-list equations-list))
	(progn
	  (push theory (rest theory-list))
	  (push (unicp~eq-create (list term1) (list term2)) (rest equations-list))
	  (values theory-list equations-list)))))

(defun cpcb=purify-insert-pred! (pterm theory theory-list equations-list)
 (declare (edited  "20-SEP-1995")
	  (authors Doelfer)
	  (input   "A Term (which is pure and in predicate notion), the theory (of all symbols in"
		   "both terms), a list of theories and a list of equations.")
	  (effect  "THEORY-LIST and EQUATIONS-LIST are changed destructively.")
	  (value   "The 'equation' PTERM is added to the equations in"
		   "EQUATIONS-LIST corresponding to THEORY.  If there are no"
		   "such equations (i.e. THEORY does not occur in THEORY-LIST),"
		   "THEORY is added to THEORY-LIST and a new equations-object"
		   "containing the new equation is added to EQUATIONS-LIST."))
 (let ((index (position theory theory-list)))
   (if index
      (let ((equations (elt equations-list index)))
	(setf (elt equations-list index) (cons pterm equations))
	(values theory-list equations-list))
      (progn
	(push theory (rest theory-list))
	(push (list pterm) (rest equations-list))
	(values theory-list equations-list)))))


;;; Choice Points

(defun cpcb=ready-p (up)
  (declare (edited  "20-FEB-1996")
	   (authors Richts)
	   (input   "A general unification problem.")
	   (effect  "None.")
	   (value   "True, iff all decision for the shared variables of UP have"
		    "been made (identification/discrimination, theory index and"
		    "ordering)."))
  (if cpcb*iterative
       (and (every #'(lambda (elem-up)
		(let ((variables (unicp~repres-list (unicp~comb-variables elem-up))))
		  (and (notany #'(lambda (var)
				   (cpcb=label-not-ready-p var (unicp~theory elem-up)))
			       variables)
		       (not (cpcb=some-var-pair
			     #'(lambda (var1 var2)
				 (cpcb=order-not-ready-p var1 var2
							 (unicp~theory elem-up)))
			     variables)))))
	    (misc~queue-list (unicp~gup-up-queue up)))
	    ;(break  "iterativ")
	    )
      (let ((variables (unicp~repres-list (unicp~shared-variables up))))
	(and (notany #'(lambda (var)
			 (cpcb=label-not-ready-p var nil))
		     variables)
	     (not (cpcb=some-var-pair #'(lambda (var1 var2)
					  (cpcb=order-not-ready-p var1 var2 nil))
				      variables))
	     (or ;(break "nicht iterativ")
		 t)))))


(defun cpcb=label-not-ready-p (var theory)
  (declare (edited  "20-NOV-1996")
	   (authors Richts)
	   (input   "A variable and a theory (or NIL).")
	   (effect  "None.")
	   (value   "True, iff the label of VAR is not ready for THEORY, i.e. iff"
		    "the label is not set and THEORY does not occur in the"
		    "NOT-THEORY-slot."
		    "(If THEORY is NIL:"
		    " True, iff the label or VAR is not set.)"))
  (not (or (unicp~var-label-theory var)
	   (member theory (unicp~var-label-not-theory var)))))


(defun cpcb=order-not-ready-p (var1 var2 theory)
  (declare (edited  "20-FEB-1996")
	   (authors Richts)
	   (input   "Two variables and a list of variables.")
	   (effect  "None.")
	   (value   "False, iff VAR1 is less than VAR2, or if VAR2 is less than"
		    "VAR1, or if VAR1 and VAR2 are allready identified or"
		    "discriminatied and they are in the same theory block,"
		    "i.e. they have the same theory index and all variables with"
		    "a different theory index are less than both or greater than"
		    "both."))
  (not (or ;; (unicp~vars-identified-p var1 var2)
	   (unicp~vars-less-p var2 var1)
	   (unicp~vars-less-p var1 var2)
	   (and (or (unicp~vars-discriminated-p var1 var2)
		    (unicp~vars-identified-p var1 var2))
		(or (and (unicp~var-label-theory var1)
			 (eq (unicp~var-label-theory var1)
			     (unicp~var-label-theory var2)))
		    (and theory
			 (not (eq (unicp~var-label-theory var1) theory))
			 (not (eq (unicp~var-label-theory var2) theory)))))
	   #+weg((unicp~vars-discriminated-p var1 var2)
		(every #'(lambda (var)
			   (or (eq (unicp~var-label-theory var1)
				   (unicp~var-label-theory var))
			       (and (unicp~vars-lesseq-p var var1)
				    (unicp~vars-lesseq-p var var2))
			       (and (unicp~vars-lesseq-p var1 var)
				    (unicp~vars-lesseq-p var2 var))))
		       variables))
	   #+weg(notany #'(lambda (th)
		       (member th (unicp~var-label-theories var1)))
		   (unicp~var-label-theories var2)))))
		       
(defun cpcb=some-var-pair (predicate variables)
  (declare (edited  "20-FEB-1996")
	   (authors Richts)
	   (input   "A predicate function and a list of variables.  The"
		    "predicate take two variables as arguments.")
	   (effect  "None.")
	   (value   "True, iff for some pair of VARIABLES (where the first"
		    "variable of the pair precedes the second one in VARIABLES)"
		    "PREDICATE returns true."))
  (if (null variables)
      nil
      (or (some #'(lambda (var)
		    (funcall predicate (first variables) var))
		(rest variables))
	  (cpcb=some-var-pair predicate (rest variables)))))

;;;
;;; Aenderung (x)
;;;


(defun cpcb=choice-point (up)
  (declare (edited  "20-FEB-1996")
	   (authors Richts)
	   (input   "A general unification problem.")
	   (effect  "The alternatives for this choice point are computed.  The"
		    "datastructures are updated with the first alternative and"
		    "the other alternatives are stored in the CHOICE-POINTS of"
		    "UP.")
	   (value   "Undefined."))
  (multiple-value-bind (idents not-idents var-indices not-var-indices const-restrs)
      (cpcb=select-choice-point up)
    (cpcb=add-choice-point up idents not-idents var-indices not-var-indices const-restrs))
  (cpcb=bktrk-next-choice up))

(defun cpcb=backtrack (up)
  (declare (edited  "20-FEB-1996")
	   (authors Richts)
	   (input   "A general unification problem.")
	   (effect  "Backtracking is performed:  The status of the last"
		    "choice-point of UP (that still has alternatives) is"
		    "restored (i.e. the status of UP, of all variables and of"
		    "the elementary unification problems in UP); then the next"
		    "alternative is chosen, i.e. the datastructures are updated.")
	   (value   "Undefined."))
  (cpcb=bktrk-restore up)
  (cpcb=bktrk-next-choice up))

(defun cpcb=bktrk-save-choice-point (up)
  (declare (edited  "20-FEB-1996")
	   (authors Richts)
	   (input   "A general unification problem.")
	   (effect  "The current status is saved so it can be restored at the"
		    "next backtracking step.  The status of the label of the"
		    "variables in UP is saved on UP's label-stack.  Each"
		    "elementary unification problem is called with the"
		    "unicp~up-save-choice-point method so it can save its status"
		    "itself.")
	   (value   "Undefined."))
  (push (mapcar #'(lambda (var)
		    (unicp~var-label-copy (unicp~orig-label var)))
		(unicp~variables up))
	(unicp~gup-label-stack up))
  (mapc #'unicp~up-save-choice-point
	(misc~queue-list (unicp~gup-up-queue up)))
  )

;;;
;;; Aenderung: lqueue mit 5 Listen, misc! (x)
;;;

(defun cpcb=bktrk-restore (up)
  (declare (edited  "20-FEB-1996")
	   (authors Richts)
	   (input   "A general unification problem.")
	   (effect  "The status of the last choice point (that still has"
		    "alternatives) is restored: The labels of the variables in"
		    "UP are set to the top of UP's label-stack.  Each elementary"
		    "unification problem in UP is called with the"
		    "unicp~up-restore-choice-point method and the results are"
		    "put in UP's up-queue.  Finally UP's constraint-queue is"
		    "filled with NILs.")
	   (value   "Undefined."))
  (mapc #'(lambda (var label)
	    (unicp~set-orig-label! var label))
	(unicp~variables up)
	(pop (unicp~gup-label-stack up)))
  (let ((up-queue (unicp~gup-up-queue up))
	(constr-queue (unicp~gup-constraint-lqueue up)))
    (mapc #'(lambda (up)
	      (misc~queue-insert! up-queue (unicp~up-restore-choice-point up))
	      (misc~queue-delete! up-queue)
	      (misc~lqueue-insert! constr-queue nil nil nil nil nil)
	      (misc~lqueue-insert! constr-queue nil nil nil nil nil)
	      (misc~lqueue-delete! constr-queue)
	      (misc~lqueue-delete! constr-queue))
	  (copy-list (misc~queue-list up-queue)))))


;;;
;;; Aenderung: lqueue mit 5 Listen, misc! (x)
;;;


(defun cpcb=bktrk-next-choice (up)
  (declare (edited  "20-FEB-1996")
	   (authors Richts)
	   (input   "A general unification problem.")
	   (effect  "The next alternative of the last choice point is popped"
		    "from UP's choice-points.  If this is the last alternative,"
		    "the complete choice point is deleted (so the next choice"
		    "point can be accessed at the next backtracking step), else"
		    "the current status is saved (since it has been deleted"
		    "during the last restore operation).  After this, the"
		    "information of the selected alternative is processed"
		    "through the order module (where the variable labels are"
		    "updated) and the resulting information is inserted into"
		    "UP's constrained-queue.")
	   (value   "Undefined."))
  (if (unicp~gup-choice-points up)
      (multiple-value-bind (ident not-ident var-index not-var-index const-restr)
	  (values-list (maplist #'(lambda (tail)
				    (pop (first tail)))
				(first (unicp~gup-choice-points up))))
	(cpcb==trace-next-choice up ident not-ident var-index not-var-index const-restr)
	(if (some #'null (first (unicp~gup-choice-points up)))
	    (if (every #'null (first (unicp~gup-choice-points up)))
		(pop (unicp~gup-choice-points up))
		(error "Programmierfehler: alle Listen in ~A muessen NIL sein."
		       (first (unicp~gup-choice-points up))))
	    (cpcb=bktrk-save-choice-point up))
	;(break "ord:")
	(multiple-value-bind (consistent ident1 not-ident1 var-index1 not-var-index1 const-restr1
					 ident2 not-ident2 var-index2 not-var-index2 const-restr2)
	    (ord~order up ident not-ident var-index not-var-index const-restr)
	  (unless consistent
	    (error "CPCB: Programmierfehler, getroffene Wahl ist nicht konsistent."))
	  (let ((ident (append ident1 ident2))
		(not-ident (append not-ident1 not-ident2))
		(var-index (append var-index1 var-index2))
		(not-var-index (append not-var-index1 not-var-index2))
		(const-restr (append const-restr1 const-restr2)))
	    (cpcb=insert-new-constraint! (unicp~gup-constraint-lqueue up)
					 ident not-ident var-index not-var-index const-restr)
	    ;(cpcb=vars-set! ident not-ident var-index nil const-restr nil)
	    )))))

;;;
;;; Aenderung: lqueue mit 5 Listen, misc! (x)
;;;


(defun cpcb=insert-new-constraint! (lqueue ident not-ident var-index not-var-index const-restr)
  (declare (edited  "20-FEB-1996")
	   (authors Richts)
	   (input   "A constraint queue (which is a list-queue) and four lists"
		    "of constraints.")
	   (effect  "The new constraints are inserted into LQUEUE and the last"
		    "list is deleted from LQUEUE.")
	   (value   "Undefined."))
  (misc~lqueue-insert! lqueue ident not-ident var-index not-var-index const-restr)
  (misc~lqueue-delete! lqueue))

#+weg(defun cpcb=insert-new-constraint! (up-queue lqueue theory
				    ident not-ident const-vars const-restr)
  (let ((up (misc~queue-front up-queue)))
    (cond ((eq theory t)
	   (misc~lqueue-insert! lqueue ident not-ident const-vars const-restr)
	   (misc~lqueue-delete! lqueue))
	  ((eq theory (unicp~theory up))
	   (misc~lqueue-insert! lqueue ident not-ident const-vars const-restr)
	   (misc~lqueue-insert! lqueue nil nil nil nil)
	   (misc~lqueue-delete! lqueue)
	   (misc~lqueue-delete! lqueue)
	   (misc~queue-insert! up-queue up)
	   (misc~queue-delete! up-queue))
	  (t
	   (misc~queue-insert! up-queue up)
	   (misc~queue-delete! up-queue)
	   (cpcb=insert-new-constraint! up-queue lqueue theory
					ident not-ident const-vars const-restr)))))

;;;
;;; Aenderung: lqueue mit 5 Listen, misc! (x)
;;;

(defun cpcb=add-choice-point (up idents not-idents const-varss not-const-varss const-restrs)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "A general unfification problem and four lists of constraints.")
	   (effect  "The constraints are pushed on UP's choice-point stack.")
	   (value   "Undefined."))
  (push (list idents not-idents const-varss not-const-varss const-restrs)
	(unicp~gup-choice-points up))
  )

;;;
;;; Aenderung (x)   
;;;


(defun cpcb=select-choice-point (up &optional (gup-list (unicp~gup-up-list up)))
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "A general unification problem.")
	   (effect  "None.")
	   (value   "The next choice point is selected, i.e. the alternatives"
		    "are computed.  The result is a multiple value of four lists"
		    "of equal length.  Each list contains the alternatives for"
		    "one constraint type (identification, discrimination, theory"
		    "index and ordering).  So the first alternative consists of"
		    "the first elements of the four values, the second"
		    "alternative of the second elements, and so on.  Since only"
		    "one decision is made in this choice point, for each"
		    "alternative all but one value is empty and the non-empty"
		    "value contains only one object."))
  (let* ((variables
	  (if cpcb*iterative
	      (unicp~repres-list (unicp~comb-variables (first gup-list)))
	      (unicp~repres-list (unicp~shared-variables up))))
	 (theory (if cpcb*iterative
		     (unicp~theory (first gup-list))
		     nil))
	 (unident1 (cpcb=some-var-pair #'(lambda (var1 var2)
					  (and (unicp~var-label-theory var1)
					       (eq (unicp~var-label-theory var1)
						   (unicp~var-label-theory var2))
					       (not (unicp~vars-identified-p var1 var2))
					       (not (unicp~vars-discriminated-p var1 var2))
					       (unicp~id-create var1 var2)))
				       ;; variables))
			    (unicp~repres-list (unicp~shared-variables up))))
	 (unident (or unident1
		      (cpcb=some-var-pair #'(lambda (var1 var2)
					      (and (not (unicp~vars-identified-p var1 var2))
						   (not (unicp~vars-discriminated-p var1 var2))
						   (unicp~id-create var1 var2)))
					  variables)))
	 (var (or unident
		  (find-if #'(lambda (var)
			       (cpcb=label-not-ready-p var theory))
			   variables)))
	 (unorder (or var
		      (cpcb=some-var-pair #'(lambda (var1 var2)
					      (if (cpcb=order-not-ready-p
						   var1 var2 theory)
						  (cons var1 var2)))
					  variables))))
    (when (and nil (<= counter 1) (equal (format nil "~A" unident) "(X1 x-10)"))
      (break))
    (cond (unident
	   (values (list (list unident) nil)
		   (list nil (list unident))
		   (list nil nil)
		   (list nil nil)
		   (list nil nil)))
	  (var
	   (let ((theory
		  (if cpcb*iterative
		      theory
		      (unicp~theory
		       (find-if-not
			#'(lambda (theory)
			    (member theory (unicp~var-label-not-theory var)))
			(misc~queue-list (unicp~gup-up-queue up))
			:key #'unicp~theory)))))
	     (if var
		 (values (list nil nil)
			 (list nil nil)
			 (list (list (unicp~vi-create var theory)) nil)
			 (list nil (list (unicp~vi-create var theory)))
			 (list nil nil))
		 (error
		  "CPCB:  Shit, found no more theories, which could assigned to that variable"))))
#|	   (let* ((theories (mapcar #'unicp~theory
				    (misc~queue-list (unicp~gup-up-queue up))))
		  (var-indices (mapcar #'(lambda (th)
					   (list (unicp~vi-create var th)))
				       theories))
		  (nil-list (make-sequence 'list (length theories)
					   :initial-element nil)))
	     (values nil-list nil-list var-indices nil-list)))
|#
          (unorder
	   (values (list nil nil) (list nil nil) (list nil nil) (list nil nil)
		   (list (list (unicp~cr-create (car unorder) (cdr unorder)))
			 (list (unicp~cr-create (cdr unorder) (car unorder))))))
	  ((or (null (rest gup-list))
	       (not cpcb*iterative))
	   (error "CPCB:  Oh oh, found no more choice point, but was told there is some."))
	  (t (cpcb=select-choice-point up (cdr gup-list))))))

    
;;; Maintaining variable information

(defun cpcb=vars-set! (ident not-ident var-index not-var-index
			     const-restr not-const-restr)
  (declare (edited  "29-SEP-1995")
	   (authors Richts)
	   (input   "Six lists of constraints.")
	   (effect  "The labels of the affected variables are updated.")
	   (value   "Undefined.")
	   (ignore not-const-restr))
  (mapc #'(lambda (var-pair)
	    (unicp~vars-identify! (unicp~id-left var-pair) (unicp~id-right var-pair)))
	ident)
  (mapc #'(lambda (var-pair)
	    (unicp~vars-discriminate! (unicp~id-left var-pair) (unicp~id-right var-pair)))
	not-ident)
  (mapc #'(lambda (var-ind)
	    (unless (unicp~var-label-theory (unicp~vi-var var-ind))
	      (setf (unicp~var-label-theory (unicp~vi-var var-ind))
		    (unicp~vi-theory var-ind))))
	var-index)
  (mapc #'(lambda (var-ind)
	    (push (unicp~vi-theory var-ind)
		  (unicp~var-label-not-theory (unicp~vi-var var-ind))))
	not-var-index)
  (mapc #'(lambda (restr)
	    (let ((var1 (unicp~cr-const restr)))
	      (mapc #'(lambda (var2)
			(unless (or (eq var1 var2)
				    (unicp~vars-identified-p var1 var2)
				    (unicp~vars-lesseq-p var1 var2))
			  (push var1 (unicp~var-label-greater var2))
			  (push var2 (unicp~var-label-less var1))))
		    (unicp~cr-varlist restr))))
	const-restr))


(defun cpcb=compute-result (up)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "A general unification problem.")
	   (effect  "None.")
	   (value   "A list containing the status of all variables in UP.  This"
		    "function is used to create an output of the algorithm."))
  (let ((variables (unicp~repres-list (unicp~variables up))))
    (list (mapcar #'unicp~var-label-identified variables)
	  (mapcar #'(lambda (var)
		      (cons var (unicp~var-label-theory var)))
		  variables)
	  (sort variables #'< :key #'(lambda (var) (length (unicp~var-label-less var))))
	  (mapcar #'(lambda (var)
		      (cons var (unicp~var-label-greater var)))
		  variables))))



;;; Initial UPs

(eval-when (load compile eval)
 (defclass cpcb+initial-up (unicp+elementary-up)
   ()
   (:documentation
    "This class is used for initial UPs, simply to store general-up, theory, and
    equations before starting some computation.  The stored information is used to
    create a UP of the class appropriate for the theory.")))


(defun cpcb=initial-up-create (theory general-up equations)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "A theory, a general unification problem, and some equations.")
	   (effect  "None.")
	   (value   "An initial unification problem for THEORY, with EQUATIONS"
		    "and GENERAL-UP."))
  (make-instance 'cpcb+initial-up
		 :theory theory :general-up general-up :equations equations))

(defun cpcb=initial-up-p (object)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "An object.")
	   (effect  "None.")
	   (value   "True, iff OBJECT is an initial unification problem."))
  (typep object 'cpcb+initial-up))

(defmethod unicp~new-constr ((elementary-up cpcb+initial-up)
			     identified discriminated constants const-restr)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "An initial unification problem and four lists of constraints.")
	   (effect  "Some changes might be made in the variables by the elementary algorithm.")
	   (value   "The unicp~theo-init method is called for the theory and"
		    "equations in ELEMENTARY-UP."))
  (unicp~theo-init (unicp~theory elementary-up)
		   (unicp~general-up elementary-up)
		   (unicp~equations elementary-up)
		   identified discriminated constants const-restr))

;;;
;;; Aenderung
;;;

(defun cpcb=new-constr (up id not-id var-index not-var-index restr)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "An elementary unification problem and four lists of constraints.")
	   (effect  "Some changes might be made by the elementary algorithm.")
	   (value   "This function is used to transform the different"
		    "representations of theory indices in the combination"
		    "algorithm and the elementary algorithms.  In the former"
		    "theory indices are consisting of the variable and the"
		    "assigned theory; in the latter theory indices simply"
		    "contain the variables that have a different theory,"
		    "i.e. that have to be treated as constants.  This function"
		    "calls unicp~new-constr with the transformed VAR-INDEX."))
  (let* ((theory (unicp~theory up))
	 (const-vars
	  (union
	   (mapcan
	    #'(lambda (vi) (if (not (eq theory (unicp~vi-theory vi))) (list (unicp~vi-var vi))))
	    var-index)
	   (mapcan
	    #'(lambda (vi) (if (eq theory (unicp~vi-theory vi)) (list (unicp~vi-var vi))))
	    not-var-index))))
    (unicp~new-constr up id not-id const-vars restr)))

(defvar cpcb*bcktrck-counter 0)

(defun cpcb==trace-backtracking (up result)
  (format t "~%Backtracking(~A): " result up)
  ;(pprint (cpcb=compute-result up))
  )

(defun cpcb==trace-next-choice (up ident not-ident var-index not-var-index const-restr)
  (when  nil (equal (format nil "~A" ident) "((X1 x-10))")
    (decf counter)
    (when (<= counter 0)
      (break)))
  (when nil
    (format t "~%(Backtracking) Next Choice" up)
    (when ident
      (format t "~%  ident: ~A" ident))
    (when not-ident
      (format t "~%  not-Ident: ~A" not-ident))
    (when var-index
      (format t "~%  var-index: ~A" var-index))
    (when not-var-index
      (format t "~%  not-var-index: ~A" not-var-index))
    (when const-restr
      (format t "~%  const-restr: ~A" const-restr))
    (cpcb==print-vars-label (unicp~variables up))))

(defun cpcb==print-vars-label (vars)
  (mapc #'(lambda (var)
	    (let ((repres (unicp~var-label-repres var)))
	      (when (eq var repres)
		(format t "~%Vars: ~A : ~A  ~%   Discrim: ~A  ~%   Order: ~A < ~A < ~A"
			(unicp~var-label-identified repres)
			(if (unicp~var-label-theory repres)
			    (if (th~symbols (unicp~var-label-theory repres))
				(first (th~symbols (unicp~var-label-theory repres)))
				"Free")
			    (mapcar #'(lambda (up) (first (th~symbols up)))
				    (unicp~var-label-not-theory repres)))
			(unicp~var-label-discrim-lists repres)
			(unicp~var-label-less repres)
			repres
			(unicp~var-label-greater repres))
		(cpcb==check-ident repres))))
	vars))

(defun cpcb==check-ident (var)
  (every #'(lambda (v)
	     (if (and (eq (unicp~var-label-theory var)
			  (unicp~var-label-theory v))
		      (or (unicp~var-label-theory var)
			  (cpcb=set-equal (unicp~var-label-not-theory var)
					  (unicp~var-label-not-theory v)))
		      (cpcb=set-equal (unicp~var-label-identified var)
				      (unicp~var-label-identified v))
		      (cpcb=set-equal (unicp~var-label-discrim-lists var)
				      (unicp~var-label-discrim-lists v)
				      :test #'cpcb=set-equal)
		      #|(cpcb=set-equal (unicp~var-label-less var)
				      (unicp~var-label-less v))
		      (cpcb=set-equal (unicp~var-label-greater var)
				      (unicp~var-label-greater v))|#)
		 t
		 (and nil (break "~A und ~A haben nicht dieselbe Information." var v) nil)))
	 (unicp~var-label-identified var)))

(defun cpcb=set-equal (set1 set2 &key (test #'eql))
  (and (subsetp set1 set2 :test test)
       (subsetp set2 set1 :test test)))



