(in-package "CLIM-USER")

(use-package (list "KEIM"))

(import '(cpframe~display-elementary-up cpframe~decide-unify)
	(find-package "KEIM"))

(setf cpframe*color-p nil)
(setf cpframe*color-p t)

(setf cpframe*gray-list (list (clim::make-gray-color 1.0)
			      (clim::make-gray-color 0.5) (clim::make-gray-color 0.8)
			      (clim::make-gray-color 0.6) (clim::make-gray-color 0.4)
			      (clim::make-gray-color 0.9) (clim::make-gray-color 0.7)
			      (clim::make-gray-color 0.2) (clim::make-gray-color 0.3) 
			      (clim::make-gray-color 0.1)
			      ))

(define-application-frame cpframe+frame ()
  ((up :initarg :up :reader cpframe~up)
   (result-flag :initform :initial :accessor cpframe=result-flag)
   (result-list :initform '(:initial) :accessor cpframe=result-list))
  (:panes (equations :application
		     :display-function 'cpframe=display-equations
		     :display-after-commands nil
		     :default-text-style '(:sans-serif :roman :large))
	  (variables :application
		     :display-function 'cpframe=display-variables
		     :display-after-commands nil
		     :default-text-style '(:sans-serif :roman :normal))
	  (queues :application
		  :display-function 'cpframe=display-queues
		  :display-after-commands nil
		  :default-text-style '(:sans-serif :roman :normal))
	  (output :application
		  :display-function 'cpframe=display-output
		  :display-after-commands nil
		  :default-text-style '(:sans-serif :roman :normal))
	  (choice-points :application
			 :display-function 'cpframe=display-choice-points
			 :display-after-commands nil
			 :default-text-style '(:sans-serif :roman :normal))
	  (elementary-up :application
			 :display-function 'cpframe=display-elementary-up
			 :display-after-commands nil
			 :default-text-style '(:sans-serif :roman :normal))
	  (menu :command-menu))
  (:layouts (standard (vertically ()
			(4/10 (horizontally ()
				(8/10 (vertically ()
					(1/4 equations)
					(3/4 queues)))
				(2/10 choice-points)))			
			(5/10 (horizontally ()
				(5/10 variables)
				(5/10 (vertically ()
					(5/10 output)
					(5/10 elementary-up)))))
			(1/10 (horizontally () (1 menu))))))
  (:menu-bar nil)
  (:command-definer cpframe=define-command)
  )


;;; Display Functions

(defmethod cpframe=display-equations ((frame cpframe+frame) stream)
  (with-slots (up) frame
    ;(print (uni~initial-problem up) stream)
    (with-drawing-options (stream :text-face :bold)
      (format stream "Equations:~%"))
    (cpframe=print-equations (uni~initial-problem up) stream)
    #+weg(mapc #'(lambda (el-up)
	      (terpri stream)
	      (terpri stream)
	      (cpframe=print-equations (unicp~equations el-up) stream))
	  (misc~queue-list (unicp~gup-up-queue up)))))


(defmethod cpframe=display-variables ((frame cpframe+frame) stream)
  (with-slots (up) frame
    (cpframe=print-graph (unicp~variables up) stream)))


(defmethod cpframe=display-queues ((frame cpframe+frame) stream)
  (with-slots (up) frame
    (let ((ql (misc~queue-list (unicp~gup-up-queue up)))
	  (cpl (labels ((cpframe=pair-list (list)
		       (if list
			   (cons (cons (first list) (second list))
				 (cpframe=pair-list (rest (rest list))))
			   nil)))
		 (cpframe=pair-list (misc~lqueue-tuples
				     (unicp~gup-constraint-lqueue up))))))
    (formatting-table (stream :x-spacing '(20 :point))
      (mapc #'(lambda (el-up constr-pair)
		(formatting-column (stream)
		  (formatting-cell (stream :align-x :center :align-y :bottom)
		    (cpframe=print-equations (unicp~equations el-up) stream))
		  (formatting-cell (stream :align-x :center)
		    (formatting-table (stream)
		      (cpframe=print-constr (car constr-pair) stream "Unif")
		      (cpframe=print-constr (cdr constr-pair) stream "Order")))))
	    ql cpl
	    #+weg(misc~queue-list (unicp~gup-up-queue up))
	    #+weg(labels ((cpframe=pair-list (list)
		       (if list
			   (cons (cons (first list) (second list))
				 (rest (rest list)))
			   nil)))
	      (cpframe=pair-list (misc~lqueue-tuples
				  (unicp~gup-constraint-lqueue up))))))))
  #+weg(terpri stream)
  #+weg(formatting-table (stream)
    (dotimes (x 256)
      (formatting-column (stream)
	(formatting-cell (stream)
	  (print x stream))
	(formatting-cell (stream)
	  (draw-text* stream (code-char x) 1 10))
	(formatting-cell (stream)
	  (print (code-char x) stream))))))

(defmethod cpframe=display-output ((frame cpframe+frame) stream)
  (let ((result-flag (cpframe=result-flag frame))
	(result-list (cpframe=result-list frame))
	(constr-list (last (misc~lqueue-tuples
				 (unicp~gup-constraint-lqueue (cpframe~up frame)))
			   2)))
    (with-drawing-options (stream :text-face :bold :text-size :larger)
      (format stream "Result:")
      (case (first result-list)
	((nil) (format stream "unsolvable"))
	(:initial nil)
	(t (terpri stream)
	   (with-drawing-options (stream :text-face :roman :text-size :smaller)
	     (formatting-table (stream)
	       (cpframe=print-constr (rest result-list) stream "Original")))))
      (terpri stream)
      (case result-flag
	(:initial (format stream "just started"))
	(:solvable (format stream "solvable"))
	(:not-solvable (format stream "not solvable"))
	((nil) (format stream "Backtracking"))
	(t
	 (when (eq result-flag :choice-point)
	   (format stream "Choice Point"))
	 (terpri stream)
	 (with-drawing-options (stream :text-face :roman :text-size :smaller)
	   (formatting-table (stream)
	     (cpframe=print-constr (first constr-list) stream "Unif")
	     (cpframe=print-constr (second constr-list) stream "Order"))))))))

(defmethod cpframe=display-choice-points ((frame cpframe+frame) stream)
  (with-slots (up) frame
    (with-drawing-options (stream :text-face :bold :text-size :larger)
      (format stream "Choice~%Points:~%"))
    (cpframe=print-choice-points (unicp~gup-choice-points up) stream)))


(defmethod cpframe=display-elementary-up ((frame cpframe+frame) stream)
  (with-slots (up) frame
    (cpframe~display-elementary-up (car (last (misc~queue-list (unicp~gup-up-queue up))))
				   stream)))

(defgeneric cpframe~display-elementary-up (up stream)
  (:method (up stream)
     (format stream "~%~%no information available (~A)"
	     (class-name (class-of (unicp~theory up))))))


;;; Printing

(defun cpframe=print-equations (equations stream)
  (formatting-table (stream :x-spacing '(5 :point))
    (mapc #'(lambda (term1 term2)
	      (formatting-row (stream)
		(formatting-cell (stream :align-x :right)
		  (cpframe=print-term term1 stream nil))
		(formatting-cell (stream :align-x :center)
		  (format stream "="))
		(formatting-cell (stream :align-x :left)
		  (cpframe=print-term term2 stream nil))))
	  (unicp~eq-lhss equations) (unicp~eq-rhss equations))))


(defun cpframe=print-term (term stream &optional supersymbol)
  (cpframe=print-term-aux (th~term-theory term) term stream supersymbol))


(defgeneric cpframe=print-term-aux (theory term stream supersymbol)
  (:method ((theory null) (var sym+var) stream supersymbol)
    (declare (ignore supersymbol))
    (cpframe=print-var var stream))
  (:method ((theory th+general) (symbol sym+sym) stream supersymbol)
    (declare (ignore supersymbol))
    (let ((color (keim~get theory :cpframe=color)))
      (with-drawing-options (stream :ink color :text-face :bold)
	 (princ symbol stream))))
  (:method ((theory th+general) (term appl+appl) stream supersymbol)
    (declare (ignore supersymbol))
    (let ((color (keim~get theory :cpframe=color)))
      (with-drawing-options (stream :ink color :text-face :bold)
	 (format stream "~A(" (appl~function term)))
      (cpframe=print-term (first (appl~arguments term)) stream)
      (mapc #'(lambda (term)
		(with-drawing-options (stream :ink color :text-face :bold)
		   (format stream ", "))
		(cpframe=print-term term stream))
	    (rest (appl~arguments term)))
      (with-drawing-options (stream :ink color :text-face :bold)
	(format stream ")" (appl~function term)))))
  (:method ((theory th+ac) (term appl+appl) stream supersymbol)
    (cpframe=print-term-infix theory term stream supersymbol))
  (:method ((theory th+aci) (term appl+appl) stream supersymbol)
    (cpframe=print-term-infix theory term stream supersymbol)))

(defun cpframe=print-var (var stream &key (theory nil supp) text)
  (let ((theory-out (if supp theory (unicp~var-label-theory var)))
	(text-out (or text (string-downcase (format nil "~A" var)))))
    (with-drawing-options (stream :text-face (if (unicp~shared-variable-p var)
						 '(:bold :italic) :bold))
      (if theory-out
	  (with-drawing-options (stream :ink (keim~get theory-out :cpframe=color))
	    (format stream text-out))
	  (format stream text-out)))))

(defun cpframe=print-term-infix (theory term stream supersymbol)
  (let ((color (keim~get theory :cpframe=color))
	(function (appl~function term)))
    (unless (keim~equal function supersymbol)
      (with-drawing-options (stream :ink color :text-face :bold)
	(format stream "(")))
    (cpframe=print-term (first (appl~arguments term)) stream function)
    (mapc #'(lambda (arg)
	      (with-drawing-options (stream :ink color :text-face :bold)
		(format stream " ~A " (appl~function term)))
	      (cpframe=print-term arg stream function))
	  (rest (appl~arguments term)))
    (unless (keim~equal function supersymbol)
      (with-drawing-options (stream :ink color :text-face :bold)
	(format stream ")" (appl~function term))))))


(defun cpframe=print-constr (constr stream &optional (header ""))
  (when (not (= 5 (length constr)))
    (format *terminal-io* "~%**** ~A" constr)
    (break))
  (formatting-column (stream)
    (formatting-cell (stream)
      (format stream header))
    (formatting-cell (stream)
      (mapc #'(lambda (ident)
		(cpframe=print-ident ident stream " "))
	    (first constr)))
    (formatting-cell (stream)
      (mapc #'(lambda (not-ident)
		(cpframe=print-not-ident not-ident stream " "))
	    (second constr)))
    (formatting-cell (stream)
      (mapc #'(lambda (var-ind)
		(cpframe=print-var-index var-ind stream " "))
	    (third constr)))
    (formatting-cell (stream)
      (mapc #'(lambda (var-ind)
		(cpframe=print-var-index var-ind stream "- "))
	    (fourth constr)))
    (formatting-cell (stream)
      (mapc #'(lambda (const-restr)
		(cpframe=print-const-restr const-restr stream " "))
	    (fifth constr)))
    #+weg(mapc #'(lambda (object)
		   (formatting-cell (stream)
		     (print object stream)))
	       constr)))


(defun cpframe=print-ident (ident stream &optional (seperator ""))
  (cpframe=print-term (unicp~id-left ident) stream)
  (format stream "=")
  (cpframe=print-term (unicp~id-right ident) stream)
  (format stream seperator))

(defun cpframe=print-not-ident (not-ident stream &optional (seperator ""))
  (cpframe=print-term (unicp~id-left not-ident) stream)
  (format stream "=/=")
  (cpframe=print-term (unicp~id-right not-ident) stream)
  (format stream seperator))

(defun cpframe=print-var-index (var-ind stream &optional (seperator ""))
  (let ((theory (unicp~vi-theory var-ind))
	(var (unicp~vi-var var-ind)))
    (cpframe=print-var var stream :theory theory)
    (format stream seperator)))

(defun cpframe=print-const-restr (const-restr stream &optional (seperator ""))
  (cpframe=print-term (first (unicp~cr-varlist const-restr)) stream)
  (mapc #'(lambda (var)
	    (format stream ", ")
	    (cpframe=print-term var stream))
	(rest (unicp~cr-varlist const-restr)))
  (format stream "<=")
  (cpframe=print-term (unicp~cr-const const-restr) stream)
  (format stream seperator))


(defun cpframe=print-choice-points (choice-points stream)
  (when choice-points
    (cpframe=print-choice-points (rest choice-points) stream)
    (formatting-table (stream :x-spacing '(5 :point))
      (mapc #'(lambda (constr)
		(cpframe=print-constr constr stream))
	    (apply #'mapcar #'list (first choice-points))))
    (terpri stream)))
		     

;;; Drawing Graphs

(defun cpframe=graph-next-nodes (var)
  (delete-duplicates
   (remove-if #'(lambda (next-var)
		  (intersection (unicp~var-label-greater var)
				(unicp~var-label-less next-var)))
	      (unicp~var-label-greater var))
   :key #'unicp~var-label-repres))

(defun cpframe=graph-print-var-node (var stream)
  (let* ((theory (unicp~var-label-theory var))
	 (text (string-downcase
		(format nil "~A" var))))
      (with-drawing-options (stream :text-face (if (unicp~shared-variable-p var)
						   '(:bold :italic) :bold))
	(if theory
	    (with-drawing-options (stream :ink (keim~get theory :cpframe=color))
	      (format stream text))
	    (format stream text)))))

(defun cpframe=graph-print-varlist-node (var stream)
  (let* ((var-list (unicp~var-label-identified var))
	 (text (string-downcase
		(if (rest var-list)
		    (format nil "(~A~{ ~A~})" (first var-list) (rest var-list))
		    (format nil "~A" var)))))
      (cpframe=print-var var stream :text text)))

(defun cpframe=graph-draw-arc (stream from-node to-node from-x from-y to-x to-y &rest rest)
  (when rest
    (break "rest: ~A" rest))
  (if (unicp~vars-discriminated-p from-node to-node)
      (draw-line* stream from-x from-y to-x to-y)
      (draw-line* stream from-x from-y to-x to-y :line-dashes t)))

(defvar cpframe*gen-sep 20)

(defun cpframe=graph-draw-arc1 (stream from-node to-node from-x from-y to-x to-y &rest rest)
  (when rest
    (break "rest: ~A" rest from-node to-node))
  (if (and (<= (+ (- from-x to-x) 10) cpframe*gen-sep)
	   (= from-y to-y))
      (let ((dir (zerop (mod (floor (- to-x from-x) cpframe*gen-sep) 2))))
	(draw-ellipse* stream
		       (/ (+ from-x to-x) 2) (/ (+ from-y to-y) 2)
		       (/ (- from-x to-x) 2) 0
		       0 (/ (- from-x to-x) 2.5)
		       :start-angle (if dir 0 pi)
		       :end-angle (if dir pi (* 2 pi))
		       :filled nil))
      (draw-line* stream from-x from-y to-x to-y)))

(defun cpframe=print-graph (variables stream)
  ;(dolist (color cpframe*gray-list))
  #+weg(dotimes (val 11)
    (with-drawing-options (stream :ink (clim::make-gray-color (* val 0.1)) :text-face :bold :text-size :larger)
      (format stream "~A Order:~%" val)))
  #+weg(let ((inks (make-contrasting-inks 7)))
    (dotimes (val 7)
    (with-drawing-options (stream :ink (elt inks val) :text-face :bold :text-size :larger)
      (format stream "~A Order:~%" val))))
  (dolist (var variables)
    (cpframe=print-var var stream)
    (format stream "(")
    (dolist (th (unicp~var-label-not-theory var))
      (cpframe=print-var var stream :theory th))
    (format stream ")  "))
  (terpri stream)
  (format-graph-from-roots
   (remove-if #'(lambda (var) (or (unicp~var-label-less var)
				  (not (eq (unicp~var-label-repres var) var))))
	      variables)
   #'cpframe=print-var
   #'cpframe=graph-next-nodes
   :stream stream
   :merge-duplicates t
   :duplicate-key #'unicp~var-label-repres
   :arc-drawer #'cpframe=graph-draw-arc
   ;:generation-separation 40
   ;:within-generation-separation 40
   :center-nodes t
   )
  (terpri stream)
  (with-drawing-options (stream :text-face :bold :text-size :larger)
    (format stream "~%Equations and Inequations:~%"))
  (let ((repres-vars (unicp~repres-list variables))
	(root-vars nil))
    (mapc #'(lambda (var1)
	      (unless (some #'(lambda (var2)
				(or (unicp~vars-identified-p var1 var2)
				    (unicp~vars-discriminated-p var1 var2)))
			    root-vars)
		(push var1 root-vars)))
	  repres-vars)
    (labels ((cpframe=next-vars (var)
	       (remove-if-not #'(lambda (var2)
				  (and (eq (unicp~var-label-repres var2) var2)
				       (unicp~vars-discriminated-p var var2)))
			      (member var repres-vars))))
      (format-graph-from-roots
       root-vars
       #'cpframe=graph-print-varlist-node
       #'cpframe=next-vars
       :stream stream
       :graph-type :dag
       :merge-duplicates t
       :duplicate-key #'unicp~var-label-repres
       :center-nodes t
       :arc-drawer #'cpframe=graph-draw-arc1))))


;;; Commands

(cpframe=define-command (com-cpframe-exit :menu "Exit") ()
  (frame-exit *application-frame*))

(cpframe=define-command (com-cpframe-nop :menu "NOP") ()
  (window-clear (get-frame-pane *application-frame* 'queues))
  (cpframe=display-queues *application-frame*
			  (get-frame-pane *application-frame* 'queues))
  (window-clear (get-frame-pane *application-frame* 'variables))
  (cpframe=display-variables *application-frame*
			  (get-frame-pane *application-frame* 'variables)))

(cpframe=define-command (com-cpframe-backtrack :menu "Backtrack") ()
  (let* ((frame *application-frame*)
	 (general-up (cpframe~up frame)))
    (keim::cpcb=backtrack general-up)
    (window-clear (get-frame-pane frame 'queues))
    (cpframe=display-queues frame
			    (get-frame-pane frame 'queues))
    (window-clear (get-frame-pane frame 'variables))
    (cpframe=display-variables frame
			       (get-frame-pane frame 'variables))
    
    (window-clear (get-frame-pane frame 'choice-points))
    (cpframe=display-choice-points frame
				   (get-frame-pane frame 'choice-points))
    (window-clear (get-frame-pane frame 'output))
    ;;(window-clear (get-frame-pane frame 'elementary-up))
    (cpframe=display-output frame
			    (get-frame-pane frame 'output))
    (cpframe=display-elementary-up frame
				   (get-frame-pane frame 'elementary-up))))

(cpframe=define-command (com-cpframe-next :menu "Next") ()
  (let ((frame *application-frame*))
    (window-clear (get-frame-pane frame 'queues))
    (cpframe=display-queues frame
			    (get-frame-pane frame 'queues))
    (window-clear (get-frame-pane frame 'variables))
    (cpframe=display-variables frame
			       (get-frame-pane frame 'variables))
    (do ((once nil t))
	((and once #+weg(member (cpframe=result-flag frame) '(:solvable nil))))
      (let* ((general-up (cpframe~up frame))
	     (result (keim::cpcb=next-step general-up))
	     (ready-p (if (eq result :choice-point) (keim::cpcb=ready-p general-up)))
	     (solvable-p (if ready-p
			     (every #'unicp~solvable-p
				    (misc~queue-list (unicp~gup-up-queue general-up))))))
	(setf (cpframe=result-flag frame) result)
	(cond ((and (eq result :choice-point)
		    ready-p)
	       (cond (solvable-p
		      (setf (cpframe=result-flag frame) :solvable))
		     ((null (unicp~gup-choice-points general-up))
		      (setf (cpframe=result-flag frame) :not-solvable))
		     (t (setf (cpframe=result-flag frame) nil)))
	       (unless (null (unicp~gup-choice-points general-up))
		 (keim::cpcb=backtrack general-up)))
	      ((and (null result) (null (unicp~gup-choice-points general-up)))
	       (setf (cpframe=result-flag frame) :not-solvable))
	      ((null result)
	       (keim::cpcb=backtrack general-up))
	      ((eq result :choice-point)
	       (keim::cpcb=choice-point general-up)))))
    (unless (member (cpframe=result-flag frame) '(:solvable nil))
      (window-clear (get-frame-pane frame 'choice-points))
      (cpframe=display-choice-points frame
				     (get-frame-pane frame 'choice-points)))
    (window-clear (get-frame-pane frame 'output))
					;(window-clear (get-frame-pane frame 'elementary-up))
    (cpframe=display-output frame
			    (get-frame-pane frame 'output))
    (cpframe=display-elementary-up frame
				   (get-frame-pane frame 'elementary-up))))


;;;

(defmethod unicp~new-constr :around (elementary-up identified discriminated
							constants const-restr)
  (let ((result-list (multiple-value-list (call-next-method))))
    (when (and (boundp '*application-frame*)
	       (typep *application-frame* 'cpframe+frame))
      (setf (cpframe=result-list *application-frame*)
	    (list (first result-list) (second result-list) (third result-list)
		  (mapcar #'(lambda (var)
			      (unicp~vi-create var (unicp~theory elementary-up)))
			  (fourth result-list))
		  (mapcar #'(lambda (var)
			      (unicp~vi-create var (unicp~theory elementary-up)))
			  (fifth result-list))
		  (sixth result-list))))    ;;; delete not-const-restr
    (values-list result-list)))

;;;;

(defun cpframe==init-variables! (vars)
  (labels ((v (n) (elt vars n)))
    (keim::cpcb=vars-set!
     (list (list (v 0) (v 1)) (list (v 6) (v 9)))
     (do ((vl (subseq vars 4) (cddr vl))
	  (nid nil (cons (list (first vl) (second vl)) nid)))
	 ((or (null vl) (null (rest vl))) nid))
     (subseq vars 0 3) nil
     (append (list (list (v 1) (v 2) (v 3) (v 4) (v 5))
		   (list (v 2) (v 5))
		   (list (v 3) (v 4) (v 5))
		   (list (v 4) (v 5)))
	     (do ((vl (subseq vars 6) (cdddr vl))
		  (cr nil (cons (subseq vl 0 (min 3 (length vl))) cr)))
		 ((null vl) cr)))
     nil)))

(defun cpframe~up-frame-create (termlist1 termlist2)
  (let* ((general-up (keim::cpcb=initial-gup termlist1 termlist2))
	 (up-list (misc~queue-list (unicp~gup-up-queue general-up))))
    (mapcar #'(lambda (up color)
		(keim~put (unicp~theory up) :cpframe=color color))
	    up-list
	    (if cpframe*color-p
		(coerce (make-contrasting-inks (length up-list)) 'list)
		cpframe*gray-list))
    ;(cpframe==init-variables! (unicp~variables general-up))
    (make-application-frame 'cpframe+frame
			    :up general-up
			    :left 100 :top 50 :width 700 :height 800)))


(defun cpframe~decide-unify (terms1 terms2)
  (let (frame)
    (unwind-protect
	(progn
	  (setf frame (cpframe~up-frame-create terms1 terms2))
	  (run-frame-top-level frame)
	  frame)
      (mapc #'(lambda (var)
		(sym~set-label! var nil))
	    (if (and frame
		     (cpframe~up frame)
		     (unicp~variables (cpframe~up frame)))
		(unicp~variables (cpframe~up frame))
		(top~free-variables (list terms1 terms2)))))))

#|

(setf up-frame (cpframe~up-frame-create keim::tl1 keim::tl2))

(run-frame-top-level up-frame)

(cpframe~decide-unify keim::tl1 keim::tl2))
|#


;;; special theories

(import 'keim::cpac*theory)

(defmethod cpframe~display-elementary-up ((up keim::cpac+up) stream)
  (format stream "~%~%")
  (let ((vars (let ((cpac*theory (unicp~theory up)))
		(declare (special cpac*theory))
		(mapcar #'cdr
			(sort (delete-if #'null
					 (mapcar #'(lambda (var)
						     (cons (keim::cpac=position var) var))
						 (unicp~variables up))
					 :key #'car)
			      #'< :key #'car)))))
    (cond ((not (keim::cpac=solutions-p up))
	   (format stream "Diophantine equations are not yet computed."))
	  (vars
	(formatting-table (stream :equalize-column-widths t)
	  (formatting-row (stream)
	    (mapc #'(lambda (var)
		      (formatting-cell (stream :align-x :center)
			(cpframe=print-var var stream)))
		  vars))
	  (mapc #'(lambda (vec)
		    (formatting-row (stream)
		      (some #'(lambda (number)
				(formatting-cell (stream :align-x :center)
				  (princ number stream))
				nil)
			    vec)))
		(keim::cpac=permanent-solutions up))
	  (formatting-row (stream)
	    (mapc #'(lambda (var)
		      (declare (ignore var))
		      (formatting-cell (stream :align-x :center)
			(format stream "-")))
		  vars))
	  (mapc #'(lambda (vec)
		    (formatting-row (stream)
		      (some #'(lambda (number)
				(formatting-cell (stream :align-x :center)
				  (princ number stream))
				nil)
			    vec)))
		(keim::cpac=solutions up))))
	  (t (format stream "AC-terms are equal.")))))

(import 'keim::cpaci*theory)

(defmethod cpframe~display-elementary-up ((up keim::cpaci+up) stream)
  (let ((cpaci*theory (unicp~theory up))
	(formset (keim::cpaci~up-formset up)))
    (declare (special cpaci*theory))
    (keim::cpaci=tab-print formset stream)))


(defmethod cpframe~display-elementary-up ((up keim::cpa+up) stream)
  (let ((cpa*theory (unicp~theory up)))
    (declare (special cpa*theory))
    (format stream "~%~%Gleichungen links: ~S ~% rechts: ~S ~%"
	  (keim::cpa=words-left up) (keim::cpa=words-right up))))
