(in-package "KEIM")


;+----------------------------------------------------------
;+ Zufallsgenerierung von Gleichungsproblemen
;+---------------------------------------------------------


(defvar genex*maxtermdepth 6
  "The maximal depth of a (randomly generated) term.")


(defun genex=some-term (maxdepth functions constants variables)
  ;; one random term of depth at most "maxdepth"
  (if (= maxdepth 1)
      (if (> (random 6) 0)
	  (elt variables (random (length variables)))
	(elt constants (random (length constants))))
    (let ((f (elt functions (random (length functions)))))
      (fo~application-create
       f (genex=some-terms (fo~arity f) (- maxdepth 1)
		     functions constants variables))))) 


(defun genex=some-terms (howmany maxdepth functions constants variables
				 &optional (generated nil))
  ;; list of "howmany" many random terms of depth at most "maxdepth"
  (if (= howmany 0)
      generated
    (genex=some-terms (- howmany 1) maxdepth functions constants variables
		      (cons (genex=some-term (+ 1 (random maxdepth))
					     functions constants variables)
			    generated)))) 


(defun genex=equation (functions constants variables)
  ;; one equation with identical top level symbols
  (let* ((f (elt functions (random (length functions))))
	 (f-stellen (fo~arity f))
	 (depth (+ 1 (random genex*maxtermdepth))))
    (list
     (fo~application-create
      f (genex=some-terms f-stellen depth functions constants variables))
     (fo~application-create
      f (genex=some-terms f-stellen depth functions constants variables))))) 


(defun genex=equations (howmany env)
  ;; list of "howmany" many enquations, randomly generated from env
  (multiple-value-bind (constants functions)
      (env~lookup-constants env)
    (genex=equations-aux
     howmany functions constants (env~lookup-variables env))))


(defun genex=equations-aux (howmany functions constants variables
				    &optional (generated nil))
  (if (= howmany 0)
      generated
    (genex=equations-aux
     (- howmany 1) functions constants variables
     (cons (genex=equation functions constants variables) generated))))



(defun env~lookup-objects (keylist env &optional (objects nil))
  ;; list of all objects from environment "env" with keys in "keylist"
  (if (endp keylist)
      objects
    (env~lookup-objects (rest keylist) env
			(cons (env~lookup-object (first keylist) env)
			      objects)))) 


(defun env~lookup-variables (env)
  ;; list of all variables from enviroment "env"
  (env~lookup-objects (env~class-keys env 'sym+var) env))


(defun env~lookup-constants (env)
  ;; list of all constants (0 arity) from environment "env"
  ;; and list of all function (> 0 arity) from environment "env"
  (env~sortout-constants
   (env~lookup-objects (env~class-keys env 'sym+const) env)))


(defun env~sortout-constants (mixedfc &optional (constants nil)
				      (functions nil))
  ;; list of all constants (= 0 arity) of "mixedfc"
  ;; and list of all functions (> 0 arity) of "mixedfc"
  (if (endp mixedfc)
      (values constants functions)
    (if (fo~constant-p (first mixedfc))
	(env~sortout-constants (rest mixedfc)
			       (cons (first mixedfc) constants) functions)
      (env~sortout-constants (rest mixedfc) constants
			     (cons (first mixedfc) functions)))))



(defun genex=test (lhs rhs)
    (declare (special counter unifiable-counter random*env))
    (let* ((ausgabe *standard-output*))
      (format ausgabe "~%~A: ~A ~A: " (incf counter) lhs rhs)
      (if (cpcb~decide-unify (list (term~read lhs random*env))
			     (list (term~read rhs random*env)))
	  (progn
	    (incf unifiable-counter)
	    (format ausgabe " +"))
	  (format ausgabe " -"))))



;+-------------------------------------------------------------------
;+ Testdateien schreiben
;+-------------------------------------------------------------------


(defun genex~make-test-file (name howmany env)
  ;; create file with "name".lisp and generates "howmany" random equations
  ;; from env
  (make-random-state t)
  (with-open-file
   (testfile name :direction :output :if-exists :supersede
	     :if-does-not-exist :create)
   (multiple-value-bind (constants functions)
       (env~lookup-constants env)
     (format testfile "~%(in-package 'KEIM)~2%")
     (format testfile "(defun run ()~% (setf counter 0 unifiable-counter 0 cpcb*bcktrck-counter 0 fo*variable-counter 1000)~% (time (progn~%")
     (genex=make-ex testfile howmany functions constants
		    (env~lookup-variables env))
     (format testfile "  ))~% (format t ~C~~%~~A unifiable   ~~A backtrack steps       iterative: ~~A~C unifiable-counter cpcb*bcktrck-counter cpcb*iterative)~%" #\" #\")
     (format testfile " (values))~%~%"))))



(defun genex=make-ex (testfile howmany functions constants variables)
  ;; writes "homany" equations to testfile
  (if (> howmany 0)
      (let ((equ (genex=equation functions constants variables)))
	(format testfile "  (genex=test (quote ~a) (quote ~a))~%"
		(first equ) (second equ))
	(genex=make-ex testfile (- howmany 1)
		       functions constants variables))))

