
(in-package :KEIM)

;;; Incremental Combination of many systems
;;; After purifying the equations, each system is
;;; handled separately

;;; For purifying equations and creation of systems,
;;; the file mcb.lisp and its functions is used

;;; a switch for using syntactic properties of theories

(defvar icb*syntactic T
  "Whether to use syntactic properties of theories.")






;;; for testing: give results of free unification
(defmethod th~decide-unify ((th th+free) tl1 tl2 restr const-vars)
  (let ((unifier (free~unif tl1 tl2 restr const-vars)))
    (if unifier
	(list unifier)
      nil)))



(defun icb~decide-unify (equations)
  (declare (authors Kepser))
  (let ((systems (mcb~purify-equations equations))
	(indices NIL) (order NIL) (non-i NIL))
    (setf systems
	  (sort  ;; free theory first
	   (mcb=combi-vars-to-theories
	    systems
	    (mcb=extract-true-combi-vars
	     (mcb=combi-vars systems)))
	   #'(lambda (th1 th2)
	       (and (typep th1 'th+free) (not (typep th2 'th+free))))
	   :key #'mcb=system-theory))
    (if icb*syntactic
	(multiple-value-bind (new-ind new-order new-non-i)
	    (icb=use-syntactic-props systems)
	  (setq indices new-ind order new-order non-i new-non-i)))
    (if (eq indices 'unsolvable)
	(values NIL)
      (progn
;;;	(format T "~A~%~A~%~%" systems indices)
	(icb=next-system-decide systems non-i indices order NIL NIL)))))



(defun icb~unify (equations)
  (declare (authors Kepser)
	   (edited  "9-Jan-96"))
  (let ((systems (mcb~purify-equations equations))
	(indices NIL) (order NIL) (non-i NIL))
    (setf systems
	  (sort  ;; free theory first
	   (mcb=combi-vars-to-theories
	    systems
	    (mcb=extract-true-combi-vars
	     (mcb=combi-vars systems)))
	   #'(lambda (th1 th2)
	       (and (typep th1 'th+free) (not (typep th2 'th+free))))
	   :key #'mcb=system-theory))
    (if icb*syntactic
	(multiple-value-bind (new-ind new-order new-non-i)
	    (icb=use-syntactic-props systems)
	  (setq indices new-ind order new-order non-i new-non-i)))
    (if (eq indices 'unsolvable)
	(values NIL)
      (progn 
;;;	(format T "~A~%~A~%~A~%~%" systems indices non-i)
	(icb=next-system-complete systems non-i indices order NIL)))))



;;; --------------------------------------------------------------

;;; decision procedure

(defun icb=next-system-decide
  (systems non-identifiables indices order vars-seen unifier)
  (declare (authors Kepser))
;;;  (format T "O: ~A~%" order)
  (cond ((endp systems)
	 unifier)
        ((not (rest systems))
	 (icb=decide-last-system (first systems) indices order unifier))
	(T
	 (let ((vars (mcb=system-combi-vars (first systems))))
	   (icb=var-identification-decide
	    (icb=get-identifiables vars)
	    non-identifiables systems indices order
	    vars-seen unifier))))) 


  
(defun icb=var-identification-decide
  (identifiables non-identifiables systems indices order vars-seen unifier)
  (declare (authors Kepser))
  (if (endp identifiables)
	   ;; call indexing
      (icb=var-indexing-decide
       (mcb=system-combi-vars (first systems))
       systems indices order non-identifiables
       (union (mcb=system-combi-vars (first systems)) vars-seen)
       unifier)
    
    (let (dom-v cod-v sys-dom sys-cod)
      (if (member (caar identifiables) vars-seen :test #'term~equal)
	  (if (and (member (cdar identifiables) vars-seen
			   :test #'term~equal)
		   (assoc (cdar identifiables) indices :test #'term~equal))
	      (setq dom-v (caar identifiables)
		    cod-v (cdar identifiables))
	    (setq dom-v (cdar identifiables)
		  cod-v (caar identifiables)))
	(setq dom-v (caar identifiables)
	      cod-v (cdar identifiables)))
      (if (or (term~equal dom-v cod-v)
	      (member (first identifiables) non-identifiables
		      :test #'mcb=var-pair-equal)
	      (and (setq sys-dom (rest (assoc dom-v indices :test
					      #'term~equal))) 
		   (setq sys-cod (rest (assoc cod-v indices :test
					      #'term~equal)))
		   (not (eq sys-dom sys-cod)))
	      (icb=is-greater dom-v cod-v order)
	      (icb=is-greater cod-v dom-v order))
	  (icb=var-identification-decide
	   (rest identifiables) non-identifiables systems indices
	   order vars-seen unifier)
	
	;; there is a pair left that we have the choice to identify
	;; don't identify
	(or (icb=var-identification-decide
	     (rest identifiables)
	     (cons (cons dom-v cod-v) non-identifiables)
	     systems indices order vars-seen unifier)

	    ;; identify
	    (let ((sub (subst~create (list dom-v) (list cod-v))))
;;;	      (format T "S: ~A  I: ~A~%"
;;;		      (mcb=system-theory (first systems)) sub)
	      (icb=var-identification-decide
	       (mcb=subst-var-pairs dom-v cod-v (rest identifiables))
	       (mcb=subst-non-ident-vars non-identifiables dom-v cod-v)
	       (subst~apply sub systems) indices
	       (if icb*syntactic (icb=identify-order dom-v cod-v order sub)
		 order) 
	       vars-seen (cons sub unifier))))))))



(defun icb=get-identifiables (vars &optional (identifiables NIL))
  (if (not (rest vars))
      identifiables
    (icb=get-identifiables
     (rest vars)
     (let ((id identifiables)
	   (var (first vars)))
       (dolist (v (rest vars) id)
	 (setq id (acons v var id))))))) 



(defun icb=identify-order (domain codomain order subst)
  (declare (authors Kepser)
	   (edited  "24-Jan-1996"))
  (let ((succ-d (rest (assoc domain order :test #'term~equal)))
	(succ-c (rest (assoc codomain order :test #'term~equal))))
    (cond ((and succ-d succ-c)
	   (acons
	    codomain
	    (delete-duplicates (union succ-c succ-d) :test #'term~equal)
	    (mapcar #'(lambda (x)
			(delete-duplicates x :test #'term~equal))
		    (subst~apply
		     subst
		     (remove-if #'(lambda (x)
				    (or (term~equal domain (first x))
					(term~equal codomain (first x))))
				order)))))
	  (succ-d
	   (acons
	    codomain succ-d
	    (mapcar #'(lambda (x)
			(delete-duplicates x :test #'term~equal))
		    (subst~apply
		     subst
		     (remove domain order :test #'term~equal :key #'first)))))
	  (T
	   (mapcar #'(lambda (x)
		       (delete-duplicates x :test #'term~equal))
		   (subst~apply subst order))))))



(defun icb=is-greater (v1 v2 order)
  ;; checks if v1 is greater than v2 in the given order
  (declare (authors Kepser)
	   (edited  "24-Jan-1996"))
  (let ((succ (rest (assoc v1 order :test #'term~equal))))
    (if succ
	(if (member v1 succ :test #'term~equal)
	    (break)
	  (or (member v2 succ :test #'term~equal)
	      (some #'(lambda (v3)
			(icb=is-greater v3 v2 order))
		    succ))))))



(defun icb=var-indexing-decide
  (vars systems indices order non-i vars-seen unifier
	&optional (consts NIL) (true-v NIL))

  (let ((var (first vars)))
    (cond ((and (endp vars) consts true-v)
	   ;; there are some consts and true-v we can order
	   ;; call ordering
	   (or ;; start with variables' block
	       (icb=order-vblock-decide
		true-v consts systems indices order non-i vars-seen
		unifier	consts NIL NIL NIL NIL)
	       ;; start with constants' block
	       (icb=order-cblock-decide
		true-v consts systems indices order non-i vars-seen
		unifier consts NIL NIL NIL NIL)))
	  ((endp vars)
	   ;; at least one of the lists consts, true-v is empty
	   ;; no ordering, call unification directly
	   (icb=unification-decide
	    systems consts NIL indices order non-i vars-seen unifier NIL))
	  ((eq (mcb=system-theory (first systems))
	       (rest (assoc var indices :test #'term~equal)))
	   ;; var is already determined for this system
	   (icb=var-indexing-decide
	    (rest vars) systems indices order non-i vars-seen unifier
	    consts (cons var true-v)))
	  ((member var indices :test #'term~equal :key #'first)
	   ;; var is already taken by an earlier system
	   (icb=var-indexing-decide
	    (rest vars) systems indices order non-i vars-seen
	    unifier (cons var consts) true-v))
	  (T ; var has no index yet
	   (or ;; var is as a true variable in this system
	       (icb=var-indexing-decide
		(rest vars) systems
		(cons (cons var (mcb=system-theory (first systems)))
		      indices)
		order non-i vars-seen unifier consts (cons var true-v))

	       ;; var is treated as a constant
	       (icb=var-indexing-decide
		(rest vars) systems indices order non-i vars-seen
		unifier (cons var consts) true-v)))))) 



(defun icb=order-vblock-decide
  (vars consts systems indices order non-i vars-seen unifier
	passed-consts vars-done out restr linear-o)
  (declare (authors Kepser))
  (if (endp consts)
      ;; call unification
      (icb=unification-decide
       systems passed-consts restr indices order non-i vars-seen unifier
       (cons vars linear-o))
    (icb=order-vblock-aux-decide
     vars consts systems indices order non-i vars-seen unifier
     passed-consts vars-done out restr linear-o T NIL NIL NIL)))



(defun icb=order-vblock-aux-decide
  (vars consts systems indices order non-i vars-seen unifier passed-consts
	vars-done out restr linear-o block-empty vars-remain marked blocked)
  (declare (authors Kepser))
  (cond ((endp vars)
	 (icb=order-cblock-decide
	  vars-remain consts systems indices order non-i vars-seen
	  unifier passed-consts vars-done out restr linear-o))
	((and block-empty (not (rest vars)))
	 (let ((sg (icb=subgraph
		    (first vars) order indices
		    (mcb=system-combi-vars (first systems)) NIL NIL out)))
	   (if sg
	       (icb=order-cblock-decide
		(set-difference vars-remain sg :test #'term~equal)
		consts systems indices order non-i vars-seen unifier
		passed-consts (append sg vars-done)
		(append sg out) restr (cons sg linear-o)))))
	(T ;; we can nondeterministically choose the var
	 (let ((var (first vars)))
	   (or ;; take it
	       (let ((sg (icb=subgraph
			  var order indices
			  (mcb=system-combi-vars (first systems))
			  marked blocked out)))
		 (if sg
		     (icb=order-vblock-aux-decide
		      (set-difference (rest vars) sg :test #'term~equal)
		      consts systems indices order non-i
		      vars-seen unifier passed-consts
		      (append sg vars-done) (append sg out) restr
		      (if block-empty
			  (cons sg linear-o)
			(cons (append sg (first linear-o)) (rest linear-o)))  
		      NIL (set-difference vars-remain sg :test #'term~equal)
		      (append (butlast sg) marked) (cons var blocked))))
	       ;; leave it out
	       (icb=order-vblock-aux-decide
		(rest vars) consts systems indices order non-i vars-seen
		unifier passed-consts vars-done out restr linear-o
		block-empty (cons var vars-remain) marked blocked))))))





(defun icb=subgraph (var order indices system-vars marked blocked out)
  (declare (authors Kepser))
  (if (assoc var order :test #'term~equal)
      (do ((stack NIL (rest stack))
	   (closed (list var))
	   (subgr NIL)
	   (index (rest (assoc var indices :test #'term~equal)))
	   (top var (first stack)))
	  ((null top) subgr)
	(cond ((member top out :test #'term~equal))
	       ;; just proceed with next element of the stack
	      ((or (member top marked :test #'term~equal)
		   (not (member top system-vars :test #'term~equal)))
	       (let ((succ (rest (assoc top order :test #'term~equal))))
		 (setq stack
		       (append (set-difference succ closed
					       :test #'term~equal)
			       stack)
		       closed (union succ closed :test #'term~equal))))
	      ((or (not (eq index
			    (rest (assoc top indices :test #'term~equal))))
		   (member top blocked :test #'term~equal))
	       (setq stack NIL
		     subgr NIL))
	      (T ;; top in current system, not yet ordered, unmarked
	       (let ((succ (rest (assoc top order :test #'term~equal))))
		 (setq stack
		       (append (set-difference succ closed
					       :test #'term~equal)
			       stack)
		       closed (union succ closed :test #'term~equal)
		       subgr (cons top subgr))))))
    (list var))) 



(defun icb=order-cblock-decide
  (vars consts systems indices order non-i vars-seen unifier passed-consts
	vars-done out restr linear-o)
  (declare (authors Kepser)
	   (edited  "21.12.1996"))
  (if (endp vars)
      ;; call unification
      (icb=unification-decide
       systems passed-consts
       (let ((r restr))
	 (dolist (c consts r)
	   (setq r (acons c vars-done r))))
       indices order non-i vars-seen unifier (cons consts linear-o))
    (icb=order-cblock-aux-decide
     vars consts systems indices order non-i vars-seen unifier passed-consts
     vars-done out restr linear-o T NIL NIL NIL))) 



(defun icb=order-cblock-aux-decide
  (vars consts systems indices order non-i vars-seen unifier passed-consts
	vars-done out restr linear-o block-empty consts-remain marked blocked)
  (declare (authors Kepser)
	   (edited  "21.12.1996"))
  (cond ((endp consts)
	 (icb=order-vblock-decide
	  vars consts-remain systems indices order non-i vars-seen
	  unifier passed-consts vars-done out restr linear-o))
	((and block-empty (not (rest consts)))
	 (let ((sg (icb=subgraph
		    (first consts) order indices
		    (mcb=system-combi-vars (first systems)) NIL NIL out)))
	   (if sg
	       (icb=order-vblock-decide
		vars (set-difference consts-remain sg :test #'term~equal)
		systems indices order non-i vars-seen unifier passed-consts
		vars-done (append sg out)
		(let ((r restr))
		  (dolist (c sg r)
		    (setq r (cons (cons c vars-done) r))))
		(cons sg linear-o)))))
	(T ;; we can nondeterministically choose the var
	 (let ((const (first consts)))
	   (or ;; take it
	       (let ((sg (icb=subgraph const order indices
				   (mcb=system-combi-vars (first systems))
				   marked blocked out)))
		 (if sg
		     (icb=order-cblock-aux-decide
		      vars (set-difference (rest consts) sg :test
					  #'term~equal)
		      systems indices order non-i vars-seen unifier
		      passed-consts vars-done (append sg out)
		      (let ((r restr))
			(dolist (c sg r)
			  (setq r (cons (cons c vars-done) r))))
		      (if block-empty
			  (cons sg linear-o)
			(cons (append sg (first linear-o)) (rest linear-o)))
		      NIL (set-difference consts-remain sg :test #'term~equal)
		      (append (butlast sg) marked) (cons const blocked))))
	       ;; leave it out
	       (icb=order-cblock-aux-decide
		vars (rest consts) systems indices order non-i vars-seen
		unifier passed-consts vars-done out restr linear-o
		block-empty (cons const consts-remain) marked blocked)))))) 



(defun icb=unification-decide
  (systems consts restr indices order non-ident vars-seen unifier linear-o)
  (declare (authors Kepser)
	   (edited  "21.12.1996"))
  (let* ((sys (first systems))
	 (uni (th~decide-unify (mcb=system-theory sys)
			       (mcb=system-lhs sys)
			       (mcb=system-rhs sys)
			       restr consts)))
;;;    (if (not (icb=check-l-o linear-o))
;;;        (break))
    (if uni
	(progn
;;;          (format T "~A ~A ~A ~A ~A~%~A ~A~%~A~%"
;;;                  (mcb=system-theory sys)
;;;                  (mcb=system-lhs sys)
;;;                  (mcb=system-rhs sys)
;;;                  restr consts order linear-o uni)
	(icb=next-system-decide
	 (rest systems) non-ident indices
	 (icb=extend-po order linear-o) vars-seen
	 (cons uni unifier)))))) 



(defun icb=extend-po (partial linear)
  (declare (authors Kepser)
	   (edited  "21.12.1996"))
  (if (not (rest linear)) ; last block
      partial
    (icb=extend-po
     (let ((block (second linear))
	   (old)
	   (new-po partial))
       (dolist (v (first linear) new-po)
         (if (setq old (rest (assoc v partial :test #'term~equal)))
	     (setq new-po
		   (acons
		    v (union block old :test #'term~equal)
		    (remove v new-po :test #'term~equal :key #'first)))
           (setq new-po (acons v block new-po)))))
     (rest linear))))



;;; -- for testing the linear order

(defun icb=check-l-o (linear &optional (block NIL) (vars-seen NIL))
  (if (endp block)
      (if (endp linear)
	  T
	(icb=check-l-o (rest linear) (first linear) vars-seen))
    (if (not (member (first block) vars-seen :test #'term~equal))
	(icb=check-l-o linear (rest block)
		       (cons (first block) vars-seen)))))



;;; -- to handle the last system

(defun icb=decide-last-system (system indices order unifier)
  (declare (authors Kepser)
	   (edited  "30-Sept-1996")
	   (input   "a system, which should be the last, "
		    "the variable indices, the partial order, "
		    "and the unifier constructed so far.")
	   (effect  "None.")
	   (value   "The result of the whole computation. If the last "
		    "system is unifiable, a unifier for the whole system"
		    ", else NIL."))
  (multiple-value-bind (consts true-v)
      (icb=vars-of-last (mcb=system-combi-vars system)
			(mcb=system-theory system) indices)
    ;; we only need an ordering
    (or ;; start with a variables' block
        (icb=order-vblock-decide
	 true-v consts (list system) indices order NIL NIL
	 unifier consts NIL NIL NIL NIL)
	;; start with a constants' block
	(icb=order-cblock-decide
	 true-v consts (list system) indices order NIL NIL
	 unifier consts NIL NIL NIL NIL))))



(defun icb=vars-of-last (vars theory indices &optional (consts NIL)
			      (true-v NIL))
  (declare (authors Kepser)
	   (edited  "2-Okt-1996")
	   (input   "Combi vars of the last system, its theory and the "
		    "variable indices.")
	   (effect  "None.")
	   (value   "The constants and true variables of the last "
		    "system according to INDICES."))
  (if (endp vars)
      (values consts true-v)
    (let ((th (rest (assoc (first vars) indices :test #'term~equal))))
      (if (or (not th) (eq th theory))
	  (icb=vars-of-last (rest vars) theory indices consts
			    (cons (first vars) true-v))
	(icb=vars-of-last (rest vars) theory indices
			  (cons (first vars) consts) true-v)))))



;;; Warshall's algorithm to compute the transitive closure of a graph

(defun warshall (graph &optional (accu graph))
  (declare (authors Warshall Kepser)
	   (edited  "1-Okt-1996")
	   (input   "A Graph represented as an a-list, each element"
		    " is a list with the head a node and the tail "
		    "a list of this node's successors.")
	   (effect  "None.")
	   (value   "The transitive Closure of the Graph, in the same"
		    " format."))
  (if (endp graph)
      accu
    (warshall
     (rest graph)
     (warshall-a accu (caar graph)
		 (rest (assoc (caar graph) accu :test #'term~equal))))))


(defun warshall-a (graph node succ &optional (new-graph NIL))
  (declare (authors Warshall Kepser)
	   (edited  "1-Okt-1996")
	   (input   "A graph represented as an a-list; "
		    "a node of the graph and its successors.")
	   (effect  "None.")
	   (value   "A graph, each node having NODE as one of its "
		    "successors will have SUCCS as successors, too."))
  (if (endp graph)
      new-graph
    (warshall-a
     (rest graph) node succ
     (if (member node (cdar graph) :test #'term~equal)
	 (acons (caar graph)
		(union succ (cdar graph) :test #'term~equal)
		new-graph)
       (cons (first graph) new-graph)))))


;;; ------------------------------------------------------------------


;;; complete sets of unifiers

(defun icb=next-system-complete
  (systems non-identifiables indices order vars-seen)
  (declare (authors Kepser))
;;;  (format T "O: ~A~%" order)
  (if (endp systems)
      (list (subst~create NIL NIL))
    (let ((vars (mcb=system-combi-vars (first systems))))
      (icb=var-identification-complete
       (icb=get-identifiables vars)
       non-identifiables systems indices order vars-seen)))) 


  
(defun icb=var-identification-complete
  (identifiables non-identifiables systems indices order vars-seen)
  (declare (authors Kepser))
  (if (endp identifiables)
	   ;; call indexing
      (icb=var-indexing-complete
       (mcb=system-combi-vars (first systems))
       systems indices order non-identifiables
       (union (mcb=system-combi-vars (first systems)) vars-seen))
    
    (let (dom-v cod-v sys-dom sys-cod)
      (if (member (caar identifiables) vars-seen :test #'term~equal)
	  (if (and (member (cdar identifiables) vars-seen
			   :test #'term~equal)
		   (assoc (cdar identifiables) indices :test #'term~equal))
	      (setq dom-v (caar identifiables)
		    cod-v (cdar identifiables))
	    (setq dom-v (cdar identifiables)
		  cod-v (caar identifiables)))
	(setq dom-v (caar identifiables)
	      cod-v (cdar identifiables)))
      (if (or (term~equal dom-v cod-v)
	      (member (first identifiables) non-identifiables
		      :test #'mcb=var-pair-equal)
	      (and (setq sys-dom (rest (assoc dom-v indices :test
					      #'term~equal))) 
		   (setq sys-cod (rest (assoc cod-v indices :test
					      #'term~equal)))
		   (not (eq sys-dom sys-cod)))) 
	  (icb=var-identification-complete
	   (rest identifiables) non-identifiables systems indices
	   order vars-seen)
	
	;; there is a pair left that we have the choice to identify
	;; don't identify
	(nconc (icb=var-identification-complete
		(rest identifiables)
		(cons (cons dom-v cod-v) non-identifiables)
		systems indices order vars-seen)
	       
	       ;; identify
	       (let ((sub (subst~create (list dom-v) (list cod-v)))
		     (unif))
;;;		 (format T "I: ~A~%" sub)
		 (setq unif
		       (icb=var-identification-complete
			(mcb=subst-var-pairs dom-v cod-v (rest identifiables))
			(mcb=subst-non-ident-vars non-identifiables
						  dom-v cod-v)
			(subst~apply sub systems) indices
			(subst~apply sub order) vars-seen))
		 (if unif
		     `((PR (,sub) ,unif)))))))))



(defun icb=var-indexing-complete
  (vars systems indices order non-i vars-seen
	&optional (consts NIL) (true-v NIL))

  (let ((var (first vars)))
    (cond ((and (endp vars) consts true-v)
	   ;; there are some consts and true-v we can order
	   ;; call ordering
	   (or ;; start with variables' block
	       (icb=order-vblock-complete
		true-v consts systems indices order non-i vars-seen
		consts NIL NIL NIL NIL)
	       ;; start with constants' block
	       (icb=order-cblock-complete
		true-v consts systems indices order non-i vars-seen
		consts NIL NIL NIL NIL)))
	  ((endp vars)
	   ;; at least one of the lists consts, true-v is empty
	   ;; no ordering, call unification directly
	   (icb=unification-complete
	    systems consts NIL indices order non-i vars-seen NIL))
	  ((eq (mcb=system-theory (first systems))
	       (rest (assoc var indices :test #'term~equal)))
	   ;; var is already determined for this system
	   (icb=var-indexing-complete
	    (rest vars) systems indices order non-i vars-seen
	    consts (cons var true-v)))
	  ((member var indices :test #'term~equal :key #'first)
	   ;; var is already taken by an earlier system
	   (icb=var-indexing-complete
	    (rest vars) systems indices order non-i vars-seen
	    (cons var consts) true-v))
	  (T  ;; var has no index yet
	   (nconc  ;; var is as a true variable in this system
	           (icb=var-indexing-complete
		    (rest vars) systems
		    (acons var (mcb=system-theory (first systems))
			   indices)
		    order non-i vars-seen consts (cons var true-v))

		   ;; var is treated as a constant
		   (icb=var-indexing-complete
		    (rest vars) systems indices order non-i vars-seen
		    (cons var consts) true-v)))))) 



(defun icb=order-vblock-complete
  (vars consts systems indices order non-i vars-seen
	passed-consts vars-done out restr linear-o)
  (declare (authors Kepser))
  (if (endp consts)
      ;; call unification
      (icb=unification-complete
       systems passed-consts restr indices order non-i vars-seen
       (cons vars linear-o))
    (icb=order-vblock-aux-complete
     vars consts systems indices order non-i vars-seen
     passed-consts vars-done out restr linear-o T NIL NIL NIL)))



(defun icb=order-vblock-aux-complete
  (vars consts systems indices order non-i vars-seen passed-consts
	vars-done out restr linear-o block-empty vars-remain marked blocked)
  (declare (authors Kepser))
  (cond ((endp vars)
	 (icb=order-cblock-complete
	  vars-remain consts systems indices order non-i vars-seen
	  passed-consts vars-done out restr linear-o))
	((and block-empty (not (rest vars)))
	 (let ((sg (icb=subgraph
		    (first vars) order indices
		    (mcb=system-combi-vars (first systems)) NIL NIL out)))
	   (if sg
	       (icb=order-cblock-complete
		(set-difference vars-remain sg :test #'term~equal)
		consts systems indices order non-i vars-seen
		passed-consts (append sg vars-done)
		(append sg out) restr (cons sg linear-o)))))
	(T ;; we can nondeterministically choose the var
	 (let ((var (first vars)))
	   (nconc ;; take it
	          (let ((sg (icb=subgraph
			     var order indices
			     (mcb=system-combi-vars (first systems))
			     marked blocked out)))
		    (if sg
			(icb=order-vblock-aux-complete
			 (set-difference (rest vars) sg :test #'term~equal)
			 consts systems indices order non-i
			 vars-seen passed-consts
			 (append sg vars-done) (append sg out) restr
			 (if block-empty
			     (cons sg linear-o)
			   (cons (append sg (first linear-o))
				 (rest linear-o)))  
			 NIL vars-remain (append (butlast sg) marked)
			 (cons var blocked))))
		  ;; leave it out
		  (icb=order-vblock-aux-complete
		   (rest vars) consts systems indices order non-i vars-seen
		   passed-consts vars-done out restr linear-o block-empty
		   (cons var vars-remain) marked blocked))))))



(defun icb=order-cblock-complete
  (vars consts systems indices order non-i vars-seen passed-consts
	vars-done out restr linear-o)
  (declare (authors Kepser)
	   (edited  "21.12.1996"))
  (if (endp vars)
      ;; call unification
      (icb=unification-complete
       systems passed-consts
       (let ((r restr))
	 (dolist (c consts r)
	   (setq r (acons c vars-done r))))
       indices order non-i vars-seen (cons consts linear-o))
    (icb=order-cblock-aux-complete
     vars consts systems indices order non-i vars-seen passed-consts
     vars-done out restr linear-o T NIL NIL NIL))) 



(defun icb=order-cblock-aux-complete
  (vars consts systems indices order non-i vars-seen passed-consts
	vars-done out restr linear-o block-empty consts-remain marked blocked)
  (declare (authors Kepser)
	   (edited  "21.12.1996"))
  (cond ((endp consts)
	 (icb=order-vblock-complete
	  vars consts-remain systems indices order non-i vars-seen
	  passed-consts vars-done out restr linear-o))
	((and block-empty (not (rest consts)))
	 (let ((sg (icb=subgraph
		    (first consts) order indices
		    (mcb=system-combi-vars (first systems)) NIL NIL out)))
	   (if sg
	       (icb=order-vblock-complete
		vars (set-difference consts-remain sg :test #'term~equal)
		systems indices order non-i vars-seen passed-consts
		vars-done (append sg out)
		(let ((r restr))
		  (dolist (c sg r)
		    (setq r (cons (cons c vars-done) r))))
		(cons sg linear-o)))))
	(T ;; we can nondeterministically choose the var
	 (let ((const (first consts)))
	   (nconc ;; take it
	          (let ((sg (icb=subgraph
			     const order indices
			     (mcb=system-combi-vars (first systems))
			     marked blocked out)))
		    (if sg
			(icb=order-cblock-aux-complete
			 vars (set-difference (rest consts) sg :test
					      #'term~equal)
			 systems indices order non-i vars-seen
			 passed-consts vars-done (append sg out)
			 (let ((r restr))
			   (dolist (c sg r)
			     (setq r (cons (cons c vars-done) r))))
			 (if block-empty
			     (cons sg linear-o)
			   (cons (append sg (first linear-o)) (rest linear-o)))
			 NIL consts-remain (append (butlast sg) marked)
			 (cons const blocked))))
		  ;; leave it out
		  (icb=order-cblock-aux-complete
		   vars (rest consts) systems indices order non-i vars-seen
		   passed-consts vars-done out restr linear-o
		   block-empty (cons const consts-remain) marked blocked)))))) 



(defun icb=unification-complete
  (systems consts restr indices order non-ident vars-seen linear-o)
  (declare (authors Kepser)
	   (edited  "21.12.1996"))
  (let* ((sys (first systems))
	 (uni (th~unify (mcb=system-theory sys)
			(mcb=system-lhs sys)
			(mcb=system-rhs sys)
			restr consts))
	 (uni-others))
    (if uni
	(progn
;;;          (format T "~A ~A ~A ~A ~A~%~A~%"
;;;                  (mcb=system-theory sys)
;;;                  (mcb=system-lhs sys)
;;;                  (mcb=system-rhs sys)
;;;                  restr consts uni)
	  (setq uni-others
		(icb=next-system-complete
		 (rest systems) non-ident indices
		 (icb=extend-po order linear-o) vars-seen))
	  (if uni-others
	      `((PR ,uni ,uni-others))))))) 




;;; ------------------------------------------------------------------


;;; Functions for using syntactic properties of theories


;;; Properties used are: collaps-free, regular

;;; class hierarchy of theories (as from theory.lisp)
;;;
;;;               th+general
;;;                /      \
;;;               /        \
;;; th+collapse-free      th+regular
;;;               \        /
;;;                \      /
;;;                th+fine
;;;                   |
;;;                   |
;;;               th+simple



(defmacro icb=non-i (var succ non-i)
  `(if ,succ
       (let ((n-i ,non-i))
	 (dolist (s ,succ n-i)
	   (setq n-i (acons ,var s n-i))))
     ,non-i))



(defun icb=use-syntactic-props (systems &optional (indices NIL) (order NIL)
					(non-i NIL))
  (declare (authors Kepser)
	   (edited  "8-Jan-96"))
  (if (endp systems)
      (values indices order non-i)
    (multiple-value-bind (indices1 order1 non-i1)
	(icb=use-syntactic-props-aux
	 (mcb=system-theory (first systems))
	 (mcb=system-lhs (first systems))
	 (mcb=system-rhs (first systems))
	 (mcb=system-combi-vars (first systems))
	 indices order non-i)
      (if (eq indices1 'unsolvable)
	  (values 'unsolvable NIL NIL)
	(multiple-value-bind (indices2 order2 non-i2)
	    (icb=use-syntactic-props (rest systems) indices1 order1 non-i1)
	  (values indices2 order2 non-i2))))))



(defgeneric icb=use-syntactic-props-aux (theory lhs rhs cv indices order non-i)
  (declare (authors Kepser)
	   (edited  "8-Jan-96"))

  (:method ((theory th+general) lhs rhs cv indices order non-i)
	   (declare (ignore lhs rhs cv))
	   (values indices order non-i))
  
  (:method ((theory th+collapse-free) lhs rhs cv indices order non-i)
	   (values (icb=use-collapse-free theory lhs rhs cv indices)
		   order non-i))
  (:method ((theory th+simple) lhs rhs cv indices order non-i)
	   (multiple-value-bind (new-indices new-order new-non-i)
	       (icb=use-simple theory lhs rhs cv indices order non-i)
	     (values new-indices new-order new-non-i))))



(defun icb=use-collapse-free (theory lhs rhs cv indices)
  (declare (authors Kepser)
	   (edited "8-Jan-96"))
  (if (endp lhs)
      indices
    (let ((l (first lhs)) (r (first rhs)))
      (cond ((and (sym~variable-p l) (member l cv :test #'term~equal))
	     (let ((th-o (rest (assoc l indices :test #'term~equal))))
	       (if (not th-o)
		   (icb=use-collapse-free
		    theory (rest lhs) (rest rhs) cv (acons l theory indices))
		 (if (eq th-o theory)
		     (icb=use-collapse-free
		      theory (rest lhs) (rest rhs) cv indices)
		   'unsolvable))))
	    ((and (sym~variable-p r) (member r cv :test #'term~equal))
	     (let ((th-o (rest (assoc r indices :test #'term~equal))))
	       (if (not th-o)
		   (icb=use-collapse-free
		    theory (rest lhs) (rest rhs) cv (acons r theory indices))
		 (if (eq th-o theory)
		     (icb=use-collapse-free
		      theory (rest lhs) (rest rhs) cv indices)
		   'unsolvable))))
	    (T  ;; l and r no combi vars
	     (icb=use-collapse-free
	      theory (rest lhs) (rest rhs) cv indices)))))) 



(defun icb=use-simple (theory lhs rhs cv indices order non-i)
  (declare (authors Kepser)
	   (edited "11-Jan-1996"))
  (if (endp lhs)
    (values indices order non-i)
    (let ((l (first lhs)) (r (first rhs)))
      (cond ((and (sym~variable-p l) (member l cv :test #'term~equal))
	     (let ((th-o (rest (assoc l indices :test #'term~equal)))
		   (succ (intersection
			  (remove-duplicates (top~free-variables r)
					     :test #'term~equal)
			  cv :test #'term~equal)))
	       (if (not th-o)
		   (multiple-value-bind (new-ind new-order new-non-i)
		       (icb=use-simple
			theory (rest lhs) (rest rhs) cv
			(acons l theory indices)
			(if succ (acons l succ order) order)
			(icb=non-i l succ non-i))
		     (values new-ind new-order new-non-i))
		 (if (eq th-o theory)
		     (multiple-value-bind (new-ind new-order new-non-i)
			 (icb=use-simple
			  theory (rest lhs) (rest rhs) cv indices
			  (icb=merge-succ l succ order)
			  (icb=non-i l succ non-i))
		       (values new-ind new-order new-non-i))
		   (values 'unsolvable NIL NIL)))))
	    ((and (sym~variable-p r) (member r cv :test #'term~equal))
	     (let ((th-o (rest (assoc r indices :test #'term~equal)))
		   (succ (intersection
			  (remove-duplicates (top~free-variables l)
					     :test #'term~equal)
			  cv :test #'term~equal)))
	       (if (not th-o)
		   (multiple-value-bind (new-ind new-order new-non-i)
		       (icb=use-simple
			theory (rest lhs) (rest rhs) cv
			(acons r theory indices)
			(if succ (acons r succ order) order)
			(icb=non-i r succ non-i))
		     (values new-ind new-order new-non-i))
		 (if (eq th-o theory)
		     (multiple-value-bind (new-ind new-order new-non-i)
			 (icb=use-simple
			  theory (rest lhs) (rest rhs) cv indices
			  (icb=merge-succ r succ order)
			  (icb=non-i r succ non-i))
		       (values new-ind new-order new-non-i))
		   (values 'unsolvable NIL NIL)))))
	    (T ;; no side a combi var
	     (multiple-value-bind (new-ind new-order new-non-i)
		 (icb=use-simple theory (rest lhs) (rest rhs)
			       cv indices order non-i)
	       (values new-ind new-order new-non-i))))))) 



(defun icb=merge-succ (var succ order)
  (declare (authors Kepser)
	   (edited  "11-Jan-1996"))
  (if (not succ)
      order
    (let ((o-succ (rest (assoc var order :test #'term~equal))))
      (if o-succ
	  (setf (rest (assoc var order :test #'term~equal))
		(union succ o-succ :test #'term~equal))
	(acons var succ order))))) 
