(in-package "KEIM")

(defun inc~decide-unify (terms1 terms2)
  (declare (edited  "29-SEP-1995" "30-MAY-1995")
	   (authors Richts Kepser)
	   (input   "Two lists of terms of equal length.")
	   (effect  "None.")
	   (value   "NIL, if the terms are not unifiable, "
		    "else non-NIL."))
  (cpcb~decide-unify terms1 terms2 :return-up t))

(defun inc~decide-unify-pred (eq-list)
  (declare (edited  "13-AUG-1997")
	   (authors Doelfer)
	   (input   "List of equations (predicate notion!)")
	   (effect  "None.")
	   (value   "NIL, if the equations are not unifiable, "
		    "else non-NIL."))
  (cpcb~decide-unify-pred eq-list  :return-up t))


(defun inc~decide-new-equations (up terms1 terms2)
  (inc=unify up terms1 terms2))

(defun inc~decide-new-relations (up eq-list)
  (inc=unify-pred up eq-list))


(defun inc=unify (general-up terms1 terms2)
  (declare (edited  "22-AUG-1997")
	   (authors Doelfer)
	   (input   "Two lists of terms of equal length.")
	   (effect  "None.")
	   (value   "NIL, if the terms are not unifieable, "
		    "else non-NIL"))
  (mapc #'(lambda (var)
	    (unicp~set-orig-label! var (keim~get var 'cpcb=label)))
	(unicp~variables general-up))
  (unwind-protect
       (progn
	 (inc=add-equations general-up terms1 terms2)
	 (if (cpcb=unify-iteration general-up :decide t)
	     general-up
	     nil))
    (mapc #'(lambda (var)
	      (keim~put var 'cpcb=label (unicp~orig-label var))
	      (unicp~set-orig-label! var nil))
	  (unicp~variables general-up))))

(defun inc=add-equations (gup terms1 terms2)
  (multiple-value-bind (theories equations-list)
      (cpcb=purify-equations terms1 terms2)
    ;(format t "~%~{~%~A~}" equations-list)
    (let* ((var-lists (mapcar #'top~free-variables equations-list))
	   (variables (top~free-variables var-lists))
	   (new-variables (set-difference variables (unicp~variables gup) :test #'term~equal)))
      (unicp~vars-label-initialize! new-variables)
      (mapl #'(lambda (var-labels-tail)
		(setf (first var-labels-tail)
		      (append (first var-labels-tail)
			      (mapcar #'(lambda (var)
					  (unicp~var-label-copy (unicp~orig-label var)))
				      new-variables))))
	    (unicp~gup-label-stack gup))
      (mapc #'(lambda (theory var-list)
		(mapc #'(lambda (var)
			  (unless (member theory (unicp~var-label-theories var))
			    (push theory (unicp~var-label-theories var))))
		      var-list))
	    (rest theories) (rest var-lists))
      (let* ((new-shared-variables
	      (sort
	       (copy-list (remove-if-not #'unicp~shared-variable-p
					 (append (unicp~variables gup) new-variables)))
	       #'string< :key #'(lambda (var) (reverse (sym~name var)))))
	     (up-list (sort (inc=new-up-list gup theories equations-list)
			    #'(lambda (theory1 theory2)
				(and (typep theory1 'th+free)
				     (not (typep theory2 'th+free))))
			    :key #'unicp~theory)))
	(setf (slot-value gup 'variables)
	      (append (unicp~variables gup) new-variables)
	      (slot-value gup 'shared-variables)
	      new-shared-variables)		   
	(do ((up-tail up-list (rest up-tail))
	     (up-queue (misc~queue-create)
		       (misc~queue-insert! up-queue (first up-tail)))
	     (lqueue (misc~lqueue-create)
		     (misc~lqueue-insert!
		      (misc~lqueue-insert! lqueue nil nil nil nil nil)
		      nil nil nil nil nil)))
	    ((null up-tail)
	       (setf (unicp~gup-up-queue gup) up-queue
		     (unicp~gup-up-list gup) up-list
		     (unicp~gup-constraint-lqueue gup) lqueue)))
	;(print (mapcar #'(lambda (up) (th~symbols (unicp~theory up))) up-list))
	gup))))


(defun inc=new-up-list (gup theories equations-list)
  (let ((up-list (misc~queue-list (unicp~gup-up-queue gup))))
    (mapc #'(lambda (theory equations)
	      (unless (member theory
			      up-list
			      :key #'unicp~theory)
		(error "Theory ~A with incremental equations ~A was not in the initial problem"
		       theory equations)))
	  (rest theories) (rest equations-list))
    (when (and (first equations-list)
	       (notany #'(lambda (up) (typep (unicp~theory up) 'th+free))
		       up-list))
      (error "Variable equations occured in incremental input~
            but theory th+free was not in the initial problem"))
    (let ((alist (mapcar #'cons (rest theories) (rest equations-list))))
      (mapcar #'(lambda (up)
		  (let ((equations (cdr (assoc (unicp~theory up) alist)))
			(var-equations (if (typep (unicp~theory up) 'th+free)
					   (first equations-list))))
		    (inc=up-create up (unicp~eq-append var-equations equations))))
	      up-list))))



(defclass inc+up (unicp+elementary-up)
  ((new-equations :initarg :new-equations :accessor inc=new-equations)
   (original-up :initarg :original-up :accessor inc=original-up)
   (initial-up :initarg :initial-up :reader inc=initial-up)
   (up-stack :initform nil :accessor inc=up-stack))
  (:documentation "This is the class of all unification problem,
                   which have been supplied with incremental equations."))

(defun inc=up-create (original-up new-equations)
  (cond ((typep original-up 'inc+up)
	 (setf (inc=new-equations original-up)
	       (unicp~eq-append (inc=new-equations original-up) new-equations)
	       (slot-value original-up 'equations)
	       (unicp~eq-append (slot-value original-up 'equations) new-equations))
	 original-up)
	(t
	 (make-instance 'inc+up
			:new-equations new-equations
			:original-up original-up
			:initial-up original-up
			:general-up (unicp~general-up original-up)
			:equations (unicp~eq-append (unicp~equations original-up)
						    new-equations)
			:theory (unicp~theory original-up)))))


(defgeneric inc~new-equations (up equations ident discrim constants const-restr)
  (:method ((up unicp+elementary-up) equations ident discrim constants const-restr)
     (declare (ignore ident discrim constants const-restr))
     (let ((theory (unicp~theory up))
	   (shared-variables (unicp~shared-variables up)))
       (mapc #'(lambda (var)
		 (setf (unicp~var-label var theory) nil))
	     (unicp~variables up))
       (unicp~theo-init
	theory
	(unicp~general-up up)
	(unicp~eq-append (unicp~equations up) equations)
	(mapcan #'(lambda (var1)
		    (mapcan #'(lambda (var2)
				(if (unicp~vars-identified-p var1 var2)
				    (list (unicp~id-create var1 var2))
				    nil))
			    (rest (member var1 shared-variables))))
		shared-variables)
	(mapcan #'(lambda (var1)
		    (mapcan #'(lambda (var2)
				(if (unicp~vars-discriminated-p var1 var2)
				    (list (unicp~id-create var1 var2))
				    nil))
			    (rest (member var1 shared-variables))))
		shared-variables)
	(remove-if #'(lambda (var)
			     (or (null (unicp~var-label-theory var))
				 (eq (unicp~var-label-theory var) theory)))
		   shared-variables)
	(mapcan #'(lambda (var1)
		    (mapcan #'(lambda (var2)
				(if (unicp~vars-less-p var1 var2)
				    (list (unicp~cr-create var2 (list var1)))
				    nil))
			    shared-variables))
		shared-variables)))))

(defmethod unicp~new-constr ((up inc+up) ident discrim constants const-restr)
  (let* ((new-equations (inc=new-equations up))
	 (result (multiple-value-list
		     (if new-equations
			 (inc~new-equations (inc=original-up up) new-equations
					    ident discrim constants const-restr)
			 (unicp~new-constr (inc=original-up up)
					   ident discrim constants const-restr)))))
    (setf (inc=new-equations up) nil)
    (cond ((first result)
	   (setf (inc=original-up up) (first result))
	   (values-list (cons up (rest result))))
	  (t (values-list result)))))


(defmethod unicp~solvable-p ((up inc+up))
  (let* ((new-equations (inc=new-equations up))
	 (new-up (if new-equations
		     (inc~new-equations (inc=original-up up) new-equations
					  nil nil nil nil)
		     (inc=original-up up))))
    (setf (inc=new-equations up) nil)
    (cond (new-up
	   (setf (inc=original-up up) new-up)
	   (unicp~solvable-p new-up))
	  (t nil))))


(defmethod unicp~more-information-p ((up inc+up))
  (or (inc=new-equations up)
      (unicp~more-information-p (inc=original-up up))))


(defmethod unicp~up-save-choice-point ((up inc+up))
  (when (inc=new-equations up)
    (error "Programmierfehler:  Obwohl es noch nicht-abgearbeitete Gleichungen gibt, ~
            tritt eine Choice-Point auf."))
  (unicp~up-save-choice-point (inc=original-up up))
  (push (inc=original-up up) (inc=up-stack up)))


(defmethod unicp~up-restore-choice-point ((up inc+up))
  (mapc #'inc=term-label-reset
	(unicp~eq-lhss (unicp~equations up)))
  (mapc #'inc=term-label-reset
	(unicp~eq-rhss (unicp~equations up)))  
  (let* ((orig-up (unicp~up-restore-choice-point
		   (or (pop (inc=up-stack up))
		       (inc=initial-up up))))
	 (new-equations (inc=eq-set-difference (unicp~equations up)
					       (unicp~equations orig-up)))
	 (theory (unicp~theory up)))
    (mapc #'(lambda (var)
	      (unless (member theory (unicp~var-label-theories var))
		(push theory (unicp~var-label-theories var))))
	  (top~free-variables new-equations))
    (setf (inc=original-up up) orig-up
	  (inc=new-equations up) new-equations))
  up)

(defun inc=term-label-reset (term)
  (unless (sym~variable-p term)
    (unicp~term-set-label! term nil)
    (when (appl~p term)
      (mapc #'inc=term-label-reset (appl~arguments term)))))

(defun inc=eq-set-difference (all-equations old-equations)
  (let* ((all-eqs (mapcar #'cons
			  (unicp~eq-lhss all-equations)
			  (unicp~eq-rhss all-equations)))
	 (old-eqs (mapcar #'cons
			  (unicp~eq-lhss old-equations)
			  (unicp~eq-rhss old-equations)))
	 (new-eqs
	  (set-difference all-eqs old-eqs
			  :test #'(lambda (eq1 eq2)
				    (and (eq (car eq1) (car eq2))
					 (eq (cdr eq1) (cdr eq2)))))))
    (if new-eqs
	(unicp~eq-create (mapcar #'car new-eqs) (mapcar #'cdr new-eqs))
	nil)))
    

;;;;

(defun cpcb==trace-backtracking (up result)
  (format t "~%Backtracking(~A): " result up)
  ;(pprint (cpcb=compute-result up))
  (incf cpcb*bcktrck-counter)
  (format t "~% Choice Points: ~A~% Label Stack: ~A~%~{ UP Stack: ~A~%~}~%"
	  (unicp~gup-choice-points up)
	  (unicp~gup-label-stack up)
	  (mapcar #'(lambda (up) (if (typep up 'inc+up) (inc=up-stack up) "-"))
		  (misc~queue-list (unicp~gup-up-queue up))))
  )


(defmethod inc~new-equations ((up cpfree+up) equations
                                 identified discriminated constants const-restr)
  (call-next-method))
