
(in-package :KEIM)

; combination of many theories

(defvar mcb*unifier T)

;;; extensions to other modules

;;;(defun term~uniq (term-list &optional (term-set NIL))
;;;  (declare (edited  "19-June-1995")
;;;           (authors Kepser)
;;;           (input   "term-list: a list of KEIM terms; "
;;;                    "term-set: a set of KEIM terms.")
;;;           (effect  "None.")
;;;           (value   "A set of KEIM terms, all the elements of "
;;;                    "term-list are adjoined to term-set."))
;;;
;;;  (if (endp term-list)
;;;      term-set
;;;    (term~uniq (rest term-list)
;;;               (adjoin (first term-list) term-set :test #'term~equal))))



;;; functions to purify equations, terms, etc.
;;; equartions t1 = t2 are supposed to be represented as
;;;  lists (t1 t2)


(defclass mcb=system ()
  ((theory :reader mcb=system-theory
	   :writer mcb=write-system-theory
	   :initarg :theory
	   :initform th*free)
   (lhs :reader mcb=system-lhs
	:writer mcb=write-system-lhs
	:initarg :lhs
	:initform NIL)
   (rhs :reader mcb=system-rhs
	:writer mcb=write-system-rhs
	:initarg :rhs
	:initform NIL)
   (combi-vars :reader mcb=system-combi-vars
	       :writer mcb=write-system-combi-vars
	       :initarg :combi-vars
	       :initform NIL)))




(defmethod print-object ((th mcb=system) stream)
  (format stream "<~A ~A ~A ~A>"
	  (mcb=system-theory th)
	  (mcb=system-lhs th)
	  (mcb=system-rhs th)
	  (mcb=system-combi-vars th)))




(defmethod top=replace-free-variables-and-rename ((system mcb=system)
						  variables)
  (make-instance 'mcb=system
		 :theory (mcb=system-theory system)
		 :lhs (top=replace-free-variables-and-rename
		       (mcb=system-lhs system) variables)
		 :rhs (top=replace-free-variables-and-rename
		       (mcb=system-rhs system) variables)
		 :combi-vars (remove-duplicates
			      (top=replace-free-variables-and-rename
			       (mcb=system-combi-vars system)
			       variables)
			      :test #'term~equal))) 



(defun mcb~purify-equation (equation equations)
  (declare (edited  "7-June-1995")
	   (authors Kepser)
	   (input   "equation: a pair (list) of KEIM terms; "
		    "equations: a list of equations still to purify.")
	   (effect  "None.")
	   (value   "The theory, left hand side and right hand side"
		    " of the purified equation and the"
		    " new equations still to purify."))

  ;;; we assume that variables have NO theory at all.
  
  (let* ((lhs (first equation))
	 (rhs (second equation))
	 (lhs-th (th~term-theory lhs))
	 (rhs-th (th~term-theory rhs)))
    (cond ((sym~variable-p lhs)
	   ;; left hand side variable
	   (multiple-value-bind (new-rhs equations1)
	       (cb=purify-term rhs rhs-th equations)
	     (values rhs-th lhs new-rhs equations1)))

	  ((sym~variable-p rhs)
	   ;; right hand side variable
	   (multiple-value-bind (new-lhs equations1)
	       (cb=purify-term lhs lhs-th equations)
	     (values lhs-th new-lhs rhs equations1)))

	  ((eq lhs-th rhs-th)
	   ;;both sides same theory (non variables)
	   (multiple-value-bind (new-lhs equations1)
	       (cb=purify-term lhs lhs-th equations)
	     (multiple-value-bind (new-rhs equations2)
		 (cb=purify-term rhs rhs-th equations1)
	       (values rhs-th new-lhs new-rhs equations2))))

	  (T  ;; sides have different theories, non-variable
	   (multiple-value-bind (new-lhs equations1)
	       (cb=purify-term lhs lhs-th equations)
	     (let ((newvar (cb~create-new-var)))
	       (values lhs-th newvar new-lhs 
		       (cons (list newvar rhs) equations1))))))))




(defun mcb~purify-equations (equations &optional (pure NIL))
  (declare (edited  "8-June-1995")
	   (authors Kepser)
	   (input   "equations: a list of equations to purify.")
	   (effect  "None.")
	   (value   "A list of purified equations for each of the"
		    " theories. Format of equations is changed"
		    "inbetween. In  equations  each equation is a"
		    "pair consisting of the lhs and rhs. The purified"
		    "equations for one theory is a triple. First"
		    "element is the theory instance, second a list"
		    "of left hand sides, third a list of"
		    "corresponding right hand sides."))

  (if (endp equations)
      pure
    (multiple-value-bind (theory lhs rhs equations1)
	(mcb~purify-equation (first equations) (rest equations))
      (mcb~purify-equations equations1 
			    (mcb=push-eq theory lhs rhs pure)))))





(defun mcb=push-eq (theory lhs rhs pure-ths &optional (new-pure NIL))
  (declare (edited  "16-June-1995")
	   (authors Kepser)
	   (input   "theory: a KEIM theory instance, the theory of the"
		    " new equation; "
		    "lhs: left hand side of the new pure equation; "
		    "rhs: right hand side of the new pure equation; "
		    "pure-ths: already purified equations, a list of"
		    " lists, each element consists of a theory and"
		    " the list for the left hand side and the the"
		    " list for the corresponding right hand side"
		    " of the pure"
		    " equations of that theory constructed so far.")
	   (effect  "None.")
	   (value   "A list of purified equations - format as in "
		    "pure-ths - containing now also the lhs-rhs eq."))

  (cond ((endp pure-ths)
	 (cons (make-instance 'mcb=system :theory theory
			      :lhs (list lhs) :rhs (list rhs))
	       new-pure))

	((eql theory (mcb=system-theory (first pure-ths)))
	 (mcb=write-system-lhs
	  (cons lhs (mcb=system-lhs (first pure-ths)))
	  (first pure-ths))
	 (mcb=write-system-rhs
	  (cons rhs (mcb=system-rhs (first pure-ths)))
	  (first pure-ths))
	 (nconc new-pure pure-ths))

	(T
	 (mcb=push-eq theory lhs rhs (rest pure-ths)
		      (cons (first pure-ths) new-pure)))))




(defclass mcb=combi-var ()
  ((variable :reader mcb=combi-var-var
	     :writer mcb=write-combi-var-var
	     :initarg :variable)
   (systems :reader mcb=combi-var-systems
	     :writer mcb=write-combi-var-systems
	     :initarg :systems
	     :initform NIL)))


(defmethod print-object ((cv mcb=combi-var) stream)
  (format stream "<~A ~A>"
	  (mcb=combi-var-var cv)
	  (mcb=combi-var-systems cv)))



(defmethod mcb=extract-vars ((term-list list) (theory th+general)
			     combi-vars-list)
  (declare (edited  "16-June-1995")
	   (authors Kepser)
	   (input   "")
	   (effect  "None.")
	   (value   ""))

  (cond ((endp term-list)
	 combi-vars-list)
	(T
	 (mcb=extract-vars (rest term-list) theory
			   (mcb=extract-vars (first term-list)
					     theory
					     combi-vars-list)))))


(defmethod mcb=extract-vars ((const sym+const) (theory th+general)
			     combi-vars-list)
  (declare (edited  "16-June-1995")
	   (authors Kepser)
	   (input   "")
	   (effect  "None.")
	   (value   ""))

  combi-vars-list)


(defmethod mcb=extract-vars ((var sym+var) (theory th+general)
			     combi-vars-list)
  (declare (edited  "16-June-1995")
	   (authors Kepser)
	   (input   "")
	   (effect  "None.")
	   (value   ""))

  (mcb=aux-extract-vars var theory combi-vars-list))


(defun mcb=aux-extract-vars (var theory combi-vars-list
			     &optional (combi-vars-so-far NIL))
  (declare (edited  "16-June-1995")
	   (authors Kepser)
	   (input   "")
	   (effect  "None.")
	   (value   ""))

  (cond ((endp combi-vars-list)
	 (cons (make-instance 'mcb=combi-var :variable var
			      :systems (list theory))
	       combi-vars-so-far))
	((term~equal var (mcb=combi-var-var (first combi-vars-list)))
	 (mcb=write-combi-var-systems (adjoin theory
					       (mcb=combi-var-systems
						(first
						 combi-vars-list)))
				       (first combi-vars-list))
	 (nconc combi-vars-so-far combi-vars-list))
	(T 
	 (mcb=aux-extract-vars var theory (rest combi-vars-list)
			   (cons (first combi-vars-list)
				 combi-vars-so-far)))))


(defmethod mcb=extract-vars ((term appl+appl) (theory th+general)
			     combi-vars-list)
  (declare (edited  "16-June-1995")
	   (authors Kepser)
	   (input   "")
	   (effect  "None.")
	   (value   ""))

  (mcb=extract-vars (appl~arguments term) theory combi-vars-list))



(defun mcb=combi-vars (eqs-list &optional (combi-vars-so-far NIL))
  (declare (edited  "16-June-1995")
	   (authors Kepser)
	   (input   "")
	   (effect  "None.")
	   (value   ""))

  (if (endp eqs-list)
      combi-vars-so-far
    (mcb=combi-vars (rest eqs-list)
		    (mcb=extract-vars
		     (mcb=system-lhs (first eqs-list))
		     (mcb=system-theory (first eqs-list))
		     (mcb=extract-vars
		      (mcb=system-rhs (first eqs-list))
		      (mcb=system-theory (first eqs-list))
		      combi-vars-so-far)))))


(defun mcb=extract-true-combi-vars (combi-vars-list)
  (declare (edited  "16-June-1995")
	   (authors Kepser)
	   (input   "A list of combination variables")
	   (effect  "None.")
	   (value   "A sublist of the input containing only those"
		    "variables which belong to at least 2 theories."))

  (remove-if #'(lambda (v) (> 2 (length (mcb=combi-var-systems v))))
	     combi-vars-list)) 




(defun mcb=combi-vars-to-theories (theories-list combi-vars-list)
  (declare (edited  "19-June-1995")
	   (authors Kepser)
	   (input   "theories-list: a list of mcb=system objects; "
		    "combi-vars-list: a list of mcb=combi-var objects; "
		    "the theories in theories list should be the same "
		    "as in combi-vars-list, thought the function works, "
		    "too, if this is not the case.")
	   (effect  "None.")
	   (value   "New theories-list with the combi-vars slot filled "
		    "acording to the information in combi-vars-list."))

  (if (endp combi-vars-list)
      theories-list
    (mcb=combi-vars-to-theories
     (mcb=aux-combi-vars-to-theories
      (mcb=combi-var-var (first combi-vars-list))
      (mcb=combi-var-systems (first combi-vars-list))
      theories-list)
     (rest combi-vars-list))))


(defun mcb=aux-combi-vars-to-theories (var th-list theories-list)
  (mapcar #'(lambda (x)
	      (if (member (mcb=system-theory x) th-list)
		  (mcb=write-system-combi-vars
		   (cons var
			 (mcb=system-combi-vars x))
		   x))
	      x)
	  theories-list)) 




(defun mcb=identifiables (systems &optional (pairs-list NIL))
  (declare (edited  "20-June-1995")
	   (authors Kepser)
	   (input   "systems: a list of mcb=system objects; "
		    "pairs-list: a list of combination variable pairs.")
	   (effect  "None.")
	   (value   "A list of dotted pairs, each pair consists of"
		    "two combination variables which may be identified"
		    " in the first step of the dekomposition."))

  (if (endp systems)
      pairs-list
    (mcb=identifiables (rest systems)
		       (mcb=construct-var-pairs
			(mcb=system-combi-vars (first systems))
			pairs-list)))) 



(defun mcb=construct-var-pairs (var-list &optional (pairs-list NIL))
  (if (< (length var-list) 2)
      pairs-list
    (mcb=construct-var-pairs
     (rest var-list)
     (let ((p-l pairs-list)
	   (v1 (first var-list)))
       (dolist (v (rest var-list) p-l)
	 (setf p-l
	       (adjoin (cons v1 v) p-l :test #'mcb=var-pair-equal)))))))


(defun mcb=var-pair-equal (pair1 pair2)
  (or (and (term~equal (car pair1) (car pair2))
	   (term~equal (cdr pair1) (cdr pair2)))
      (and (term~equal (car pair1) (cdr pair2))
	   (term~equal (cdr pair1) (car pair2))))) 




(defun mcb=var-identification-decide
  (identifiable-vars non-ident-vars systems)

  (declare (edited  "21-June-1995")
	   (authors Kepser)
	   (input   "identifiable-vars: a list of combination variables"
		    " which can/may be identified; "
		    "non-ident-vars: a list of combination variables"
		    " which may NOT be identified; "
		    "systems: a list of mcb=system objects,"
		    " pass-through argument.")
	   (effect  "None.")
	   (value   "Non-NIL, iff elements of systems are"
		    " unifiable."))

  (let ((vl (caar identifiable-vars))
	(vr (cdar identifiable-vars)))
    (cond ((endp identifiable-vars)
	   ; call indexing
	   (let ((combi-vars (mcb=create-combi-vars-list systems)))
	     (mcb=var-indexing-decide combi-vars systems combi-vars)))
	  ((or (term~equal vl vr)
	       (member (first identifiable-vars) non-ident-vars
		       :test #'mcb=var-pair-equal))
	   (mcb=var-identification-decide
	    (rest identifiable-vars) non-ident-vars systems))
	  (T ; there is a pair left that we have the choice to
	     ; identify
	   ; don't identify
	   (or (mcb=var-identification-decide
		(rest identifiable-vars)
		(cons (first identifiable-vars) non-ident-vars)
		systems)
	      
	   
	     ; identify
	       (let ((su (subst~create (list vl) (list vr))))
		 (mcb=var-identification-decide
		  (mcb=subst-var-pairs vl vr (rest identifiable-vars))
		  (mcb=subst-non-ident-vars non-ident-vars vl vr)
		  (subst~apply su systems)))))))) 



(defun mcb=var-identification-complete
  (identifiable-vars non-ident-vars systems-list)

  (declare (edited  "21-June-1995")
	   (authors Kepser)
	   (input   "identifiable-vars: a list of combination variables"
		    " which can/may be identified; "
		    "non-ident-vars: a list of combination variables"
		    " which may NOT be identified; "
		    "systems-list: a list of mcb=system objects,"
		    " pass-through argument.")
	   (effect  "None.")
	   (value   "The list of unifiers for systems-list."))

  (let ((vl (caar identifiable-vars))
	(vr (cdar identifiable-vars)))
    (cond ((endp identifiable-vars)
	   ; call indexing
	   (let ((combi-vars (mcb=create-combi-vars-list systems-list)))
	     (mcb=var-indexing-complete combi-vars systems-list combi-vars)))
	  ((or (term~equal vl vr)
	       (member (first identifiable-vars) non-ident-vars
		       :test #'mcb=var-pair-equal))
	   (mcb=var-identification-complete
	    (rest identifiable-vars) non-ident-vars systems-list))
	  (T ; there is a pair left that we have the choice to
	     ; identify
	   (nconc
 	    ; don't identify
	    (mcb=var-identification-complete
		   (rest identifiable-vars)
		   (cons (first identifiable-vars) non-ident-vars)
		   systems-list)
	      
            ; identify
	    (let* ((su (subst~create (list vl) (list vr)))
		   (unif
		    (mcb=var-identification-complete
		     (mcb=subst-var-pairs vl vr (rest identifiable-vars))
		     (mcb=subst-non-ident-vars non-ident-vars vl vr)
		     (subst~apply su systems-list))))
	      (if unif
		  (list (cons su unif)))))))))
  


(defun mcb=subst-non-ident-vars
  (non-ident-vars dom codom &optional (new-non-ident-vars NIL))
  (declare (edited  "21-June-1995")
	   (authors Kepser)
	   (input   "")
	   (effect  "None.")
	   (value   ""))

  (let ((vl (caar non-ident-vars))
	(vr (cdar non-ident-vars)))
    (if (endp non-ident-vars)
	new-non-ident-vars
      (mcb=subst-non-ident-vars
       (rest non-ident-vars) dom codom
       (cond ((term~equal dom vl)
	      (cons (cons codom vr)
		    new-non-ident-vars))
	     ((term~equal dom vr)
	      (cons (cons vl codom)
		    new-non-ident-vars))
	     (T
	      (cons (first non-ident-vars)
		    new-non-ident-vars)))))))



(defun mcb=subst-var-pairs (dom codom var-pair-list &optional
				(result NIL))
  (let ((v-l (caar var-pair-list))
	(v-r (cdar var-pair-list)))
    (cond ((endp var-pair-list)
	   result)
	  ((term~equal v-l v-r)
	   (mcb=subst-var-pairs dom codom (rest var-pair-list)
				result))
	  ((and (term~equal v-l dom) (term~equal v-r codom))
	   (mcb=subst-var-pairs dom codom (rest var-pair-list)
				result))
	  ((term~equal v-l dom) ; and (not (term~equal v-r codom))
	   (mcb=subst-var-pairs dom codom (rest var-pair-list)
				(acons (term~copy codom)
				       (term~copy v-r)
				       result)))
	  ((and (term~equal v-r dom) (term~equal v-l codom))
	   (mcb=subst-var-pairs dom codom (rest var-pair-list)
				result))
	  ((term~equal v-r dom) ; and (not (term~equal v-l codom))
	   (mcb=subst-var-pairs dom codom (rest var-pair-list)
				(acons (term~copy v-l)
				       (term~copy codom)
				       result)))
	  (T
	   (mcb=subst-var-pairs dom codom (rest var-pair-list)
				(acons (term~copy v-l)
				       (term~copy v-r)
				       result))))))



(defun mcb=create-combi-vars-list (th-list &optional (combi-vars NIL))
  (declare (edited  "21-June-1995")
	   (authors Kepser)
	   (input   "")
	   (effect  "None.")
	   (value   ""))

  (if (endp th-list)
      combi-vars
    (mcb=create-combi-vars-list
     (rest th-list)
     (mcb=insert-combi-vars
      (mcb=system-combi-vars (first th-list))
      (mcb=system-theory (first th-list))
      combi-vars)))) 


(defun mcb=insert-combi-vars (new-vars theory combi-vars)
  (declare (edited  "22-June-1995")
	   (authors Kepser)
	   (input   "")
	   (effect  "None.")
	   (value   ""))

  (if (endp new-vars)
      combi-vars
    (mcb=insert-combi-vars
     (rest new-vars) theory
     (let ((v (find (first new-vars) combi-vars
		    :test #'term~equal :key #'mcb=combi-var-var)))
       (if v
	   (progn
	     (mcb=write-combi-var-systems
	      (cons theory
		    (mcb=combi-var-systems v))
	      v)
	     combi-vars)
	 (cons (make-instance 'mcb=combi-var
			      :variable (first new-vars)
			      :systems (list theory))
	       combi-vars))))))



(defun mcb=var-indexing-decide (combi-vars systems-l combi-vars-context
				&optional (indices NIL)
					  (vars-nat-order NIL)) 
  (declare (edited  "22-June-1995")
	   (authors Kepser)
	   (input   "combi-vars: a list of mcb=combi-var objects; "
		    "systems-l: a list of mcb=system objects; "
		    "indices: an a-list of variable-theory pairs.")
	   (effect  "None.")
	   (value   "Non-NIL, iff the elements of systems-l "
		    "are unifiable."))

  (if (endp combi-vars)
      (mcb=order-decide systems-l indices vars-nat-order 
			  combi-vars-context NIL -1 
			  (mcb=initialise-restr-collector systems-l))
    (do ((unifiable NIL)
	 (c-v (mcb=combi-var-var (first combi-vars)))
	 (c-t (first (mcb=combi-var-systems (first combi-vars)))
	      (first c-tl))
	 (c-tl (rest (mcb=combi-var-systems (first combi-vars)))
	       (rest c-tl)))
	((or unifiable (null c-t)) unifiable)
      (setf unifiable
	    (mcb=var-indexing-decide (rest combi-vars) systems-l
				     combi-vars-context
				     (acons c-v c-t indices)
				     (cons c-v vars-nat-order)))))) 



(defun mcb=var-indexing-complete (combi-vars systems-l combi-vars-context
				  &optional (indices NIL)
				            (vars-nat-order NIL)) 
  (declare (edited  "22-June-1995")
	   (authors Kepser)
	   (input   "combi-vars: a list of mcb=combi-var objects; "
		    "systems-l: a list of mcb=system objects; "
		    "indices: an a-list of variable-theory pairs.")
	   (effect  "None.")
	   (value   "The list of unifiers for systems-l."))

  (if (endp combi-vars)
	(mcb=order-complete systems-l indices vars-nat-order
			     combi-vars-context NIL -1
			     (mcb=initialise-restr-collector systems-l))
    (do ((unifiers NIL)
	 (c-v (mcb=combi-var-var (first combi-vars)))
	 (c-t (first (mcb=combi-var-systems (first combi-vars)))
	      (first c-tl))
	 (c-tl (rest (mcb=combi-var-systems (first combi-vars)))
	       (rest c-tl)))
	((null c-t) unifiers)
      (setf unifiers
	    (nconc
	     (mcb=var-indexing-complete (rest combi-vars) systems-l
					combi-vars-context
					(acons c-v c-t indices)
					(cons c-v vars-nat-order))
	     unifiers))))) 



  


(defun mcb=unification-decide (systems-l restr-collector)
  (do ((unifiable T)
	(unifier NIL)
	(th (first systems-l) (first t-r))
	(t-r (rest systems-l) (rest t-r)))
      ((or (not unifiable) (null th))
       (cond ((not unifiable)
	      NIL)
	     (mcb*unifier
	      unifier)
	     (T T)))
    (let ((collected-restr (find (mcb=system-theory th)
				 restr-collector
				 :key #'mcb=restr-collect-theory)))
      (setq unifiable
	    (th~decide-unify (mcb=system-theory th)
			     (mcb=system-lhs th)
			     (mcb=system-rhs th)
			     (mcb=restr-collect-restr collected-restr)
			     (mcb=restr-collect-consts collected-restr))
	    unifier (cons unifiable unifier)))))



(defun mcb=get-vars (combi-vars indices theory
				&optional (vars NIL) (consts NIL))
  (declare (edited "14-July-1995")
	   (authors Kepser)
	   (input   "")
	   (effect  "")
	   (value   ""))

  (cond ((endp combi-vars)
	 (values vars consts))
	((eq theory (rest (assoc (first combi-vars) indices
				 :test #'term~equal)))
	 (mcb=get-vars (rest combi-vars) indices theory
		       (cons (first combi-vars) vars)
		       consts))
	(T
	 (mcb=get-vars (rest combi-vars) indices theory vars
		       (cons (first combi-vars) consts)))))





(defun mcb~decide-unify (equations)
  (let ((systems-l (mcb~purify-equations equations)))
    (setf systems-l
	  (mcb=combi-vars-to-theories
	   systems-l
	   (mcb=extract-true-combi-vars
	    (mcb=combi-vars systems-l))))
    (mcb=var-identification-decide
     (mcb=identifiables systems-l)
     NIL systems-l)))


;----------------------------------------------------------------------


(defun mcb=unification-complete (systems-l restr-collector)
  (do ((unifiable T)
       (unifier NIL)
       (th (first systems-l) (first t-r))
       (t-r (rest systems-l) (rest t-r)))
      ((or (not unifiable) (null th))
       (if unifiable
	   (list unifier)
	 NIL))
    (let ((collected-restr (find (mcb=system-theory th)
				 restr-collector
				 :key #'mcb=restr-collect-theory)))
      (setq unifiable
	    (th~unify (mcb=system-theory th)
		      (mcb=system-lhs th)
		      (mcb=system-rhs th)
		      (mcb=restr-collect-restr collected-restr)
		      (mcb=restr-collect-consts collected-restr))
	    unifier (cons unifiable unifier)))))









(defun mcb~unify (equations)
  (let ((systems-l (mcb~purify-equations equations)))
    (setf systems-l
	  (mcb=combi-vars-to-theories
	   systems-l
	   (mcb=extract-true-combi-vars
	    (mcb=combi-vars systems-l))))
    (mcb=var-identification-complete
     (mcb=identifiables systems-l)
     NIL systems-l)))



;;; ---------------------------------------------------------------------


(defstruct (mcb=restr-collect
	    (:copier mcb=copy-restr-collect))
  theory vars consts restr)



(defun mcb=initialise-restr-collector (systems-l)
  (let ((restr-collector NIL))
    (dolist (system systems-l restr-collector)
      (setf restr-collector
	(cons (make-mcb=restr-collect 
	       :theory (mcb=system-theory system)
	       :vars NIL
	       :consts NIL
	       :restr NIL)
	      restr-collector)))))



(defun mcb=update-restr-collector (restr-collector sys var sys-of-var
				   &optional (result-restr NIL))
  (cond ((endp restr-collector)
	 result-restr)
	((member (mcb=restr-collect-theory (first restr-collector))
		 sys-of-var)
	 (let ((new-collected-restr (mcb=copy-restr-collect
				     (first restr-collector))))
	   (if (eq sys (mcb=restr-collect-theory new-collected-restr))
	       (setf (mcb=restr-collect-vars new-collected-restr)
		     (cons var (mcb=restr-collect-vars new-collected-restr)))
	     (setf (mcb=restr-collect-consts new-collected-restr)
		   (cons var (mcb=restr-collect-consts new-collected-restr))
		   (mcb=restr-collect-restr new-collected-restr)
		   (cons (cons var (mcb=restr-collect-vars
				    new-collected-restr))
			 (mcb=restr-collect-restr new-collected-restr))))
	   (mcb=update-restr-collector
	    (rest restr-collector) sys var sys-of-var
	    (cons new-collected-restr result-restr))))
	(T ; restr of system not in sys-of-var
	 (mcb=update-restr-collector
	  (rest restr-collector) sys var sys-of-var
	  (cons (first restr-collector) result-restr)))))



(defun mcb=order-decide (systems-l indices vars-natural-order combi-vars-l
			 last-theory last-position restr-collector) 
  (if (endp indices)
      ;; call unification
      (mcb=unification-decide systems-l restr-collector)
    (do ((var (caar indices)
	      (caar ind-aux))
	 (index (cdar indices)
		(cdar ind-aux))
	 (ind-aux (rest indices)
		  (rest ind-aux))
	 (position)
	 (unifiable NIL))
	((or unifiable (null var)) unifiable)
      (setq position (position var vars-natural-order))
      (when (or (not (eq index last-theory))
		(< last-position position))
	(setq unifiable
	      (mcb=order-decide
	       systems-l (remove var indices :key #'first) vars-natural-order
	       combi-vars-l index (if (eq index last-theory) position -1) 
	       (mcb=update-restr-collector
		restr-collector index var
		(mcb=combi-var-systems
		 (find var combi-vars-l :test #'term~equal
		                        :key  #'mcb=combi-var-var))))))))) 




(defun mcb=order-complete (systems-l indices vars-natural-order combi-vars-l
			   last-theory last-position restr-collector) 
  (if (endp indices)
      ;; call unification
      (mcb=unification-complete systems-l restr-collector)
    (do ((var (caar indices)
	      (caar ind-aux))
	 (index (cdar indices)
		(cdar ind-aux))
	 (ind-aux (rest indices)
		  (rest ind-aux))
	 (position)
	 (unifier NIL))
	((null var) unifier)
      (setq position (position var vars-natural-order))
      (when (or (not (eq index last-theory))
		(< last-position position))
	(setq unifier
	      (nconc
	       (mcb=order-complete
		systems-l (remove var indices :key #'first) vars-natural-order
		combi-vars-l index (if (eq index last-theory) position -1) 
		(mcb=update-restr-collector
		 restr-collector index var
		 (mcb=combi-var-systems
		  (find var combi-vars-l :test #'term~equal
			                 :key  #'mcb=combi-var-var))))
	       unifier)))))) 

