(in-package "KEIM")



(defmacro ord=push-list! (list place)
  (declare (edited  "22-FEB-1996")
	   (authors Richts)
	   (input   "A list and a place that can be modified with SETF.")
	   (effect  "LIST is NCONCed with the value of PLACE and PLACE is set to"
		    "this new list.")
	   (value   "The new list, i.e. the new value of PLACE."))
  `(setf ,place (nconc ,list ,place)))



;;;
;;; Aenderung: Konsistenz var-ind, not-var-ind, Abschluss bilden
;;;



(defun ord~order (up ident discrim var-ind not-var-ind const-restr)
  (declare (edited  "22-FEB-1996")
	   (authors Richts)
	   (input   "A general unification problem and five lists of constraints.")
	   (effect  "New information that follows from the argument constraints"
		    "is computed; redundant information in the arguments is"
		    "removed.  The labels of the variables are updated with all"
		    "this information.")
	   (value   "A multiple value:"
		    "1. A flag signalling if the constraints are still consistent."
		    "2. - 5. Five lists of constraints (identification,"
		    "  discrimination, theory indices, negative theory"
		    "  indices  and constant restrictions)"
		    "  that are a subset of the arguments, but without redundant"
		    "  information."
		    "6. - 9. Five lists of constraints (identification,"
		    "  discrimination, theory indices, negative theory"
		    "  indices and constant restrictions)"
		    "  that are consequences of the aguments."))
  (let (real-ident real-discrim real-var-ind real-not-var-ind real-const-restr
        new-ident  new-discrim  new-var-ind new-not-var-ind new-const-restr)
    (labels ((ord=order-new-ident! (ident &optional recursive)
	       ;(break "anfang")
	       (every #'(lambda (id)
			  (let ((var1 (unicp~id-left id))
				(var2 (unicp~id-right id)))
			    (cond ((ord=id-inconsistent-p var1 var2) nil)
				  ((unicp~vars-identified-p var1 var2) t)
				  (t
				   (if recursive
				       (push id new-ident)
				       (push id real-ident))
				   (ord=order-new-discrim! (ord=id-new-discrim var1 var2)
							   t)
				   (ord=push-list! (unicp~vars-identify! var1 var2)
						   new-ident)
				   (ord=push-list! (ord=id-new-var-ind! var1 var2)
						   new-var-ind)
				   (ord=push-list! (ord=id-new-not-var-ind! var1 var2)
						   new-not-var-ind)
				   (ord=order-new-const-restr! (ord=id-new-const-restr var1 var2)
							       t)
				   t))))
		      ident))
	     (ord=order-new-discrim! (discrim &optional recursive)
	       (every #'(lambda (dis)
			  (let ((var1 (unicp~id-left dis))
				(var2 (unicp~id-right dis)))
			    (cond ((unicp~vars-identified-p var1 var2)
				   nil)
				  ((unicp~vars-discriminated-p var1 var2)
				   t)
				  (t
				   (if recursive
				       (push dis new-discrim)
				       (push dis real-discrim))
				   (ord=push-list! (ord=new-discrim! var1 var2)
						   new-discrim)
				   t))))
		      discrim))
	     (ord=order-new-var-ind! (var-ind &optional recursive)
	       (declare (ignore recursive))
	       (every #'(lambda (vi)
			  (let* ((var (unicp~vi-var vi))
				 (old-theory (unicp~var-label-theory var)))
			    (cond ((member (unicp~vi-theory vi) (unicp~var-label-not-theory var)) nil)
				  ((null old-theory)
				   (push vi real-var-ind)
				   (setf (unicp~var-label-theory var) (unicp~vi-theory vi))
				   (ord=push-list! (ord=vi-new-var-ind! var (unicp~vi-theory vi))
						   new-var-ind)
				   (ord=order-new-discrim!
				    (ord=vi-new-discrim var (unicp~vi-theory vi)
							(unicp~variables up))
				    t)
				   t)
				  ((eq old-theory (unicp~vi-theory vi)) t)
				  (t nil))))
		      var-ind))
	     (ord=order-new-not-var-ind! (not-var-ind &optional recursive)
	       (declare (ignore recursive))
	       (every
		#'(lambda (vi)
		    (let* ((var (unicp~vi-var vi))
			   (new-theory (unicp~vi-theory vi))
			   (old-theory (unicp~var-label-theory var))
			   (up-list (misc~queue-list (unicp~gup-up-queue up))))
		      (cond ((eq new-theory old-theory) nil)
			    ((not (member new-theory (unicp~var-label-not-theory var)))
			     (push new-theory (unicp~var-label-not-theory var))
			     (cond ((= (1+ (length (unicp~var-label-not-theory var)))
				       (length up-list))
				    (let ((remaining-theory
					   (unicp~theory
					    (find-if-not
					     #'(lambda (theory)
						 (member theory (unicp~var-label-not-theory var)))
					     up-list
					     :key #'unicp~theory))))
				      (ord=order-new-var-ind!
				       (list (unicp~vi-create var remaining-theory)))))
				   (t (push vi real-not-var-ind)
				      (ord=push-list! (ord=vi-new-not-var-ind! var new-theory)
						      new-not-var-ind)
				      t)))
			    (t t))))
		not-var-ind))
	     (ord=order-new-const-restr! (const-restr &optional recursive)
	       ;(format *trace-output* "~% ENTER ord=order-new-const-restr! ~A ~A." const-restr recursive)
	       (every
		#'(lambda (res)
		    (let ((const (unicp~cr-const res)))
		      (every
		       #'(lambda (var)
			   (cond ((unicp~vars-less-p const var) nil)
				 ((unicp~vars-identified-p const var) t)
				 ((unicp~vars-lesseq-p const var)
				  (ord=order-new-ident! (list (unicp~id-create const var)) t))
				 ((unicp~vars-lesseq-p var const) t)
				 (t
				  (if recursive
				      (push (unicp~cr-create const var)
					    new-const-restr)
				      (push (unicp~cr-create const var)
					    real-const-restr))
				  (prog1
				      (ord=order-new-discrim! (ord=cr-new-discrim var const)
							      t)
				    (ord=push-list!
				     (ord=new-const-restr! var const)
				     new-const-restr)))))
		       (unicp~cr-varlist res))))
		const-restr)))
      (and
       (ord=order-new-var-ind! var-ind)
       ;;(print "Variable Index: ok." *trace-output*)
       (ord=order-new-not-var-ind! not-var-ind)
       ;;(print "Negative Variable Index: ok." *trace-output*)
       (ord=order-new-discrim! discrim);(break "4")
       ;;(print *trace-output* "Discrimination: ok." *trace-output*)
       (ord=order-new-ident! ident);(break "5")
       ;;(print "Identification: ok." *trace-output*)
       (ord=order-new-const-restr! const-restr)
       ;;(print "Constant Restrictions: ok." *trace-output*)
       (values t
	       real-ident real-discrim real-var-ind real-not-var-ind real-const-restr
	       new-ident new-discrim new-var-ind new-not-var-ind new-const-restr)))))


(defun ord=vi-new-discrim (var theory variables)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "A variable, a theory, and a list of variables.")
	   (effect  "None.")
	   (value   "A list of all variables in VARIABLES that have a theory"
		    "index different from THEORY and that have not been"
		    "discriminated from VAR, yet.  These variables must be"
		    "discriminated from VAR since THEORY will be VAR's theory"
		    "index."))
  (mapcan #'(lambda (v)
	      (when (and (unicp~var-label-theory v)
			 (not (eq (unicp~var-label-theory v) theory))
			 (not (unicp~vars-discriminated-p v var)))
		(list (unicp~id-create v var))))
	  variables))

(defun ord=vi-new-var-ind! (var theory)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "A variable and a theory.")
	   (effect  "The theory index of all variables identified with VAR is"
		    "set to THEORY.")
	   (value   "A list of all variables identified with VAR that had no"
		    "theory index before."))
  (mapcan #'(lambda (v)
	      (unless (unicp~var-label-theory v)
		(setf (unicp~var-label-theory v) theory)
		(list (unicp~vi-create v theory))))
	  (unicp~var-label-identified var)))

(defun ord=vi-new-not-var-ind! (var theory)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "A variable and a theory.")
	   (effect  "The theory is added to the negative theory indices of all variables"
		    "identified with VAR")
	   (value   "A list of all variables identified with VAR that hadn't theory as negative"
		    "theory index before."))
  (mapcan #'(lambda (v)
	      (unless (member theory (unicp~var-label-not-theory var))
		(push theory (unicp~var-label-not-theory var))
		(list (unicp~vi-create v theory))))
	  (unicp~var-label-identified var)))

(defun ord=new-const-restr! (less greater)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "Two variables.")
	   (effect  "If LESS and GREATER are not identified and have not been"
		    "ordered, the following action is performed:  First, the"
		    "ordering of LESS and GREATER is set in their labels.  Then"
		    "for each pair of variables, where the first variable is"
		    "less than or identified with LESS and the second variable"
		    "is greater than or identified with GREATER, the ordering is"
		    "set in their labels (the first variable is less than the"
		    "second one), if they have not been ordered or identified so"
		    "far.")
	   (value   "A list containing all pairs of variables (except the pair"
		    "of LESS and GREATER) for which the ordering was set."))
  (if (ord=add-const-restr-p! less greater)
      (mapcan #'(lambda (l)
		  (mapcan #'(lambda (g)
			      (when (ord=add-const-restr-p! l g)
				(list (unicp~cr-create g l))))
			  (append (unicp~var-label-identified greater)
				  (unicp~var-label-greater greater))))
	      (append (unicp~var-label-identified less)
		      (unicp~var-label-less less)))))


(defun ord=cr-new-discrim (less greater)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "Two variables.")
	   (effect  "None.")
	   (value   "If LESS is already less than GREATER, NIL. Else a list of"
		    "new discriminations that contains the following pairs:  For"
		    "each variable that is less than LESS and discriminated from"
		    "LESS, all discriminations between this variable and all"
		    "variables greater than GREATER (including GREATER); and"
		    "anlogously for variables greater than GREATER and"
		    "discriminated from GREATER."))
  (let ((less-list (cons less (unicp~var-label-less less)))
	(greater-list (cons greater (unicp~var-label-greater greater))))
    (cond ((unicp~vars-lesseq-p less greater)
	   nil)
	  ((unicp~vars-discriminated-p less greater)
	   (mapcan #'(lambda (l)
		       (mapcan #'(lambda (g)
				   (unless (unicp~vars-discriminated-p l g)
				     (list (unicp~id-create l g))))
			       greater-list))
		   less-list))
	  (t
	   (nconc
	    (mapcan #'(lambda (l)
			(if (unicp~vars-discriminated-p l less)
			    (mapcan #'(lambda (g)
					(unless (unicp~vars-discriminated-p l g)
					  (list (unicp~id-create l g))))
				  greater-list)))
		    (unicp~var-label-less less))
	    (mapcan #'(lambda (g)
			(if (unicp~vars-discriminated-p g greater)
			    (mapcan #'(lambda (l)
					(unless (unicp~vars-discriminated-p l g)
					(list (unicp~id-create l g))))
				    less-list)))
		    (unicp~var-label-greater greater)))))))


(defun ord=add-const-restr-p! (less greater)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "Two variables.")
	   (effect  "If less and greater are not identified and one is not"
		    "greater than the other, then LESS is added to the less slot"
		    "of GREATER's label-object and GREATER is added to the"
		    "greater slot of LESS's label-object.")
	   (value   "True, iff slots have been modified, i.e. if the condition"
		    "is true."))
  (unless (or (unicp~vars-identified-p less greater)
	      (unicp~vars-lesseq-p less greater)
	      (unicp~vars-lesseq-p greater less))
    (push less (unicp~var-label-less greater))
    (push greater (unicp~var-label-greater less))
    T))


(defun ord=new-discrim! (var1 var2)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "Two variables.")
	   (effect  "First VAR1 and VAR2 are discriminated, i.e. their labels"
		    "are updated.  Then, if VAR1 and VAR2 are ordered, all"
		    "variables less than the less one are discriminated from all"
		    "variables greater than the greater one (including the"
		    "greater one and the less one, respectively).")
	   (value   "A list of all discriminations that are new."))
  (nconc
   (unicp~vars-discriminate! var1 var2)
   (let* ((less-1-2 (unicp~vars-lesseq-p var1 var2))
	  (less-2-1 (unicp~vars-lesseq-p var2 var1))
	  (less (if less-1-2 var1 var2))
	  (greater (if less-1-2 var2 var1)))
     (if (or less-1-2 less-2-1)
	 (mapcan #'(lambda (l)
		     (mapcan #'(lambda (g)
				 (when (and (not (eq l g))
					    (ord=add-discrim-p! l g))
				   (list (unicp~id-create g l))))
			     (append (unicp~var-label-identified greater)
				     (unicp~var-label-greater greater))))
		 (append (unicp~var-label-identified less)
			 (unicp~var-label-less less)))))))



(defun ord=add-discrim-p! (var1 var2)
  (declare (edited  "21-FEB-1996")
	   (authors Richts)
	   (input   "Two variable that shall be discriminated.")
	   (effect  "If VAR1 and VAR2 are not yet discriminated, they are set to"
		    "be discriminated.")
	   (value   "True, iff VAR1 and VAR2 have not been discriminated before."))
  (cond ((unicp~vars-identified-p var1 var2)
	 (error "ORDER: Programmierfehler, die Variablen ~A und ~A sollen diskriminiert werden, ~
                       obwohl sie schon identifiziert sind; ~
                       es haette zuvor ein Widerspruch entdeckt werden muessen!"
		var1 var2))
	((unicp~vars-discriminated-p var1 var2)
	 nil)
	(t
	 (unicp~vars-discriminate! var1 var2)
	 T)))

(defun ord=id-inconsistent-p (var1 var2)
  (declare (edited  "25-NOV-1996")
	   (authors Richts)
	   (input   "Two variables which are going to be identified.")
	   (effect  "None.")
	   (value   "True iff identification of VAR1 and VAR2 is inconsistent,"
		    "i.e. they are discriminated or their label information"
		    "cannot be joined."))
  (let ((theory1 (unicp~var-label-theory var1))
	(theory2 (unicp~var-label-theory var2)))
    (or (unicp~vars-discriminated-p var1 var2)
	(and theory1
	     theory2
	     (not (eq theory1 theory2)))
	(member theory1 (unicp~var-label-not-theory var2))
	(member theory2 (unicp~var-label-not-theory var1)))))

(defun ord=id-new-discrim (var1 var2)
     ;;; called BEFORE identification of VAR1 and VAR2
  (declare (edited  "22-FEB-1996")
	   (authors Richts)
	   (input   "Two variables that are going to be identified.")
	   (effect  "None.")
	   (value   "A list of new discirminations containing pairs of variables"
		    "identified with VAR1 and variables discriminated from VAR2"
		    "and pairs of variables identified with VAR2 and variables"
		    "discriminated from VAR1."))
  (nconc
   (mapcan #'(lambda (var)
	       (mapcan #'(lambda (other-var)
			   (unless (unicp~vars-discriminated-p var other-var)
			     (list (unicp~id-create var other-var))))
		       (unicp~vars-discriminated var2)))
	   (unicp~var-label-identified var1))
   (mapcan #'(lambda (var)
	       (mapcan #'(lambda (other-var)
			   (unless (unicp~vars-discriminated-p var other-var)
			     (list (unicp~id-create var other-var))))
		       (unicp~vars-discriminated var1)))
	   (unicp~var-label-identified var2))))

(defun ord=id-new-var-ind! (var1 var2)
     ;;; called AFTER identification of VAR1 and VAR2
  (declare (edited  "22-FEB-1996")
	   (authors Richts)
	   (input   "Two variables that have been identified.  If VAR2 has a"
		    "theory index assigned and VAR1 has not, the function is"
		    "called with the arguments swapped immediatly.  So if one of"
		    "the variables has a theory index assigned, then VAR1 has"
		    "one.")
	   (effect  "If VAR1 has a theory index and VAR2 has a different one,"
		    "then an error is signaled.  Else, if VAR1 has a theory"
		    "index and VAR2 has none (i.e. it is NIL), then the theory"
		    "index of all variables identified with VAR2 (including"
		    "VAR2) is set to the the theory index of VAR1.")
	   (value   "The list of new theory indices."))
  (let ((theory (unicp~var-label-theory var1)))
    (cond ((and (null theory) (unicp~var-label-theory var2))
	   (ord=id-new-var-ind! var2 var1))
	  ((unicp~var-label-theory var2)
	   (if (eq (unicp~var-label-theory var2) theory)
	       nil
	       (error "ORDER: Programmierfehler, ~A und ~A sollen identifiziert werden, ~
                              haben aber verschiedene Variablenindizes" var1 var2)))
	  ((null theory) nil)
	  (t
	   (mapcan #'(lambda (var)
		       (unless (unicp~var-label-theory var)
			 (setf (unicp~var-label-theory var) theory)
			 (list (unicp~vi-create var theory))))
		   (unicp~var-label-identified var2))))))

(defun ord=id-new-not-var-ind! (var1 var2)
     ;;; called AFTER identification of VAR1 and VAR2
  (declare (edited  "22-FEB-1996")
	   (authors Richts)
	   (input   "Two variables that have been identified.")
	   (effect  "")
	   (value   ""))
  (let ((theory-list (union (unicp~var-label-not-theory var1)
			    (unicp~var-label-not-theory var2))))
    (mapcan #'(lambda (var)
		(mapcan #'(lambda (theory)
			    (unless (member theory (unicp~var-label-not-theory var))
			      (push theory (unicp~var-label-not-theory var))
			      (list (unicp~vi-create var theory))))
			theory-list))
	    (unicp~var-label-identified var2))))

(defun ord=id-new-const-restr (var1 var2)
     ;;; called AFTER identification of VAR1 and VAR2
  (declare (edited  "22-FEB-1996")
	   (authors Richts)
	   (input   "Two variables that have been identified.")
	   (effect  "None.")
	   (value   "A list of new constant restrictions (i.e. ordering"
		    "information for pairs of variables).  It considers all"
		    "pairs between variables that are less than VAR1 or VAR2 or"
		    "identified with both and variables that are greater than"
		    "VAR1 or VAR2 or identified with both.  Such a pair is"
		    "contained in the result if it is not ordered or identified"
		    "so far."))
  (mapcan #'(lambda (l)
	      (mapcan #'(lambda (g)
			  (unless (or (unicp~vars-identified-p l g)
				      ;(unicp~vars-lesseq-p g l)
				      (unicp~vars-lesseq-p l g))
			    (list (unicp~cr-create g l))))
		      (append (unicp~var-label-identified var1)
			      (unicp~var-label-greater var1)
			      (unicp~var-label-greater var2))))
	  (append (unicp~var-label-identified var1)
		  (unicp~var-label-less var1)
		  (unicp~var-label-less var2))))
  

#|
(defun ord==test (real-discrim new-discrim)
  (when (or (member "(x-112 x-106)" (append real-discrim new-discrim)
		    :test #'(lambda (str id)
			      (string= str (format nil "~A" id))))
	    (member "(x-106 x-112)" (append real-discrim new-discrim)
		    :test #'(lambda (str id)
			      (string= str (format nil "~A" id)))))
    (format t "~% *** ~A" real-discrim)
    (format t "~% *** ~A" new-discrim)))
|#
