
(in-package "KEIM")

;;; Algorithm

(defvar diocd*trace nil)

(defun diocd-solve.equations (matrix)
  (let ((columns (diocd=matrix.columns matrix))
	(sol-tree nil))
    (do ((tree nil nil)
	 (proposals
	   (diocd=start.proposals columns)
	   (mapcan #'(lambda (proposal)
		       (do ((i 0 (1+ i))
			    (column-tail columns (cdr column-tail))
			    (ddd t (when diocd*trace 
				     (format diocd*trace  "~%Prop: ~A  , Col: ~A   "
					     (my-print proposal nil) (my-print (car column-tail) nil))))
			    (successors
			      nil
			      (if (minusp (diocd=vector.scalar.product (cdr proposal) (car column-tail)))
				  (let* ((new-proposal (diocd=proposal.copy proposal))
					 (new-vector (car new-proposal))
					 (new-defect (cdr new-proposal)))
				    (incf (elt new-vector i))
				    (multiple-value-bind (new-tree member-p)
					(diocd=tree.insert tree new-vector)
				      (setf tree new-tree)
				      (if (or member-p (not (diocd=tree.minimal sol-tree new-vector)))
					  successors
					  (progn
					    (diocd=vector.+ new-defect (car column-tail))
					    (when diocd*trace (format diocd*trace  "  -> ~A" (my=print new-proposal)))
					    (if (diocd=vector.zerop new-defect)
						(progn (setf sol-tree (diocd=tree.insert sol-tree new-vector))
						       (when diocd*trace (format diocd*trace "~%Solutions: ~A"
										 (my=print sol-tree)))
						       successors)
						(cons new-proposal successors))))))
				  successors)))
			   ((null column-tail) successors)))
		   proposals)))
	((null proposals))
      (when diocd*trace (format diocd*trace "~%Naechste Stufe: ~%  ~A" (my=print proposals))))
    (diocd=tree.members sol-tree)))

(defun diocd=start.proposals (columns)
  (let ((n (length columns)))
    (do ((i 0 (1+ i))
	 (column-tail columns (cdr column-tail))
	 (proposals nil ;(list (diocd=proposal.make n i (car column-tail)))
		    (cons (diocd=proposal.make n i (car column-tail)) proposals)))
	 ((null column-tail) proposals))))

(defun diocd=proposal.make (n i defect)
  (let ((vector (make-array n :initial-element 0)))
    (setf (elt vector i) 1)
    (cons vector defect)))

(defun diocd=proposal.copy (proposal)
  (cons (copy-seq (car proposal)) (copy-seq (cdr proposal))))


;;; Trees

(defun diocd=tree.insert (tree vector)
  (if tree
      (let ((member-p (diocd=tree.insert.rec tree vector 0)))
	(values tree member-p))
      (values (acons (elt vector 0) vector nil))))

(defun diocd=tree.insert.rec (tree vector pos)
  (let* ((coeff (elt vector pos))
	 (subtree-pair (assoc coeff tree))
	 (subtree (cdr subtree-pair)))
    (cond ((null subtree-pair) (push (cons coeff vector) (cdr tree))
	   nil)
	  ((vectorp subtree)
	   (if (diocd=vector.equal vector subtree)
	       t
	       (progn (setf (cdr subtree-pair) (acons (elt subtree (1+ pos)) subtree (acons (elt vector (1+ pos)) vector nil)))
		      nil)))
	  (t (diocd=tree.insert.rec subtree vector (1+ pos))))))

(defun diocd=tree.minimal (tree vector &optional (pos 0))
  (cond ((null tree) t)
	((vectorp tree) (not (diocd=solution.<= tree vector)))
	(t (every #'(lambda (subtree-pair)
		      (or (< (elt vector pos) (car subtree-pair))
			  (diocd=tree.minimal (cdr subtree-pair) vector (1+ pos))))
		  tree))))

(defun diocd=tree.members (tree)
  (cond ((null tree) nil)
	((vectorp tree) (list tree))
	(t (nconc (diocd=tree.members (cdar tree)) (diocd=tree.members (cdr tree))))))

;;; Matrices

(defun diocd=matrix.columns (matrix)
  (declare (edited  "31-JUL-1991 18:57")
	   (authors RICHTS)
	   (input   "A list of lists of integers of equal length.")
	   (effect  "None.")
	   (value   "A list containing the columns (vectors) of matrix."))
  (let* ((m (length matrix))
	 (n (length (elt matrix 0)))
	 (new-matrix (make-list n)))
    (do ((i 0 (1+ i)))
	((= i n))
      (do ((new-row (make-array m))
	   (j 0 (1+ j)))
	  ((= j m)
	   (setf (elt new-matrix i) new-row))
	(setf (elt new-row j) (elt (elt matrix j) i))))
    new-matrix))

(defun diocd=vector.scalar.product (vector1 vector2)
  (do* ((i (length vector1) (1- i))
	(sum 0 (+ sum (* (elt vector1 i) (elt vector2 i)))))
       ((= i 0) sum)))

(defun diocd=vector.+ (vector1 vector2)
  (do ((i (1- (length vector1)) (1- i)))
      ((< i 0) vector1)
    (setf (elt vector1 i) (+ (elt vector1 i) (elt vector2 i)))))

(defun diocd=vector.zerop (vector)
  (every #'zerop vector))

(defun diocd=vector.equal (vector1 vector2)
  (every #'= vector1 vector2))

(defun diocd=solution.<= (solution1 solution2)
  (do* ((i (length solution1) (1- i))
	(sol1.<=.sol2 t (and sol1.<=.sol2 (<= (elt solution1 i) (elt solution2 i)))))
       ((or (zerop i) (not sol1.<=.sol2))
	sol1.<=.sol2)))


;;; printing

(defun my-print (object &optional (stream t))
  (if (equal stream t)
      (format t "~%~A" (my=print object))
      (format stream "~A" (my=print object))))

(defun my=print (object)
  (cond ((vectorp object) (format nil "[~{~A ~}]" (map 'list #'(lambda (object) (my=print object)) object)))
	((consp object) (cons (my=print (car object)) (my=print (cdr object))))
	((listp object) (mapcar #'(lambda (object) (my=print object)) object))
	(t object)))


#|

(my-print (setf res (diocd-solve.equations '((1 2 -3 -2 -4) (2 -1 -3 2 5)))))    ;Bsp. 1:  10 loesungen
(my-print (setf res (diocd-solve.equations '((10 -7 -8 3 -11) (12 -9 -7 3 13)))))      ;Bsp. 2: 240 loesungen
(my-print (setf res (diocd-solve.equations '((1 2 -1 0 -2 -1) (0 -1 -2 2 0 1) (2 0 1 -1 -2 0)))))      ;Bsp. 3: 13 loesungen
(my-print (setf res (diocd-solve.equations '((-10 0 20 -1 -21) (9 1 -17 2 19)))))      ;Bsp. 3: 0 loesungen
(my-print (setf res (diocd-solve.equations '((1 2 -2 -1) (1 0 -2 0)))))              ;Bsp. CD: 2 Loesungen
 |#
