;;; -*- Package: HADES; Syntax: Common-lisp; Mode: LISP -*-


(in-package 'keim)

(defmacro test= (number epression result &optional (test 'eq))
  `(let ((res ,epression))
     (if (funcall #',test res ',result)
	 (format t "~%~A" ,number)
       (format t "~%Test ~A results in ~A, but ~A was expected." ,number res ',result))))

(defvar dio*alg :aed)

(defun seq-equal (seq1 seq2)
  (and (= (length seq1) (length seq2))
       (every #'eql seq1 seq2)))

(defun dio.test-solve.equation (matrix)
  (declare (special dio*alg))
  (cond ((eq dio*alg t) (break "Algorithmus nicht fuer Systeme von diophantischen Gleichungen."))
	((eq dio*alg :fc) (break "Algorithmus nicht fuer Systeme von diophantischen Gleichungen."))
	((eq dio*alg :ed) (dioed-solve.equations matrix))
	((eq dio*alg :aed) (dioaed-solve.equations matrix))
	((eq dio*alg :cd) (diocd-solve.equations matrix))
	((eq dio*alg :lp) (diolp-solve.equations matrix))
	(t (break "Unbekannter Algorithmus"))))


(defun dio.test-solutions.= (solutions1 solutions2)
  (declare (special dio*alg))
  (if (numberp solutions2)
      (= (length solutions1) solutions2)
      (and (= (length solutions1) (length solutions2))
	   (every #'(lambda (sol) (member sol solutions2 :test #'seq-equal)) solutions1))))

;(time (progn

(test= 1
       (dio.test-solve.equation '((1 1 -1 -1)))
       ((1 0 1 0) (1 0 0 1) (0 1 1 0) (0 1 0 1))
       dio.test-solutions.=)

(test= 2
       (dio.test-solve.equation '((3 -1 -1 -1)))
       ((1 0 0 3) (1 0 3 0) (1 3 0 0) (1 2 1 0) (1 2 0 1) (1 1 2 0) (1 0 2 1) (1 1 0 2) (1 0 1 2) (1 1 1 1))
       dio.test-solutions.=)

(test= 3
       (dio.test-solve.equation '((1 4 -2 -3)))
       ((3 0 0 1) (2 1 0 2) (2 0 1 0) (1 1 1 1) (1 2 0 3) (0 2 1 2) (0 1 2 0) (0 3 0 4))
       dio.test-solutions.=)

(test= 4
       (dio.test-solve.equation '((3 -2 -3)))
       ((1 0 1) (2 3 0))
       dio.test-solutions.=)

(test= 5
       (dio.test-solve.equation '((1 1 2 -1 -2 -10)))
       43
       dio.test-solutions.=)

#+weg(test= 6
       (dio.test-solve.equation '((1 2 10 -1 -1 -1 -2 -2)))
       349
       dio.test-solutions.=)

#+weg(test= 7
       (dio.test-solve.equation '((2 2 2 3 3 3 -2 -2 -2 -3 -3 -3)))
       138
       dio.test-solutions.=)

(test= 8
       (dio.test-solve.equation '((1 4 -3 -5)))
       ((1 1 0 1 ) (0 2 1 1 ) (2 1 2 0 ) (1 2 3 0 ) (0 5 0 4 ) (5 0 0 1 ) (0 3 4 0 ) (3 0 1 0 ))
       dio.test-solutions.=)




(test= 9
       (dio.test-solve.equation '((1 2 -3 -2 -4) (2 -1 -3 2 5)))
       ((0 10 2 3 2) (3 7 3 0 2) (0 11 2 6 1) (2 5 2 1 1) (6 5 4 0 1)
	(1 3 1 2 0) (5 3 3 1 0) (0 9 2 0 3) (0 12 2 9 0) (9 3 5 0 0))
       dio.test-solutions.=)

(test= 10
       (dio.test-solve.equation '((10 -7 -8 3 -11) (12 -9 -7 3 13)))
       240
       dio.test-solutions.=)

(test= 11
       (dio.test-solve.equation '((1 2 -1 0 -2 -1) (0 -1 -2 2 0 1) (2 0 1 -1 -2 0)))
       ((9 4 1 3 8 0) (8 4 2 4 7 0) (7 4 3 5 6 0) (6 4 4 6 5 0) (5 4 5 7 4 0)
	(4 4 6 8 3 0) (3 4 7 9 2 0) (2 4 8 10 1 0) (0 8 2 0 1 12) (1 1 0 0 1 1)
	(0 1 1 1 0 1) (10 4 0 2 9 0) (1 4 9 11 0 0))
       dio.test-solutions.=)

(test= 12
       (dio.test-solve.equation '((-10 0 20 -1 -21) (9 1 -17 2 19)))
       0
       dio.test-solutions.=)

(test= 13
       (dio.test-solve.equation '((1 2 -2 -1) (1 0 -2 0)))
       ((2 0 1 0) (0 1 0 2))
       dio.test-solutions.=)


;))
