
(in-package "KEIM")

(mod~defmod th
	    :uses ()
	    :exports (th+general
		      th+collapse-free
		      th+regular
		      th+fine
		      th+simple
		      th~theory-symbols
		      th~theory-axioms
		      th+unification-class
		      th+uc-general
		      th~combine-unif-classes
		      th~combine-unif-class-list))


#{
Each symbol has a theory slot.  The theories are ordered hierarchically.
#}

(defclass th+general (keim+unif-class)
  ((symbolnames :allocation :class)
   (symbols)
   (axioms))
  (:documentation "This is the superclass of all equational theories."))

(defclass th+collapse-free (th+general)
  ()
  (:documentation "This is the superclass of all collapse-free equational theories."))

(defclass th+regular (th+general)
  ()
  (:documentation "This is the superclass of all regular equational theories."))

(defclass th+fine (th+collapse-free th+regular)
  ()
  (:documentation "This is the superclass of all collapse-free and regular equational theories."))

(defclass th+simple (th+fine)
  ()
  (:documentation "This is the superclass of all simple equational theories."))


(defgeneric th~theory-symbols (theory)
  (declare (edited  "12-DEC-1994")
	   (authors Richts)
	   (input   "A theory or a symbol.")
	   (effect  "None.")
	   (value   "A list with the symbols of the theory.")))

(defgeneric th~theory-axioms (theory)
  (declare (edited  "12-DEC-1994")
	   (authors Richts)
	   (input   "A theory or a symbol.")
	   (effect  "None.")
	   (value   "The list of axioms of the theory.")))


#{Each term has a unification class.  The unification class determines, which
unification algorithm must be used. It is computed from the theories in a
term. If the term contains only symbols of one theory (plus variables), the
theory itself is the unification class (i.e. {\tt th+general} is a subclass of
{\tt th+unification-class}).#}

(defclass th+unification-class (keim+name)
  ((type :allocation :class))
  (:documentation "This is the superclass of all unification classes."))


(defclass th+uc-general (th+unification-class)
  ((type :initform :infinitary :allocation :class))
  (:documentation "This is the unification class for general unification."))


(defgeneric th~combine-unif-classes (unif-class1 unif-class2)
  (declare (edited  "12-DEC-1994")
	   (authors Richts)
	   (input   "Two unification classes or two terms.")
	   (effect  "None.")
	   (value   "An object of the unification class, that results if the two unification classes are combined.")))

(defun th~combine-unif-class-list (unif-class-list)
  (declare (edited  "12-DEC-1994")
	   (authors Richts)
	   (input   "A list of unification classes or terms.")
	   (effect  "None.")
	   (value   "An object of the unification class, that combines all unification classes in the list.")))

#{
Theories and instances of theories are defined in the \post declarations:


\begin{postsyntax}
\syntax{\nt{theory-decl} ::= (theories (\nt{name} (\nt{type}+) \nt{axiom}*)*).}
\end{postsyntax}

\begin{postsyntax}
\syntax{\nt{theory-inst-decl} ::= (theory-inst (\nt{theory-name} \nt{function-name}+)+).}
\end{postsyntax}


#}
